package gov.usgs.earthquake.nshmp;

import static com.google.common.base.Preconditions.checkArgument;
import static gov.usgs.earthquake.nshmp.data.DoubleData.checkInRange;

import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.Locations;

/**
 * Constants and utility methods pertaining to faults.
 *
 * @author U.S. Geological Survey
 */
public final class Faults {

  private Faults() {}

  /** Allowed aseismicity factor range: {@code [0..1]}. */
  public static final Range<Double> ASEISMICITY_RANGE = Range.closed(0.0, 1.0);

  /** Allowed fault dip range: {@code [10..90]°}. */
  public static final Range<Double> DIP_RANGE = Range.closed(10.0, 90.0);

  /** Allowed fault rake range: {@code [-180..180]°}. */
  public static final Range<Double> RAKE_RANGE = Range.closed(-180.0, 180.0);

  /** Allowed fault slip rate range: {@code (0..60] mm/yr}. */
  public static final Range<Double> SLIP_RATE_RANGE = Range.openClosed(0.0, 60.0);

  /** Allowed fault strike range: {@code [0..360)°}. */
  public static final Range<Double> STRIKE_RANGE = Range.closedOpen(0.0, 360.0);

  /**
   * Ensure {@code 10° ≤ dip ≤ 90°}.
   *
   * @param dip to validate
   * @return the validated dip
   * @throws IllegalArgumentException if {@code dip} is outside the range
   *         {@code [10..90]°}
   */
  public static double checkDip(double dip) {
    return checkInRange(DIP_RANGE, "Dip", dip);
  }

  /**
   * Ensure {@code strike < dipDirection < strike + 180°}.
   *
   * @param trace to validate against
   * @param dipDirection to validate
   * @return the validated dipDirection
   * @throws IllegalArgumentException if {@code dipDirection} is outside the
   *         range {@code [0..360]°} or is incompatible with trace, assuming
   *         geologic right-hand-rule
   */
  public static double checkDipDirection(LocationList trace, double dipDirection) {
    checkInRange(STRIKE_RANGE, "Dip direction", dipDirection);
    double strikeMin = strike(trace);
    double strikeMax = (strikeMin + 180.0) % 360.0;
    checkArgument((dipDirection > strikeMin && dipDirection < strikeMax) ||
        (dipDirection > strikeMax || dipDirection < strikeMin),
        "Dip direction [%s] unlikely for strike [%s]", dipDirection, strikeMin);
    return dipDirection;
  }

  /**
   * Ensure {@code 0° ≤ strike < 360°}.
   *
   * @param strike to validate
   * @return the validated strike
   * @throws IllegalArgumentException if {@code strike} is outside the range
   *         {@code [0..360)°}
   */
  public static double checkStrike(double strike) {
    return checkInRange(STRIKE_RANGE, "Strike", strike);
  }

  /**
   * Ensure {@code -180° ≤ rake ≤ 180°}.
   *
   * @param rake to validate
   * @return the validated rake
   * @throws IllegalArgumentException if {@code rake} is outside the range
   *         {@code [-180..180]°}
   */
  public static double checkRake(double rake) {
    return checkInRange(RAKE_RANGE, "Rake", rake);
  }

  /**
   * Ensure {@code 0 < rate ≤ 60 mm/yr}.
   *
   * @param rate to validate
   * @return the validated rate
   * @throws IllegalArgumentException if {@code rate} is outside the range
   *         {@code (0..60] mm/yr}
   */
  public static double checkSlipRate(double rate) {
    return checkInRange(SLIP_RATE_RANGE, "Slip rate", rate);
  }

  /**
   * Ensure {@code trace} contains at least two points.
   *
   * @param trace to validate
   * @return the validated trace
   * @throws IllegalArgumentException if {@code trace.size() < 2}
   */
  public static LocationList checkTrace(LocationList trace) {
    checkArgument(trace.size() > 1, "Fault trace must have at least 2 points");
    return trace;
  }

  /**
   * Ensure {@code lower ≥ upper} seismogenic depth.
   *
   * @param lower depth
   * @param upper depth
   * @throws IllegalArgumentException if {@code upper > lower}
   */
  public static void checkDepths(double lower, double upper) {
    checkArgument(lower >= upper, "Lower must be deeper than upper seismogenic depth ");
  }

  /**
   * Ensure {@code 0.0 ≤ aseismicity ≤ 1.0}.
   *
   * @param aseismicity factor to validate
   * @return the validated aseismicity factor
   * @throws IllegalArgumentException if {@code aseismicity} is outside the
   *         range {@code [0..1]}
   */
  public static double checkAseismicityFactor(double aseismicity) {
    return checkInRange(ASEISMICITY_RANGE, "Aseismicity factor", aseismicity);
  }

  /**
   * Generic model for hypocentral depth returns a value that is halfway between
   * the top and bottom of a fault, parameterized by its dip, width, and depth.
   * This method performs no input validation.
   *
   * @param dip of the fault plane in decimal degrees
   * @param width of the fault plane
   * @param zTor depth to the fault plane (positive down)
   */
  public static double hypocentralDepth(double dip, double width, double zTor) {
    return zTor + Math.sin(dip * Maths.TO_RADIANS) * width / 2.0;
  }

  /**
   * Project a vertically measured value (e.g. slip rate, depth) onto a dipping
   * surface.
   *
   * @param dip of surface (i.e. a fault)
   * @param value to project (vertical)
   */
  public static double projectVertical(double dip, double value) {
    return value / Math.sin(dip * Maths.TO_RADIANS);
  }

  /**
   * Compute the strike in decimal degrees of the supplied fault trace by
   * connecting the endpoints. This approach has been shown to be as accurate as
   * length-weighted angle averaging and is significantly faster.
   *
   * @param trace for which to compute strike
   * @return strike direction in the range [0°..360°)
   * @see #strikeRad(LocationList)
   */
  public static double strike(LocationList trace) {
    return strike(trace.first(), trace.last());
  }

  /**
   * Compute the strike in decimal degrees of the line connecting the supplied
   * locations.
   *
   * @param start location
   * @param end location
   * @return strike direction in the range [0°..360°)
   * @see #strikeRad(Location, Location)
   */
  public static double strike(Location start, Location end) {
    return Locations.azimuth(start, end);
  }

  /**
   * Compute the strike in radians of the supplied fault trace by connecting the
   * endpoints. This approach has been shown to be as accurate as
   * length-weighted angle averaging and is significantly faster.
   *
   * @param trace for which to compute strike
   * @return strike direction in the range [0..2π)
   * @see #strike(LocationList)
   */
  public static double strikeRad(LocationList trace) {
    return strikeRad(trace.first(), trace.last());
  }

  /**
   * Compute the strike in radians of the line connecting the supplied
   * locations.
   *
   * @param start location
   * @param end location
   * @return strike direction in the range [0..2π)
   * @see #strike(Location, Location)
   */
  public static double strikeRad(Location start, Location end) {
    return Locations.azimuthRad(start, end);
  }

  /**
   * Compute the dip direction in decimal degrees for the supplied fault trace
   * assuming the right-hand rule (strike + 90°).
   *
   * @param trace for which to compute dip direction
   * @return dip direction in the range [0°..360°)
   * @see #dipDirectionRad(LocationList)
   */
  public static double dipDirection(LocationList trace) {
    return dipDirection(strike(trace));
  }

  /**
   * Compute the dip direction in decimal degrees of the line connecting the
   * supplied locations assuming the right-hand rule (strike + 90°).
   *
   * @param start location
   * @param end location
   * @return dip direction in the range [0°..360°)
   * @see #dipDirectionRad(Location, Location)
   */
  public static double dipDirection(Location start, Location end) {
    return dipDirection(strike(start, end));
  }

  /**
   * Compute the dip direction in radians for the supplied fault trace assuming
   * the right-hand rule (strike + π/2).
   *
   * @param trace for which to compute dip direction
   * @return dip direction in the range [0..2π)
   * @see #dipDirection(LocationList)
   */
  public static double dipDirectionRad(LocationList trace) {
    return dipDirectionRad(strikeRad(trace));
  }

  /**
   * Compute the dip direction in radians of the line connecting the supplied
   * locations assuming the right-hand rule (strike + π/2).
   *
   * @param start location
   * @param end location
   * @return dip direction in the range [0..2π)
   * @see #dipDirection(Location, Location)
   */
  public static double dipDirectionRad(Location start, Location end) {
    return dipDirectionRad(strikeRad(start, end));
  }

  /**
   * Compute the dip direction for the supplied strike (strike + 90°).
   *
   * @param strike (in decimal degrees) for which to compute dip direction
   * @return dip direction in the range [0°..360°)
   */
  public static double dipDirection(double strike) {
    return (strike + 90.0) % 360.0;
  }

  /**
   * Compute the dip direction for the supplied strike (strike + π/2).
   *
   * @param strike (in radians) for which to compute dip direction
   * @return dip direction in the range [0°..360°)
   */
  public static double dipDirectionRad(double strike) {
    return (strike + Maths.PI_BY_2) % Maths.TWO_PI;
  }

}
