package gov.usgs.earthquake.nshmp;

import static gov.usgs.earthquake.nshmp.data.DoubleData.checkInRange;
import static gov.usgs.earthquake.nshmp.data.DoubleData.multiply;
import static gov.usgs.earthquake.nshmp.data.DoubleData.sum;
import static java.lang.Math.exp;
import static java.lang.Math.log;

import java.math.BigDecimal;
import java.math.MathContext;
import java.math.RoundingMode;
import java.util.Arrays;

import com.google.common.base.Converter;
import com.google.common.collect.Range;

/**
 * Miscellaneous math utilities.
 *
 * @author U.S. Geological Survey
 */
public final class Maths {

  private Maths() {}

  /** Constant for π/2. */
  public static final double PI_BY_2 = Math.PI / 2;

  /** Constant for 2π. */
  public static final double TWO_PI = 2.0 * Math.PI;

  /** Conversion multiplier for degrees to radians. */
  public static final double TO_RADIANS = Math.toRadians(1.0);

  /** Conversion multiplier for radians to degrees. */
  public static final double TO_DEGREES = Math.toDegrees(1.0);

  /**
   * The precomputed √<span style="border-top:1px solid; padding:0 0.1em;"
   * >2</span>.
   */
  public static final double SQRT_2 = Math.sqrt(2.0);

  /**
   * The precomputed √<span style="border-top:1px solid; padding:0 0.1em;"
   * >2π</span>.
   */
  public static final double SQRT_2PI = Math.sqrt(TWO_PI);

  /**
   * Standardized normal variate {@code ε = (x - μ) / σ}.
   *
   * @param m (μ) mean
   * @param s (σ) standard deviation
   */
  public static double epsilon(double m, double s, double x) {
    return (x - m) / s;
  }

  /**
   * Error function approximation of Abramowitz and Stegun, formula 7.1.26 in
   * the <em>Handbook of Mathematical Functions with Formulas, Graphs, and
   * Mathematical Tables</em>. Although the approximation is only valid for
   * {@code x ≥ 0}, because {@code erf(x)} is an odd function,
   * {@code erf(x) = −erf(−x)} and negative values are supported.
   */
  public static double erf(double x) {
    return x < 0.0 ? -erfBase(-x) : erfBase(x);
  }

  private static final double P = 0.3275911;
  private static final double A1 = 0.254829592;
  private static final double A2 = -0.284496736;
  private static final double A3 = 1.421413741;
  private static final double A4 = -1.453152027;
  private static final double A5 = 1.061405429;

  private static double erfBase(double x) {
    double t = 1 / (1 + P * x);
    double tsq = t * t;
    return 1 - (A1 * t +
        A2 * tsq +
        A3 * tsq * t +
        A4 * tsq * tsq +
        A5 * tsq * tsq * t) * Math.exp(-x * x);
  }

  /**
   * Same as {@link Math#hypot(double, double)} without regard to intermediate
   * under/over flow prior to internal call to {@link Math#sqrt(double)}.
   *
   * @param v1 first value
   * @param v2 second value
   * @see Math#hypot(double, double)
   */
  public static double hypot(double v1, double v2) {
    return Math.sqrt(v1 * v1 + v2 * v2);
  }

  /**
   * Three argument form of {@link Maths#hypot(double, double)}. Does not
   * protect against intermediate under/over flow prior to internal call to
   * {@link Math#sqrt(double)}.
   *
   * @param v1 first value
   * @param v2 second value
   * @param v3 second value
   * @see #hypot(double, double)
   */
  public static double hypot(double v1, double v2, double v3) {
    return Math.sqrt(v1 * v1 + v2 * v2 + v3 * v3);
  }

  /**
   * Variadic form of {@link Maths#hypot(double, double)}. Does not protect
   * against intermediate under/over flow prior to internal call to
   * {@link Math#sqrt(double)}.
   *
   * @param values to process
   * @see #hypot(double, double)
   */
  public static double hypot(double... values) {
    return Math.sqrt(Arrays.stream(values)
        .map(v -> v * v)
        .sum());
  }

  /**
   * Weighted square-root of the sum of squares.
   *
   * @param values to combine; commonly standard deviation components or models
   * @param weights to apply
   */
  public static double srssWeighted(double[] values, double[] weights) {
    double[] variances = Arrays.stream(values).map(v -> v * v).toArray();
    return Math.sqrt(sum(multiply(variances, weights)));
  }

  /**
   * Normal complementary cumulative distribution function.
   *
   * @param m (μ) mean
   * @param s (σ) standard deviation
   * @param x variate
   */
  public static double normalCcdf(double m, double s, double x) {
    return (1.0 + erf((m - x) / (s * SQRT_2))) * 0.5;
  }

  /**
   * Normal probability density function.
   *
   * @param m (μ) mean
   * @param s (σ) standard deviation
   * @param x variate
   */
  public static double normalPdf(double m, double s, double x) {
    return Math.exp((m - x) * (x - m) / (2 * s * s)) / (s * SQRT_2PI);
  }

  /**
   * Step function for which {@code f(x) = }&#123;
   * {@code 1 if x ≤ μ; 0 if x > μ }&#125;.
   *
   * @param m (μ) mean
   * @param x variate
   */
  public static double stepFunction(double m, double x) {
    return x < m ? 1.0 : 0.0;
  }

  /**
   * Round a double to a specified number of digits (according to
   * {@link RoundingMode#HALF_UP}. The first digit in a double value is the
   * first non-zero number, regardless of representation (e.g. scientific or
   * decimal).
   *
   * @param value to round
   * @param digits the number of digits in the result
   * @see #round(double, int)
   */
  public static double roundToDigits(double value, int digits) {
    return BigDecimal.valueOf(value)
        .round(new MathContext(digits, RoundingMode.HALF_UP))
        .doubleValue();
  }

  /**
   * Round a double to a specified number of decimal places according to
   * {@link RoundingMode#HALF_UP}. Internally this method uses the scaling and
   * rounding capabilities of {@link BigDecimal}. A negative {@code scale} will
   * round {@code value} to the specified number of places above the decimal.
   * The scale value is applied relative to zero. Therefore, if the supplied
   * value is represented in scientific notation the result may be zero (e.g.
   * {@code round(1.2345E-4, 2) = 0.0} as opposed to
   * {@code round(1.2345E-4, 6) = 1.23E-4}.
   *
   * @param value to round
   * @param scale the number of decimal places in the result
   * @see #roundToDigits(double, int)
   */
  public static double round(double value, int scale) {
    return round(value, scale, RoundingMode.HALF_UP);
  }

  /**
   * Round a double to a specified number of decimal places according to the
   * supplied {@link RoundingMode}. Internally this method uses the scaling and
   * rounding capabilities of {@link BigDecimal}. A negative {@code scale} will
   * round {@code value} to the specified number of places above the decimal.
   *
   * @param value to round
   * @param scale the number of decimal places in the result
   */
  public static double round(double value, int scale, RoundingMode mode) {
    return BigDecimal.valueOf(value).setScale(scale, mode).doubleValue();
  }

  /**
   * Return a converter between decimal values and percentages. The forward
   * operation of this converter <i>may</i> be lossy in as much as the supplied
   * decimal value will be rounded to the specified scale after conversion to a
   * percentage. For instance:
   *
   * <p>{@code decimalToProbabilityConverter(2).convert(0.23456) = 23.46}
   *
   * <p>Reverse operations do not affect precision at this time.
   *
   * @param percentScale the number of decimal places in percentages output by
   *        this converter
   */
  public static Converter<Double, Double> decimalToProbabilityConverter(int percentScale) {
    return new DecimalToProbabilityConverter(percentScale);
  }

  private static class DecimalToProbabilityConverter extends Converter<Double, Double> {

    private final int percentScale;

    DecimalToProbabilityConverter(int percentScale) {
      this.percentScale = percentScale;
    }

    @Override
    protected Double doBackward(Double percent) {
      return percent / 100.0;
    }

    @Override
    protected Double doForward(Double decimal) {
      return round(decimal * 100.0, percentScale);
    }
  }

  /**
   * Convert an annual rate of occurrence to a Poisson probability of occurrence
   * over the specified time period.
   *
   * @param rate (annual) of occurrence
   * @param timespan of interest
   * @return the Poisson probability of occurrence over the specified
   *         {@code timespan}
   */
  public static double rateToProbability(double rate, double timespan) {
    return 1 - exp(-rate * timespan);
  }

  /**
   * Convert a Poisson probability of occurrence of some event over a specified
   * time period to the equivalent annual rate.
   *
   * @param P the Poisson probability
   * @param timespan of interest
   * @return the annnual rate of occurrence
   */
  public static double probabilityToRate(double P, double timespan) {
    return -log(1 - P) / timespan;
  }

  /**
   * Return a converter between annual rate and Poisson probability over a
   * 1-year time span. The converter expects probabilities to be represented as
   * fractions of 1.0.
   */
  public static Converter<Double, Double> annualRateToProbabilityConverter() {
    return new AnnRateToPoissProbConverter(1.0);
  }

  /**
   * Return a converter between annual rate and Poisson probability over the
   * specified time span. The converter expects probabilities to be represented
   * as fractions of 1.0.
   */
  public static Converter<Double, Double> annualRateToProbabilityConverter(double timespan) {
    return new AnnRateToPoissProbConverter(timespan);
  }

  /**
   * Supported timespans for Poisson probabilities: {@code [1..10000] years}.
   */
  public static final Range<Double> TIMESPAN_RANGE = Range.closed(1.0, 10000.0);

  private static final class AnnRateToPoissProbConverter extends Converter<Double, Double> {

    private final double timespan;

    AnnRateToPoissProbConverter(double timespan) {
      checkInRange(TIMESPAN_RANGE, "Timespan", timespan);
      this.timespan = timespan;
    }

    @Override
    protected Double doForward(Double rate) {
      return rateToProbability(rate, timespan);
    }

    @Override
    protected Double doBackward(Double prob) {
      return probabilityToRate(prob, timespan);
    }
  }

}
