package gov.usgs.earthquake.nshmp.calc;

import static com.google.common.util.concurrent.Futures.allAsList;
import static com.google.common.util.concurrent.Futures.immediateFuture;
import static com.google.common.util.concurrent.Futures.transform;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executor;

import com.google.common.util.concurrent.ListenableFuture;

import gov.usgs.earthquake.nshmp.calc.Transforms.ClusterCurveConsolidator;
import gov.usgs.earthquake.nshmp.calc.Transforms.ClusterToCurves;
import gov.usgs.earthquake.nshmp.calc.Transforms.CurveConsolidator;
import gov.usgs.earthquake.nshmp.calc.Transforms.CurveSetConsolidator;
import gov.usgs.earthquake.nshmp.calc.Transforms.ParallelSystemToCurves;
import gov.usgs.earthquake.nshmp.calc.Transforms.SourceToCurves;
import gov.usgs.earthquake.nshmp.calc.Transforms.SystemToCurves;
import gov.usgs.earthquake.nshmp.model.ClusterRuptureSet;
import gov.usgs.earthquake.nshmp.model.ClusterSource;
import gov.usgs.earthquake.nshmp.model.GmmTree;
import gov.usgs.earthquake.nshmp.model.HazardModel;
import gov.usgs.earthquake.nshmp.model.RuptureSet;
import gov.usgs.earthquake.nshmp.model.Source;
import gov.usgs.earthquake.nshmp.model.SystemRuptureSet;

/**
 * Static utility methods for performing single-threaded and asynchronous hazard
 * calculations via functions in {@link Transforms}.
 *
 * @author U.S. Geological Survey
 * @see Transforms
 * @see AsyncList
 */
final class CalcFactory {

  private CalcFactory() {}

  /* Compute hazard curves for a RuptureSet. */
  static HazardCurveSet sourcesToCurves(
      RuptureSet<? extends Source> ruptures,
      Site site,
      CalcConfig config) {

    GmmTree gmmTree = getGmmTree(ruptures, site);
    SourceToCurves sourceToCurves = new SourceToCurves(gmmTree, site, config);
    List<HazardCurves> curvesList = new ArrayList<>();
    for (Source source : ruptures.iterableForLocation(site.location())) {
      curvesList.add(sourceToCurves.apply(source));
    }
    CurveConsolidator cc = new CurveConsolidator(ruptures, gmmTree, config);
    return cc.apply(curvesList);
  }

  /* Asynchronously compute hazard curves for a RuptureSet. */
  static ListenableFuture<HazardCurveSet> sourcesToCurves(
      RuptureSet<? extends Source> ruptures,
      Site site,
      CalcConfig config,
      Executor ex) {

    GmmTree gmmTree = getGmmTree(ruptures, site);
    SourceToCurves sourceToCurves = new SourceToCurves(gmmTree, site, config);
    AsyncList<HazardCurves> curvesList = AsyncList.create();
    for (Source source : ruptures.iterableForLocation(site.location())) {
      ListenableFuture<HazardCurves> curves = transform(
          immediateFuture(source),
          sourceToCurves::apply,
          ex);
      curvesList.add(curves);
    }
    CurveConsolidator cc = new CurveConsolidator(ruptures, gmmTree, config);
    return transform(allAsList(curvesList), cc::apply, ex);
  }

  /* Compute hazard curves for a SystemRuptureSet. */
  static HazardCurveSet systemToCurves(
      SystemRuptureSet ruptures,
      Site site,
      CalcConfig config) {

    GmmTree gmmTree = getGmmTree(ruptures, site);
    return new SystemToCurves(gmmTree, site, config).apply(ruptures);
  }

  /* Asynchronously compute hazard curves for a SystemRuptureSet. */
  static ListenableFuture<HazardCurveSet> systemToCurves(
      SystemRuptureSet ruptures,
      Site site,
      CalcConfig config,
      final Executor ex) {

    GmmTree gmmTree = getGmmTree(ruptures, site);
    ParallelSystemToCurves psc = new ParallelSystemToCurves(gmmTree, site, config, ex);
    return transform(immediateFuture(ruptures), psc::apply, ex);
  }

  /* Compute hazard curves for a ClusterRuptureSet. */
  static HazardCurveSet clustersToCurves(
      ClusterRuptureSet ruptures,
      Site site,
      CalcConfig config) {

    GmmTree gmmTree = getGmmTree(ruptures, site);
    ClusterToCurves clusterToCurves = new ClusterToCurves(gmmTree, site, config);
    List<ClusterCurves> curvesList = new ArrayList<>();
    for (ClusterSource source : ruptures.iterableForLocation(site.location())) {
      curvesList.add(clusterToCurves.apply(source));
    }
    ClusterCurveConsolidator consolidateFn = new ClusterCurveConsolidator(
        ruptures, gmmTree, config);
    return consolidateFn.apply(curvesList);
  }

  /* Asynchronously compute hazard curves for a ClusterRuptureSet. */
  static ListenableFuture<HazardCurveSet> clustersToCurves(
      ClusterRuptureSet ruptures,
      Site site,
      CalcConfig config,
      Executor ex) {

    GmmTree gmmTree = getGmmTree(ruptures, site);
    ClusterToCurves clusterToCurves = new ClusterToCurves(gmmTree, site, config);
    AsyncList<ClusterCurves> curvesList = AsyncList.create();
    for (ClusterSource source : ruptures.iterableForLocation(site.location())) {
      ListenableFuture<ClusterCurves> curves = transform(
          immediateFuture(source),
          clusterToCurves::apply,
          ex);
      curvesList.add(curves);
    }
    ClusterCurveConsolidator ccc = new ClusterCurveConsolidator(ruptures, gmmTree, config);
    return transform(allAsList(curvesList), ccc::apply, ex);
  }

  /* Reduce hazard curves to a result. */
  static Hazard toHazardResult(
      HazardModel model,
      Site site,
      CalcConfig config,
      List<HazardCurveSet> curveSets) {

    return new CurveSetConsolidator(model, config, site).apply(curveSets);
  }

  /* Asynchronously reduce hazard curves to a result. */
  static Hazard toHazardResult(
      HazardModel model,
      Site site,
      CalcConfig config,
      AsyncList<HazardCurveSet> curveSets,
      Executor ex) throws InterruptedException, ExecutionException {

    CurveSetConsolidator csc = new CurveSetConsolidator(model, config, site);
    return transform(allAsList(curveSets), csc::apply, ex).get();
  }

  private static GmmTree getGmmTree(RuptureSet<? extends Source> ruptures, Site site) {
    Optional<GmmTree> siteGmmTree = site.gmmTree(
        ruptures.setting(),
        ruptures.type());
    return siteGmmTree.isPresent()
        ? siteGmmTree.orElseThrow()
        : ruptures.gmmTree();
  }

}
