package gov.usgs.earthquake.nshmp.calc;

import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.data.XySequence.copyOf;

import java.util.EnumMap;
import java.util.Map;
import java.util.TreeMap;

import gov.usgs.earthquake.nshmp.data.MutableXySequence;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.gmm.Imt;

/**
 * Container class for the combined hazard curves derived from the individual
 * {@code Source}s in a {@code ClusterSource}, one for each
 * {@code GroundMotionModel} used. The curves will have been scaled by source
 * and rupture weights, but not by {@code GroundMotionModel} weights.
 *
 * @author U.S. Geological Survey
 */
final class ClusterCurves {

  final ClusterGroundMotions clusterGroundMotions;
  final Map<Imt, Map<Gmm, XySequence>> curveMap;
  final Map<Imt, TreeMap<Double, MutableXySequence>> magMap;

  private ClusterCurves(Builder builder) {
    this.clusterGroundMotions = builder.clusterGroundMotions;
    this.curveMap = builder.curveMap;
    this.magMap = builder.magCurveMap;
  }

  static Builder builder(ClusterGroundMotions clusterGroundMotions) {
    return new Builder(clusterGroundMotions);
  }

  static class Builder {

    private static final String ID = "ClusterCurves.Builder";
    private boolean built = false;

    private final ClusterGroundMotions clusterGroundMotions;
    private final Map<Imt, Map<Gmm, XySequence>> curveMap;
    private final Map<Imt, TreeMap<Double, MutableXySequence>> magCurveMap;

    private Builder(ClusterGroundMotions clusterGroundMotions) {
      this.clusterGroundMotions = clusterGroundMotions;
      // look at first HazardGM to determine curve table dimensions
      GroundMotions model = clusterGroundMotions.get(0);
      curveMap = new EnumMap<>(Imt.class);
      magCurveMap = new EnumMap<>(Imt.class);
      for (Imt imt : model.gmMap.keySet()) {
        Map<Gmm, XySequence> gmmMap = new EnumMap<>(Gmm.class);
        curveMap.put(imt, gmmMap);
        TreeMap<Double, MutableXySequence> magMap = HazardCurves.initMagMap();
        magCurveMap.put(imt, magMap);
      }
    }

    /* Add an immutable copy of the supplied curve. */
    Builder addGmmCurve(Imt imt, Gmm gmm, XySequence curve) {
      curveMap.get(imt).put(gmm, copyOf(curve));
      return this;
    }

    /* Put an immutable copy of the supplied curve. */
    Builder addMagCurve(Imt imt, double Mw, XySequence curve) {
      TreeMap<Double, MutableXySequence> magMap = magCurveMap.get(imt);
      Double k = magMap.floorKey(Mw);
      MutableXySequence v = magMap.get(k);
      if (v == null) {
        magMap.put(k, MutableXySequence.copyOf(curve));
      } else {
        v.add(curve);
      }
      return this;
    }

    ClusterCurves build() {
      checkState(!built, "This %s instance has already been used", ID);
      built = true;
      return new ClusterCurves(this);
    }
  }
}
