package gov.usgs.earthquake.nshmp.calc;

import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.model.SourceType.FAULT_SYSTEM;
import static java.util.Comparator.comparingDouble;
import static java.util.stream.Collectors.toList;

import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.calc.DisaggExport.ContributionFilter;
import gov.usgs.earthquake.nshmp.data.DoubleData;
import gov.usgs.earthquake.nshmp.data.IntervalArray;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.model.ClusterSource;
import gov.usgs.earthquake.nshmp.model.Rupture;
import gov.usgs.earthquake.nshmp.model.RuptureSet;
import gov.usgs.earthquake.nshmp.model.Source;
import gov.usgs.earthquake.nshmp.model.SourceType;

/**
 * Skeletal class and implementations used for contribution to hazard in a
 * disaggregation. A contribution may be keyed to a {@code RuptureSet}, an
 * individual {@code Source}, or some other representation, e.g. optimized
 * (table-based) {@code GridSource}s or {@code SystemSource} fault sections.
 *
 * @author U.S. Geological Survey
 */
abstract class DisaggContributor {

  final double rate;
  final double residual;
  final double rScaled;
  final double mScaled;
  final double εScaled;

  DisaggContributor(Builder b) {
    this.rate = b.rate;
    this.residual = b.residual;
    this.rScaled = b.rScaled;
    this.mScaled = b.mScaled;
    this.εScaled = b.εScaled;
  }

  double total() {
    return rate + residual;
  }

  /* Convenience array for use when combining children. */
  private double[] values() {
    return new double[] { rate, residual, rScaled, mScaled, εScaled };
  }

  abstract List<JsonContributor> toJson(ContributionFilter filter);

  abstract static class Builder {

    double rate;
    double residual;
    double rScaled;
    double mScaled;
    double εScaled;

    /*
     * Multiple ways to add contributing disagg values are provided to limit
     * value copying to the extent possible.
     */

    Builder addValues(
        double rate,
        double residual,
        double rScaled,
        double mScaled,
        double εScaled) {

      this.rate += rate;
      this.residual += residual;
      this.rScaled += rScaled;
      this.mScaled += mScaled;
      this.εScaled += εScaled;
      return this;
    }

    Builder addValues(double[] values) {
      this.rate += values[0];
      this.residual += values[1];
      this.rScaled += values[2];
      this.mScaled += values[3];
      this.εScaled += values[4];
      return this;
    }

    Builder addValues(DisaggContributor contrib) {
      this.rate += contrib.rate;
      this.residual += contrib.residual;
      this.rScaled += contrib.rScaled;
      this.mScaled += contrib.mScaled;
      this.εScaled += contrib.εScaled;
      return this;
    }

    /*
     * Optional, single-call operation. May only be called once with subsequent
     * calls to addChild() throwing an IllegalStateException.
     */
    Builder multiply(double scale) {
      throw new UnsupportedOperationException();
    }

    /*
     * Optional operation; some implementations may throw an
     * UnsupportedOperationException.
     */
    Builder addChild(DisaggContributor.Builder contributor) {
      throw new UnsupportedOperationException();
    }

    /*
     * Build implementations should compile rate and statistic values for this
     * from any children instead of allowing them to be add()-ed directly.
     */
    abstract DisaggContributor build();

  }

  /* Implementations. */

  /* RuptureSet contributor with references to child contributors. */
  static final class RuptureSetContributor extends DisaggContributor {

    final RuptureSet<? extends Source> ruptures;
    final List<DisaggContributor> children;

    private RuptureSetContributor(Builder b) {
      super(b);
      this.ruptures = b.ruptures;
      this.children = b.children;
    }

    @Override
    List<JsonContributor> toJson(ContributionFilter filter) {
      ArrayList<JsonContributor> jsonList = new ArrayList<>();
      double total = total();
      double rBar = rScaled / total;
      double mBar = mScaled / total;
      double εBar = εScaled / total;
      JsonContributor jc = JsonContributor.createMulti(
          ruptures.name(),
          ruptures.type(),
          Maths.round(filter.toPercent(total()), 2),
          rBar,
          mBar,
          εBar);
      jsonList.add(jc);
      for (DisaggContributor child : children) {
        if (filter.test(child)) {
          jsonList.addAll(child.toJson(filter));
          continue;
        }
        break;
      }
      return jsonList;
    }

    static final class Builder extends DisaggContributor.Builder {

      RuptureSet<? extends Source> ruptures;
      List<DisaggContributor.Builder> childBuilders = new ArrayList<>();
      List<DisaggContributor> children;

      Builder ruptures(RuptureSet<? extends Source> ruptures) {
        this.ruptures = ruptures;
        return this;
      }

      @Override
      Builder addChild(DisaggContributor.Builder contributor) {
        childBuilders.add(contributor);
        return this;
      }

      @Override
      RuptureSetContributor build() {
        children = buildAndSort(childBuilders);
        double[] childValues = sumChildValues(children);
        addValues(childValues);
        return new RuptureSetContributor(this);
      }
    }
  }

  /*
   * Builders should have better state checking. e.g. once the source has been
   * set for a source contributor, it may not be set again. Likewise, add() can
   * not be called unless source is set.
   */

  /* Generic Source contributor. Location and azimuth are site-specific. */
  static class SourceContributor extends DisaggContributor {

    final Source source;
    final Location location;
    final double azimuth;

    private SourceContributor(Builder b) {
      super(b);
      this.source = b.source;
      this.location = b.location;
      this.azimuth = b.azimuth;
    }

    @Override
    List<JsonContributor> toJson(ContributionFilter filter) {
      double total = total();
      double rBar = rScaled / total;
      double mBar = mScaled / total;
      double εBar = εScaled / total;
      JsonContributor jc = JsonContributor.createSingle(
          source.name(),
          source.type(),
          Maths.round(filter.toPercent(total()), 2),
          -1,
          rBar,
          mBar,
          εBar,
          azimuth,
          location.latitude,
          location.longitude);
      return List.of(jc);
    }

    static final class Builder extends DisaggContributor.Builder {

      Source source;
      Location location;
      double azimuth;
      boolean scaled = false;

      Builder source(Source source, Location location, double azimuth) {
        this.source = source;
        this.location = location;
        this.azimuth = azimuth;
        return this;
      }

      @Override
      Builder addValues(DisaggContributor contrib) {
        checkState(!scaled);
        super.addValues(contrib);
        return this;
      }

      @Override
      Builder multiply(double scale) {
        checkState(!scaled);
        rate *= scale;
        residual *= scale;
        rScaled *= scale;
        mScaled *= scale;
        εScaled *= scale;
        scaled = true;
        return this;
      }

      @Override
      SourceContributor build() {
        return new SourceContributor(this);
      }
    }
  }

  /* ClusterSource contributor. */
  static final class ClusterContributor extends DisaggContributor {

    final ClusterSource cluster;
    final List<DisaggContributor> faults;
    final Location location;
    final double azimuth;

    ClusterContributor(Builder b) {
      super(b);
      this.cluster = b.cluster;
      this.faults = b.faults;
      this.location = b.location;
      this.azimuth = b.azimuth;
    }

    @Override
    List<JsonContributor> toJson(ContributionFilter filter) {
      double total = total();
      double rBar = rScaled / total;
      double mBar = mScaled / total;
      double εBar = εScaled / total;
      JsonContributor jc = JsonContributor.createSingle(
          cluster.name(),
          cluster.type(),
          Maths.round(filter.toPercent(total()), 2),
          -2,
          rBar,
          mBar,
          εBar,
          azimuth,
          location.latitude,
          location.longitude);
      return List.of(jc);
    }

    static final class Builder extends DisaggContributor.Builder {

      ClusterSource cluster;
      List<DisaggContributor.Builder> faultBuilders = new ArrayList<>();
      List<DisaggContributor> faults;
      Location location;
      double azimuth;

      Builder cluster(ClusterSource cluster, Location location, double azimuth) {
        this.cluster = cluster;
        this.location = location;
        this.azimuth = azimuth;
        return this;
      }

      @Override
      Builder multiply(double scale) {
        for (DisaggContributor.Builder fault : faultBuilders) {
          fault.multiply(scale);
        }
        return this;
      }

      @Override
      Builder addChild(DisaggContributor.Builder contributor) {
        faultBuilders.add(contributor);
        return this;
      }

      @Override
      ClusterContributor build() {
        faults = buildAndSort(faultBuilders);
        double[] childValues = sumChildValues(faults);
        addValues(childValues);
        return new ClusterContributor(this);
      }
    }
  }

  static final class SystemContributor extends DisaggContributor {

    final SectionSource section;
    final Location location;
    final double azimuth;
    final IntervalArray mfd;

    SystemContributor(Builder b) {
      super(b);
      this.section = b.section;
      this.location = b.location;
      this.azimuth = b.azimuth;
      this.mfd = b.mfd;
    }

    @Override
    List<JsonContributor> toJson(ContributionFilter filter) {
      double total = total();
      double rBar = rScaled / total;
      double mBar = mScaled / total;
      double εBar = εScaled / total;
      JsonContributor jc = JsonContributor.createSingle(
          section.name(),
          section.type(),
          Maths.round(filter.toPercent(total()), 2),
          -3,
          rBar,
          mBar,
          εBar,
          azimuth,
          location.latitude,
          location.longitude);
      return List.of(jc);
    }

    static final class Builder extends DisaggContributor.Builder {

      SectionSource section;
      Location location;
      double azimuth;
      IntervalArray.Builder mfdBuilder;
      IntervalArray mfd;

      Builder section(
          SectionSource section,
          Location location,
          double azimuth,
          IntervalArray.Builder mfdBuilder) {

        this.section = section;
        this.location = location;
        this.azimuth = azimuth;
        this.mfdBuilder = mfdBuilder;
        return this;
      }

      /*
       * The mfd data could be backed out on every call to add, but this likely
       * requires a bin index lookup to be repeated across all relevant Gmms.
       */
      Builder addToMfd(int index, double rate) {
        mfdBuilder.add(index, rate);
        return this;
      }

      Builder addMfd(IntervalArray mfd) {
        mfdBuilder.add(mfd);
        return this;
      }

      @Override
      DisaggContributor build() {
        mfd = mfdBuilder.build();
        return new SystemContributor(this);
      }
    }
  }

  /*
   * Custom source that encapsulates the fact that relevant data in the
   * disaggregation of SystemRuptureSet is associated with individual fault
   * sections rather than sources/ruptures.
   */
  static final class SectionSource extends AbstractList<Rupture> implements Source {

    final int index;
    final String name;

    SectionSource(int index, String name) {
      this.index = index;
      this.name = name;
    }

    @Override
    public String name() {
      return name;
    }

    @Override
    public int size() {
      // perhaps this should return the number of participating ruptures
      return 1;
    }

    @Override
    public Rupture get(int index) {
      throw new UnsupportedOperationException();
    }

    @Override
    public int id() {
      return index;
    }

    @Override
    public SourceType type() {
      return FAULT_SYSTEM;
    }

    @Override
    public Location location(Location site) {
      return null; // unused
    }

    @Override
    public List<Mfd> mfds() {
      return null;
      // this needs to return something, not sure what yet
    }

    @Override
    public Iterator<Rupture> iterator() {
      throw new UnsupportedOperationException();
    }

    /*
     * With the switch to Source implementing List, when we add a source to a
     * Map, hashCode() is called on the contents of the source. System sources
     * for disagg have special handling and so we revert to the system hash code
     * and equals. See also SystemRuptureSet.
     */

    @Override
    public int hashCode() {
      return System.identityHashCode(this);
    }

    @Override
    public boolean equals(Object obj) {
      return (this == obj);
    }

  }

  /* Wrapper of contributor data suitable for JSON serialization. */
  static final class JsonContributor {
    String name;
    String source;
    JsonContributorType type;
    Double contribution;
    Integer id;
    Double r;
    Double m;
    Double ε;
    Double azimuth;
    Double latitude;
    Double longitude;

    /* Use for RuptureSets. */
    static JsonContributor createMulti(
        String name,
        SourceType source,
        double contribution,
        double r,
        double m,
        double ε) {

      JsonContributor jc = new JsonContributor();
      jc.name = name;
      jc.source = source.toString();
      jc.type = JsonContributorType.SET;
      jc.contribution = contribution;
      jc.r = r;
      jc.m = m;
      jc.ε = ε;
      return jc;
    }

    /* Use for more detailed individual sources. */
    static JsonContributor createSingle(
        String name,
        SourceType source,
        double contribution,
        int id,
        double r,
        double m,
        double ε,
        double azimuth,
        double latitude,
        double longitude) {

      JsonContributor jc = new JsonContributor();
      jc.name = name;
      jc.source = source.toString();
      jc.type = JsonContributorType.SINGLE;
      jc.contribution = contribution;
      jc.id = id;
      jc.r = r;
      jc.m = m;
      jc.ε = ε;
      jc.azimuth = azimuth;
      jc.latitude = latitude;
      jc.longitude = longitude;
      return jc;
    }

    @Override
    public String toString() {
      StringBuilder sb = new StringBuilder(type == JsonContributorType.SINGLE ? "  " : "");
      sb.append(name).append(" ").append(source).append(" ").append(contribution);
      return sb.toString();
    }
  }

  private static enum JsonContributorType {
    SET,
    SINGLE;
  }

  /* Convert a builder list to sorted contributor list. */
  private static List<DisaggContributor> buildAndSort(
      Collection<DisaggContributor.Builder> builders) {
    return builders.stream()
        .map(Builder::build)
        .sorted(comparingDouble(DisaggContributor::total).reversed())
        .collect(toList());
  }

  private static double[] sumChildValues(Collection<DisaggContributor> contributors) {
    double[] sum = new double[5];
    for (DisaggContributor contributor : contributors) {
      DoubleData.add(sum, contributor.values());
    }
    return sum;
  }
}
