package gov.usgs.earthquake.nshmp.calc;

import static java.lang.Math.exp;

import java.util.ArrayList;
import java.util.List;
import java.util.function.Predicate;
import java.util.stream.Collectors;

import com.google.gson.annotations.SerializedName;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.calc.CalcConfig.Disagg.Bins;
import gov.usgs.earthquake.nshmp.calc.DisaggContributor.JsonContributor;
import gov.usgs.earthquake.nshmp.data.DoubleData;
import gov.usgs.earthquake.nshmp.gmm.Imt;

/**
 * Disaggregation dataset exporter. This class handles String/Text and JSON
 * output of DisaggDatasets. JSON output is supported through the serialization
 * of custom data containers. String output is supported through the
 * {@code toString()} methods of those objects.
 *
 * @author U.S. Geological Survey
 */
final class DisaggExport {

  final transient DisaggDataset ddTotal;
  final transient DisaggDataset dd;
  final transient DisaggConfig dc;

  @SerializedName("component")
  final String id;
  final List<RmBin> data;
  final List<SummaryElement> summary;
  final List<JsonContributor> sources;

  /*
   * All component DisaggDatasets require data from the final total
   * DisaggDataset to correctly calculate contributions and represent summary
   * data that is not specific to the component dataset.
   */

  DisaggExport(
      DisaggDataset ddTotal,
      DisaggDataset dd,
      DisaggConfig dc,
      String id,
      boolean εDataFlag) {

    this.ddTotal = ddTotal;
    this.dd = dd;
    this.dc = dc;
    this.id = id;

    summary = createSummaryElements(ddTotal, dd, dc);
    sources = createJsonContributorList(ddTotal, dd, dc.contributorLimit);
    data = εDataFlag ? createDistanceMagnitudeData(ddTotal, dd) : null;
  }

  /*
   * Disaggregation data.
   *
   * Create a container of distance-magnitude-epsilon data bins. This object
   * serializes to a JSON array of distance-magnitude (r-m) bins, each of which
   * contans an array of non-zero ε bins. The r-m bins are sorted ascending in
   * distance but descending in magnitude to facilitate rendering order in a
   * disaggregation 3D histogram.
   */
  private static List<RmBin> createDistanceMagnitudeData(
      DisaggDataset ddTotal,
      DisaggDataset dd) {

    List<RmBin> rmBins = new ArrayList<>();
    List<Double> distances = dd.rmε.rows();
    List<Double> magnitudes = dd.rmε.columns();
    double toPercent = percentScalar(ddTotal);

    // iterate distances ascending, magnitudes descending
    for (int ri = 0; ri < distances.size(); ri++) {
      double r = distances.get(ri);
      for (int mi = magnitudes.size() - 1; mi >= 0; mi--) {
        double rmBinWeight = dd.rmWeights.get(ri, mi);
        // skip empty bins
        if (rmBinWeight == 0.0) {
          continue;
        }
        double m = magnitudes.get(mi);
        double rBar = dd.rmrScaled.get(ri, mi) / rmBinWeight;
        double mBar = dd.rmmScaled.get(ri, mi) / rmBinWeight;
        double εBar = dd.rmεScaled.get(ri, mi) / rmBinWeight;
        double[] εValues = dd.rmε.column(ri, mi).yValues().toArray();
        DoubleData.multiply(toPercent, εValues);
        List<εData> εDataList = new ArrayList<>();
        for (int i = 0; i < εValues.length; i++) {
          double εValue = εValues[i];
          if (εValue <= 0.0) {
            continue;
          }
          εDataList.add(new εData(i, εValue));
        }
        RmBin rmBin = new RmBin(
            r, m,
            round(rBar, 2),
            round(mBar, 2),
            round(εBar, 2),
            List.copyOf(εDataList));
        rmBins.add(rmBin);
      }
    }
    return List.copyOf(rmBins);
  }

  /*
   * Data container for a single distance-magnitude (r-m) bin.
   */
  @SuppressWarnings("unused")
  private static final class RmBin {

    final double r;
    final double m;
    final double rBar;
    final double mBar;
    final double εBar;
    final List<εData> εdata;

    private RmBin(
        double r,
        double m,
        double rBar,
        double mBar,
        double εBar,
        List<εData> εdata) {

      this.r = r;
      this.m = m;
      this.rBar = rBar;
      this.mBar = mBar;
      this.εBar = εBar;
      this.εdata = εdata;
    }
  }

  /*
   * Single ε value bin.
   */
  @SuppressWarnings("unused")
  private static final class εData {

    final int εbin;
    final double value;

    private εData(int εbin, double value) {
      this.εbin = εbin;
      this.value = value;
    }
  }

  private static final double TRACE_LIMIT = 0.01;
  private static final int RME_ROUNDING = 2;
  private static final String DISCRETIZATION_FMT = "min = %.1f, max = %.1f, Δ = %.1f";

  /*
   * Disaggregation summary information for a DisaggDataset.
   */
  private static List<SummaryElement> createSummaryElements(
      DisaggDataset total,
      DisaggDataset dd,
      DisaggConfig dc) {

    double toPercent = percentScalar(total);
    int ri, mi, εi;

    /* targets */
    double recoveredRate = total.binned + total.residual;
    double recoveredReturnPeriod = 1.0 / recoveredRate;

    /* totals */
    double ddTotal = dd.binned * toPercent;
    double ddResidual = dd.residual * toPercent;
    double ddTrace = traceContribution(dd);

    /* modes: largest r-m bin */
    int[] rmIndex = dd.rmWeights.maxIndex();
    ri = rmIndex[0];
    mi = rmIndex[1];
    double rmBinWeight = dd.rmWeights.get(ri, mi);
    double rmrMode = dd.rmrScaled.get(ri, mi) / rmBinWeight;
    double rmmMode = dd.rmmScaled.get(ri, mi) / rmBinWeight;
    double rmεMode = dd.rmεScaled.get(ri, mi) / rmBinWeight;
    double rmModeContrib = rmBinWeight * toPercent;

    /* modes: largest ε bin */
    int[] rmεIndex = dd.rmε.maxIndex();
    ri = rmεIndex[0];
    mi = rmεIndex[1];
    εi = rmεIndex[2];
    double εBinWeight = dd.rmε.get(ri, mi, εi);
    double εrMode = dd.rScaled.get(ri, mi, εi) / εBinWeight;
    double εmMode = dd.mScaled.get(ri, mi, εi) / εBinWeight;
    double εεMode = dd.εScaled.get(ri, mi, εi) / εBinWeight;
    double εModeContrib = εBinWeight * toPercent;

    /* discretization */
    Bins b = dc.bins;
    String rDiscr = String.format(DISCRETIZATION_FMT, b.rMin, b.rMax, b.Δr);
    String mDiscr = String.format(DISCRETIZATION_FMT, b.mMin, b.mMax, b.Δm);
    String εDiscr = String.format(DISCRETIZATION_FMT, b.εMin, b.εMax, b.Δε);

    /* custom imt formatting */
    Imt imt = dc.imt;
    String imlLabel = (imt.isSA() ? imt.period() + " s SA" : imt.name()) + " ground motion";

    List<SummaryElement> summaryElements = List.of(

        new SummaryElement("Disaggregation targets", true, List.of(
            new SummaryItem("Return period", dc.returnPeriod, "yrs"),
            new SummaryItem("Exceedance rate", dc.rate, "yr⁻¹"),
            new SummaryItem(imlLabel, exp(dc.iml), dc.imt.units()))),

        new SummaryElement("Recovered targets", true, List.of(
            new SummaryItem("Return period", recoveredReturnPeriod, "yrs"),
            new SummaryItem("Exceedance rate", recoveredRate, "yr⁻¹"))),

        new SummaryElement("Totals", true, List.of(
            new SummaryItem("Binned", round(ddTotal, RME_ROUNDING), "%"),
            new SummaryItem("Residual", round(ddResidual, RME_ROUNDING), "%"),
            new SummaryItem("Trace", round(ddTrace, RME_ROUNDING), "%"))),

        new SummaryElement("Mean (over all sources)", true, List.of(
            new SummaryItem("m", round(dd.mBar, RME_ROUNDING), null),
            new SummaryItem("r", round(dd.rBar, RME_ROUNDING), "km"),
            new SummaryItem("ε₀", round(dd.εBar, RME_ROUNDING), "σ"))),

        new SummaryElement("Mode (largest m-r bin)", true, List.of(
            new SummaryItem("m", round(rmmMode, RME_ROUNDING), null),
            new SummaryItem("r", round(rmrMode, RME_ROUNDING), "km"),
            new SummaryItem("ε₀", round(rmεMode, RME_ROUNDING), "σ"),
            new SummaryItem("Contribution", round(rmModeContrib, RME_ROUNDING), "%"))),

        new SummaryElement("Mode (largest m-r-ε₀ bin)", true, List.of(
            new SummaryItem("m", round(εmMode, RME_ROUNDING), null),
            new SummaryItem("r", round(εrMode, RME_ROUNDING), "km"),
            new SummaryItem("ε₀", round(εεMode, RME_ROUNDING), "σ"),
            new SummaryItem("Contribution", round(εModeContrib, RME_ROUNDING), "%"))),

        new SummaryElement("Discretization", true, List.of(
            new SummaryItem("r", rDiscr, "km"),
            new SummaryItem("m", mDiscr, null),
            new SummaryItem("ε", εDiscr, "σ"))),

        new SummaryElement("Epsilon keys", true, toSummaryItems(dc.εBins)));

    return summaryElements;
  }

  private static double traceContribution(DisaggDataset dd) {
    double toPercent = percentScalar(dd);
    double traceContribution = 0.0;
    for (int ri = 0; ri < dd.rmε.rows().size(); ri++) {
      for (int mi = 0; mi < dd.rmε.columns().size(); mi++) {
        if (dd.rmWeights.get(ri, mi) <= 0.0) {
          continue;
        }
        for (double εValue : dd.rmε.column(ri, mi).yValues().toArray()) {
          double rmεBinContribution = εValue * toPercent;
          if (rmεBinContribution < TRACE_LIMIT) {
            traceContribution += rmεBinContribution;
          }
        }
      }
    }
    return traceContribution;
  }

  private static double percentScalar(DisaggDataset dd) {
    return 100.0 / (dd.binned + dd.residual);
  }

  /*
   * When all sources are out of range for a dataset (i.e. residual ≠ 0.0) the
   * rBar, mBar, and εBar fields are NaN as a result of dividing-by-zero rates,
   * as are other summary statistics. For the purpose of presentation and JSON
   * serialization we convert to a Double which may be null that can be handled
   * during serialization and in toString().
   */
  private static Double round(double value, int scale) {
    return Double.isNaN(value) ? null : Maths.round(value, scale);
  }

  @SuppressWarnings("unused")
  private static final class SummaryElement {

    final String name;
    final boolean display;
    final List<SummaryItem> data;

    SummaryElement(String name, boolean display, List<SummaryItem> data) {
      this.name = name;
      this.display = display;
      this.data = data;
    }
  }

  @SuppressWarnings("unused")
  private static final class SummaryItem {

    final String name;
    final Object value;
    final String units;

    SummaryItem(String name, Object value, String units) {
      this.name = name;
      this.value = value;
      this.units = units;
    }
  }

  /* Filters out contributors whose contribution is below supplied limit. */
  static final class ContributionFilter implements Predicate<DisaggContributor> {

    private final double limit;
    private final double toPercent;

    ContributionFilter(double limit, double toPercent) {
      this.limit = limit;
      this.toPercent = toPercent;
    }

    double toPercent(double rate) {
      return rate * toPercent;
    }

    @Override
    public boolean test(DisaggContributor contributor) {
      return contributor.total() * toPercent >= limit;
    }
  }

  private List<JsonContributor> createJsonContributorList(
      DisaggDataset ddTotal,
      DisaggDataset dd,
      double contributorLimit) {

    ContributionFilter contributionFilter = new ContributionFilter(
        contributorLimit,
        percentScalar(ddTotal));

    List<JsonContributor> jsonContributors = new ArrayList<>();
    for (DisaggContributor contributor : dd.contributors) {
      if (contributionFilter.test(contributor)) {
        jsonContributors.addAll(contributor.toJson(contributionFilter));
        continue;
      }
      break;
    }

    return List.copyOf(jsonContributors);
  }

  private static List<SummaryItem> toSummaryItems(List<εBin> εBins) {
    return εBins.stream()
        .map(DisaggExport::εBinToSummary)
        .collect(Collectors.toList());
  }

  private static SummaryItem εBinToSummary(εBin bin) {
    return new SummaryItem("ε" + bin.id, bin.toString(), null);
  }

  static final class εBin {

    final int id;
    final Double min;
    final Double max;

    εBin(int id, Double min, Double max) {
      this.id = id;
      this.min = min;
      this.max = max;
    }

    @Override
    public String toString() {
      return new StringBuilder("[")
          .append((min == null) ? "-∞" : Double.toString(min))
          .append(" .. ")
          .append((max == null) ? "+∞" : Double.toString(max))
          .append((max == null) ? "]" : ")")
          .toString();
    }
  }
}
