package gov.usgs.earthquake.nshmp.calc;

import static com.google.common.base.Preconditions.checkArgument;

import java.util.EnumMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.Callable;
import java.util.function.Consumer;

import com.google.common.base.Converter;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Maps;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.calc.CalcConfig.Rate.Bins;
import gov.usgs.earthquake.nshmp.data.IntervalArray;
import gov.usgs.earthquake.nshmp.data.IntervalArray.Builder;
import gov.usgs.earthquake.nshmp.data.MutableXySequence;
import gov.usgs.earthquake.nshmp.data.Sequences;
import gov.usgs.earthquake.nshmp.data.XyPoint;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.model.ClusterRuptureSet;
import gov.usgs.earthquake.nshmp.model.ClusterSource;
import gov.usgs.earthquake.nshmp.model.Distance;
import gov.usgs.earthquake.nshmp.model.HazardModel;
import gov.usgs.earthquake.nshmp.model.Rupture;
import gov.usgs.earthquake.nshmp.model.RuptureSet;
import gov.usgs.earthquake.nshmp.model.Source;
import gov.usgs.earthquake.nshmp.model.SourceTree;
import gov.usgs.earthquake.nshmp.model.SourceType;
import gov.usgs.earthquake.nshmp.model.SystemRuptureSet;

/**
 * General purpose earthquake rate and probability data container. This class
 * makes no distinction between incremental or cumulative MFDs, or whether
 * values are stored as annual-rate or poisson probability.
 *
 * @author U.S. Geological Survey
 */
public class EqRate {

  /** The site of interest. */
  public final Site site;

  /** The total MFD of interest. */
  public final XySequence totalMfd;

  /** The MFDs for each contributing source type. */
  public final Map<SourceType, XySequence> typeMfds;

  private EqRate(
      Site site,
      XySequence totalMfd,
      Map<SourceType, XySequence> typeMfds) {

    this.site = site;
    this.totalMfd = totalMfd;
    this.typeMfds = typeMfds;
  }

  /**
   * Create a new earthquake rate data container at a {@code site} for the
   * supplied {@code model} and {@code config}.
   *
   * <p><b>Note:</b> any model initialization settings in {@code config} will be
   * ignored as the supplied model will already have been initialized.
   *
   * @param model of interest
   * @param config calculation configuration
   * @param site of interest
   */
  public static EqRate create(
      HazardModel model,
      CalcConfig config,
      Site site) {

    CalcConfig.Rate rateConfig = config.rate;

    Bins mBins = rateConfig.bins;
    IntervalArray modelMfd = IntervalArray.Builder
        .withRows(
            mBins.mMin,
            mBins.mMax,
            mBins.Δm)
        .build();

    EqRate rates = createIncremental(model, site, rateConfig.distance, modelMfd);
    if (rateConfig.distributionFormat == DistributionFormat.CUMULATIVE) {
      rates = toCumulative(rates);
    }
    if (rateConfig.valueFormat == ValueFormat.POISSON_PROBABILITY) {
      rates = toPoissonProbability(rates, rateConfig.timespan);
    }
    return rates;
  }

  /**
   * Wraps {@link #create(HazardModel, CalcConfig, Site)} in a {@link Callable}
   * for processing multiple sites concurrently.
   *
   * @param model to process
   * @param config calculation configuration
   * @param site of interest
   */
  @Deprecated
  public static Callable<EqRate> callable(
      HazardModel model,
      CalcConfig config,
      Site site) {

    return new RateTask(model, config, site);
  }

  private static EqRate createIncremental(
      HazardModel model,
      Site site,
      double distance,
      IntervalArray modelMfd) {

    /* Initialize SourceType mfd builders. */
    Map<SourceType, IntervalArray.Builder> typeMfdBuilders = new EnumMap<>(SourceType.class);
    for (SourceType type : model.types()) {
      typeMfdBuilders.put(type, IntervalArray.Builder.fromModel(modelMfd));
    }

    /* Populate builders. */
    for (SourceTree tree : model) {
      for (var branch : tree) {
        RuptureSet<? extends Source> ruptures = branch.value();
        IntervalArray rupturesMfd = mfd(ruptures, site.location(), distance, modelMfd);
        typeMfdBuilders.get(ruptures.type()).add(rupturesMfd);
      }
    }

    /* Compute total and convert to sequences. */
    IntervalArray.Builder totalMfd = IntervalArray.Builder.fromModel(modelMfd);
    ImmutableMap.Builder<SourceType, XySequence> typeMfds = ImmutableMap.builder();
    for (Entry<SourceType, IntervalArray.Builder> entry : typeMfdBuilders.entrySet()) {
      IntervalArray typeMfd = entry.getValue().build();
      typeMfds.put(entry.getKey(), typeMfd.values());
      totalMfd.add(typeMfd);
    }

    return new EqRate(
        site,
        totalMfd.build().values(),
        typeMfds.build());
  }

  /**
   * Create a new earthquake rate container with cumulative values.
   *
   * @param incremental rate source to convert
   */
  public static EqRate toCumulative(EqRate incremental) {

    XySequence cumulativeTotal = Sequences.toCumulative(incremental.totalMfd);
    ImmutableMap.Builder<SourceType, XySequence> cumulativeTypes = ImmutableMap.builder();
    for (Entry<SourceType, XySequence> entry : incremental.typeMfds.entrySet()) {
      cumulativeTypes.put(
          entry.getKey(),
          Sequences.toCumulative(entry.getValue()));
    }
    return new EqRate(
        incremental.site,
        cumulativeTotal,
        cumulativeTypes.build());
  }

  /**
   * Create a new earthquake rate container with Poisson probability values.
   *
   * @param annualRates source to convert
   * @param timespan of interest for annual rate to Poisson probability
   *        conversion
   */
  public static EqRate toPoissonProbability(EqRate annualRates, double timespan) {
    Converter<Double, Double> converter =
        Maths.annualRateToProbabilityConverter(timespan)
            .andThen(Maths.decimalToProbabilityConverter(2));
    Consumer<XyPoint> pointConvert = p -> p.set(converter.convert(p.y()));

    XySequence totalMfd = MutableXySequence
        .copyOf(annualRates.totalMfd)
        .transform(pointConvert);

    EnumMap<SourceType, XySequence> typeMfds = new EnumMap<>(SourceType.class);
    for (Entry<SourceType, XySequence> entry : annualRates.typeMfds.entrySet()) {
      typeMfds.put(
          entry.getKey(),
          MutableXySequence
              .copyOf(entry.getValue())
              .transform(pointConvert));
    }
    return new EqRate(
        annualRates.site,
        XySequence.copyOf(totalMfd),
        Maps.immutableEnumMap(typeMfds));
  }

  /**
   * Create a new earthquake rate container with the sum of the supplied
   * {@code rates}.
   *
   * <p><b>NOTE:</b> This operation is additive and will produce meaningless
   * results if {@code rates} have already been converted to
   * {@link #toPoissonProbability(EqRate, double) probabilities}, or are not all
   * of {@link DistributionFormat#INCREMENTAL} or
   * {@link DistributionFormat#CUMULATIVE} distribution format.
   *
   * <p>Buyer beware.
   *
   * @param rates to combine
   */
  @Deprecated
  public static EqRate combine(EqRate... rates) {
    Site referenceSite = rates[0].site;
    MutableXySequence totalMfd = MutableXySequence.emptyCopyOf(rates[0].totalMfd);
    EnumMap<SourceType, MutableXySequence> typeMfds = new EnumMap<>(SourceType.class);
    for (EqRate rate : rates) {
      checkArgument(
          rate.site.location().equals(referenceSite.location()),
          "Site locations are not the same:\n\ts1: %s\n\ts2: %s",
          referenceSite, rate.site);
      totalMfd.add(rate.totalMfd);
      for (Entry<SourceType, XySequence> entry : rate.typeMfds.entrySet()) {
        XySequence.addToMap(entry.getKey(), typeMfds, entry.getValue());
      }
    }
    return new EqRate(
        referenceSite,
        XySequence.copyOf(totalMfd),
        Maps.immutableEnumMap(typeMfds));
  }

  private static IntervalArray mfd(
      RuptureSet<? extends Source> ruptures,
      Location location,
      double distance,
      IntervalArray modelMfd) {

    switch (ruptures.type()) {
      case GRID:
        return gridMfd(ruptures, location, distance, modelMfd);
      case SLAB:
        return gridMfd(ruptures, location, distance, modelMfd);
      case FAULT_CLUSTER:
        return clusterMfd((ClusterRuptureSet) ruptures, location, distance,
            modelMfd);
      case FAULT_SYSTEM:
        return systemMfd((SystemRuptureSet) ruptures, location, distance,
            modelMfd);
      default:
        return faultMfd(ruptures, location, distance, modelMfd);
    }
  }

  /*
   * Short-circuit GridRuptureSet by summing the relevant node mfds. Handles
   * both GRID and SLAB types.
   */
  private static IntervalArray gridMfd(
      RuptureSet<? extends Source> ruptures,
      Location location,
      double distance,
      IntervalArray modelMfd) {

    IntervalArray.Builder rupturesMfd = IntervalArray.Builder.fromModel(modelMfd);
    for (Source source : ruptures.iterableForLocation(location, distance)) {
      for (Mfd mfd : source.mfds()) {
        rupturesMfd.addEach(mfd.data());
      }
    }
    return rupturesMfd.multiply(ruptures.weight()).build();
  }

  /*
   * Special case ClusterRuptureSet.
   *
   * Nested fault rates are in fact weights that need to be scaled by the
   * cluster rate.
   */

  private static IntervalArray clusterMfd(
      ClusterRuptureSet ruptures,
      Location location,
      double distance,
      IntervalArray modelMfd) {

    IntervalArray.Builder srcSetMfd = IntervalArray.Builder.fromModel(modelMfd);
    for (Source source : ruptures.iterableForLocation(location, distance)) {
      ClusterSource clusterSource = (ClusterSource) source;
      for (RuptureSet<? extends Source> rs : clusterSource.ruptureSets()) {
        IntervalArray.Builder faultMfd = Builder
            .copyOf(faultMfd(
                rs,
                location,
                distance,
                modelMfd))
            .multiply(clusterSource.rate());
        srcSetMfd.add(faultMfd.build());
      }
    }
    return srcSetMfd.multiply(ruptures.weight()).build();
  }

  /*
   * Special case and delegate to SystemRuptureSet.
   */
  private static IntervalArray systemMfd(
      SystemRuptureSet ruptures,
      Location location,
      double distance,
      IntervalArray modelMfd) {

    IntervalArray mfd = SystemRuptureSet
        .toRatesFunction(location, distance, modelMfd)
        .apply(ruptures);

    return IntervalArray.Builder.copyOf(mfd)
        .multiply(ruptures.weight())
        .build();
  }

  /*
   * Default approach: distance filter on ruptures.
   */
  private static IntervalArray faultMfd(
      RuptureSet<? extends Source> ruptures,
      Location location,
      double distance,
      IntervalArray modelMfd) {

    IntervalArray.Builder rupturesMfd = IntervalArray.Builder.fromModel(modelMfd);
    for (Source source : ruptures.iterableForLocation(location, distance)) {
      for (Rupture rupture : source) {
        Distance d = rupture.surface().distanceTo(location);
        if (d.rJB <= distance) {
          rupturesMfd.add(rupture.mag(), rupture.rate());
        }
      }
    }
    return rupturesMfd.multiply(ruptures.weight()).build();
  }

  private static final class RateTask implements Callable<EqRate> {

    private final HazardModel model;
    private final CalcConfig config;
    private final Site site;

    RateTask(
        HazardModel model,
        CalcConfig config,
        Site site) {

      this.model = model;
      this.config = config;
      this.site = site;
    }

    @Override
    public EqRate call() throws Exception {
      return create(model, config, site);
    }
  }

}
