package gov.usgs.earthquake.nshmp.calc;

import static gov.usgs.earthquake.nshmp.Text.Delimiter.COMMA;
import static java.nio.file.StandardOpenOption.APPEND;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.List;
import java.util.function.Function;
import java.util.stream.Collectors;

import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;

import gov.usgs.earthquake.nshmp.Text;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.model.HazardModel;
import gov.usgs.earthquake.nshmp.model.SourceType;

/**
 * Earthquake rate and probability exporter.
 *
 * @author U.S. Geological Survey
 */
public final class EqRateExport {

  private static final String RATE_FORMAT = "%.8g";
  private static final String PROB_FORMAT = "%.2f";
  private static final String RATE_FILE = "rates.csv";
  private static final String PROB_FILE = "probs.csv";

  private final Path out;
  private final String file;
  private final String valueFormat;
  private final HazardModel model;
  private final boolean exportSource;

  private final boolean namedSites;
  private boolean firstRecord = true;

  private EqRateExport(
      HazardModel model,
      CalcConfig config,
      List<Site> sites,
      Path out) {

    // whether or not rates or probabilities have been calculated
    boolean rates = config.rate.valueFormat == ValueFormat.ANNUAL_RATE;

    this.out = out;
    this.file = rates ? RATE_FILE : PROB_FILE;
    this.valueFormat = rates ? RATE_FORMAT : PROB_FORMAT;
    this.model = model;
    this.exportSource = config.output.dataTypes.contains(DataType.SOURCE);

    Site demoSite = sites.get(0);
    this.namedSites = demoSite.name() != Site.NO_NAME;
  }

  /**
   * Create a new results handler.
   *
   * @param model being run
   * @param config that specifies output options and formats
   * @param sites reference to the sites to be processed (not retained)
   * @param out results directory; this may have been been modified from
   *        {@code config.output.directory} by calling program
   */
  public static EqRateExport create(
      HazardModel model,
      CalcConfig config,
      List<Site> sites,
      Path out) {

    return new EqRateExport(model, config, sites, out);
  }

  /**
   * Add an {@code EqRate} to this exporter.
   *
   * @param rate data container to add
   */
  public void write(EqRate rate) throws IOException {

    double[] emptyValues = new double[rate.totalMfd.size()];
    Function<Double, String> formatter = Text.formatDoubleFunction(valueFormat);

    /* Can't init output until we have a demo rate record for mfd x-values */
    if (firstRecord) {
      init(rate);
      firstRecord = false;
    }

    String name = namedSites ? rate.site.name() : null;
    Location loc = rate.site.location();

    String locStr = name + "," +
        String.format("%.5f", loc.longitude) + "," +
        String.format("%.5f", loc.latitude);

    double[] values = rate.totalMfd.yValues().toArray();
    String totalLine = toLine(locStr, values, formatter);
    String emptyLine = toLine(locStr, emptyValues, formatter);

    /* write/append */
    Path totalFile = out.resolve(file);
    HazardExport.writeLine(totalFile, totalLine, APPEND);
    if (exportSource) {
      Path parentDir = out.resolve(HazardExport.TYPE_DIR);
      for (SourceType type : model.types()) {
        Path typeFile = parentDir.resolve(type.name()).resolve(file);
        String typeLine = emptyLine;
        if (rate.typeMfds.containsKey(type)) {
          XySequence typeRate = rate.typeMfds.get(type);
          typeLine = toLine(locStr, typeRate.yValues().toArray(), formatter);
        }
        HazardExport.writeLine(typeFile, typeLine, APPEND);
      }
    }
  }

  private void init(EqRate demo) throws IOException {
    Iterable<?> headerElements = Iterables.concat(
        Lists.newArrayList(namedSites ? "name" : null, "lon", "lat"),
        demo.totalMfd.xValues().boxed().collect(Collectors.toList()));
    String header = COMMA.joiner().join(headerElements);
    Path totalFile = out.resolve(file);
    HazardExport.writeLine(totalFile, header);

    if (exportSource) {
      Path parentDir = out.resolve(HazardExport.TYPE_DIR);
      Files.createDirectory(parentDir);
      for (SourceType type : model.types()) {
        Path typeDir = parentDir.resolve(type.name());
        Files.createDirectory(typeDir);
        Path typeFile = typeDir.resolve(file);
        HazardExport.writeLine(typeFile, header);
      }
    }
  }

  private static String toLine(
      String locStr,
      double[] values,
      Function<Double, String> formatter) {

    StringBuilder sb = new StringBuilder(locStr);
    Arrays.stream(values).forEach(v -> sb.append("," + v));
    return sb.toString();
  }
}
