package gov.usgs.earthquake.nshmp.calc;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.util.concurrent.Futures.allAsList;
import static com.google.common.util.concurrent.Futures.immediateFuture;
import static com.google.common.util.concurrent.Futures.transform;
import static gov.usgs.earthquake.nshmp.calc.CalcFactory.clustersToCurves;
import static gov.usgs.earthquake.nshmp.calc.CalcFactory.sourcesToCurves;
import static gov.usgs.earthquake.nshmp.calc.CalcFactory.systemToCurves;
import static gov.usgs.earthquake.nshmp.calc.CalcFactory.toHazardResult;
import static gov.usgs.earthquake.nshmp.model.SourceType.FAULT;
import static gov.usgs.earthquake.nshmp.model.SourceType.FAULT_CLUSTER;
import static gov.usgs.earthquake.nshmp.model.SourceType.FAULT_SYSTEM;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executor;
import java.util.function.Predicate;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.google.common.base.Stopwatch;

import gov.usgs.earthquake.nshmp.model.ClusterRuptureSet;
import gov.usgs.earthquake.nshmp.model.GridRuptureSet;
import gov.usgs.earthquake.nshmp.model.HazardModel;
import gov.usgs.earthquake.nshmp.model.RuptureSet;
import gov.usgs.earthquake.nshmp.model.Source;
import gov.usgs.earthquake.nshmp.model.SourceTree;
import gov.usgs.earthquake.nshmp.model.SourceType;
import gov.usgs.earthquake.nshmp.model.SystemRuptureSet;
import gov.usgs.earthquake.nshmp.tree.Branch;

/**
 * Static probabilistic seismic hazard analysis calculators.
 *
 * @author U.S. Geological Survey
 */
public class HazardCalcs {

  /*
   * Developer notes:
   *
   * -------------------------------------------------------------------------
   * Method argument order in this class, CalcFactory, and Transforms follow the
   * general rule of model (or model derived objects), calc config, site, and
   * then any others.
   * -------------------------------------------------------------------------
   * All calculations are done in log space, only on export are x-values
   * returned to linear space.
   * -------------------------------------------------------------------------
   * The ability to run a full calculation on the current thread facilitates
   * debugging.
   * -------------------------------------------------------------------------
   * Single threaded calcs monitor and log calculation duration of each
   * RuptureSet.
   * -------------------------------------------------------------------------
   * Although NPEs will be thrown once any method argument in this class is
   * used, because multiple arguments may be passed on the same line thereby
   * complicating identification of a null value, preemtive checks are
   * performed.
   * -------------------------------------------------------------------------
   * Notes on empty objects: In the default calculation pipeline, calculation
   * tasks are initiated when RuptureSet.iterableForLocation() yields relevant
   * sources for a site. Similarly, optimized (table-based) grid sources, once
   * created for a site, will not yield sources if the mfd table is empty.
   * System source sets, on the other hand, create a list of inputs on a
   * separate thread using means other than iterableForLocation. They are
   * therefore short-circuited returning an empty HazardCurveSet if no sources
   * are within range of a site. Cluster source sets have similar issue.
   *
   * Follow HazardCurveSet.Builder.empty() calls because it seems like clusters
   * shouldn't have to use it as the iterator in clustersToInputs should just
   * not be submitting any sources in any event if they are all too far away.
   * -------------------------------------------------------------------------
   */

  /**
   * Compute probabilistic seismic hazard curves at a {@code site} using the
   * supplied {@code model} and {@code config}. If an {@code executor} is
   * supplied,the calculation will be distributed; otherwise, it will run on the
   * current thread. Be sure to shutdown any supplied executor after a
   * calculation completes.
   *
   * <p><b>Note:</b> any model initialization settings in {@code config} will be
   * ignored as the supplied model will already have been initialized.
   *
   * @param model to use
   * @param config calculation configuration
   * @param site of interest
   * @param exec optional {@code Executor} to distribute calculation
   */
  public static Hazard hazard(
      HazardModel model,
      CalcConfig config,
      Site site,
      Executor exec) {

    checkNotNull(model);
    checkNotNull(config);
    checkNotNull(site);
    checkNotNull(exec);

    try {
      if (config.performance.threadCount == ThreadCount.ONE) {
        Logger log = Logger.getLogger(HazardCalcs.class.getName());
        return hazardCurve(model, config, site, log);
      }
      return asyncHazardCurve(model, config, site, exec);
    } catch (ExecutionException | InterruptedException e) {
      throw new RuntimeException(e);
    }
  }

  /*
   * Run a hazard curve calculation in parallel.
   */
  private static Hazard asyncHazardCurve(
      HazardModel model,
      CalcConfig config,
      Site site,
      Executor ex) throws InterruptedException, ExecutionException {

    AsyncList<HazardCurveSet> curveSets = AsyncList.create();
    AsyncList<RuptureSet<? extends Source>> gridTables = AsyncList.create();
    var settingFilter = model.settingFilter(config, site.location());
    var typeFilter = new SourceTypeFilter(config);

    for (SourceTree tree : model) {

      for (Branch<RuptureSet<? extends Source>> branch : tree) {
        RuptureSet<? extends Source> ruptures = branch.value();

        // check if source is applicable to site or should be skipped
        if (settingFilter.test(ruptures) && typeFilter.test(ruptures)) {

          switch (ruptures.type()) {

            case GRID:
              GridRuptureSet gss = (GridRuptureSet) ruptures;
              if (config.performance.optimizeGrids && gss.optimizable()) {
                gridTables.add(transform(
                    immediateFuture(gss),
                    GridRuptureSet.optimizer(
                        site.location(),
                        config.performance.smoothGrids,
                        config.hazard.gridFocalMechUpdate)::apply,
                    ex));
                break;
              }
              curveSets.add(sourcesToCurves(ruptures, site, config, ex));
              break;

            case FAULT_CLUSTER:
              curveSets.add(clustersToCurves(
                  (ClusterRuptureSet) ruptures, site, config, ex));
              break;

            case FAULT_SYSTEM:
              curveSets.add(systemToCurves(
                  (SystemRuptureSet) ruptures, site, config, ex));
              break;

            case INTERFACE_SYSTEM:
              curveSets.add(systemToCurves(
                  (SystemRuptureSet) ruptures, site, config, ex));
              break;

            default:
              curveSets.add(sourcesToCurves(ruptures, site, config, ex));
              break;
          }
        }

      }

    }
    /*
     * If grid optimization is enabled, grid calculations were deferred (above)
     * while table based source sets were initialized. Submit once all other
     * source types have been submitted.
     */
    for (RuptureSet<? extends Source> ruptures : allAsList(gridTables).get()) {
      curveSets.add(sourcesToCurves(ruptures, site, config, ex));
    }
    return toHazardResult(model, site, config, curveSets, ex);
  }

  /*
   * Run a hazard curve calculation on the current thread.
   */
  private static Hazard hazardCurve(
      HazardModel model,
      CalcConfig config,
      Site site,
      Logger log) {

    List<HazardCurveSet> curveSets = new ArrayList<>();
    var settingsFilter = model.settingFilter(config, site.location());
    var typeFilter = new SourceTypeFilter(config);

    log.info("HazardCurve: (single-threaded)");
    Stopwatch swTotal = Stopwatch.createStarted();
    Stopwatch swSource = Stopwatch.createStarted();

    for (SourceTree tree : model) {

      for (Branch<RuptureSet<? extends Source>> branch : tree) {
        RuptureSet<? extends Source> ruptures = branch.value();

        // check if source is applicable to site or should be skipped
        if (settingsFilter.test(ruptures) && typeFilter.test(ruptures)) {

          switch (ruptures.type()) {
            case GRID:
              GridRuptureSet gss = (GridRuptureSet) ruptures;
              if (config.performance.optimizeGrids && gss.optimizable()) {
                ruptures = GridRuptureSet.optimizer(
                    site.location(),
                    config.performance.smoothGrids,
                    config.hazard.gridFocalMechUpdate).apply(gss);
                log(log, MSSG_GRID_INIT, ruptures.name(), duration(swSource));
              }
              curveSets.add(sourcesToCurves(ruptures, site, config));
              log(log, MSSG_COMPLETED, ruptures.name(), duration(swSource));
              break;

            case FAULT_CLUSTER:
              curveSets.add(clustersToCurves(
                  (ClusterRuptureSet) ruptures, site, config));
              log(log, MSSG_COMPLETED, ruptures.name(), duration(swSource));
              break;

            case FAULT_SYSTEM:
              curveSets.add(systemToCurves(
                  (SystemRuptureSet) ruptures, site, config));
              log(log, MSSG_COMPLETED, ruptures.name(), duration(swSource));
              break;

            case INTERFACE_SYSTEM:
              curveSets.add(systemToCurves(
                  (SystemRuptureSet) ruptures, site, config));
              log(log, MSSG_COMPLETED, ruptures.name(), duration(swSource));
              break;

            default:
              curveSets.add(sourcesToCurves(ruptures, site, config));
              log(log, MSSG_COMPLETED, ruptures.name(), duration(swSource));
              break;
          }
        }
      }
    }

    log.log(Level.INFO, String.format(" %s: %s", MSSG_DURATION, duration(swTotal)));
    swTotal.stop();
    swSource.stop();
    return toHazardResult(model, site, config, curveSets);
  }

  /*
   * Support methods and fields for single-threaded calculations with more
   * verbose output.
   */

  private static String duration(Stopwatch sw) {
    String duration = sw.toString();
    sw.reset().start();
    return duration;
  }

  private static final String MSSG_GRID_INIT = "Init grid table";
  private static final String MSSG_COMPLETED = "      Completed";
  private static final String MSSG_DURATION = "     Total time";

  private static void log(Logger log, String leader, String source, String duration) {
    log.log(Level.INFO, String.format(" %s: %s %s", leader, duration, source));
  }

  /*
   * A filter that skips ruptures with dource types that do not match the types
   * specified in calc-config.
   */
  static class SourceTypeFilter implements Predicate<RuptureSet<? extends Source>> {

    private Set<SourceType> types;

    SourceTypeFilter(CalcConfig config) {
      // start with all source types included in model specified in config
      types = config.hazard.sourceTypes.isEmpty()
          ? EnumSet.allOf(SourceType.class)
          : EnumSet.copyOf(config.hazard.sourceTypes);
      if (types.contains(FAULT)) {
        types.add(FAULT_CLUSTER);
        types.add(FAULT_SYSTEM);
      }
    }

    @Override
    public boolean test(RuptureSet<? extends Source> ruptures) {
      return types.contains(ruptures.type());
    }
  }

}
