package gov.usgs.earthquake.nshmp.calc;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.data.MutableXySequence.emptyCopyOf;

import java.util.ArrayList;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;

import gov.usgs.earthquake.nshmp.data.MutableXySequence;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.model.GmmTree;
import gov.usgs.earthquake.nshmp.model.RuptureSet;
import gov.usgs.earthquake.nshmp.model.Source;
import gov.usgs.earthquake.nshmp.model.SourceType;
import gov.usgs.earthquake.nshmp.tree.Branch;

/**
 * Container class for hazard curves derived from a {@code RuptureSet}. Class
 * stores the {@code GroundMotions}s associated with each {@code Source} used in
 * a hazard calculation and the individual curves for each
 * {@code GroundMotionModel} used.
 *
 * <p>The {@code Builder} for this class is used to aggregate the HazardCurves
 * associated with each {@code Source} in a {@code RuptureSet}, scaled by the
 * {@code RuptureSet} weight. It also scales curves by their associated
 * {@code GroundMotionModel} weight, using distance-dependent weights when
 * appropriate. Note that this may lead to the dropping of some curves that are
 * not appropriate at large distances.
 *
 * <p>A HazardCurveSet is compatible with all types of {@code RuptureSet}s,
 * including {@code ClusterRuptureSet}s, which are handled differently in hazard
 * calculations. This container marks a point in the calculation pipeline where
 * results from cluster and other sources may be recombined into a single
 * {@code Hazard} result, regardless of {@code RuptureSet.type()} for all
 * relevant {@code RuptureSet}s.
 *
 * <p>For reasons related to disaggregation, this class also retains the curves
 * for each {@code ClusterSource} in a {@code ClusterRuptureSet}. Specifically,
 * disaggregation of cluster sources is not a straightforward decomposition of
 * contributions from individual sources and its helpful to have the total curve
 * for each {@code ClusterSource} available. Given the relatively specializaed
 * and infrequent use of {@code ClusterSource}s, this incurs little additional
 * overhead.
 *
 * <p>Although a {@code RuptureSet} stores a reference to the GMMs that were
 * used to compute gournd motions, GMM logic trees can be overridde in certain
 * regions and so a HazardCurveSet also stores a reference to the relevant
 * {@code GmmTree}.
 *
 * @author U.S. Geological Survey
 */
final class HazardCurveSet {

  final RuptureSet<? extends Source> ruptures;
  final GmmTree gmmTree;
  final List<GroundMotions> hazardGroundMotionsList;
  final List<ClusterGroundMotions> clusterGroundMotionsList;
  final Map<Imt, List<Map<Gmm, XySequence>>> clusterCurveLists;
  final Map<Imt, Map<Gmm, MutableXySequence>> curveMap;
  final Map<Imt, XySequence> totalCurves;
  final Map<Imt, TreeMap<Double, MutableXySequence>> magCurves;

  // Separate references by what is needed for hazard vs disagg
  // disagg of cluster and system types requires us to hold onto some
  // objects created when generating hazard curves

  private HazardCurveSet(Builder builder) {
    this.ruptures = builder.ruptures;
    this.gmmTree = builder.gmmTree;
    this.hazardGroundMotionsList = builder.hazardGroundMotionsList;
    this.clusterGroundMotionsList = builder.clusterGroundMotionsList;
    this.clusterCurveLists = builder.clusterCurveLists;
    this.curveMap = builder.curveMap;
    this.totalCurves = builder.totalCurves;
    this.magCurves = builder.magCurves;
  }

  private HazardCurveSet(RuptureSet<? extends Source> ruptures) {
    this.ruptures = ruptures;
    this.gmmTree = null;
    this.hazardGroundMotionsList = null;
    this.clusterGroundMotionsList = null;
    this.clusterCurveLists = null;
    this.curveMap = null;
    this.totalCurves = null;
    this.magCurves = null;
  }

  static Builder builder(
      RuptureSet<? extends Source> ruptures,
      GmmTree gmmTree,
      Map<Imt, XySequence> modelCurves) {

    return new Builder(ruptures, gmmTree, modelCurves);
  }

  /*
   * A HazardCurveSet is relatively heavyweight wrt initializing its builder.
   * This provides a placeholder for functions where a HazardCurveSet is
   * required but for which there may be no HazardCurves because all sources
   * were out of range of a site.
   */
  static HazardCurveSet empty(RuptureSet<? extends Source> ruptures) {
    return new HazardCurveSet(ruptures);
  }

  boolean isEmpty() {
    return totalCurves == null;
  }

  static class Builder {

    private static final String ID = "HazardCurveSet.Builder";
    private boolean built = false;

    private final RuptureSet<? extends Source> ruptures;
    private final GmmTree gmmTree;
    private final Map<Imt, XySequence> modelCurves;

    private final List<GroundMotions> hazardGroundMotionsList;
    private final List<ClusterGroundMotions> clusterGroundMotionsList;
    private final Map<Imt, List<Map<Gmm, XySequence>>> clusterCurveLists;
    private final Map<Imt, Map<Gmm, MutableXySequence>> curveMap;
    private final Map<Imt, XySequence> totalCurves;
    private final Map<Imt, TreeMap<Double, MutableXySequence>> magCurves;

    private Builder(
        RuptureSet<? extends Source> ruptures,
        GmmTree gmmTree,
        Map<Imt, XySequence> modelCurves) {

      this.ruptures = ruptures;
      this.gmmTree = gmmTree;
      this.modelCurves = modelCurves;

      boolean cluster = ruptures.type() == SourceType.FAULT_CLUSTER;

      if (cluster) {
        hazardGroundMotionsList = null;
        clusterGroundMotionsList = new ArrayList<>();
        clusterCurveLists = new EnumMap<>(Imt.class);
      } else {
        hazardGroundMotionsList = new ArrayList<>();
        clusterGroundMotionsList = null;
        clusterCurveLists = null;
      }
      curveMap = new EnumMap<>(Imt.class);
      magCurves = new EnumMap<>(Imt.class);

      Set<Gmm> gmms = gmmTree.gmms();
      Set<Imt> imts = modelCurves.keySet();

      for (Imt imt : imts) {

        Map<Gmm, MutableXySequence> gmmMap = new EnumMap<>(Gmm.class);
        for (Gmm gmm : gmms) {
          MutableXySequence emptyCurve = emptyCopyOf(modelCurves.get(imt));
          gmmMap.put(gmm, emptyCurve);
        }
        curveMap.put(imt, gmmMap);

        if (cluster) {
          List<Map<Gmm, XySequence>> clusterCurveList = new ArrayList<>();
          clusterCurveLists.put(imt, clusterCurveList);
        }

        TreeMap<Double, MutableXySequence> magMap = HazardCurves.initMagMap();
        magCurves.put(imt, magMap);

      }
      totalCurves = new EnumMap<>(Imt.class);
    }

    Builder addCurves(HazardCurves curvesIn) {
      checkNotNull(hazardGroundMotionsList, "%s only supports ClusterCurves", ID);
      hazardGroundMotionsList.add(curvesIn.groundMotions);
      // loop Imts based on what's been calculated
      for (Imt imt : curvesIn.curveMap.keySet()) {
        Map<Gmm, XySequence> curveMapIn = curvesIn.curveMap.get(imt);
        Map<Gmm, MutableXySequence> curveMapBuild = curveMap.get(imt);

        for (Branch<Gmm> branch : gmmTree.tree()) {
          Gmm gmm = branch.value();

          // TODO demote gmm weighting to transforms
          double weight = branch.weight() * ruptures.weight();
          curveMapBuild.get(gmm).add(
              MutableXySequence.copyOf(curveMapIn.get(gmm)).multiply(weight));
        }

        TreeMap<Double, MutableXySequence> dest = magCurves.get(imt);
        curvesIn.magMap.get(imt).entrySet().stream()
            .filter(e -> e.getValue() != null)
            .forEach(e -> dest.merge(
                e.getKey(),
                e.getValue().multiply(ruptures.weight()),
                MutableXySequence::add));
      }
      return this;
    }

    Builder addCurves(ClusterCurves curvesIn) {
      checkNotNull(clusterGroundMotionsList, "%s only supports HazardCurves", ID);
      clusterGroundMotionsList.add(curvesIn.clusterGroundMotions);
      // loop Imts based on what's been calculated
      for (Imt imt : curvesIn.curveMap.keySet()) {
        Map<Gmm, XySequence> curveMapIn = curvesIn.curveMap.get(imt);
        Map<Gmm, MutableXySequence> curveMapBuild = curveMap.get(imt);

        /*
         * Retain references to the total curve for each cluster source for
         * disaggregation in clusterCurveLists. These lists of maps by GMM will
         * contain only those curves for the GMMs approprate for the source-site
         * distance. When disaggregating, the same distance cutoff will be
         * considered so retrieval of curves from the maps should never throw a
         * NPE.
         */

        Map<Gmm, XySequence> clusterCurves = new EnumMap<>(Gmm.class);
        // loop Gmms based on what's supported at this distance
        for (Branch<Gmm> branch : gmmTree.tree()) {
          // (cluster per-section mags have already been collapsed, UNDO)
          Gmm gmm = branch.value();
          double weight = branch.weight() * ruptures.weight();

          XySequence clusterCurve = MutableXySequence.copyOf(curveMapIn.get(gmm)).multiply(weight);
          curveMapBuild.get(gmm).add(clusterCurve);
          clusterCurves.put(gmm, clusterCurve);
        }
        clusterCurveLists.get(imt).add(clusterCurves);

        TreeMap<Double, MutableXySequence> dest = magCurves.get(imt);
        curvesIn.magMap.get(imt).entrySet().stream()
            .filter(e -> e.getValue() != null)
            .forEach(e -> dest.merge(
                e.getKey(),
                e.getValue().multiply(ruptures.weight()),
                MutableXySequence::add));

      }
      return this;
    }

    HazardCurveSet build() {
      checkState(!built, "This %s instance has already been used", ID);
      built = true;
      computeFinal();
      return new HazardCurveSet(this);
    }

    /*
     * Create the final wieghted (Gmm) combined curve. The Gmm curves were
     * scaled by their weights while building (above).
     */
    private void computeFinal() {
      for (Entry<Imt, Map<Gmm, MutableXySequence>> entry : curveMap.entrySet()) {
        Imt imt = entry.getKey();
        MutableXySequence totalCurve = emptyCopyOf(modelCurves.get(imt));
        for (XySequence curve : entry.getValue().values()) {
          totalCurve.add(curve);
        }
        totalCurves.put(imt, XySequence.copyOf(totalCurve));
      }
    }
  }

}
