package gov.usgs.earthquake.nshmp.calc;

import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.data.XySequence.copyOf;
import static java.util.stream.Collectors.toList;

import java.util.Arrays;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;

import com.google.common.collect.Iterables;

import gov.usgs.earthquake.nshmp.data.MutableXySequence;
import gov.usgs.earthquake.nshmp.data.Sequences;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.gmm.Imt;

/**
 * Container class for the combined hazard curves derived from the
 * {@code Rupture}s in an individual {@code Source}, one for each
 * {@code GroundMotionModel} and {@code Imt} of interest. The curves will have
 * been scaled by the associated Mfd or rupture weights, but not by
 * {@code GroundMotionModel} weights.
 *
 * @author U.S. Geological Survey
 */
final class HazardCurves {

  final GroundMotions groundMotions;
  final Map<Imt, Map<Gmm, XySequence>> curveMap;
  final Map<Imt, TreeMap<Double, MutableXySequence>> magMap;

  // Map<Imt, Map<Gmm, List<LogicTree<XySequence>>>> curveTreeMap;

  private HazardCurves(Builder builder) {
    this.groundMotions = builder.groundMotions;
    this.curveMap = builder.curveMap;
    this.magMap = builder.magCurveMap;
  }

  static Builder builder(GroundMotions groundMotions) {
    return new Builder(groundMotions);
  }

  /*
   * Specialized constructor that creates a single HazardCurves from the results
   * of processing a partitioned InputList.
   */
  static HazardCurves combine(
      InputList inputs,
      List<HazardCurves> curvesList,
      boolean magCurves) {

    List<GroundMotions> groundMotionsList = curvesList.stream()
        .map(curves -> curves.groundMotions)
        .collect(toList());
    GroundMotions groundMotions = GroundMotions.combine(inputs, groundMotionsList);
    return builder(groundMotions)
        .addPartitioned(curvesList, magCurves)
        .build();
  }

  static class Builder {

    private static final String ID = "HazardCurves.Builder";
    private boolean built = false;

    private final GroundMotions groundMotions;
    private final Map<Imt, Map<Gmm, XySequence>> curveMap;
    private final Map<Imt, TreeMap<Double, MutableXySequence>> magCurveMap;

    private Builder(GroundMotions groundMotions) {
      this.groundMotions = groundMotions;
      curveMap = new EnumMap<>(Imt.class);
      magCurveMap = new EnumMap<>(Imt.class);
      for (Imt imt : groundMotions.gmMap.keySet()) {
        Map<Gmm, XySequence> gmmMap = new EnumMap<>(Gmm.class);
        curveMap.put(imt, gmmMap);
        TreeMap<Double, MutableXySequence> magMap = initMagMap();
        magCurveMap.put(imt, magMap);
      }
    }

    /* Put an immutable copy of the supplied curve, one per GMM. */
    Builder addGmmCurve(Imt imt, Gmm gmm, XySequence curve) {
      curveMap.get(imt).put(gmm, copyOf(curve));
      return this;
    }

    /* Put a copy or add the supplied magnitude curve. */
    Builder addMagCurve(Imt imt, double Mw, XySequence curve) {
      TreeMap<Double, MutableXySequence> magMap = magCurveMap.get(imt);
      Double k = magMap.floorKey(Mw);
      MutableXySequence v = magMap.get(k);
      if (v == null) {
        magMap.put(k, MutableXySequence.copyOf(curve));
      } else {
        v.add(curve);
      }
      return this;
    }

    HazardCurves build() {
      checkState(!built, "This %s instance has already been used", ID);
      built = true;
      return new HazardCurves(this);
    }

    /*
     * For internal use only. Adds multiple ordered HazardCurves that result
     * from InputList partitioning. Can only be called once after intializing
     * the builder with a combined GroundMotions object previously created from
     * the supplied HazardCurves.
     */
    private Builder addPartitioned(List<HazardCurves> curvesList, boolean magCurves) {

      /* Add GMM curves. */
      Map<Imt, Map<Gmm, XySequence>> model = curvesList.get(0).curveMap;
      for (Entry<Imt, Map<Gmm, XySequence>> imtEntry : model.entrySet()) {
        Imt imt = imtEntry.getKey();
        for (Entry<Gmm, XySequence> gmmEntry : imtEntry.getValue().entrySet()) {
          Gmm gmm = gmmEntry.getKey();
          MutableXySequence sequence = MutableXySequence.copyOf(gmmEntry.getValue());
          for (HazardCurves curves : Iterables.skip(curvesList, 1)) {
            sequence.add(curves.curveMap.get(imt).get(gmm));
          }
          addGmmCurve(imt, gmm, sequence);
        }
      }

      if (magCurves) {
        /* Add magnitude curves. */
        for (HazardCurves curves : curvesList) {
          for (Imt imt : curves.magMap.keySet()) {
            TreeMap<Double, MutableXySequence> dest = magCurveMap.get(imt);
            curves.magMap.get(imt).entrySet().stream()
                .filter(e -> e.getValue() != null)
                .forEach(e -> dest.merge(e.getKey(), e.getValue(), MutableXySequence::add));
          }
        }
      }

      return this;
    }
  }

  private static final double[] MAGS = Sequences.arrayBuilder(4.4, 9.4, 0.2)
      .centered()
      .scale(1)
      .build();

  /*
   * Depending on what type of source is being processed we won't likely need
   * all magnitude bins and so we lazily initialize the receiver
   * MutableXySequences in addMagCurve(). Alternative would be to initialize
   * receiver sequences with model sequence:
   *
   * m -> MutableXySequence.copyOf(model)
   */
  static TreeMap<Double, MutableXySequence> initMagMap() {
    TreeMap<Double, MutableXySequence> map = new TreeMap<>();
    Arrays.stream(MAGS)
        .boxed()
        .forEach(m -> map.put(m, null));
    return map;
  }
}
