package gov.usgs.earthquake.nshmp.calc;

import static com.google.common.base.Preconditions.checkState;
import static com.google.common.util.concurrent.Futures.allAsList;
import static com.google.common.util.concurrent.Futures.getUnchecked;
import static com.google.common.util.concurrent.Futures.immediateFuture;
import static com.google.common.util.concurrent.Futures.transform;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.instances;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.Executor;
import java.util.function.Function;

import com.google.common.collect.ListMultimap;
import com.google.common.collect.MultimapBuilder;

import gov.usgs.earthquake.nshmp.Faults;
import gov.usgs.earthquake.nshmp.data.DoubleData;
import gov.usgs.earthquake.nshmp.data.MutableXySequence;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.fault.surface.RuptureSurface;
import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.gmm.GmmInput;
import gov.usgs.earthquake.nshmp.gmm.GroundMotion;
import gov.usgs.earthquake.nshmp.gmm.GroundMotionModel;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.model.ClusterRuptureSet;
import gov.usgs.earthquake.nshmp.model.ClusterSource;
import gov.usgs.earthquake.nshmp.model.Distance;
import gov.usgs.earthquake.nshmp.model.GmmTree;
import gov.usgs.earthquake.nshmp.model.HazardModel;
import gov.usgs.earthquake.nshmp.model.Rupture;
import gov.usgs.earthquake.nshmp.model.RuptureSet;
import gov.usgs.earthquake.nshmp.model.Source;
import gov.usgs.earthquake.nshmp.model.SystemRuptureSet;
import gov.usgs.earthquake.nshmp.tree.LogicTree;
import gov.usgs.earthquake.nshmp.tree.Trees;

/**
 * Data transform {@link Function}s. Package private classes are used
 * exclusively by {@link CalcFactory}.
 *
 * <p>These functions perform all the work of computing probabilistic seismic
 * hazard and facilitate the reduction of calculations into component tasks that
 * may be executed concurrently.
 *
 * <p>For convenience, some functions coalesce multiple tasks. For example,
 * {@link SourceToCurves} merges the work of {@link SourceToInputs},
 * {@link InputsToGroundMotions}, and {@link GroundMotionsToCurves}. There are
 * also custom functions to handle special case cluster and system
 * {@code Source}s.
 *
 * @author U.S. Geological Survey
 * @see CalcFactory
 */
final class Transforms {

  /*
   * Source --> InputList
   *
   * Create a list of ground motion inputs from a source.
   */
  private static final class SourceToInputs implements Function<Source, InputList> {

    private final Site site;

    SourceToInputs(Site site) {
      this.site = site;
    }

    @Override
    public SourceInputList apply(Source source) {
      SourceInputList hazardInputs = new SourceInputList(source);

      for (Rupture rup : source) {

        RuptureSurface surface = rup.surface();

        Distance distances = surface.distanceTo(site.location());
        double dip = surface.dip();
        double width = surface.width();
        double zTor = surface.depth();
        double zHyp = Faults.hypocentralDepth(dip, width, zTor);

        HazardInput input = new HazardInput(
            rup.rate(),
            rup.mag(),
            distances.rJB,
            distances.rRup,
            distances.rX,
            dip,
            width,
            zTor,
            zHyp,
            rup.rake(),
            site.vs30(),
            site.z1p0(),
            site.z2p5(),
            site.zSed());
        hazardInputs.add(input);
      }
      return hazardInputs;
    }
  }

  /*
   * InputList --> GroundMotions
   *
   * Calculate ground motions for a list of ground motion inputs.
   */
  private static final class InputsToGroundMotions implements Function<InputList, GroundMotions> {

    private final Map<Imt, Map<Gmm, GroundMotionModel>> gmmInstances;
    private final double dampingRatio;
    private final boolean damping;

    InputsToGroundMotions(CalcConfig config, Map<Imt, Map<Gmm, GroundMotionModel>> gmmInstances) {
      this.gmmInstances = gmmInstances;
      this.dampingRatio = config.gmm.dampingRatio;
      this.damping = dampingRatio != 5.0;
    }

    @Override
    public GroundMotions apply(InputList inputs) {

      Set<Imt> imtKeys = gmmInstances.keySet();
      Set<Gmm> gmmKeys = gmmInstances.get(imtKeys.iterator().next()).keySet();
      GroundMotions.Builder builder = GroundMotions.builder(
          inputs,
          imtKeys,
          gmmKeys);

      for (Imt imt : imtKeys) {
        for (Gmm gmm : gmmKeys) {
          GroundMotionModel model = gmmInstances.get(imt).get(gmm);
          for (GmmInput gmmInput : inputs) {
            LogicTree<GroundMotion> gms = damping
                ? model.dampedCalc(gmmInput, gmm.type(), dampingRatio)
                : model.calc(gmmInput);
            builder.add(imt, gmm, gms);
          }
        }
      }
      return builder.build();
    }
  }

  /*
   * GroundMotions --> HazardCurves
   *
   * Derive hazard curves for a set of ground motions.
   */
  private static final class GroundMotionsToCurves implements
      Function<GroundMotions, HazardCurves> {

    private final Map<Imt, XySequence> modelCurves;
    private final ExceedanceModel exceedanceModel;
    private final double truncationLevel;
    private final double σScale;
    private final GmmTree gmms;
    private final boolean magCurves;

    GroundMotionsToCurves(CalcConfig config, GmmTree gmms) {
      this.modelCurves = config.hazard.logModelCurves();
      this.exceedanceModel = config.hazard.exceedanceModel;
      this.truncationLevel = config.hazard.truncationLevel;
      this.σScale = config.gmm.sigmaScale;
      this.gmms = gmms;
      this.magCurves = config.output.dataTypes.contains(DataType.MAGNITUDE);
    }

    @Override
    public HazardCurves apply(GroundMotions gms) {

      HazardCurves.Builder curveBuilder = HazardCurves.builder(gms);

      for (Entry<Imt, Map<Gmm, List<LogicTree<GroundMotion>>>> imtEntry : gms.gmMap.entrySet()) {

        Imt imt = imtEntry.getKey();
        XySequence modelCurve = modelCurves.get(imt);
        MutableXySequence gmmCurve = MutableXySequence.copyOf(modelCurve);
        MutableXySequence utilCurve = MutableXySequence.emptyCopyOf(modelCurve);

        /*
         * Exceedance methods always put result in supplied curve and are
         * responsible for 'clearing' it before use if needed.
         */

        for (Entry<Gmm, List<LogicTree<GroundMotion>>> gmmEntry : imtEntry.getValue().entrySet()) {

          gmmCurve.clear();
          List<LogicTree<GroundMotion>> gmTrees = gmmEntry.getValue();
          Gmm gmm = gmmEntry.getKey();

          for (int i = 0; i < gmTrees.size(); i++) {

            exceedanceModel.treeExceedanceCombined(
                gmTrees.get(i),
                truncationLevel,
                σScale,
                imt,
                utilCurve);

            // TODO scale gmms here instead of upon addition
            // HazardCurveSet -- gmms.gmmWeight(gmm)
            utilCurve.multiply(gms.inputs.get(i).rate);
            gmmCurve.add(utilCurve);

            if (magCurves) {
              // utilCurve has been added to a receiver so we can modify it here
              // to be added to the magnitude curve map
              curveBuilder.addMagCurve(
                  imt,
                  gms.inputs.get(i).Mw,
                  utilCurve.multiply(gmms.gmmWeight(gmm)));
            }
          }
          curveBuilder.addGmmCurve(imt, gmm, gmmCurve);
        }
      }
      return curveBuilder.build();
    }
  }

  /*
   * Source --> HazardCurves
   *
   * Compute hazard curves for a source. This function coalesces the four
   * preceeding functions into one.
   */
  static final class SourceToCurves implements Function<Source, HazardCurves> {

    private final Function<Source, InputList> sourceToInputs;
    private final Function<InputList, GroundMotions> inputsToGroundMotions;
    private final Function<GroundMotions, HazardCurves> groundMotionsToCurves;

    SourceToCurves(
        GmmTree gmms,
        Site site,
        CalcConfig config) {

      Map<Imt, Map<Gmm, GroundMotionModel>> gmmInstances = instances(
          config.hazard.imts,
          gmms.gmms());

      this.sourceToInputs = new SourceToInputs(site);
      this.inputsToGroundMotions = new InputsToGroundMotions(config, gmmInstances);
      this.groundMotionsToCurves = new GroundMotionsToCurves(config, gmms);
    }

    @Override
    public HazardCurves apply(Source source) {
      return sourceToInputs
          .andThen(inputsToGroundMotions)
          .andThen(groundMotionsToCurves)
          .apply(source);
    }
  }

  /*
   * List<HazardCurves> --> HazardCurveSet
   *
   * Reduce multiple source curves.
   */
  static final class CurveConsolidator implements Function<List<HazardCurves>, HazardCurveSet> {

    private final RuptureSet<? extends Source> ruptures;
    private final GmmTree gmms;
    private final Map<Imt, XySequence> modelCurves;

    CurveConsolidator(
        RuptureSet<? extends Source> ruptures,
        GmmTree gmms,
        CalcConfig config) {

      this.ruptures = ruptures;
      this.gmms = gmms;
      this.modelCurves = config.hazard.logModelCurves();
    }

    @Override
    public HazardCurveSet apply(List<HazardCurves> curvesList) {

      if (curvesList.isEmpty()) {
        return HazardCurveSet.empty(ruptures);
      }

      HazardCurveSet.Builder curveSetBuilder = HazardCurveSet.builder(
          ruptures,
          gmms,
          modelCurves);

      for (HazardCurves curves : curvesList) {
        curveSetBuilder.addCurves(curves);
      }
      return curveSetBuilder.build();
    }
  }

  /*
   * A note on system sources:
   *
   * SystemRuptureSets contain many single sources and the functions here handle
   * them collectively. Rather than creating lists of input lists for each
   * source, one large input list is created. This may change if it is decided
   * to apply a magnitude distribution on each system source. This motivated
   * reordering of the SourceType enum such that SystemRuptureSets are processed
   * first and granted a thread early in the calculation process.
   */

  /*
   * SYSTEM: SystemRuptureSet --> HazardCurveSet
   *
   * Compute hazard curves for system sources. This function derives all inputs
   * for an entire SystemRuptureSet before being composing them with standard
   * ground motion and hazard curve functions.
   */
  static final class SystemToCurves implements Function<SystemRuptureSet, HazardCurveSet> {

    private final GmmTree gmms;
    private final Site site;
    private final CalcConfig config;

    SystemToCurves(
        GmmTree gmms,
        Site site,
        CalcConfig config) {

      this.gmms = gmms;
      this.site = site;
      this.config = config;
    }

    @Override
    public HazardCurveSet apply(SystemRuptureSet ruptures) {

      InputList inputs = SystemRuptureSet.toInputsFunction(site).apply(ruptures);
      if (inputs.isEmpty()) {
        return HazardCurveSet.empty(ruptures);
      }

      Map<Imt, Map<Gmm, GroundMotionModel>> gmmInstances = instances(
          config.hazard.imts,
          gmms.gmms());

      InputsToGroundMotions inputsToGm = new InputsToGroundMotions(config, gmmInstances);
      GroundMotions gms = inputsToGm.apply(inputs);

      Function<GroundMotions, HazardCurves> gmToCurves = new GroundMotionsToCurves(config, gmms);
      HazardCurves curves = gmToCurves.apply(gms);

      CurveConsolidator consolidator = new CurveConsolidator(
          ruptures, gmms, config);
      return consolidator.apply(List.of(curves));
    }
  }

  /*
   * SYSTEM: SystemRuptureSet --> HazardCurveSet
   *
   * Compute hazard curves for system sources concurrently. This function
   * derives all inputs for an entire SystemRuptureSet and partitions them
   * before composing them with standard ground motion and hazard curve
   * functions.
   */
  static final class ParallelSystemToCurves implements Function<SystemRuptureSet, HazardCurveSet> {

    private final GmmTree gmms;
    private final Site site;
    private final CalcConfig config;
    private final Executor ex;

    ParallelSystemToCurves(
        GmmTree gmms,
        Site site,
        CalcConfig config,
        Executor ex) {

      this.gmms = gmms;
      this.site = site;
      this.config = config;
      this.ex = ex;
    }

    @Override
    public HazardCurveSet apply(SystemRuptureSet ruptures) {

      // create input list
      InputList master = SystemRuptureSet.toInputsFunction(site).apply(ruptures);
      if (master.isEmpty()) {
        return HazardCurveSet.empty(ruptures);
      }

      // calculate curves from list in parallel
      SystemInputsToCurves inputsToCurves = new SystemInputsToCurves(gmms, site, config);
      AsyncList<HazardCurves> asyncCurvesList = AsyncList.create();
      int size = config.performance.systemPartition;
      for (InputList partition : master.partition(size)) {
        asyncCurvesList.add(transform(
            immediateFuture(partition),
            inputsToCurves::apply,
            ex));
      }
      List<HazardCurves> curvesList = getUnchecked(allAsList(asyncCurvesList));

      // combine and consolidate
      HazardCurves hazardCurves = HazardCurves.combine(
          master,
          curvesList,
          config.output.dataTypes.contains(DataType.MAGNITUDE));
      CurveConsolidator consolidator = new CurveConsolidator(
          ruptures, gmms, config);

      return consolidator.apply(List.of(hazardCurves));
    }
  }

  /*
   * SYSTEM: InputList --> HazardCurves
   *
   * Compute hazard curves from an input list. Although this function is
   * generally applicable to all source types, it is presently only used to
   * process partitioned input lists derived from a system source.
   */
  private static final class SystemInputsToCurves implements Function<InputList, HazardCurves> {

    private final Function<InputList, GroundMotions> inputsToGroundMotions;
    private final Function<GroundMotions, HazardCurves> groundMotionsToCurves;

    SystemInputsToCurves(
        GmmTree gmms,
        Site site,
        CalcConfig config) {

      Map<Imt, Map<Gmm, GroundMotionModel>> gmmInstances = instances(
          config.hazard.imts,
          gmms.gmms());

      this.inputsToGroundMotions = new InputsToGroundMotions(config, gmmInstances);
      this.groundMotionsToCurves = new GroundMotionsToCurves(config, gmms);
    }

    @Override
    public HazardCurves apply(InputList inputs) {
      return groundMotionsToCurves.apply(inputsToGroundMotions.apply(inputs));
    }
  }

  /*
   * CLUSTER: ClusterSource --> ClusterInputs
   *
   * Create a list of ground motion inputs from a cluster source.
   */
  private static final class ClusterSourceToInputs implements
      Function<ClusterSource, ClusterInputs> {

    private final SourceToInputs transform;

    ClusterSourceToInputs(Site site) {
      transform = new SourceToInputs(site);
    }

    @Override
    public ClusterInputs apply(ClusterSource clusterSource) {
      ClusterInputs clusterInputs = new ClusterInputs(clusterSource);
      for (RuptureSet<? extends Source> ruptureSet : clusterSource.ruptureSets()) {
        checkState(ruptureSet.size() == 1);
        clusterInputs.add(transform.apply(ruptureSet.iterator().next()));
      }
      return clusterInputs;
    }
  }

  /*
   * CLUSTER: ClusterInputs --> ClusterGroundMotions
   *
   * Calculate ground motions for a list of cluster ground motion inputs.
   */
  private static final class ClusterInputsToGroundMotions implements
      Function<ClusterInputs, ClusterGroundMotions> {

    private final InputsToGroundMotions transform;

    ClusterInputsToGroundMotions(
        CalcConfig config,
        Map<Imt, Map<Gmm, GroundMotionModel>> gmmInstances) {
      transform = new InputsToGroundMotions(config, gmmInstances);
    }

    @Override
    public ClusterGroundMotions apply(ClusterInputs clusterInputs) {
      ClusterGroundMotions clusterGroundMotions = new ClusterGroundMotions(clusterInputs.parent);
      for (SourceInputList hazardInputs : clusterInputs) {
        clusterGroundMotions.add(transform.apply(hazardInputs));
      }
      return clusterGroundMotions;
    }
  }

  /*
   * CLUSTER: ClusterGroundMotions --> ClusterCurves
   *
   * Collapse magnitude variants and compute the joint probability of exceedence
   * for sources in a cluster. Note that this is only to be used with cluster
   * sources as the weight of each magnitude variant is stored in the
   * HazardInput.rate field, which is kinda KLUDGY, but works.
   *
   * TODO: Investigate sensitivity to collapsing magnitude variants on each
   * fault in a cluster [F1(M1,M2,M3), F2(M1 M2,M3), ...] versus doing each
   * magnitude branch separately [(F1M1,F2M1,...), (F1M2,F2M2,...)]
   */
  private static final class ClusterGroundMotionsToCurves implements
      Function<ClusterGroundMotions, ClusterCurves> {

    private final Map<Imt, XySequence> logModelCurves;
    private final ExceedanceModel exceedanceModel;
    private final double truncationLevel;
    private final double σScale;
    private final GmmTree gmms;
    private final boolean magCurves;

    ClusterGroundMotionsToCurves(CalcConfig config, GmmTree gmms) {
      this.logModelCurves = config.hazard.logModelCurves();
      this.exceedanceModel = config.hazard.exceedanceModel;
      this.truncationLevel = config.hazard.truncationLevel;
      this.σScale = config.gmm.sigmaScale;
      this.gmms = gmms;
      this.magCurves = config.output.dataTypes.contains(DataType.MAGNITUDE);
    }

    @Override
    public ClusterCurves apply(ClusterGroundMotions clusterGroundMotions) {

      ClusterCurves.Builder curveBuilder = ClusterCurves.builder(clusterGroundMotions);

      for (Entry<Imt, XySequence> entry : logModelCurves.entrySet()) {

        XySequence modelCurve = entry.getValue();
        Imt imt = entry.getKey();

        /*
         * Developer notes:
         *
         * A holdover from earlier codes is the storing of the magnitude branch
         * weight in the rate field of the GmmInputs.
         */

        /* Averaged magnitudes over each fault in a cluster. */
        List<Double> avgMag = new ArrayList<>();

        /* Aggregator of curves for each fault in a cluster. */
        ListMultimap<Gmm, LogicTree<MutableXySequence>> faultCurves = MultimapBuilder
            .enumKeys(Gmm.class)
            .arrayListValues(clusterGroundMotions.size())
            .build();
        MutableXySequence utilCurve = MutableXySequence.emptyCopyOf(modelCurve);

        /*
         * Collapse magnitude branches down to lists of logic trees of epistemic
         * curves for each GMM.
         */
        for (GroundMotions groundMotions : clusterGroundMotions) {

          /* Epistemic ground motion trees for each fault section magnitude. */
          Map<Gmm, List<LogicTree<GroundMotion>>> gmmGmMap = groundMotions.gmMap.get(imt);

          for (Gmm gmm : gmmGmMap.keySet()) {
            List<LogicTree<GroundMotion>> gmTrees = gmmGmMap.get(gmm);

            /* Epistemic curve trees for each fault section magnitude. */
            List<LogicTree<MutableXySequence>> magCurves = new ArrayList<>(gmTrees.size());

            /* Loop fault magnitude variants. */
            for (int i = 0; i < gmTrees.size(); i++) {

              /* Tree of exceedance curves for each magnitude variant. */
              LogicTree<GroundMotion> gmTree = gmTrees.get(i);
              LogicTree<MutableXySequence> magTreeCurves = exceedanceModel.treeExceedance(
                  gmTree,
                  truncationLevel,
                  σScale,
                  imt,
                  utilCurve);

              /* Scale by magnitude weight and aggregate. */
              double magWt = groundMotions.inputs.get(i).rate;
              magTreeCurves.forEach(branch -> branch.value().multiply(magWt));
              magCurves.add(magTreeCurves);
            }

            /*
             * Combine magnitude variants and collect. faultTreeCurves contains
             * the epistemic GMM branch curves combined over the magnitude
             * branches for a single fault in the cluster. faultCurves lists are
             * the same size as the number of faults in the cluster.
             */
            LogicTree<List<MutableXySequence>> transposed = Trees.transpose(magCurves);
            LogicTree<MutableXySequence> faultTreeCurves = Trees.transform(
                transposed,
                Transforms::sum,
                Optional.empty());
            faultCurves.put(gmm, faultTreeCurves);
          }

          if (magCurves) {
            /*
             * The weighted average magnitudes of each rupture in the cluster
             * for magnitude curve collection.
             */
            double wtMag = 0.0;
            for (HazardInput in : groundMotions.inputs) {
              wtMag += in.Mw * in.rate; // Mw weight is stored in rate
            }
            avgMag.add(wtMag);
          }
        }

        /* Average magnitude over a cluster. NaN if !magCurves */
        double clusterMag = DoubleData.sum(avgMag) / avgMag.size();

        /* Cluster exceedance */
        double rate = clusterGroundMotions.parent.rate();

        for (Gmm gmm : faultCurves.keySet()) {

          /* Combine cluster fault exceedances on each gmm epistemic branch. */
          LogicTree<List<MutableXySequence>> transposed = Trees.transpose(faultCurves.get(gmm));
          LogicTree<MutableXySequence> clusterTreeCurves = Trees.transform(
              transposed,
              ExceedanceModel::clusterExceedance,
              Optional.empty());

          /* Scale gmm branches by weight and combine. */
          MutableXySequence clusterCurve = collapse(clusterTreeCurves);

          /* Scale to cluster rate and save. */

          // TODO scale gmms here instead of upon addition
          // HazardCurveSet -- gmms.gmmWeight(gmm)
          clusterCurve.multiply(rate);
          curveBuilder.addGmmCurve(imt, gmm, clusterCurve);

          if (magCurves) {
            curveBuilder.addMagCurve(
                imt,
                clusterMag,
                clusterCurve.multiply(gmms.gmmWeight(gmm)));
          }
        }
      }
      return curveBuilder.build();
    }
  }

  /*
   * Returns a new sequence with the sum of the supplied sequences. Method
   * assumes all sequences are the same size and not empty. Also returns mutable
   * sequence, XySequence.combine(...) does not.
   */
  private static MutableXySequence sum(List<MutableXySequence> sequences) {
    MutableXySequence reduced = MutableXySequence.emptyCopyOf(sequences.get(0));
    sequences.forEach(reduced::add);
    return reduced;
  }

  private static MutableXySequence collapse(LogicTree<MutableXySequence> tree) {
    MutableXySequence collapsed = MutableXySequence.emptyCopyOf(tree.get(0).value());
    tree.forEach(br -> collapsed.add(br.value().multiply(br.weight())));
    return collapsed;
  }

  /*
   * CLUSTER: ClusterSource --> ClusterCurves
   *
   * Compute hazard curves for a cluster source. This function coalesces the
   * three preceeding functions into one.
   */
  static final class ClusterToCurves implements Function<ClusterSource, ClusterCurves> {

    private final Function<ClusterSource, ClusterInputs> sourceToInputs;
    private final Function<ClusterInputs, ClusterGroundMotions> inputsToGroundMotions;
    private final Function<ClusterGroundMotions, ClusterCurves> groundMotionsToCurves;

    ClusterToCurves(
        GmmTree gmms,
        Site site,
        CalcConfig config) {

      Map<Imt, Map<Gmm, GroundMotionModel>> gmmInstances = instances(
          config.hazard.imts,
          gmms.gmms());

      this.sourceToInputs = new ClusterSourceToInputs(site);
      this.inputsToGroundMotions = new ClusterInputsToGroundMotions(config, gmmInstances);
      this.groundMotionsToCurves = new ClusterGroundMotionsToCurves(config, gmms);
    }

    @Override
    public ClusterCurves apply(ClusterSource source) {
      return groundMotionsToCurves.apply(
          inputsToGroundMotions.apply(
              sourceToInputs.apply(source)));
    }
  }

  /*
   * CLUSTER: List<ClusterCurves> --> HazardCurveSet
   *
   * Reduce multiple cluster source curves.
   */
  static final class ClusterCurveConsolidator implements
      Function<List<ClusterCurves>, HazardCurveSet> {

    private final ClusterRuptureSet ruptures;
    private final GmmTree gmms;
    private final Map<Imt, XySequence> modelCurves;

    ClusterCurveConsolidator(
        ClusterRuptureSet ruptures,
        GmmTree gmms,
        CalcConfig config) {

      this.ruptures = ruptures;
      this.gmms = gmms;
      this.modelCurves = config.hazard.logModelCurves();
    }

    @Override
    public HazardCurveSet apply(List<ClusterCurves> curvesList) {

      if (curvesList.isEmpty()) {
        return HazardCurveSet.empty(ruptures);
      }

      HazardCurveSet.Builder curveSetBuilder = HazardCurveSet.builder(
          ruptures,
          gmms,
          modelCurves);

      for (ClusterCurves curves : curvesList) {
        curveSetBuilder.addCurves(curves);
      }
      return curveSetBuilder.build();
    }
  }

  /*
   * ALL: List<HazardCurveSet> --> HazardResult
   *
   * Final 'fan-in' consolidator function used for all source types.
   */
  static final class CurveSetConsolidator implements Function<List<HazardCurveSet>, Hazard> {

    private final HazardModel model;
    private final CalcConfig config;
    private final Site site;

    CurveSetConsolidator(
        HazardModel model,
        CalcConfig config,
        Site site) {

      this.model = model;
      this.config = config;
      this.site = site;
    }

    @Override
    public Hazard apply(List<HazardCurveSet> curveSetList) {

      Hazard.Builder resultBuilder = Hazard.builder(config)
          .model(model)
          .site(site);

      for (HazardCurveSet curveSet : curveSetList) {
        if (curveSet.isEmpty()) {
          continue;
        }
        resultBuilder.addCurveSet(curveSet);
      }
      return resultBuilder.build();
    }
  }

}
