package gov.usgs.earthquake.nshmp.data;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkElementIndex;
import static gov.usgs.earthquake.nshmp.Text.NEWLINE;
import static java.util.stream.Collectors.joining;

import java.util.Arrays;
import java.util.Iterator;
import java.util.Objects;
import java.util.stream.DoubleStream;

/**
 * An immutable, array-backed implementation of {@code XySequence}.
 *
 * @author U.S. Geological Survey
 */
class ArrayXySequence implements XySequence {

  final double[] xs;
  final double[] ys;

  ArrayXySequence(double[] xs, double[] ys) {
    this.xs = xs;
    this.ys = ys;
  }

  ArrayXySequence(XySequence sequence, boolean clear) {
    /*
     * This constructor provides the option to 'clear' (or zero-out) the
     * y-values when copying. In practice, it is only ever used when creating
     * mutable instances.
     *
     * The covariant cast below is safe as all current implementations descend
     * from this class.
     */
    ArrayXySequence s = (ArrayXySequence) sequence;
    xs = s.xs;
    ys = clear ? new double[xs.length] : Arrays.copyOf(s.ys, s.ys.length);
  }

  @Override
  public final double x(int index) {
    checkElementIndex(index, xs.length);
    return xUnchecked(index);
  }

  final double xUnchecked(int index) {
    return xs[index];
  }

  @Override
  public final double y(int index) {
    checkElementIndex(index, ys.length);
    return yUnchecked(index);
  }

  final double yUnchecked(int index) {
    return ys[index];
  }

  @Override
  public final int size() {
    return xs.length;
  }

  @Override
  public XyPoint min() {
    return new Point(0);
  }

  @Override
  public XyPoint max() {
    return new Point(size() - 1);
  }

  @Override
  public final boolean isClear() {
    return DoubleData.areZeroValued(ys);
  }

  @Override
  public XySequence trim() {
    return Sequences.trim(this);
  }

  @Override
  public final boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }
    if (!(obj instanceof ArrayXySequence)) {
      return false;
    }
    ArrayXySequence that = (ArrayXySequence) obj;
    return Arrays.equals(this.xs, that.xs) && Arrays.equals(this.ys, that.ys);
  }

  @Override
  public final int hashCode() {
    return Objects.hash(Arrays.hashCode(xs), Arrays.hashCode(ys));
  }

  @Override
  public DoubleStream xValues() {
    return Arrays.stream(xs);
  }

  @Override
  public DoubleStream yValues() {
    return Arrays.stream(ys);
  }

  /*
   * Check the x-value object references; if mismatched, compare values.
   */
  ArrayXySequence validateSequence(ArrayXySequence that) {
    checkArgument(Arrays.equals(this.xs, that.xs));
    return that;
  }

  @Override
  public String toString() {
    return new StringBuilder(getClass().getSimpleName())
        .append(":")
        .append(NEWLINE)
        .append(stream()
            .map(XyPoint::toString)
            .collect(joining(NEWLINE)))
        .toString();
  }

  class Point implements XyPoint {

    final int index;

    Point(int index) {
      this.index = index;
    }

    @Override
    public double x() {
      return xUnchecked(index);
    }

    @Override
    public double y() {
      return yUnchecked(index);
    }

    @Override
    public int index() {
      return index;
    }

    @Override
    public String toString() {
      return " [" + x() + ", " + y() + "]";
    }

  }

  @Override
  public Iterator<XyPoint> iterator() {
    return new XyIterator();
  }

  class XyIterator implements Iterator<XyPoint> {
    private int caret = 0;

    @Override
    public boolean hasNext() {
      return caret < size();
    }

    @Override
    public XyPoint next() {
      return new Point(caret++);
    }

    @Override
    public void remove() {
      throw new UnsupportedOperationException();
    }
  }

}
