package gov.usgs.earthquake.nshmp.data;

import java.util.Arrays;
import java.util.Collection;
import java.util.Optional;
import java.util.function.Consumer;

import com.google.common.primitives.Doubles;

/**
 * An {@code XySequence} with mutable y-values. Structural modifications are not
 * permitted, however, and {@code Iterator.remove()} throws an
 * {@code UnsupportedOperationException}.
 *
 * <p>Methods and classes in this package will always return a
 * {@code MutableXySequence} if the instance is, in fact, mutable. Mutable
 * sequences should not be considered thread safe.
 *
 * <p>The mutator methods ({@link #set(int, double)}, {@link #add(double)},
 * etc...) of this class return a reference to {@code this} sequence for use
 * inline. For x- and y-dependent value modifications, use
 * {@link Iterable#forEach(Consumer)} or {@link #transform(Consumer)}. The
 * latter returns a reference to {@code this} sequence for use inline.
 *
 * <p>All data supplied to the factory methods in this interface is copied
 * unless it is not necessary to do so. Users should consider
 * {@link #copyOf(XySequence)} where possible as x-values will never be
 * replicated in memory.
 *
 * @author U.S. Geological Survey
 * @see XySequence
 */
public interface MutableXySequence extends XySequence {

  /**
   * Create a new sequence with mutable y-values from the supplied value arrays.
   * If the optional y-value array is empty, y-values are initialized to 0.
   *
   * @param xs x-values to initialize sequence with
   * @param ys optional y-values to initialize sequence with
   * @throws IllegalArgumentException if {@code xs} is empty
   * @throws IllegalArgumentException if {@code xs} and {@code ys} are not the
   *         same size
   * @throws IllegalArgumentException if {@code xs} does not increase
   *         monotonically or contains repeated values
   */
  static MutableXySequence create(
      double[] xs,
      Optional<double[]> ys) {

    return construct(Arrays.copyOf(xs, xs.length), (ys.isPresent())
        ? Arrays.copyOf(ys.orElseThrow(), ys.orElseThrow().length)
        : new double[xs.length]);
  }

  /**
   * Create a new sequence with mutable y-values from the supplied value
   * collections. If the optional y-value collection is empty, y-values are
   * initialized to 0.
   *
   * @param xs x-values to initialize sequence with
   * @param ys optional y-values to initialize sequence with
   * @throws IllegalArgumentException if {@code xs} is empty
   * @throws IllegalArgumentException if {@code xs} and {@code ys} are not the
   *         same size
   * @throws IllegalArgumentException if {@code xs} does not increase
   *         monotonically or contains repeated values
   */
  static MutableXySequence create(
      Collection<? extends Number> xs,
      Optional<Collection<? extends Number>> ys) {

    return construct(Doubles.toArray(xs), (ys.isPresent())
        ? Doubles.toArray(ys.orElseThrow())
        : new double[xs.size()]);
  }

  /**
   * Create a mutable copy of the supplied sequence.
   *
   * @param sequence to copy
   */
  static MutableXySequence copyOf(XySequence sequence) {
    return new MutableArrayXySequence(sequence, false);
  }

  /**
   * Create a mutable copy of the supplied sequence with all y-values reset to
   * zero.
   *
   * @param sequence to copy
   * @return a mutable copy of the supplied sequence
   */
  static MutableXySequence emptyCopyOf(XySequence sequence) {
    return new MutableArrayXySequence(sequence, true);
  }

  /**
   * Sets the y-value at index.
   *
   * @param index of y-value to set
   * @param value to set
   * @return {@code this} sequence, for use inline
   * @throws IndexOutOfBoundsException if the index is out of range
   */
  MutableXySequence set(int index, double value);

  /**
   * Adds a {@code term} to the y-values of this sequence in place.
   *
   * @param term to add
   * @return {@code this} sequence, for use inline
   */
  MutableXySequence add(double term);

  /**
   * Adds the supplied y-values to the y-values of this sequence in place.
   *
   * @param ys y-values to add
   * @return {@code this} sequence, for use inline
   * @throws IllegalArgumentException if {@code ys.length != this.size()}
   */
  MutableXySequence add(double[] ys);

  /**
   * Adds the y-values of a sequence to the y-values of this sequence in place.
   *
   * @param sequence to add
   * @return {@code this} sequence, for use inline
   * @throws IllegalArgumentException if
   *         {@code sequence.xValues() != this.xValues()}
   */
  MutableXySequence add(XySequence sequence);

  /**
   * Multiplies ({@code scale}) the y-values of this sequence in place.
   *
   * @param scale factor
   * @return {@code this} sequence, for use inline
   */
  MutableXySequence multiply(double scale);

  /**
   * Multiplies the y-values of this sequence by the supplied y-values in place.
   *
   * @param ys y-values to add
   * @return {@code this} sequence, for use inline
   * @throws IllegalArgumentException if {@code ys.length != this.size()}
   */
  MutableXySequence multiply(double[] ys);

  /**
   * Multiplies the y-values of this sequence by the y-values of another
   * sequence in place.
   *
   * @param sequence to multiply {@code this} sequence by
   * @return {@code this} sequence, for use inline
   * @throws IllegalArgumentException if
   *         {@code sequence.xValues() != this.xValues()}
   */
  MutableXySequence multiply(XySequence sequence);

  /**
   * Sets the y-values of this sequence to their complement in place [
   * {@code 1 - y}]. Assumes this is a probability distribution limited to the
   * domain [0..1].
   *
   * @return {@code this} sequence, for use inline
   */
  MutableXySequence complement();

  /**
   * Sets all y-values to 0.
   *
   * @return {@code this} sequence, for use inline
   */
  MutableXySequence clear();

  @Override
  MutableXySequence trim();

  /**
   * Converts {@code this} to an immutable XySequence.
   */
  default XySequence toImmutable() {
    return XySequence.copyOf(this);
  };

  /**
   * Transforms all y-values of this sequence in place by applying the supplied
   * action to each {@link XyPoint}. Internally this method delegates to
   * {@link Iterable#forEach(Consumer)}.
   *
   * @param action to perform on each element
   * @return {@code this} sequence, for use inline
   */
  default MutableXySequence transform(Consumer<? super XyPoint> action) {
    forEach(action);
    return this;
  }

  /* Assumes array copies have already been created. */
  private static MutableXySequence construct(
      double[] xs,
      double[] ys) {
    Sequences.validateArrays(xs, ys);
    return new MutableArrayXySequence(xs, ys);
  }

}
