package gov.usgs.earthquake.nshmp.data;

import static com.google.common.base.Preconditions.checkArgument;
import static gov.usgs.earthquake.nshmp.data.DoubleData.areMonotonic;

import java.math.BigDecimal;
import java.util.Arrays;
import java.util.stream.IntStream;

import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.Maths;

/**
 * Static support methods for sequences.
 *
 * @author U.S. Geological Survey
 */
public class Sequences {

  private Sequences() {}

  /**
   * Check that the supplied arrays are not empty, are the same size, and that
   * the supplied xs increase monotonically.
   */
  static void validateArrays(double[] xs, double[] ys) {
    checkArgument(xs.length > 0, "x-values may not be empty");
    checkArgument(xs.length == ys.length, "x- and y-values are different sizes");
    if (xs.length > 1) {
      checkArgument(areMonotonic(true, true, xs), "x-values do not increase monotonically");
    }
  }

  /**
   * Trim points from the start or end of a sequence for which y-values are
   * zero. If no such leading or trailing points exist, sequence contains only a
   * single point, or all y-values are zero, method returns the supplied
   * sequence. If supplied sequence is mutable, returned sequence is mutable;
   * otherwise returned sequence is immutable.
   */
  static XySequence trim(ArrayXySequence xy) {
    if (xy.size() == 1 || xy.isClear()) {
      return xy;
    }
    if (xy.min().y() > 0.0 && xy.max().y() > 0.0) {
      return xy;
    }
    int iMin = firstNonZeroIndex(xy.ys);
    int iMax = lastNonZeroIndex(xy.ys) + 1;
    double[] xs = Arrays.copyOfRange(xy.xs, iMin, iMax);
    double[] ys = Arrays.copyOfRange(xy.ys, iMin, iMax);

    return (xy instanceof MutableArrayXySequence)
        ? new MutableArrayXySequence(xs, ys)
        : new ArrayXySequence(xs, ys);
  }

  /** Index of the first non-zero value (from start), or -1. */
  static int firstNonZeroIndex(double[] values) {
    for (int i = 0; i < values.length; i++) {
      if (values[i] != 0.0) {
        return i;
      }
    }
    return -1;
  }

  /** Index of the first non-zero value (from end), or -1. */
  static int lastNonZeroIndex(double[] values) {
    for (int i = values.length - 1; i >= 0; i--) {
      if (values[i] != 0.0) {
        return i;
      }
    }
    return -1;
  }

  /**
   * Computes the expected size of a sequence.
   *
   * This method rounds {@code (max-min)/Δ} to 1e-6 precision before casting to
   * an integer to determine the number of elements in the sequence. If
   * (max-min)/Δ is reasonably close to an integer, including 0, but max is off
   * by some double-precision error (e.g. 6.7999999999), the returned size would
   * be large enough to include a maximum value of 6.8.
   *
   * @param min the first value in the sequence
   * @param max the last value in the sequence
   * @param delta sequence step size
   * @throws IllegalArgumentException if (1) {@code min}, {@code max}, and
   *         {@code Δ} are not finite (2) {@code max < min}, or (3)
   *         {@code Δ <= 0}
   */
  public static int size(double min, double max, double delta) {
    DoubleData.checkFinite("Sequence minimum", min);
    DoubleData.checkFinite("Sequence maximum", max);
    DoubleData.checkFinite("Sequence Δ", delta);
    checkArgument(
        Maths.round(max, 6) >= Maths.round(min, 6),
        "Sequence max [%s] < min [%s]", max, min);
    checkArgument(delta > 0.0, "Sequence Δ [%s] < 0", delta);
    return ((int) Maths.round(((max - min) / delta), 6)) + 1;
  }

  /**
   * Create a new XySequence with reverse cumulative values.
   */
  public static XySequence toCumulative(XySequence incremental) {
    MutableXySequence cumulative = MutableXySequence.copyOf(incremental);
    double sum = 0.0;
    for (int i = incremental.size() - 1; i >= 0; i--) {
      sum += incremental.y(i);
      cumulative.set(i, sum);
    }
    return XySequence.copyOf(cumulative);
  }

  /**
   * Create a new sequential value array builder. If {@code ((max - min) / Δ)}
   * is reasonably close to an integer (i.e. the value may be off by some
   * expected double precision error), {@code max} will be included in the
   * sequence. If supplied {@code min} and {@code max} values are 'centered' and
   * {@code (max - min) < Δ}, or they are bin edges and not 'centered' and
   * {@code ((max - Δ/2) - (min + Δ/2)) < Δ}, the sequence will contain only one
   * value.
   *
   * @param min value for sequence; if sequence is 'centered', {@code min} will
   *        be the first value, otherwise first value will be {@code min + Δ/2}
   * @param max value for sequence; if sequence is 'centered', {@code max} will
   *        be the last value, otherwise last value will be {@code max - Δ/2}
   * @param delta sequence step size
   * @throws IllegalArgumentException if (1) {@code min}, {@code max}, and
   *         {@code Δ} are not finite (2) {@code max <= min}, or (3)
   *         {@code Δ <= 0}
   */
  public static ArrayBuilder arrayBuilder(double min, double max, double delta) {
    return new ArrayBuilder(min, max, delta);
  }

  /**
   * Sequential value array builder. Monotonically increasing or decreasing
   * arrays with uniformely spaced values are commonly used to mark the bin
   * centers of a histogram, for example the x-values or magnitudes in a
   * magnitude-frequency distribution (MFD).
   */
  public static class ArrayBuilder {

    final double min;
    final double max;
    final double Δ;

    boolean descending = false;
    boolean centered = false;
    int scale = -1;

    private ArrayBuilder(double min, double max, double Δ) {
      this.min = min;
      this.max = max;
      this.Δ = Δ;
    }

    /**
     * Indicates the supplied {@code min} and {@code max} are bin centers as
     * opposed to the lowermost and uppermost bin edges, respectively.
     *
     * @return this builder
     */
    public ArrayBuilder centered() {
      this.centered = true;
      return this;
    }

    /**
     * Indicates that this sequence should descend from {@code max} to
     * {@code min}.
     *
     * @return this builder
     */
    public ArrayBuilder descending() {
      this.descending = true;
      return this;
    }

    /**
     * Sets the precision (or {@link BigDecimal#scale()}) of the values in the
     * sequence.
     *
     * @param scale of values (precision)
     * @throws IllegalArgumentException if scale is not in the range
     *         {@code [0..8]}
     * @return this builder
     */
    public ArrayBuilder scale(int scale) {
      Indexing.checkInRange(Range.closed(0, 8), "scale", scale);
      this.scale = scale;
      return this;
    }

    /**
     * Create a new {@code double[]} containing a monotonically increasing or
     * decreasing sequence of values.
     */
    public double[] build() {
      double min = centered ? this.min : this.min + Δ / 2.0;
      double max = centered ? this.max : this.max - Δ / 2.0;
      int size = size(min, max, Δ);
      double[] sequence = descending
          ? DoubleData.flip(buildSequence(-max, Δ, size))
          : buildSequence(min, Δ, size);
      return (scale >= 0)
          ? DoubleData.round(scale, sequence)
          : sequence;
    }

    private static double[] buildSequence(double min, double Δ, int size) {
      checkArgument(Range.openClosed(0, 100000).contains(size), "sequence size");
      return IntStream.range(0, size)
          .mapToDouble(i -> min + Δ * i)
          .toArray();
    }
  }

}
