package gov.usgs.earthquake.nshmp.geo;

import static com.google.common.base.Preconditions.checkArgument;

import java.util.Arrays;
import java.util.Objects;

/**
 * A rectangular (in Mercator projection) bounding box specified by a lower-left
 * coordinate ({@link #min}) and an upper-right coordinate ({@link #max}).
 *
 * <p>Bounds are 2-dimensional in that the depth component of the corners will
 * always be zero. The bounds of any {@code Iterable<Location>} may be computed
 * using {@link Locations#bounds(Iterable)}.
 *
 * @author U.S. Geological Survey
 */
public final class Bounds {

  /** The lower left coordinate (minimum longitude and latatide). */
  public final Location min;

  /** The upper right coordinate (maximum longitude and latitude). */
  public final Location max;

  /**
   * Create a new {@code Bounds} object.
   *
   * @param minLat minimum latitude
   * @param minLon minimum longitude
   * @param maxLat maximum latitude
   * @param maxLon maximum longitude
   */
  public static Bounds create(
      double minLat,
      double minLon,
      double maxLat,
      double maxLon) {

    checkArgument(minLat < maxLat, "minLat >= maxLat");
    checkArgument(minLon < maxLon, "minLon >= maxLon");

    return new Bounds(
        Location.create(minLon, minLat),
        Location.create(maxLon, maxLat));
  }

  private Bounds(Location min, Location max) {
    this.min = min;
    this.max = max;
  }

  /**
   * Return this {@code Bounds} as a {@code LocationList} of five vertices,
   * starting with {@link #min} and winding counter-clockwise.
   */
  public LocationList toList() {
    return LocationList.of(
        min,
        Location.create(max.longitude, min.latitude),
        max,
        Location.create(min.longitude, max.latitude),
        min);
  }

  /**
   * Return the values of this {@code Bounds} object in the form
   * {@code [min.longitude, min.latitude, max.longitude, max.latitude]}.
   */
  public double[] toArray() {
    return new double[] {
        min.longitude,
        min.latitude,
        max.longitude,
        max.latitude
    };
  }

  @Override
  public boolean equals(Object obj) {
    if (obj == null) {
      return false;
    }
    if (!(obj instanceof Bounds)) {
      return false;
    }
    Bounds other = (Bounds) obj;
    return min.equals(other.min) && max.equals(other.max);
  }

  @Override
  public int hashCode() {
    return Objects.hash(min, max);
  }

  /**
   * Returns the string representation of {@link #toArray()}.
   */
  @Override
  public String toString() {
    return Arrays.toString(toArray());
  }

}
