package gov.usgs.earthquake.nshmp.geo;

import static gov.usgs.earthquake.nshmp.Earthquakes.checkDepth;
import static gov.usgs.earthquake.nshmp.geo.Coordinates.checkLatitude;
import static gov.usgs.earthquake.nshmp.geo.Coordinates.checkLongitude;
import static java.lang.Double.valueOf;

import java.util.Objects;

import gov.usgs.earthquake.nshmp.Maths;

/**
 * A location represents a point with reference to the earth's ellipsoid. It is
 * expressed in terms of longitude and latitude in decimal degrees, and depth in
 * km. Location depth is always positive-down, consistent with seismological
 * convention. Locations may be defined using longitude values in the range:
 * (-360°..360°). Location instances are immutable.
 *
 * <p>Note that constructors and static factory methods take arguments in the
 * order: {@code [lon, lat, depth]}, which is consistent with {@code String}
 * representations used in KML, GeoJSON, and other digital coordinate formats
 * that match standard plotting coordinate order: {@code [x, y, z]}.
 *
 * @author U.S. Geological Survey
 */
public final class Location implements Comparable<Location> {

  /*
   * Developer notes:
   *
   * Historically, this class has at different times stored lon-lat values in
   * decimal degrees or radians and then converted as needed for math operations
   * and string representations. However, such conversions can be lossy due to
   * double-precision operations. We now preserve both the original decimal
   * degree values used to initialize the location along with radian formatted
   * values for computational efficiency within the geo package.
   *
   * Serialization: Most serialization of locations moving forward will be
   * to/from GeoJSON coordinate arrays. The Location.fromString() method and
   * GeoJSON utilities in this package ensure that the radian based values,
   * which are not serialized, are recreated as needed. If 'standard' object
   * deserialization is required, users should implement a deserilaizer that
   * calls a constructor.
   */

  /** The latitude of this location in decimal degrees. */
  public final double latitude;

  /** The longitude of this location in decimal degrees. */
  public final double longitude;

  /** The depth of this location in kilometers. */
  public final double depth;

  final double latRad;
  final double lonRad;

  private Location(double longitude, double latitude, double depth) {
    this.latitude = checkLatitude(latitude);
    this.longitude = checkLongitude(longitude);
    this.depth = checkDepth(depth);
    this.lonRad = longitude * Maths.TO_RADIANS;
    this.latRad = latitude * Maths.TO_RADIANS;
  }

  /**
   * Create a new location from the supplied longitude and latitude with a depth
   * of 0 km.
   *
   * @param longitude in decimal degrees
   * @param latitude in decimal degrees
   * @throws IllegalArgumentException if any supplied values are out of range
   * @see Coordinates
   */
  public static Location create(double longitude, double latitude) {
    return create(longitude, latitude, 0);
  }

  /**
   * Create a new location with the supplied longitude, latitude, and depth.
   *
   * @param longitude in decimal degrees
   * @param latitude in decimal degrees
   * @param depth in km (positive down)
   * @throws IllegalArgumentException if any supplied values are out of range
   * @see Coordinates
   */
  public static Location create(double longitude, double latitude, double depth) {
    return new Location(longitude, latitude, depth);
  }

  /**
   * Create a new location from the comma-separated values in the supplied
   * string. Method performs the reverse of {@link #toString()}.
   *
   * @param s string to parse
   * @throws NumberFormatException if any values in {@code s} are unparseable
   * @throws IndexOutOfBoundsException if {@code s} contains fewer than 3
   *         comma-separated values; any additional values in the supplied
   *         string are ignored
   * @see #toString()
   */
  public static Location fromString(String s) {
    String[] v = s.trim().split(",");
    return create(valueOf(v[0]), valueOf(v[1]), valueOf(v[2]));
  }

  /**
   * Return a KML and GeoJSON compatible string tuple:
   * {@code "longitude,latitude,depth"} (no spaces).
   *
   * @see #fromString(String)
   */
  @Override
  public String toString() {
    return String.format("%s,%s,%s", longitude, latitude, depth);
  }

  /**
   * Returns {@code true} if this location is equal to the supplied object,
   * {@code false} otherwise.
   */
  @Override
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj == null) {
      return false;
    }
    if (!(obj instanceof Location)) {
      return false;
    }
    Location loc = (Location) obj;
    return this.longitude == loc.longitude &&
        this.latitude == loc.latitude &&
        this.depth == loc.depth;
  }

  /**
   * Returns a hash code based on the longitude, latitude, and depth of this
   * location.
   */
  @Override
  public int hashCode() {
    return Objects.hash(longitude, latitude, depth);
  }

  /**
   * Compare this location to another sorting first by latitude then longitude.
   * When sorting a list of locations, the resultant ordering is left-to-right,
   * bottom-to-top.
   *
   * @param location to compare
   * @return a negative integer, zero, or a positive integer if this location is
   *         less than, equal to, or greater than the supplied {@code location}.
   */
  @Override
  public int compareTo(Location location) {
    double d = (latitude == location.latitude)
        ? longitude - location.longitude
        : latitude - location.latitude;
    return (d != 0) ? (d < 0) ? -1 : 1 : 0;
  }
}
