package gov.usgs.earthquake.nshmp.geo.json;

import static com.google.common.base.CaseFormat.LOWER_HYPHEN;
import static com.google.common.base.CaseFormat.UPPER_UNDERSCORE;
import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;

import java.lang.reflect.Type;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Optional;
import java.util.OptionalDouble;
import java.util.OptionalInt;

import com.google.common.base.CaseFormat;
import com.google.common.base.Converter;
import com.google.common.collect.BoundType;
import com.google.common.collect.Range;
import com.google.gson.JsonElement;
import com.google.gson.reflect.TypeToken;

/**
 * GeoJSON properties helper class.
 *
 * <p>GeoJSON properties serialize to/from a {@code Map<String, Object>}. To
 * simplify repeated property map creation where only one property might be
 * changing, this class provides a reusable builder. The class also provides
 * methods to help get and set <a
 * href="https://github.com/mapbox/simplestyle-spec/tree/master/1.1.0">
 * simplestyle-spec</a> (v1.1.0) properties that may be considered when
 * rendering GeoJSON.
 *
 * <p>If a property map contains arrays or other nested objects as values, use
 * {@link #get(String, Class)} or {@link #get(String, Type)}.
 *
 * @author U.S. Geological Survey
 */
public final class Properties {

  private Map<String, JsonElement> source;

  Properties(Map<String, JsonElement> source) {
    this.source = source;
  }

  /**
   * Return an {@code Optional} describing the {@code JsonElement} for the
   * specified key, or an empty {@code Optional} if the key is absent. The
   * returned {@code JsonElement} may be a {@code JsonNull}.
   *
   * @param key to get value for
   */
  public Optional<JsonElement> getJsonElement(String key) {
    checkNotNull(key);
    return Optional.ofNullable(source.get(key));
  }

  /**
   * Return an {@code Optional} describing a value of the specified class for
   * the specified key, or an empty {@code Optional} if the key is absent, its
   * value is {@code null} or its value cannot be interpreted as an object of
   * class {@code T}. Use this method for non-generic objects.
   *
   * @param <T> the type of the desired object
   * @param key to get value for
   * @param classOfT the class of T
   */
  public <T> Optional<T> get(String key, Class<T> classOfT) {
    try {
      JsonElement element = getJsonElement(key).orElseThrow();
      T obj = GeoJson.GSON_DEFAULT.fromJson(element, classOfT);
      return Optional.of(obj);
    } catch (Exception e) {
      return Optional.empty();
    }
  }

  /**
   * Return an {@code Optional} describing a value of the specified type for the
   * specified key, or an empty {@code Optional} if the key is absent, its value
   * is {@code null} or its value cannot be interpreted as an object of type
   * {@code T}. Use this method for generic objects.
   *
   * @param <T> the type of the desired object
   * @param key to get value for
   * @param typeOfT the genericized type of src. You can obtain this type by
   *        using the {@link TypeToken} class. For example, to get the type for
   *        {@code Collection<Foo>}, you should use:
   *
   *        <pre>
   *        Type typeOfT = new TypeToken&lt;Collection&lt;Foo&gt;&gt;() {}.getType();
   *        </pre>
   */
  public <T> Optional<T> get(String key, Type typeOfT) {
    try {
      JsonElement element = getJsonElement(key).orElseThrow();
      T obj = GeoJson.GSON_DEFAULT.fromJson(element, typeOfT);
      return Optional.of(obj);
    } catch (Exception e) {
      return Optional.empty();
    }
  }

  /**
   * Return an {@code Optional} describing the value for the specified key, or
   * an empty {@code Optional} if the key is absent, its value is {@code null}
   * or its value cannot be interpreted as a {@code boolean}.
   *
   * @param key to get value for
   */
  public Optional<Boolean> getBoolean(String key) {
    try {
      return Optional.of(getJsonElement(key).orElseThrow().getAsBoolean());
    } catch (Exception e) {
      return Optional.empty();
    }
  }

  /**
   * Return an {@code OptionalInt} describing the value for the specified key,
   * or an empty {@code OptionalInt} if the key is absent, its value is
   * {@code null} or its value cannot be interpreted as an {@code int}.
   *
   * @param key to get value for
   */
  public OptionalInt getInt(String key) {
    try {
      return OptionalInt.of(getJsonElement(key).orElseThrow().getAsInt());
    } catch (Exception e) {
      return OptionalInt.empty();
    }
  }

  /**
   * Return an {@code OptionalDouble} describing the value for the specified
   * key, or an empty {@code OptionalDouble} if the key is absent, its value is
   * {@code null} or its value cannot be interpreted as a {@code double}.
   *
   * @param key to get value for
   */
  public OptionalDouble getDouble(String key) {
    try {
      return OptionalDouble.of(getJsonElement(key).orElseThrow().getAsDouble());
    } catch (Exception e) {
      return OptionalDouble.empty();
    }
  }

  /**
   * Return an {@code Optional} describing the value for the specified key, or
   * an empty {@code Optional} if the key is absent, its value is {@code null}
   * or its value cannot be interpreted as a {@code String}.
   *
   * @param key to get value for
   */
  public Optional<String> getString(String key) {
    try {
      return Optional.of(getJsonElement(key).orElseThrow().getAsString());
    } catch (Exception e) {
      return Optional.empty();
    }
  }

  /**
   * Return the backing map.
   */
  public Map<String, JsonElement> map() {
    return new LinkedHashMap<>(source);
  }

  @Override
  public String toString() {
    return "Properties: " + source.toString();
  }

  /**
   * Return a reusable property map builder that preserves the order in which
   * properties are added. Note that {@code build()} returns a copy of the
   * internal map instance.
   */
  public static Builder builder() {
    return new Builder();
  }

  /**
   * Return a reusable property map builder initialized with the properties of
   * the supplied feature.
   *
   * @param feature from which to copy properties
   */
  public static Builder fromFeature(Feature feature) {
    /* Note: properties are copied into builder as JsonElements. */
    Builder builder = builder();
    builder.map.putAll(feature.properties().source);
    return builder;
  }

  /**
   * Reusable property map builder. Although the internal map is copied on
   * {@code build()}, the objects in the map may be altered if they are
   * themselves mutable. A property map does not permit {@code null} values or
   * the {@code null} key.
   *
   * @see Properties#builder()
   */
  public static class Builder {

    /*
     * Insertion order preserving map. On build(), we create a copy so that in
     * the unlikely event that a previous property map is hanging around and
     * hasn't been written it won't be modified.
     */
    private final Map<String, Object> map = new LinkedHashMap<>();

    private Builder() {}

    /**
     * Add a property to this builder.
     *
     * @param key for value
     * @param value for key; may be {@code null}
     * @return this builder
     */
    public Builder put(String key, Object value) {
      map.put(checkNotNull(key), value);
      return this;
    }

    /**
     * Remove a property from this builder.
     *
     * @param key of entry to remove
     * @return this builder
     */
    public Builder remove(String key) {
      map.remove(checkNotNull(key));
      return this;
    }

    /**
     * Add a SimpleStyle property to this builder.
     *
     * @param key for value
     * @param value for key
     * @return this builder
     */
    public Builder put(Style key, Object value) {
      return put(checkNotNull(key).toString(), checkStyle(key, value));
    }

    /**
     * Return a mutable copy of the internal builder map.
     */
    public Map<String, Object> build() {
      return new LinkedHashMap<>(map);
    }
  }

  /**
   * Identifiers for <a
   * href="https://github.com/mapbox/simplestyle-spec/tree/master/1.1.0">
   * simplestyle-spec</a> (v1.1.0) properties. These properties are rendering
   * hints that may improve the appearance of GeoJSON features in some rendering
   * environments. The description for each identifier indicates the expected
   * value type when adding style-keyed properties.
   *
   * <p><a name="color-rules"/><b>Color rules:</b><ul><li>Colors can be in short
   * form: {@code "#ace"}<li>Or long form: {@code "#aaccee"}<li>With or without
   * the prefix: {@code #}<li>Colors are interpreted the same as in CSS...<li>In
   * either {@code #RRGGBB} or {@code #RGB} order<li>Other color formats or
   * named colors are not supported</ul>
   */
  public enum Style {

    /**
     * Value: {@code String}
     */
    DESCRIPTION,

    /**
     * Value: color {@code String} (see <a href="#color-rules">color rules</a>)
     */
    FILL,

    /**
     * Value: {@code 0.0 ≤ double ≤ 1.0}
     */
    FILL_OPACITY,

    /**
     * Value: color {@code String} (see <a href="#color-rules">color rules</a>)
     */
    MARKER_COLOR,

    /**
     * Value: {@code ["small", "medium, "large"]}
     */
    MARKER_SIZE,

    /**
     * Value: <a href="https://www.mapbox.com/maki-icons/">named icon</a>
     * {@code String}, {@code 0 ≤ int ≤ 9}, or any lowercase character
     * {@code String}, {@code a-z}
     */
    MARKER_SYMBOL,

    /**
     * Value: color {@code String} (see <a href="#color-rules">color rules</a>)
     */
    STROKE,

    /**
     * Value: {@code 0.0 ≤ double ≤ 1.0}
     */
    STROKE_OPACITY,

    /**
     * Value: {@code 0.0 ≤ double}
     */
    STROKE_WIDTH,

    /**
     * Value: {@code String}
     */
    TITLE;

    private static final Converter<String, String> CONVERTER =
        UPPER_UNDERSCORE.converterTo(LOWER_HYPHEN);

    /**
     * Returns the {@link #name()} of this identifier converted to
     * {@link CaseFormat#LOWER_HYPHEN}.
     */
    @Override
    public String toString() {
      return checkNotNull(CONVERTER.convert(this.name()));
    }

  }

  static Object checkStyle(Style key, Object value) {
    checkNotNull(key);
    checkNotNull(value);

    switch (key) {
      case DESCRIPTION:
      case TITLE:
        return checkStyleString(key, value);
      case FILL:
      case MARKER_COLOR:
      case STROKE:
        return checkStyleColor(key, value);
      case MARKER_SIZE:
        return checkStyleMarkerSize(key, value);
      case FILL_OPACITY:
      case STROKE_OPACITY:
        return checkStyleDoubleRange(key, value, Range.closed(0.0, 1.0));
      case MARKER_SYMBOL:
        return checkStyleSymbol(key, value);
      case STROKE_WIDTH:
        return checkStyleDoubleRange(key, value, Range.downTo(0.0, BoundType.CLOSED));
      default:
        throw new IllegalArgumentException(
            String.format("Style [%s] not supported", key.toString()));
    }
  }

  private static Object checkStyleColor(Style key, Object value) {
    checkStyleString(key, value);
    checkArgument(
        ((String) value).matches("^#?(([0-9a-fA-F]{2}){3}|([0-9a-fA-F]){3})$"),
        "Color [%s] must be in short or long HEX form", value);
    return value;
  }

  private static Object checkStyleDoubleRange(Style key, Object value, Range<Double> range) {
    checkArgument(
        value instanceof Double,
        "Style [%s] must be a double",
        key);
    checkArgument(
        range.contains((double) value),
        String.format("Value [%f] out of range: %s", (double) value, range.toString()));
    return value;
  }

  private static Object checkStyleMarkerSize(Style key, Object value) {
    checkStyleString(key, value);
    checkArgument(
        value.equals("small") || value.equals("medium") || value.equals("large"),
        "Marker size [%s] not allowed",
        value);
    return value;
  }

  private static Object checkStyleString(Style key, Object value) {
    checkArgument(
        value instanceof String,
        "Style [%s] must be a string",
        key);
    return value;
  }

  private static Object checkStyleSymbol(Style key, Object value) {
    if (value instanceof String) {
      var symbol = ((String) value).trim();
      checkArgument(!symbol.isEmpty());
      if (symbol.length() == 1) {
        checkArgument(symbol.matches("[a-z]"));
      }
    } else if (value instanceof Integer) {
      checkArgument(
          Range.closed(0, 9).contains((int) value),
          "Value [%d] not in range: 0 ≤ int ≤ 9",
          value);
    } else {
      throw new IllegalArgumentException("Marker symbol must be a string or int");
    }
    return value;
  }
}
