package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.Maths.hypot;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.AM_09_INTERFACE;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.AM_09_INTERFACE_BASIN;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.AM_09_INTERFACE_BASIN_CONUS_2018;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.AM_09_INTERFACE_BASIN_M9;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.AM_09_INTERFACE_CONUS_2014;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.BA_08_BASE;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.CB_14_BASE;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RRUP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static gov.usgs.earthquake.nshmp.gmm.GmmUtils.BASE_10_TO_E;
import static gov.usgs.earthquake.nshmp.gmm.GmmUtils.LN_G_CM_TO_M;
import static gov.usgs.earthquake.nshmp.gmm.Imt.PGA;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P01;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P02;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P03;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P05;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P075;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P1;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P15;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P2;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P25;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P3;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA1P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA1P5;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA2P0;
import static java.lang.Math.exp;
import static java.lang.Math.log;
import static java.lang.Math.log10;

import java.util.Map;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Maps;
import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Implementation of the subduction interface ground motion model by Atkinson
 * &amp; Macias (2009). This implementation matches that used in the 2014 USGS
 * NSHMP.
 *
 * <p><b>Note:</b> Direct instantiation of {@code GroundMotionModel}s is
 * prohibited. Use {@link Gmm#instance(Imt)} to retrieve an instance for a
 * desired {@link Imt}.
 *
 * <p><b>Implementation notes:</b><ul>
 *
 * <li>NSHM fortran implementation converts 0.13Hz to 7.7s; this implementation
 * uses 7.5s instead.
 *
 * <li>Model uses a magnitude dependent depth term and so does not impose 20km
 * hypocentral depth as other subduction interface models do.
 *
 * <li>Support for spectral period 0.01s is provided using the same coefficients
 * as PGA.
 *
 * <li>Support for spectral periods 0.02s, 0.03s, 0.075s, 0.15s, 0.25s, and 1.5s
 * is provided via interpolation of ground motion and sigma of adjacent periods
 * for which there are coefficients.
 *
 * <li>Support for basin effects is provided per documentation of the 2018
 * Conterminous U.S. NSHM update (<a
 * href="https://doi.org/10.1177%2F8755293019878199">Petersen et al., 2020</a>).
 * Implementations include USGS developed depth tapering and restriction to long
 * periods ({@code T &geq; 1 s}), as well as a Seattle-specific basin
 * implementation that makes further adjustments based on M9 simulations for
 * interface events at sites over the deepest depths ({@code Z2.5 > 6 km}) and
 * at long periods ({@code T &qeq; 2 s}). These are region-specific behaviors
 * and are managed through the use of a gmm-region defined in the
 * {@code data/site} directory of recent {@code nshm-conus} models.</ul>
 *
 * <p><b>Reference:</b> Atkinson, G.M. and Macias, D.M., 2009, Predicted ground
 * motions for great interface earthquakes in the Cascadia subduction zone:
 * Bulletin of the Seismological Society of America, v. 99, p. 1552-1578.
 *
 * <p><b>doi:</b><a href="http://dx.doi.org/10.1785/0120080147">
 * 10.1785/0120080147</a>
 *
 * <p><b>Component:</b> geometric mean of two horizontal components
 *
 * @author U.S. Geological Survey
 * @see Gmm#AM_09_INTERFACE
 * @see Gmm#AM_09_INTERFACE_BASIN
 * @see Gmm#AM_09_INTERFACE_BASIN_M9
 * @see Gmm#AM_09_INTERFACE_BASIN_CONUS_2018
 * @see Gmm#AM_09_INTERFACE_CONUS_2014
 */
public class AtkinsonMacias_2009 implements GroundMotionModel {

  /*
   * 0.75s interpolated period coefficients added that should be removed if a
   * viable on-the-fly interpolation algorithm is added.
   */
  static final String NAME = "Atkinson & Macias (2009) : Interface";

  static final Constraints CONSTRAINTS = Constraints.builder()
      .set(MW, Range.closed(5.0, 9.5))
      .set(RRUP, Range.closed(0.0, 1000.0))
      .set(VS30, Range.closed(150.0, 1500.0))
      .build();

  static final CoefficientContainer COEFFS = new CoefficientContainer("AM09.csv");

  private static final Map<Imt, Range<Imt>> INTERPOLATED_IMTS = Maps.immutableEnumMap(
      ImmutableMap.<Imt, Range<Imt>> builder()
          .put(SA0P02, Range.closed(SA0P01, SA0P05))
          .put(SA0P03, Range.closed(SA0P01, SA0P05))
          .put(SA0P075, Range.closed(SA0P05, SA0P1))
          .put(SA0P15, Range.closed(SA0P1, SA0P2))
          .put(SA0P25, Range.closed(SA0P2, SA0P3))
          .put(SA1P5, Range.closed(SA1P0, SA2P0))
          .build());

  private static final class Coefficients {

    final Imt imt;
    final double c0, c1, c2, c3, c4, σ;

    Coefficients(Imt imt, CoefficientContainer cc) {
      this.imt = imt;
      Map<String, Double> coeffs = cc.get(imt);
      c0 = coeffs.get("c0");
      c1 = coeffs.get("c1");
      c2 = coeffs.get("c2");
      c3 = coeffs.get("c3");
      c4 = coeffs.get("c4");
      σ = coeffs.get("sig");
    }
  }

  private final Coefficients coeffs;
  private final Coefficients coeffsPGA;
  private final BooreAtkinson_2008 siteAmp;
  private final CampbellBozorgnia_2014 cb14;

  // interpolatedGmm = null if !interpolated
  private final boolean interpolated;
  private final GroundMotionModel interpolatedGmm;
  private final Gmm subtype;

  AtkinsonMacias_2009(Imt imt) {
    this(imt, AM_09_INTERFACE);
  }

  @Override
  public Imt imt() {
    return coeffs.imt;
  }

  AtkinsonMacias_2009(Imt imt, Gmm subtype) {
    this.subtype = subtype;
    coeffs = new Coefficients(imt, COEFFS);
    coeffsPGA = new Coefficients(PGA, COEFFS);
    siteAmp = (BooreAtkinson_2008) BA_08_BASE.instance(imt);
    cb14 = (CampbellBozorgnia_2014) CB_14_BASE.instance(imt);
    interpolated = INTERPOLATED_IMTS.containsKey(imt);
    interpolatedGmm = interpolated
        ? new InterpolatedGmm(subtype, imt, INTERPOLATED_IMTS.get(imt))
        : null;
  }

  /* Use USGS deep basin, long period amplification model. */
  boolean usgsBasin() {
    return false;
  }

  /* Apply Seattle M9 adjustment. */
  boolean m9() {
    return false;
  }

  /* Maintain natural log μPga for site term. */
  boolean lnPgaRock() {
    return false;
  }

  @Override
  public final LogicTree<GroundMotion> calc(GmmInput in) {

    if (interpolated) {
      return interpolatedGmm.calc(in);
    }

    if (coeffs.imt == Imt.PGV) {
      GroundMotion pgv = UsgsPgvSupport.calcAB20Pgv(subtype, in);
      return GroundMotions.createTree(pgv.mean(), pgv.sigma());
    }

    double σ = coeffs.σ * BASE_10_TO_E;
    double μRef = calcMean(coeffs, in);
    double μPga = calcMean(coeffsPGA, in);
    double site = siteAmp.siteAmp(lnPgaRock() ? μPga : exp(μPga), in.vs30);
    double μAm = μRef + site;

    /* Add (possibly depth-tapered) CB14 deep basin term or M9 adjustment. */
    double fb = 0.0;
    if (usgsBasin()) {
      fb = cb14.deepBasinScaling(in.z2p5);
      if (m9() && in.z2p5 > 6.0 && (coeffs.imt.isSA() && coeffs.imt.period() > 1.9)) {
        fb = log(2.0); // M9 scaling
      }
    }
    μAm += fb;

    return GroundMotions.createTree(μAm, σ);
  }

  private static final double calcMean(Coefficients c, GmmInput in) {
    double Mw = in.Mw;
    double h = (Mw * Mw) - (3.1 * Mw) - 14.55;
    double dM = Mw - 8.0;
    double gnd = c.c0 + (c.c3 * dM) + (c.c4 * dM * dM);
    double r = hypot(in.rRup, h);
    gnd += c.c1 * log10(r) + c.c2 * r;
    return gnd * BASE_10_TO_E - LN_G_CM_TO_M;
  }

  /*
   * Developer note: In most GMMs, subtype constructors, if present, need only
   * the IMT argument to initialize their parent. To support several
   * interpolated spectral periods, the parent also needs to know the specific
   * subtype Gmm identifier in order to obtain concrete instances of the
   * bounding spectral periods.
   */

  static class Basin extends AtkinsonMacias_2009 {
    static final String NAME = AtkinsonMacias_2009.NAME + " : Basin";

    Basin(Imt imt) {
      super(imt, AM_09_INTERFACE_BASIN);
    }

    protected Basin(Imt imt, Gmm subtype) {
      super(imt, subtype);
    }

    @Override
    final boolean usgsBasin() {
      return true;
    }
  }

  static final class BasinM9 extends Basin {
    static final String NAME = Basin.NAME + " : M9";

    BasinM9(Imt imt) {
      super(imt, AM_09_INTERFACE_BASIN_M9);
    }

    @Override
    final boolean m9() {
      return true;
    }
  }

  /*
   * Developer note (10/2024): The following two implementations maintain an
   * error where the natural log of the PGA rock reference value is passed to
   * the BA08 site term, but should be exp(μPga). The published 2018 CONUS model
   * (5.2) has been updated to use the AM_09_INTERFACE_BASIN_CONUS_2018. The
   * updated model (5.3) uses the corrected AM_09_INTERFACE_BASIN
   * implementation. Although the *_CONUS_2018 implementation could also be used
   * for 2014 because there will never be any z2p5 values supplied (no basin
   * term would ever be calculated), for clarity we've added
   * AM_09_INTERFACE_CONUS_2014 as well.
   */

  static final class BasinConus2018 extends Basin {
    static final String NAME = Basin.NAME + " (conus-2018)";

    BasinConus2018(Imt imt) {
      super(imt, AM_09_INTERFACE_BASIN_CONUS_2018);
    }

    @Override
    final boolean lnPgaRock() {
      return true;
    }
  }

  static final class Conus2014 extends AtkinsonMacias_2009 {
    static final String NAME = AtkinsonMacias_2009.NAME + " (conus-2014)";

    Conus2014(Imt imt) {
      super(imt, AM_09_INTERFACE_CONUS_2014);
    }

    @Override
    final boolean lnPgaRock() {
      return true;
    }
  }

}
