package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RAKE;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RJB;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static java.lang.Math.log;
import static java.lang.Math.sqrt;

import java.util.Map;

import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.Faults;
import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Alaska-specific implementation of the Boore, Joyner &amp; Fumal (1997) ground
 * motion model for shallow earthquakes in active continental crust. This
 * implementation was created to match fortran implementation with custom
 * coefficients for different tectonic settings and styles of faulting. It's not
 * clear what the relationship between the published 'b1' coefficients is and
 * the values included in the fortran input files. This implementation is only
 * for soft rock sites (Vs30 = 760 m/s).
 *
 * <p><b>Component:</b> random horizontal
 *
 * @author U.S. Geological Survey
 * @see Gmm#BJF_97_AK
 */
public class BooreEtAl_1997_Alaska implements GroundMotionModel {

  static final String NAME = "Boore et al. (1997) : Alaska";

  static final Constraints CONSTRAINTS = Constraints.builder()
      .set(MW, Range.closed(4.0, 8.0))
      .set(RJB, Range.closed(0.0, 80.0))
      .set(RAKE, Faults.RAKE_RANGE)
      .set(VS30, Range.singleton(760.0))
      .build();

  static final CoefficientContainer COEFFS = new CoefficientContainer("BJF97-AK.csv");

  private static final class Coefficients {

    final Imt imt;
    final double b1, b1mt, b2, b3, b5, h, σ;

    Coefficients(Imt imt, CoefficientContainer cc) {
      this.imt = imt;
      Map<String, Double> coeffs = cc.get(imt);
      b1 = coeffs.get("b1");
      b1mt = coeffs.get("b1mt");
      b2 = coeffs.get("b2");
      b3 = coeffs.get("b3");
      b5 = coeffs.get("b5");
      h = coeffs.get("h");
      σ = coeffs.get("sigma");
    }
  }

  private final Coefficients coeffs;

  BooreEtAl_1997_Alaska(Imt imt) {
    coeffs = new Coefficients(imt, COEFFS);
  }

  boolean interfaceZone() {
    return false;
  }

  @Override
  public Imt imt() {
    return coeffs.imt;
  }

  @Override
  public LogicTree<GroundMotion> calc(GmmInput in) {
    double μ = calcMean(coeffs, in);
    return GroundMotions.createTree(μ, coeffs.σ);
  }

  private final double calcMean(Coefficients c, GmmInput in) {
    double b1 = interfaceZone() ? c.b1mt : c.b1;
    double mFac = in.Mw - 6.0;
    double r = sqrt(in.rJB * in.rJB + c.h * c.h);
    double μ = b1 + c.b2 * mFac + c.b3 * mFac * mFac + c.b5 * log(r);
    return μ;
  }

  static final class InterfaceZone extends BooreEtAl_1997_Alaska {

    static final String NAME = BooreEtAl_1997_Alaska.NAME + " : Interface";

    InterfaceZone(Imt imt) {
      super(imt);
    }

    @Override
    boolean interfaceZone() {
      return true;
    }
  }

}
