package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RAKE;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RRUP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static java.lang.Math.exp;
import static java.lang.Math.log;
import static java.lang.Math.max;
import static java.lang.Math.sqrt;

import java.util.Map;

import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.Faults;
import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Alaska-specific implementation of the Boore, Joyner &amp; Fumal (1997) ground
 * motion model for shallow earthquakes in active continental crust. This
 * implementation was created to match fortran implementation with custom
 * coefficients such as 'csite'.
 *
 * <p><b>Component:</b> average horizontal
 *
 * @author U.S. Geological Survey
 * @see Gmm#CB_03
 */
public class CampbellBozorgnia_2003_Alaska implements GroundMotionModel {

  static final String NAME = "Campbell & Bozorgnia (2003) : Alaska";

  static final Constraints CONSTRAINTS = Constraints.builder()
      .set(MW, Range.closed(5.8, 8.0))
      .set(RRUP, Range.closed(0.0, 300.0))
      .set(RAKE, Faults.RAKE_RANGE)
      .set(VS30, Range.singleton(760.0))
      .build();

  static final CoefficientContainer COEFFS;

  static {
    COEFFS = new CoefficientContainer("CB03-AK.csv");
  }

  private static final class Coefficients {

    final Imt imt;
    final double c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, c11, csite, c15, c16;

    Coefficients(Imt imt, CoefficientContainer cc) {
      this.imt = imt;
      Map<String, Double> coeffs = cc.get(imt);
      c1 = coeffs.get("c1");
      c2 = coeffs.get("c2");
      c3 = coeffs.get("c3");
      c4 = coeffs.get("c4");
      c5 = coeffs.get("c5");
      c6 = coeffs.get("c6");
      c7 = coeffs.get("c7");
      c8 = coeffs.get("c8");
      c9 = coeffs.get("c9");
      c10 = coeffs.get("c10");
      c11 = coeffs.get("c11");
      csite = coeffs.get("csite");
      c15 = coeffs.get("c15");
      c16 = coeffs.get("c16");
    }
  }

  private final Coefficients coeffs;
  private final double gS;

  CampbellBozorgnia_2003_Alaska(Imt imt) {
    coeffs = new Coefficients(imt, COEFFS);
    gS = coeffs.c5 + 0.5 * coeffs.c6 + 0.5 * coeffs.c7;
  }

  @Override
  public Imt imt() {
    return coeffs.imt;
  }

  @Override
  public LogicTree<GroundMotion> calc(GmmInput in) {
    double μ = calcMean(coeffs, in);
    double σ = calcStdDev(coeffs.c16, in.Mw);
    return GroundMotions.createTree(μ, σ);
  }

  double calcMean(
      Coefficients c,
      GmmInput in) {

    double Mw = in.Mw;

    // magnitude scaling
    double mTerm = (8.5 - Mw) * (8.5 - Mw);
    double f1 = c.c2 * Mw + c.c3 * mTerm;

    // distance scaling - per verbal communication with Ken Campbell it
    // is ok to approximate rSeis by constraining rRup to be >= 3.0 km.
    // (see also Campbell 1997)
    double rSeis = max(in.rRup, 3.0);
    double rTerm = gS * exp(c.c8 * Mw + c.c9 * mTerm);
    double f2 = rSeis * rSeis + rTerm * rTerm;

    // // focal mechanism
    // FaultStyle style = GmmUtils.rakeToFaultStyle_NSHMP(in.rake);
    // double f3 = (style == REVERSE) ? revStyleTerm : (style == UNKNOWN) ?
    // unkStyleTerm : 0.0;

    // far-source effect of local site conditions
    // (pre-calculated)

    // // hanging-wall effect
    // double HW = (in.rJB < 5.0 && in.dip <= 70.0) ? (5.0 - in.rJB) / 5.0 :
    // 0.0;
    // double hwM = (Mw > 6.5) ? 1.0 : (Mw >= 5.5) ? Mw - 5.5 : 0.0;
    // double hwR = (rSeis < 8.0) ? c.c15 * (rSeis / 8.0) : c.c15;
    // double f5 = HW * f3 * hwM * hwR;

    return c.c1 + f1 + c.c4 * log(sqrt(f2)) + c.csite; // + f3 + f5;
  }

  private static final double calcStdDev(double c16, double Mw) {
    return c16 - ((Mw < 7.4) ? 0.07 * Mw : 0.518);
  }

  public static class Gridded extends CampbellBozorgnia_2003_Alaska {

    static final String NAME = CampbellBozorgnia_2003_Alaska.NAME + " : Gridded";

    Gridded(Imt imt) {
      super(imt);
    }

    @Override
    double calcMean(
        Coefficients c,
        GmmInput in) {

      double unkStyleTerm = 0.25 * c.c10 + 0.25 * c.c11;
      return super.calcMean(c, in) + unkStyleTerm;
    }

  }

}
