package gov.usgs.earthquake.nshmp.gmm;

import static com.google.common.base.Preconditions.checkNotNull;
import static gov.usgs.earthquake.nshmp.Text.Delimiter.COMMA;
import static java.nio.charset.StandardCharsets.UTF_8;
import static java.util.stream.Collectors.toList;
import static java.util.stream.Collectors.toMap;

import java.io.IOException;
import java.net.URL;
import java.util.EnumSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.google.common.collect.ArrayTable;
import com.google.common.collect.ImmutableTable;
import com.google.common.collect.Table;
import com.google.common.io.Resources;

/**
 * Class loads and manages {@code GroundMotionModel} coefficients.
 *
 * <p>Coefficients are loaded from CSV files. When such files are updated, it
 * may be necessary to edit certain {@code Imt} designations that are commonly
 * coded as integers (e.g. -1 = PGV, usually) or coefficient IDs that contain
 * illegal characters (e.g those with units labels in parentheses).
 *
 * @author U.S. Geological Survey
 */
final class CoefficientContainer {

  private static final String C_DIR = "gmm/coeffs/";
  private Table<Imt, String, Double> table;

  /**
   * Create a new coefficent container from a comma-delimited coefficient
   * resource.
   *
   * @param resource coefficent csv text resource
   */
  CoefficientContainer(String resource) {
    try {
      table = ImmutableTable.copyOf(load(resource));
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }

  /**
   * Return the value of the coefficient for the supplied name and intensity
   * measure type.
   *
   * @param imt intensity measure type
   * @param name of the coefficient to look up
   */
  double get(Imt imt, String name) {
    return checkNotNull(table.get(imt, name));
  }

  /**
   * Return a {@code Map} of all coefficient for the supplied intensity measure
   * type.
   *
   * @param imt intensity measure type
   */
  Map<String, Double> get(Imt imt) {
    return table.row(imt);
  }

  /**
   * Returns the {@code Set} of intensity measure types (IMTs) for which
   * coefficients are supplied. The returned Set is a always a new mutable
   * EnumSet.
   *
   * @return the {@code Set} of supported IMTs
   */
  Set<Imt> imts() {
    return EnumSet.copyOf(table.rowKeySet());
  }

  private static Table<Imt, String, Double> load(String resource) throws IOException {
    URL url = Resources.getResource(C_DIR + resource);
    List<String> lines = Resources.readLines(url, UTF_8);

    // build coeff name list from first line
    List<String> names = COMMA.splitter().splitToStream(lines.get(0))
        .skip(1)
        .collect(toList());

    // build Imt value arrays from remaining lines
    Map<Imt, double[]> valueMap = lines.stream()
        .skip(1)
        .filter(line -> !line.startsWith("#"))
        .map(COMMA.splitter()::splitToList)
        .collect(toMap(
            elements -> parseImt(elements),
            elements -> parseValues(elements)));

    // create and load table
    Table<Imt, String, Double> table = ArrayTable.create(valueMap.keySet(), names);
    for (Imt imt : valueMap.keySet()) {
      double[] values = valueMap.get(imt);
      int i = 0;
      for (String name : names) {
        table.put(imt, name, values[i++]);
      }
    }
    return table;
  }

  /*
   * Parses IMT strings from from a line in a coefficient file. Method expects
   * Imt.name() for specifically named intensity measure types, e.g. "PGA", and
   * double value strings for spectral periods, e.g. "0.2". This method is NOT
   * the same as {@link Imt#valueOf(String)}. Method will throw a
   * NumberFormatException or IllegalArgumentException if the supplied string is
   * not parseable into a known IMT.
   */
  private static Imt parseImt(List<String> elements) {
    String imtStr = elements.get(0);
    if (imtStr.equals("PGA") || imtStr.equals("PGV") || imtStr.equals("PGD")) {
      return Imt.valueOf(imtStr);
    }
    double period = Double.parseDouble(imtStr);
    return Imt.fromPeriod(period);
  }

  /* Parses coefficient values from a line in a coefficient file. */
  private static double[] parseValues(List<String> elements) {
    return elements.stream()
        .skip(1)
        .mapToDouble(Double::parseDouble)
        .toArray();
  }
}
