package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RRUP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.ZTOR;
import static java.lang.Math.exp;
import static java.lang.Math.log;
import static java.lang.Math.min;
import static java.lang.Math.pow;

import java.util.Map;

import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.Earthquakes;
import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Implementation of the subduction interface ground motion model by Geomatrix
 * consultants (1993) to support the 2007 Alaska NSHM. In the fortran codes used
 * at the time, this GMM is noted as a 'Youngs and Coppersmith variation'.
 * Although similar to Youngs et al. (1997), that GMM has quite a few add ons
 * for site terms not supported or required by the 2007 Alaska NSHM such that it
 * is simpler to have a standalone implementation.
 *
 * <p><b>Note:</b> Direct instantiation of {@code GroundMotionModel}s is
 * prohibited. Use {@link Gmm#instance(Imt)} to retrieve an instance for a
 * desired {@link Imt}.
 *
 * <p><b>Reference:</b> Youngs, R.R., Chiou, S.-J., Silva, W.J., and Humphrey,
 * J.R., 1997, Strong ground motion ground motion models for subduction zone
 * earthquakes: Seismological Research Letters, v. 68, p. 58-73.
 *
 * <p><b>doi:</b> <a href="http://dx.doi.org/10.1785/gssrl.68.1.58">
 * 10.1785/gssrl.68.1.58</a>
 *
 * <p><b>Component:</b> Geometric mean of two horizontal components
 *
 * @author U.S. Geological Survey
 * @see Gmm#GEOMATRIX_93_INTERFACE
 * @see Gmm#GEOMATRIX_93_SLAB
 */
public class Geomatrix_1993 implements GroundMotionModel {

  static final String NAME = "Geomatrix (1993)";

  static final Constraints CONSTRAINTS = Constraints.builder()
      .set(MW, Range.closed(5.0, 9.5))
      .set(RRUP, Range.closed(0.0, 1000.0))
      .set(ZTOR, Earthquakes.SLAB_DEPTH_RANGE)
      .set(VS30, Range.singleton(760.0))
      .build();

  static final CoefficientContainer COEFFS = new CoefficientContainer("Geomatrix93.csv");

  // centering term
  private static final double C0 = 0.2418;
  // magnitude scaling
  private static final double CM = 1.414;
  // distance scaling term constants
  // hazSUB and hazgrid have different values
  private static final double CRI = 1.782;
  private static final double CRS = 1.7818;
  private static final double CRE = 0.554;
  // slab scaling constant
  private static final double C_SLAB = 0.3846;
  // depth scaling
  private static final double CZI = 0.1215; // * 20 km (fixed)
  private static final double CZS = 0.00607; // * zTor

  private static final class Coefficients {

    final Imt imt;
    final double gc1, gc2, gc3, gc4, gc5;

    Coefficients(Imt imt, CoefficientContainer cc) {
      this.imt = imt;
      Map<String, Double> coeffs = cc.get(imt);
      gc1 = coeffs.get("gc1");
      gc2 = coeffs.get("gc2");
      gc3 = coeffs.get("gc3");
      gc4 = coeffs.get("gc4");
      gc5 = coeffs.get("gc5");
    }
  }

  private final Coefficients coeffs;

  Geomatrix_1993(Imt imt) {
    coeffs = new Coefficients(imt, COEFFS);
  }

  @Override
  public Imt imt() {
    return coeffs.imt;
  }

  @Override
  public final LogicTree<GroundMotion> calc(GmmInput in) {
    double μ = calcMean(coeffs, slab(), in);
    double σ = calcStdDev(coeffs, in.Mw);
    return GroundMotions.createTree(μ, σ);
  }

  /* Intraslab (true) or interface (false) model. */
  boolean slab() {
    return false;
  }

  private static final double calcMean(
      Coefficients c,
      boolean slab,
      GmmInput in) {

    double cR = slab ? CRS : CRI; // fortran discrepancy
    double cSlab = slab ? C_SLAB : 0.0;
    double cZ = slab ? in.zTor * CZS : CZI;

    double μ = C0 + (CM * in.Mw) +
        c.gc1 + c.gc2 * pow(10.0 - in.Mw, 3) +
        c.gc3 * log(in.rRup + cR * exp(CRE * in.Mw)) +
        cZ + cSlab;

    return μ;
  }

  private static final double calcStdDev(Coefficients c, double Mw) {
    return c.gc4 + c.gc5 * min(8.0, Mw);
  }

  static final class Interface extends Geomatrix_1993 {
    static final String NAME = Geomatrix_1993.NAME + " : Interface";

    Interface(Imt imt) {
      super(imt);
    }
  }

  static final class Slab extends Geomatrix_1993 {
    static final String NAME = Geomatrix_1993.NAME + " : Slab";

    Slab(Imt imt) {
      super(imt);
    }

    @Override
    final boolean slab() {
      return true;
    }
  }
}
