package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.FaultStyle.NORMAL;
import static gov.usgs.earthquake.nshmp.gmm.FaultStyle.REVERSE;
import static gov.usgs.earthquake.nshmp.gmm.FaultStyle.STRIKE_SLIP;
import static gov.usgs.earthquake.nshmp.gmm.FaultStyle.UNKNOWN;
import static gov.usgs.earthquake.nshmp.gmm.Imt.PGA;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P02;
import static java.lang.Math.log;

import java.util.Map;

import com.google.common.primitives.Doubles;

import gov.usgs.earthquake.nshmp.gmm.GroundMotionTables.GroundMotionTable;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Ground motion model (Gmm) utilities.
 * @author U.S. Geological Survey
 */
public final class GmmUtils {

  /* Name for gmm logic tree; same as in Deserialize. */
  private static final String TREE_NAME = "gmm-tree";

  /*
   * Base-10 to base-e conversion factor commonly used with ground motion lookup
   * tables that supply log10 instead of natural log values. This conversion
   * factor is 2.302585092994046 and is commonly referenced as SFAC in legacy
   * NSHMP fortran subroutines.
   */
  static final double BASE_10_TO_E = log(10.0);

  /*
   * Natural log space conversion of gravity from cm to m.
   */
  static final double LN_G_CM_TO_M = log(980.0);

  enum CeusSiteClass {
    SOFT_ROCK,
    HARD_ROCK;
  }

  /**
   * Returns the NSHMP interpretation of fault type based on rake; divisions are
   * on 45° diagonals.
   *
   * <p><b>Note:</b> This is inconsistent with next generation attenuation
   * relationship (NGAW1 and NGAW2) recommendations.
   *
   * @param rake to convert (in degrees)
   * @return the corresponding {@code FaultStyle}
   */
  public static FaultStyle rakeToFaultStyle_NSHMP(double rake) {
    if (Double.isNaN(rake)) {
      return UNKNOWN;
    }
    return (rake >= 45 && rake <= 135) ? REVERSE
        : (rake >= -135 && rake <= -45) ? NORMAL : STRIKE_SLIP;
  }

  /**
   * Convert a map of GMMs and weights to a logic tree of instances for the
   * supplied IMT.
   */
  static LogicTree<GroundMotionModel> weightMapToInstanceTree(
      Map<Gmm, Double> gmms,
      Imt imt) {

    LogicTree.EnumBuilder<Gmm, GroundMotionModel> tree = LogicTree.enumBuilder(TREE_NAME);
    gmms.entrySet().forEach(e -> tree.addBranch(
        e.getKey(),
        e.getKey().instance(imt),
        e.getValue()));
    return tree.build();
  }

  /* Derived from z2p5 via mmoschetti's 2018 ngaw2 db regressions */
  static final double BASIN_Z1P0_UPPER = 0.3;
  static final double BASIN_Z1P0_LOWER = 0.5;

  /* Based on Campbell & Bozorgnia (2014) */
  static final double BASIN_Z2P5_UPPER = 1.0;
  static final double BASIN_Z2P5_LOWER = 3.0;

  /* Return true if T > 0.5s and z > zMin km */
  static boolean checkBasin(Imt imt, double z, double zMin) {
    return !Double.isNaN(z) &&
        z > zMin &&
        imt.ordinal() > Imt.SA0P5.ordinal();
  }

  /* Compute linear scaling between two depth values */
  static double basinScale(double z, double upper, double lower) {
    double constrained = Doubles.constrainToRange(z, upper, lower);
    return (constrained - upper) / (lower - upper);
  }

  /* Compute the Δz1.0 scaling for z1p0 (in km) based GMMs. */
  static double deltaZ1scale(Imt imt, double z1p0) {
    if (GmmUtils.checkBasin(imt, z1p0, GmmUtils.BASIN_Z1P0_UPPER)) {
      double zScale = GmmUtils.basinScale(
          z1p0,
          GmmUtils.BASIN_Z1P0_UPPER,
          GmmUtils.BASIN_Z1P0_LOWER);
      return (imt == Imt.SA0P75) ? zScale * 0.585 : zScale;
    }
    return 0.0;
  }

  /* Compute the Δz2.5 scaling for z2p5 (in km) based GMMs. */
  static double deltaZ25scale(Imt imt, double z2p5) {
    if (GmmUtils.checkBasin(imt, z2p5, GmmUtils.BASIN_Z2P5_UPPER)) {
      double zScale = GmmUtils.basinScale(
          z2p5,
          GmmUtils.BASIN_Z2P5_UPPER,
          GmmUtils.BASIN_Z2P5_LOWER);
      return (imt == Imt.SA0P75) ? zScale * 0.585 : zScale;
    }
    return 0.0;
  }

  static boolean cybershakeImt(Imt imt) {
    return imt.isSA() && imt.period() > 1.9;
  }

  /**
   * Period dependent ground motion clipping. For PGA ({@code period} = 0.0s),
   * method returns {@code Math.min(ln(1.5g), μ)}; for
   * {@code 0.02s < period < 0.5s}, method returns {@code Math.min(ln(3.0g), μ)}
   * .
   *
   * @param imt of interest
   * @param μ natural log of ground motion
   * @return the clipped ground motion if required by the supplied
   *         {@code period}, {@code μ} otherwise
   */
  static double ceusMeanClip(Imt imt, double μ) {
    // ln(1.5) = 0.405; ln(3.0) = 1.099
    if (imt == Imt.PGA || imt == Imt.PGV || imt.period() <= 0.01) {
      return Math.min(0.405, μ);
    }
    if (imt.period() < 0.5) {
      return Math.min(μ, 1.099);
    }
    return μ;
  }

  /**
   * Returns a site class identifier for use with CEUS GMMs.
   *
   * @param vs30
   * @return the site class corresponding to the supplied vs30
   */
  static CeusSiteClass ceusSiteClass(double vs30) {
    if (vs30 == 760.0) {
      return CeusSiteClass.SOFT_ROCK;
    }
    if (vs30 == 2000.0) {
      return CeusSiteClass.HARD_ROCK;
    }
    throw new IllegalArgumentException("Unsupported CEUS vs30: " + vs30);
  }

  /**
   * Scales μ with distance at high frequencies and soft rock site conditions.
   */
  static double atkinsonTableValue(GroundMotionTable table, Imt imt,
      double m, double r, double vs30, double bcfac) {

    double μ = table.get(r, m);

    if (ceusSiteClass(vs30) == CeusSiteClass.SOFT_ROCK) {
      // NOTE this scaling is very atkinson table specific
      // we should probably be checking that period <= 0.02
      // Also, notes in fortran say gm scales with R for PGA
      // but Steve has applied it to 0.02 sec as well
      if (imt == PGA || imt == SA0P02) {
        // is there a citation for this distance decay??
        μ += -0.3 + 0.15 * Math.log10(r);
      } else {
        μ += bcfac;
      }
    }

    μ *= BASE_10_TO_E;
    if (imt != Imt.PGV) {
      μ -= LN_G_CM_TO_M;
    }
    return μ;
  }
}
