package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RRUP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static java.lang.Math.log;
import static java.lang.Math.sqrt;

import java.util.Map;

import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Implementation of the Gómez (202) ground motion model for crustal and
 * non-crustal regions of Puerto Rico.
 *
 * <p><b>Note:</b> Direct instantiation of {@code GroundMotionModel}s is
 * prohibited. Use {@link Gmm#instance(Imt)} to retrieve an instance for a
 * desired {@link Imt}.
 *
 * <p><b>Implementation note:</b>
 *
 * <li>Site term uses site class definitions from Japan (after Zhao et al.,
 * 2006) and defines two binary variables based on these Vs30 bins.</li>
 *
 * <p><b>Reference:</b> Gómez, D.F.C., 2022, Ground motion prediction equations
 * for Puerto Rico, Dissertation, Univ. Puerto Rico, Mayagüez campus, 317p.
 *
 * <p><b>doi:</b></a>
 *
 * <p><b>Component:</b> not specified (avg. horizontal implied)
 *
 * @author U.S. Geological Survey
 * @see Gmm#GOMEZ22_CRUSTAL
 * @see Gmm#GOMEZ22_NONCRUSTAL
 */
public abstract class Gomez_2022 implements GroundMotionModel {

  static final String NAME = "Gomez (2022) BETA";

  /* TODO: set constraints */
  static final Constraints CONSTRAINTS_CRUSTAL = Constraints.builder()
      .set(MW, Range.closed(4.0, 9.5))
      .set(RRUP, Range.closed(0.0, 1000.0))
      .set(VS30, Range.closed(150.0, 1000.0))
      .build();
  static final Constraints CONSTRAINTS_NONCRUSTAL = Constraints.builder()
      .set(MW, Range.closed(3.8, 9.5)) // M 3.8 - 5.0
      .set(RRUP, Range.closed(0.0, 1000.0)) // 100 - 400 km
      .set(VS30, Range.closed(150.0, 1000.0))
      .build();

  static final CoefficientContainer COEFFS = new CoefficientContainer("Gomez2022.csv");

  /* distance saturation term, section 3.5.2, p. 38 */
  private static final double H_CRUSTAL = 2.3;
  private static final double H_NONCRUSTAL = 44.0;

  private static final class Coefficients {

    final Imt imt;
    final double b1, b2, b3, b4, b5, b6, sigma;
    final double h;

    Coefficients(Imt imt, CoefficientContainer cc, boolean crustal) {
      this.imt = imt;
      Map<String, Double> coeffs = cc.get(imt);
      if (crustal) {
        b1 = coeffs.get("b1_c");
        b2 = coeffs.get("b2_c");
        b3 = coeffs.get("b3_c");
        b4 = coeffs.get("b4_c");
        b5 = coeffs.get("b5_c");
        b6 = coeffs.get("b6_c");
        sigma = coeffs.get("sigma_c");
      } else {
        b1 = coeffs.get("b1_nc");
        b2 = coeffs.get("b2_nc");
        b3 = coeffs.get("b3_nc");
        b4 = coeffs.get("b4_nc");
        b5 = coeffs.get("b5_nc");
        b6 = coeffs.get("b6_nc");
        sigma = coeffs.get("sigma_nc");
      }

      h = crustal ? H_CRUSTAL : H_NONCRUSTAL;
    }
  }

  private final Coefficients coeffs;

  Gomez_2022(Imt imt) {
    coeffs = new Coefficients(imt, COEFFS, crustal());
  }

  /* Crustal (true) or non-crustal (false) model. */
  boolean crustal() {
    return true;
  }

  @Override
  public Imt imt() {
    return coeffs.imt;
  }

  @Override
  public final LogicTree<GroundMotion> calc(GmmInput in) {

    double μ = calcMean(coeffs, in.Mw, in.rRup, in.vs30);
    double σ = coeffs.sigma;
    return GroundMotions.createTree(μ, σ);
  }

  private static final double calcMean(Coefficients c, double Mw, double rRup,
      double vs30) {
    /*
     * D = sqrt(d*d + h*h), where d is hypocentral distance and h is saturation
     * term
     *
     * USING RRUP for RHYP
     */
    double D = sqrt(rRup * rRup + c.h * c.h);

    double siteTerm = getSiteTerm(vs30, c.b5, c.b6);

    /* Eq. 4 */
    return c.b1 + c.b2 * Mw + c.b3 * log(D) + c.b4 * D + siteTerm;
  }

  private static final double getSiteTerm(double vs30, double b5, double b6) {
    /*
     * Site terms (part of Eq. 4) = b5 * S + b6 * H
     *
     * Site classes SC I - IV are defined in Table 3.8 and values of S & H are
     * defined in section 3.5.1 (p. 38).
     */
    double S, H;
    if (vs30 > 600.0) {
      /* SC I (A + B) */
      S = 0.0;
      H = 0.0;
    } else if (vs30 > 300.0 && vs30 <= 600.0) {
      /* SC II (C) */
      S = 0.0;
      H = 1.0;
    } else {
      /* SC III (D): 200 < vs30 <= 300 */
      /* SC IV (E): vs30 <= 200 */
      S = 1.0;
      H = 0.0;
    }
    return b5 * S + b6 * H;
  }

  static final class Crustal extends Gomez_2022 {
    static final String NAME = Gomez_2022.NAME + " : CRUSTAL";

    Crustal(Imt imt) {
      super(imt);
    }
  }

  static final class NonCrustal extends Gomez_2022 {
    static final String NAME = Gomez_2022.NAME + " : NONCRUSTAL";

    NonCrustal(Imt imt) {
      super(imt);
    }

    @Override
    boolean crustal() {
      return false;
    }
  }
}
