package gov.usgs.earthquake.nshmp.gmm;

import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Interface implemented by all ground motion models (GMMs); these are also
 * commonly referred to as ground motion prediction equations (GMPEs) or
 * attenuation relationships. Direct instantiation of GMMs is discouraged in
 * concrete implementations in favor of using the corresponding {@link Gmm}
 * {@code enum} identifier and its {@link Gmm#instance(Imt)} method. Concrete
 * implementations are public solely for the purpose of documentation.
 *
 * <p>Models generally have a single concrete implementation. However, for those
 * supplying region- or source-specific variants, there will typically be an
 * abstract base-model implementation and subclasses to handle each flavor (e.g.
 * {@link ZhaoEtAl_2006}. Each flavor has a unique {@link Gmm} identifier.
 *
 * @author U.S. Geological Survey
 * @see Gmm
 */
public interface GroundMotionModel {

  /**
   * The intensity measure type ({@link Imt IMT}) of this GMM instance.
   */
  Imt imt();

  /**
   * Computes a logic tree of scalar ground motions and standard deviations for
   * the supplied arguments. This method is implemented by GMMs to provide the
   * horizontal peak ground acceleration and (optionally) velocity (PGA, PGV) as
   * well as 5% damped pseudo-spectral acceleration for periods between 0.01 s
   * and 10 s. For most modern GMMs, ground motions are computed as the average
   * horizontal component (RotD50) but many older GMMs use some other average
   * horizontal measure (e.g. GMRotI50).
   *
   * @param in a ground motion model input argument container
   */
  LogicTree<GroundMotion> calc(GmmInput in);

  /**
   * Computes a logic tree of scalar ground motions via {@link #calc(GmmInput)},
   * and then applies any damping scaling. This method is called by most
   * applications and services.
   *
   * @param in a ground motion model input argument container
   * @param type the GMM type
   * @param damping
   */
  default LogicTree<GroundMotion> dampedCalc(
      GmmInput in,
      Gmm.Type type,
      double dampingRatio) {

    LogicTree<GroundMotion> gmTree = calc(in);
    gmTree = UsgsDampingScaling.apply(imt(), in, gmTree, type, dampingRatio);
    return gmTree;
  }

}
