package gov.usgs.earthquake.nshmp.gmm;

import static com.google.common.base.Preconditions.checkArgument;

import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.data.Interpolator;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Class provides internal support for interpolation between spectral periods.
 * Interpolation requires spectral periods bounding the target spectral period
 * and is linear in period and log in ground motion.
 *
 * @author U.S. Geological Survey
 */
class InterpolatedGmm implements GroundMotionModel {

  private final Imt imt;
  private final double tLo;
  private final double tHi;
  private final double tTarget;
  private final GroundMotionModel gmmLo;
  private final GroundMotionModel gmmHi;

  InterpolatedGmm(
      Gmm gmm,
      Imt targetImt,
      Range<Imt> imtRange) {

    this.imt = targetImt;

    /*
     * All Imt arguments must be spectral accelerations and the supplied range
     * must contain the target.
     */
    checkArgument(imtRange.hasLowerBound());
    checkArgument(imtRange.hasUpperBound());

    Imt saLo = imtRange.lowerEndpoint();
    Imt saHi = imtRange.upperEndpoint();

    checkArgument(saLo.isSA());
    checkArgument(saHi.isSA());
    checkArgument(imtRange.contains(targetImt));

    tLo = saLo.period();
    tHi = saHi.period();
    tTarget = targetImt.period();
    gmmLo = gmm.instance(saLo);
    gmmHi = gmm.instance(saHi);
  }

  @Override
  public Imt imt() {
    return imt;
  }

  @Override
  public LogicTree<GroundMotion> calc(GmmInput in) {
    GroundMotion sgmLo = GroundMotions.combine(gmmLo.calc(in));
    GroundMotion sgmHi = GroundMotions.combine(gmmHi.calc(in));
    double iμ = Interpolator.findY(tLo, sgmLo.mean(), tHi, sgmHi.mean(), tTarget);
    double iσ = Interpolator.findY(tLo, sgmLo.sigma(), tHi, sgmHi.sigma(), tTarget);
    return GroundMotions.createTree(iμ, iσ);
  }
}
