package gov.usgs.earthquake.nshmp.gmm;

import static com.google.common.base.Preconditions.checkArgument;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.BA_08_BASE;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RRUP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static gov.usgs.earthquake.nshmp.gmm.GmmUtils.BASE_10_TO_E;
import static gov.usgs.earthquake.nshmp.gmm.GmmUtils.LN_G_CM_TO_M;
import static gov.usgs.earthquake.nshmp.gmm.Imt.PGA;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P01;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P02;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P03;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P05;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P06;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P075;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P08;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P1;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P12;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P15;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P16;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P2;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P5;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P6;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P75;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P8;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA1P0;
import static java.lang.Math.log10;
import static java.lang.Math.sqrt;

import java.util.Map;
import java.util.Optional;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Maps;
import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.data.Interpolator;
import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Implementation of the ground motion relations for Puerto Rico by Motazedian
 * and Atkinson (2005) in order to evaluate the model for use in the 2025 Puerto
 * Rico and U.S. Virgin Islands NSHM.
 *
 * <p><b>Note:</b> Direct instantiation of {@code GroundMotionModel}s is
 * prohibited. Use {@link Gmm#instance(Imt)} to retrieve an instance for a
 * desired {@link Imt}.
 *
 * <p><b>Implementation notes:</b><ul>
 *
 * <li>Published coefficients are defined in frequency-space. Frequencies are
 * converted to the closest existing spectral period in IMT.</li>
 *
 * <li>Support for spectral period 0.01s is provided using the same coefficients
 * as PGA.</li>
 *
 * <li>Support for spectral periods 0.02, 0.03, 0.05, 0.075, 0.15, and 0.75 s is
 * provided via interpolation of ground motion of adjacent periods for which
 * there are coefficients. Sigma is fixed and period independent.</li>
 *
 * <li>Base model is for soft rock site classes only (Vs30 = 760 m/s).</li>
 *
 * <li>USGS subclass adds site-scaling model of BA08.</li>
 *
 * <p><b>Reference:</b> Motazedian, D. and Atkinson, G., 2005, Ground-motion
 * relations for Puerto Rico, in Mann, P., ed., Active tectonics and seismic
 * hazards of Puerto Rico, the Virgin Islands, and offshore areas: Geological
 * Society of America Special Paper 385, p. 61-80.
 *
 * <p><b>doi: </b> <a href="https://doi.org/10.1130/SPE385">10.1130/SPE385</a>
 *
 * <p><b>Component:</b> not specified (avg horizontal implied)
 *
 * @author U.S. Geological Survey
 * @see Gmm#MA_05_BASE
 * @see Gmm#MA_05
 */
public class MotazedianAtkinson_2005 implements GroundMotionModel {

  static final String NAME = "Motazedian & Atkinson (2005) BASE BETA";

  static final Constraints CONSTRAINTS = Constraints.builder()
      .set(MW, Range.closed(4.0, 9.5))
      .set(RRUP, Range.closed(0.0, 1000.0))
      .set(VS30, Range.closedOpen(150.0, 1500.0))
      .build();

  /*
   * Published coefficients are given for frequency (Hz). These are converted to
   * the closest available IMT period. Published PGA coefficients are duplicated
   * for 0.01 sec. Non-MPRS coefficients for 1.25, 2.5, and 6 s commented out.
   */
  static final CoefficientContainer COEFFS =
      new CoefficientContainer("MotazedianAtkinson05.csv");

  /* Interpolated IMTs to support MPRS. */
  private static final Map<Imt, Range<Imt>> INTERPOLATED_IMTS =
      Maps.immutableEnumMap(
          ImmutableMap.<Imt, Range<Imt>> builder()
              .put(SA0P02, Range.closed(SA0P01, SA0P06))
              .put(SA0P03, Range.closed(SA0P01, SA0P06))
              .put(SA0P05, Range.closed(SA0P01, SA0P06))
              .put(SA0P075, Range.closed(SA0P06, SA0P08))
              .put(SA0P15, Range.closed(SA0P12, SA0P16))
              .put(SA0P75, Range.closed(SA0P6, SA0P8))
              .build());

  // distance centering coefficients
  private static final double C5 = -7.333;
  private static final double C6 = 2.333;
  // hinge function coefficients
  private static final double C7 = -1.8;
  private static final double C8 = 0.1;
  private static final double DIST_HINGE_LOWER = 75.0;
  private static final double DIST_HINGE_UPPER = 100.0;

  // convert log10 sigma to ln
  private static final double SIGMA = 0.28 * BASE_10_TO_E;

  private static final class Coefficients {

    final Imt imt;
    final double c1, c2, c3, c4;

    Coefficients(Imt imt, CoefficientContainer cc) {
      this.imt = imt;
      Map<String, Double> coeffs = cc.get(imt);
      c1 = coeffs.get("c1");
      c2 = coeffs.get("c2");
      c3 = coeffs.get("c3");
      c4 = coeffs.get("c4");
    }
  }

  private final Coefficients coeffs;
  private final boolean interpolated;
  private final GroundMotionModel interpolatedGmm;

  MotazedianAtkinson_2005(Imt imt) {
    coeffs = new Coefficients(imt, COEFFS);
    interpolated = INTERPOLATED_IMTS.containsKey(imt);
    interpolatedGmm = interpolated
        ? new InterpolatedGmm(Gmm.MA_05_BASE, imt, INTERPOLATED_IMTS.get(imt))
        : null;
  }

  @Override
  public Imt imt() {
    return coeffs.imt;
  }

  @Override
  public LogicTree<GroundMotion> calc(GmmInput in) {
    if (interpolated) {
      return interpolatedGmm.calc(in);
    }
    return calc(coeffs, in);
  }

  private static LogicTree<GroundMotion> calc(
      Coefficients c,
      GmmInput in) {

    double μ = calcMean(c, in.rRup, in.Mw);
    return GroundMotions.createTree(μ, SIGMA);
  }

  private static final double calcMean(
      Coefficients c,
      double rRup,
      double Mw) {

    /* Eq. 6, log10(PSA) in cm/s/s */

    double r = getR(rRup, Mw);

    double hingeFunction = getHingeFunction(r, Mw);

    double μ = c.c1 + c.c2 * (Mw - 6) + c.c3 * (Mw - 6) * (Mw - 6) + hingeFunction + c.c4 * r;
    // convert from base 10 to base e and from cm/s/s to g
    return μ * BASE_10_TO_E - LN_G_CM_TO_M;
  }

  private static final double getR(double rRup, double Mw) {
    // Equation and coefficients defined in Table 2 caption
    double delta = C5 + C6 * Mw;
    return sqrt(rRup * rRup + delta * delta);
  }

  private static final double getHingeFunction(double dist, double Mw) {
    // Equation and coefficients defined in Table 2 caption
    double mFac = C7 + C8 * Mw;
    if (dist <= DIST_HINGE_LOWER) {
      // R <= 75 km
      return mFac * log10(dist);
    } else if (dist <= DIST_HINGE_UPPER) {
      // 75 km < R <= 100 km
      return mFac * log10(DIST_HINGE_LOWER);
    } else {
      // R >= 100 km
      return mFac * log10(DIST_HINGE_LOWER) - 0.5 * log10(dist / DIST_HINGE_UPPER);
    }
  }

  /*
   * USGS implementation that adds site scaling using BA08 model.
   *
   * MA05 supports a variety of non-MPRS spectral periods required for
   * interpolation that are, in turn, not supported by BA08. We therefore
   * interpolate site terms as needed for the six IMTs missing from BA08.
   */
  static class Usgs extends MotazedianAtkinson_2005 {

    static final String NAME = "Motazedian & Atkinson (2005) BETA";

    private static final Map<Imt, Range<Imt>> INTERPOLATED_SITE_IMTS =
        Maps.immutableEnumMap(
            ImmutableMap.<Imt, Range<Imt>> builder()
                .put(SA0P06, Range.closed(SA0P05, SA0P075))
                .put(SA0P08, Range.closed(SA0P075, SA0P1))
                .put(SA0P12, Range.closed(SA0P1, SA0P15))
                .put(SA0P16, Range.closed(SA0P15, SA0P2))
                .put(SA0P6, Range.closed(SA0P5, SA0P75))
                .put(SA0P8, Range.closed(SA0P75, SA1P0))
                .build());

    private final Coefficients coeffsPGA;
    private final Optional<BooreAtkinson_2008> siteAmp;

    Usgs(Imt imt) {
      super(imt);
      coeffsPGA = new Coefficients(PGA, COEFFS);
      siteAmp = BooreAtkinson_2008.COEFFS.imts().contains(imt)
          ? Optional.of((BooreAtkinson_2008) BA_08_BASE.instance(imt))
          : Optional.empty();
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      // Get μ from MA_05_BASE for current IMT (interpolate if necessary)
      double μ = GroundMotions.combine(super.calc(in)).mean();

      double pgaRef = Math.exp(calcMean(coeffsPGA, in.rRup, in.Mw));
      double site = siteAmp.isPresent()
          ? siteAmp.get().siteAmp(pgaRef, in.vs30)
          : calcInterpolatedSite(super.coeffs.imt, pgaRef, in.vs30);
      return GroundMotions.createTree(μ + site, SIGMA);
    }

    private static double calcInterpolatedSite(Imt imt, double pgaRef, double vs30) {
      Range<Imt> imtRange = INTERPOLATED_SITE_IMTS.get(imt);
      Imt imtLo = imtRange.lowerEndpoint();
      Imt imtHi = imtRange.upperEndpoint();
      // basic checks following InterpolatedGmm to catch bad Range<Imt>
      checkArgument(imtLo.isSA());
      checkArgument(imtHi.isSA());
      checkArgument(imtRange.contains(imt));

      double tLo = imtLo.period();
      double tHi = imtHi.period();
      double tTarget = imt.period();
      BooreAtkinson_2008 ba08lo = (BooreAtkinson_2008) BA_08_BASE.instance(imtLo);
      BooreAtkinson_2008 ba08hi = (BooreAtkinson_2008) BA_08_BASE.instance(imtHi);
      double siteLo = ba08lo.siteAmp(pgaRef, vs30);
      double siteHi = ba08hi.siteAmp(pgaRef, vs30);
      return Interpolator.findY(tLo, siteLo, tHi, siteHi, tTarget);
    }
  }
}
