package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RJB;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static gov.usgs.earthquake.nshmp.gmm.GmmUtils.CeusSiteClass.HARD_ROCK;
import static gov.usgs.earthquake.nshmp.gmm.MagConverter.MB_TO_MW_ATKIN_BOORE;
import static gov.usgs.earthquake.nshmp.gmm.MagConverter.MB_TO_MW_JOHNSTON;
import static java.lang.Math.exp;
import static java.lang.Math.log;
import static java.lang.Math.sqrt;

import java.util.Map;

import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.gmm.GmmUtils.CeusSiteClass;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Implementation of the Toro et al. (1997) ground motion model for stable
 * continental regions with 2002 updates. This implementation matches that used
 * in the 2008 USGS NSHMP and comes in two flavors (mb and Mw) to support the
 * 2008 central and eastern US model.
 *
 * <p><b>Note:</b> Direct instantiation of {@code GroundMotionModel}s is
 * prohibited. Use {@link Gmm#instance(Imt)} to retrieve an instance for a
 * desired {@link Imt}.
 *
 * <p><b>Implementation note:</b> Mean values are clamped per
 * {@link GmmUtils#ceusMeanClip(Imt, double)}.
 *
 * <p><b>Reference:</b> Toro, G.R., 2002, Modification of the Toro et al. (1997)
 * attenuation relations for large magnitudes and short distances: Risk
 * Engineering, Inc. <a href=
 * "http://www.ce.memphis.edu/7137/PDFs/attenuations/Toro_2001_(modification_1997).pdf">
 * Report</a>
 *
 * <p><b>Reference:</b> Toro, G.R., Abrahamson, N.A., and Schneider, J.F., 1997,
 * A model of strong ground motions from earthquakes in central and eastern
 * North America: Best estimates and uncertainties: Seismological Research
 * Letters, v. 68, p. 41–57.
 *
 * <p><b>doi:</b> <a href="http://dx.doi.org/10.1785/gssrl.68.1.41">
 * 10.1785/gssrl.68.1.41</a>
 *
 * <p><b>Component:</b> not specified
 *
 * @author U.S. Geological Survey
 * @see Gmm#TORO_97_MB
 * @see Gmm#TORO_97_MW
 */
public abstract class ToroEtAl_1997 implements GroundMotionModel {

  static final String NAME = "Toro et al. (1997)";

  static final Constraints CONSTRAINTS = Constraints.builder()
      .set(MW, Range.closed(4.0, 8.0))
      .set(RJB, Range.closed(0.0, 1000.0))
      .set(VS30, Range.closed(760.0, 2000.0))
      .build();

  static final CoefficientContainer COEFFS_MW = new CoefficientContainer("Toro97Mw.csv");
  static final CoefficientContainer COEFFS_MB = new CoefficientContainer("Toro97Mb.csv");

  private static final class Coefficients {

    final Imt imt;
    final double t1, t1h, t2, t3, t4, t5, t6, th, tσ;

    Coefficients(Imt imt, CoefficientContainer cc) {
      this.imt = imt;
      Map<String, Double> coeffs = cc.get(imt);
      t1 = coeffs.get("t1");
      t1h = coeffs.get("t1h");
      t2 = coeffs.get("t2");
      t3 = coeffs.get("t3");
      t4 = coeffs.get("t4");
      t5 = coeffs.get("t5");
      t6 = coeffs.get("t6");
      th = coeffs.get("th");
      tσ = coeffs.get("tsigma");
    }
  }

  private final Coefficients coeffs;

  ToroEtAl_1997(Imt imt) {
    coeffs = new Coefficients(imt, isMw() ? COEFFS_MW : COEFFS_MB);
  }

  @Override
  public Imt imt() {
    return coeffs.imt;
  }

  @Override
  public final LogicTree<GroundMotion> calc(GmmInput in) {
    double μ = calcMean(coeffs, isMw(), in);
    return GroundMotions.createTree(μ, coeffs.tσ);
  }

  abstract boolean isMw();

  private static final double calcMean(Coefficients c, boolean isMw,
      GmmInput in) {

    double mag = in.Mw;
    double rJB = in.rJB;

    CeusSiteClass siteClass = GmmUtils.ceusSiteClass(in.vs30);

    double thsq = c.th * c.th;

    // magnitude correction: with Toro model, you change the coefficients
    // appropriate to the magnitude.
    //
    // New, Nov 2006: the finite-fault correction, affects the fictitious
    // depth or bending point; from Toro Paducah paper. Mod. Dec 2007, mblg
    // to Mw for the correction.

    double mCorr;

    if (isMw) {
      mCorr = Math.exp(-1.25 + 0.227 * mag);
    } else {
      double magJ = MB_TO_MW_JOHNSTON.convert(mag);
      double cor1 = exp(-1.25 + 0.227 * magJ);
      double magAB = MB_TO_MW_ATKIN_BOORE.convert(mag);
      double cor2 = exp(-1.25 + 0.227 * magAB);
      mCorr = sqrt(cor1 * cor2); // geo mean
    }

    double dist = Math.sqrt(rJB * rJB + thsq * mCorr * mCorr);

    // default to SOFT_ROCK values
    double gnd = (siteClass == HARD_ROCK) ? c.t1h : c.t1;
    gnd += c.t2 * (mag - 6.0) + c.t3 * ((mag - 6.0) * (mag - 6.0));
    gnd += -c.t4 * log(dist) - c.t6 * dist;

    double factor = log(dist / 100.0);
    if (factor > 0) {
      gnd = gnd - (c.t5 - c.t4) * factor;
    }

    return GmmUtils.ceusMeanClip(c.imt, gnd);
  }

  static final class Mb extends ToroEtAl_1997 {
    static final String NAME = ToroEtAl_1997.NAME + " : mb";

    Mb(Imt imt) {
      super(imt);
    }

    @Override
    boolean isMw() {
      return false;
    }
  }

  static final class Mw extends ToroEtAl_1997 {
    static final String NAME = ToroEtAl_1997.NAME + " : Mw";

    Mw(Imt imt) {
      super(imt);
    }

    @Override
    boolean isMw() {
      return true;
    }
  }

}
