package gov.usgs.earthquake.nshmp.gmm;

import static com.google.common.base.Preconditions.checkArgument;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.AG_20_GLOBAL_INTERFACE;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.AG_20_GLOBAL_SLAB;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.ASK_14_BASE;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.BSSA_14_BASE;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.CB_14_BASE;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.CY_14_BASE;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.KBCG_20_GLOBAL_INTERFACE;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.KBCG_20_GLOBAL_SLAB;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.PSBAH_20_GLOBAL_INTERFACE;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.PSBAH_20_GLOBAL_SLAB;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.Type.ACTIVE_CRUST;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.Type.SUBDUCTION_INTERFACE;
import static gov.usgs.earthquake.nshmp.gmm.Gmm.Type.SUBDUCTION_SLAB;
import static gov.usgs.earthquake.nshmp.gmm.GroundMotionTables.getPrviEpi;
import static gov.usgs.earthquake.nshmp.gmm.GroundMotions.EPI_IDS;
import static gov.usgs.earthquake.nshmp.gmm.GroundMotions.EPI_WTS;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.Type.INTERFACE;
import static gov.usgs.earthquake.nshmp.gmm.SubductionZone.Type.SLAB;
import static java.lang.Math.exp;
import static java.lang.Math.log;
import static java.lang.Math.sqrt;

import java.util.Map;

import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.gmm.GroundMotionTables.GroundMotionTable;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * PRVI GMC class for backbone GMM implementations.
 *
 * Copied from CombinedGmm:
 *
 * Convenience class for combined GMM implementations that compute weight
 * averaged ground motions and sigmas for a logic tree of ground motion models.
 * These are NOT intended for use in hazard calculations, but are made available
 * to support comparative deterministic analysis of GMM logic trees.
 */
public abstract class UsgsPrviBackbone2025 implements GroundMotionModel {

  private static final String NAME = "USGS PRVI Backbone (2025)";

  static final CoefficientContainer COEFFS_PHIS2S_ACTIVE_CRUST =
      new CoefficientContainer("prvi-25-backbone-phiS2S_crustal.csv");
  static final CoefficientContainer COEFFS_PHISS_ACTIVE_CRUST =
      new CoefficientContainer("prvi-25-backbone-phiSS_crustal.csv");
  static final CoefficientContainer COEFFS_SIGMA_SUBDUCTION =
      new CoefficientContainer("prvi-25-backbone-sigma_subduction.csv");
  static final CoefficientContainer COEFFS_TAU_ACTIVE_CRUST =
      new CoefficientContainer("prvi-25-backbone-tau_crustal.csv");

  // TODO consider separating calc method in NGA models so that we can get the
  // mean directly instead of unwrapping from singleton ground motion logic
  // tree

  /*
   * Bias adjustment coefficients from https://github.com/gem/oq-engine/blob/
   * 576bf0c6853ca896aae84e68ad848d39de09a396/openquake/hazardlib/gsim/usgs_prvi
   * .py last updated Jun 25, 2024
   */
  // COEFFS_DATA_ADJUSTMENT is currently used to get the IMT set for USGS_PRVI_*
  // models
  static final CoefficientContainer COEFFS_DATA_ADJUSTMENT =
      new CoefficientContainer("prvi-25-backbone-adjustments.csv");

  public static final double DATA_ADJUST_CRUSTAL = -0.3;
  public static final double DATA_ADJUST_SUBDUCTION = -0.4;

  public static final String MODEL_BASE_ID = "base";
  public static final String MODEL_ADJUST_ID = "adjust";

  public static final String SIGMA_NGA_ID = "σ-nga";
  public static final String SIGMA_PRVI_ID = "σ-prvi";
  private static final String[] SIGMA_IDS = { SIGMA_NGA_ID, SIGMA_PRVI_ID };
  private static final double[] SIGMA_WTS = { 0.5, 0.5 };

  private static final class Coefficients {

    final double τ1, τ2, τ3, τ4;
    final double a, b;
    final double φs2sPrvi, φs2sNga;
    final double τSub, φSub, φSubSS, φSubS2S;

    Coefficients(
        Imt imt,
        CoefficientContainer ccTau,
        CoefficientContainer ccPhiSS,
        CoefficientContainer ccPhiS2S,
        CoefficientContainer ccSubSigma) {

      Map<String, Double> coeffsTau = ccTau.get(imt);
      τ1 = coeffsTau.get("tau_mean_4.5");
      τ2 = coeffsTau.get("tau_mean_5.0");
      τ3 = coeffsTau.get("tau_mean_5.5");
      τ4 = coeffsTau.get("tau_mean_6.5");

      Map<String, Double> coeffsPhiSS = ccPhiSS.get(imt);
      a = coeffsPhiSS.get("a_smoothed_Mod");
      b = coeffsPhiSS.get("b_smoothed_Mod");

      Map<String, Double> coeffsPhiS2S = ccPhiS2S.get(imt);
      φs2sPrvi = coeffsPhiS2S.get("phiS2S_PRVI");
      φs2sNga = coeffsPhiS2S.get("phiS2S_WUS");

      Map<String, Double> coeffsSubSigma = ccSubSigma.get(imt);
      τSub = coeffsSubSigma.get("tau_mean");
      φSub = coeffsSubSigma.get("phi_mean");
      φSubSS = coeffsSubSigma.get("PhiSS_PRVI");
      φSubS2S = coeffsSubSigma.get("phiS2S_PRVI_smooth");

    }
  }

  private final Coefficients coeffs;

  private final LogicTree<GroundMotionModel> tree;
  private final Imt imt;
  private final GroundMotionTable[] epiTables; // 0=lower, 1=upper
  private final double dataAdjust;

  /* Supply map of ground motion models initialized to the required IMT. */
  UsgsPrviBackbone2025(
      Imt imt,
      String name,
      Map<Gmm, Double> gmms,
      GroundMotionTable[] epiTables,
      double dataAdjust) {
    this.imt = imt;
    this.epiTables = epiTables;
    this.dataAdjust = dataAdjust;
    LogicTree.Builder<GroundMotionModel> b = LogicTree.builder(name);
    gmms.entrySet().stream().forEach(e -> b.addBranch(
        e.getKey().name(),
        e.getKey().instance(imt),
        e.getValue()));
    this.tree = b.build();

    coeffs = new Coefficients(
        imt,
        COEFFS_TAU_ACTIVE_CRUST,
        COEFFS_PHISS_ACTIVE_CRUST,
        COEFFS_PHIS2S_ACTIVE_CRUST,
        COEFFS_SIGMA_SUBDUCTION);
  }

  @Override
  public Imt imt() {
    return imt;
  }

  @Override
  public LogicTree<GroundMotion> calc(GmmInput in) {
    return calc(in, true, SigmaType.TREE);
  }

  LogicTree<GroundMotion> calc(GmmInput in, boolean epi, SigmaType σType) {

    GroundMotion bgm = calcBackboneGroundMotion(in);
    double μ = bgm.mean() + dataAdjust;
    double[] σs = calcSigmas(in); // NGA=0, PRVI=1

    if (epi) {
      double[] μs = applyEpistemic(μ, in.Mw, in.rRup);
      if (σType == SigmaType.TREE) {
        return GroundMotions.createTree(
            EPI_IDS, μs, EPI_WTS,
            SIGMA_IDS, σs, SIGMA_WTS);
      } else {
        double σ = (σType == SigmaType.NGA) ? σs[0] : σs[1];
        return GroundMotions.createTree(EPI_IDS, μs, EPI_WTS, σ);
      }
    }

    // single mean with multi sigma not supported
    checkArgument(σType != SigmaType.TREE);
    double σ = (σType == SigmaType.NGA) ? σs[0] : σs[1];
    return GroundMotions.createTree(μ, σ);
  }

  /* Compute backbone ground motions. */
  GroundMotion calcBackboneGroundMotion(GmmInput in) {
    double median = 0.0;
    double sigmaSq = 0.0;
    for (Branch<GroundMotionModel> b : tree) {
      // NGA GMMs currently returning mix of singleton trees and epi branches
      // TODO would be preferable to get the mean directly w/o epi branches
      LogicTree<GroundMotion> gmTree = b.value().calc(in);
      GroundMotion gm = (gmTree.size() == 1)
          ? gmTree.get(0).value()
          : gmTree.get(1).value();
      median += exp(gm.mean()) * b.weight();
      sigmaSq += gm.sigma() * gm.sigma() * b.weight();
    }
    return GroundMotion.create(log(median), sqrt(sigmaSq));

  }

  double[] applyEpistemic(double μ, double Mw, double rRup) {
    double εLo = epiTables[0].get(rRup, Mw);
    double εHi = epiTables[1].get(rRup, Mw);
    return new double[] { μ + εLo, μ, μ + εHi };
  }

  static enum SigmaType {
    NGA,
    PRVI,
    TREE;
  }

  abstract double[] calcSigmas(GmmInput in);

  /*
   * Implementations. For each, coefficients are only used to get the set of
   * supported IMTs and therefore reference a model that supports the
   * intersection of all support IMTs.
   */

  /* PRVI 2025 Active Crust */
  static class ActiveCrust extends UsgsPrviBackbone2025 {

    static final String NAME = UsgsPrviBackbone2025.NAME + " : Active Crust";
    static final Constraints CONSTRAINTS = AbrahamsonEtAl_2014.CONSTRAINTS;

    static final Map<Gmm, Double> GMM_MAP = Map.of(
        ASK_14_BASE, 0.25,
        BSSA_14_BASE, 0.25,
        CB_14_BASE, 0.25,
        CY_14_BASE, 0.25);

    ActiveCrust(Imt imt) {
      this(imt, NAME, 0.0);
    }

    ActiveCrust(Imt imt, String name, double dataAdjust) {
      super(imt, name, GMM_MAP, getPrviEpi(ACTIVE_CRUST, imt), dataAdjust);
    }

    @Override
    double[] calcSigmas(GmmInput in) {
      Coefficients c = super.coeffs;
      double τ = calcTau(in.Mw, c);
      double φSS = calcPhiSS(in.Mw, c);
      double σNga = sqrt(τ * τ + c.φs2sNga * c.φs2sNga + φSS * φSS);
      double σPrvi = sqrt(τ * τ + c.φs2sPrvi * c.φs2sPrvi + φSS * φSS);
      return new double[] { σNga, σPrvi };
    }

    // active crust tau
    static double calcTau(double Mw, Coefficients c) {
      if (Mw <= 4.5) {
        return c.τ1;
      } else if (Mw <= 5.0) {
        return c.τ1 + (c.τ2 - c.τ1) * (Mw - 4.5) / 0.5;
      } else if (Mw <= 5.5) {
        return c.τ2 + (c.τ3 - c.τ2) * (Mw - 5.0) / 0.5;
      } else if (Mw <= 6.5) {
        return c.τ4 + (c.τ4 - c.τ3) * (Mw - 5.5);
      } else {
        return c.τ4;
      }
    }

    // active crust phiSS
    static double calcPhiSS(double Mw, Coefficients c) {
      return Mw <= 5.0 ? c.a : Mw > 6.5 ? c.b : c.a + (Mw - 5.0) * ((c.b - c.a) / 1.5);
    }
  }

  /* PRVI 2025 Active Crust - sigma NGA */
  @Deprecated
  static final class ActiveCrustSigmaNga extends ActiveCrust {

    static final String NAME = ActiveCrust.NAME + " (NGA sigma)";

    ActiveCrustSigmaNga(Imt imt) {
      super(imt, NAME, 0.0);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      return calc(in, true, SigmaType.NGA);
    }
  }

  /* PRVI 2025 Active Crust - data adjusted */
  static final class ActiveCrustAdjusted extends ActiveCrust {

    static final String NAME = ActiveCrust.NAME + " (adjusted)";

    ActiveCrustAdjusted(Imt imt) {
      super(imt, NAME, DATA_ADJUST_CRUSTAL);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      return calc(in, true, SigmaType.TREE);
    }
  }

  /* PRVI 2025 Active Crust - data adjusted, no epi, sigma NGA */
  @Deprecated
  static final class ActiveCrustAdjustedNoEpiSigmaNga extends ActiveCrust {

    static final String NAME = ActiveCrust.NAME + " (adjusted, no epi, NGA sigma)";

    ActiveCrustAdjustedNoEpiSigmaNga(Imt imt) {
      super(imt, NAME, DATA_ADJUST_CRUSTAL);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      return calc(in, false, SigmaType.NGA);
    }
  }

  /* PRVI 2025 Active Crust - no epi, sigma NGA */
  @Deprecated
  static final class ActiveCrustNoEpiSigmaNga extends ActiveCrust {

    static final String NAME = ActiveCrust.NAME + " (no epi, NGA sigma)";

    ActiveCrustNoEpiSigmaNga(Imt imt) {
      super(imt, NAME, 0.0);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      return calc(in, false, SigmaType.NGA);
    }
  }

  /* PRVI 2025 Active Crust - no epi, sigma PRVI */
  @Deprecated
  static final class ActiveCrustNoEpiSigmaPrvi extends ActiveCrust {

    static final String NAME = ActiveCrust.NAME + " (no epi, PRVI sigma)";

    ActiveCrustNoEpiSigmaPrvi(Imt imt) {
      super(imt, NAME, 0.0);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      return calc(in, false, SigmaType.PRVI);
    }
  }

  @Deprecated
  static final class ActiveCrustTotal implements GroundMotionModel {
    static final String NAME = ActiveCrust.NAME + " (total tree)";
    static final Constraints CONSTRAINTS = AbrahamsonEtAl_2014.CONSTRAINTS;

    private final Imt imt;
    private final ActiveCrust gmm1;
    private final ActiveCrustAdjusted gmm2;

    ActiveCrustTotal(Imt imt) {
      this.imt = imt;
      gmm1 = new ActiveCrust(imt);
      gmm2 = new ActiveCrustAdjusted(imt);
    }

    @Override
    public Imt imt() {
      return imt;
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      LogicTree<GroundMotion> treeBase = gmm1.calc(in);
      LogicTree<GroundMotion> treeAdj = gmm2.calc(in);
      return combine(treeBase, treeAdj, "active-crust");
    }
  }

  /* PRVI 2025 Interface */
  static class Interface extends UsgsPrviBackbone2025 {

    static final String NAME = UsgsPrviBackbone2025.NAME + " : " + INTERFACE;
    static final Constraints CONSTRAINTS = AbrahamsonGulerce_2020.CONSTRAINTS_INTERFACE;

    static final Map<Gmm, Double> GMM_MAP = Map.of(
        AG_20_GLOBAL_INTERFACE, 0.3333,
        KBCG_20_GLOBAL_INTERFACE, 0.3333,
        PSBAH_20_GLOBAL_INTERFACE, 0.3334);

    Interface(Imt imt) {
      this(imt, NAME, 0.0);
    }

    Interface(Imt imt, String name, double dataAdjust) {
      super(imt, name, GMM_MAP, getPrviEpi(SUBDUCTION_INTERFACE, imt), dataAdjust);

    }

    @Override
    double[] calcSigmas(GmmInput in) {
      Coefficients c = super.coeffs;
      double σNga = sqrt(c.τSub * c.τSub + c.φSub * c.φSub);
      double σPrvi = sqrt(c.τSub * c.τSub + c.φSubS2S * c.φSubS2S + c.φSubSS * c.φSubSS);
      return new double[] { σNga, σPrvi };
    }
  }

  /* PRVI 2025 Interface - NGA sigma */
  @Deprecated
  static final class InterfaceSigmaNga extends Interface {

    static final String NAME = Interface.NAME + " (NGA sigma)";

    InterfaceSigmaNga(Imt imt) {
      super(imt, NAME, 0.0);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      return calc(in, true, SigmaType.NGA);
    }
  }

  /* PRVI 2025 Interface - data adjusted */
  static final class InterfaceAdjusted extends Interface {

    static final String NAME = Interface.NAME + " (adjusted)";

    InterfaceAdjusted(Imt imt) {
      super(imt, NAME, DATA_ADJUST_SUBDUCTION);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      return calc(in, true, SigmaType.TREE);
    }
  }

  /* PRVI 2025 Interface - data adjusted, no epi, NGA sigma */
  @Deprecated
  static final class InterfaceAdjustedNoEpiSigmaNga extends Interface {

    static final String NAME = Interface.NAME + " (adjusted, no epi, NGA sigma)";

    InterfaceAdjustedNoEpiSigmaNga(Imt imt) {
      super(imt, NAME, DATA_ADJUST_SUBDUCTION);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      return calc(in, false, SigmaType.NGA);
    }
  }

  /* PRVI 2025 Interface - no epi, NGA sigma */
  @Deprecated
  static final class InterfaceNoEpiSigmaNga extends Interface {

    static final String NAME = Interface.NAME + " (no epi, NGA sigma)";

    InterfaceNoEpiSigmaNga(Imt imt) {
      super(imt, NAME, 0.0);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      return calc(in, false, SigmaType.NGA);
    }
  }

  /* PRVI 2025 Interface - no epi, PRVI sigma */
  @Deprecated
  static final class InterfaceNoEpiSigmaPrvi extends Interface {

    static final String NAME = Interface.NAME + " (no epi, PRVI sigma)";

    InterfaceNoEpiSigmaPrvi(Imt imt) {
      super(imt, NAME, 0.0);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      return calc(in, false, SigmaType.PRVI);
    }
  }

  @Deprecated
  static final class InterfaceTotal implements GroundMotionModel {
    static final String NAME = Interface.NAME + " (total tree)";
    static final Constraints CONSTRAINTS = AbrahamsonGulerce_2020.CONSTRAINTS_INTERFACE;

    private final Imt imt;
    private final Interface gmm1;
    private final InterfaceAdjusted gmm2;

    InterfaceTotal(Imt imt) {
      this.imt = imt;
      gmm1 = new Interface(imt);
      gmm2 = new InterfaceAdjusted(imt);
    }

    @Override
    public Imt imt() {
      return imt;
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      LogicTree<GroundMotion> treeBase = gmm1.calc(in);
      LogicTree<GroundMotion> treeAdj = gmm2.calc(in);
      return combine(treeBase, treeAdj, "interface");
    }
  }

  /* PRVI 2025 Intraslab */
  static class Slab extends UsgsPrviBackbone2025 {

    static final String NAME = UsgsPrviBackbone2025.NAME + " : " + SLAB;
    static final Constraints CONSTRAINTS = AbrahamsonGulerce_2020.CONSTRAINTS_SLAB;

    static final Map<Gmm, Double> GMM_MAP = Map.of(
        AG_20_GLOBAL_SLAB, 0.3333,
        KBCG_20_GLOBAL_SLAB, 0.3333,
        PSBAH_20_GLOBAL_SLAB, 0.3334);

    Slab(Imt imt) {
      this(imt, NAME, 0.0);
    }

    Slab(Imt imt, String name, double dataAdjust) {
      super(imt, name, GMM_MAP, getPrviEpi(SUBDUCTION_SLAB, imt), dataAdjust);
    }

    @Override
    double[] calcSigmas(GmmInput in) {
      Coefficients c = super.coeffs;
      double σNga = sqrt(c.τSub * c.τSub + c.φSub * c.φSub);
      double σPrvi = sqrt(c.τSub * c.τSub + c.φSubS2S * c.φSubS2S + c.φSubSS * c.φSubSS);
      return new double[] { σNga, σPrvi };
    }
  }

  /* PRVI 2025 Intraslab - NGA sigma */
  @Deprecated
  static final class SlabSigmaNga extends Slab {

    static final String NAME = Slab.NAME + " (NGA sigma)";

    SlabSigmaNga(Imt imt) {
      super(imt, NAME, 0.0);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      return calc(in, true, SigmaType.TREE);
    }
  }

  /* PRVI 2025 Intraslab - data adjusted */
  static final class SlabAdjusted extends Slab {

    static final String NAME = Slab.NAME + " (adjusted)";

    SlabAdjusted(Imt imt) {
      super(imt, NAME, DATA_ADJUST_SUBDUCTION);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      return calc(in, true, SigmaType.TREE);
    }
  }

  /* PRVI 2025 Intraslab - data adjusted, no epi, NGA sigma */
  @Deprecated
  static final class SlabAdjustedNoEpiSigmaNga extends Slab {

    static final String NAME = Slab.NAME + " (adjusted, no epi, NGA sigma)";

    SlabAdjustedNoEpiSigmaNga(Imt imt) {
      super(imt, NAME, DATA_ADJUST_SUBDUCTION);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      return calc(in, false, SigmaType.NGA);
    }
  }

  /* PRVI 2025 Intraslab - no epi, NGA sigma */
  @Deprecated
  static final class SlabNoEpiSigmaNga extends Slab {

    static final String NAME = Slab.NAME + " (no epi, NGA sigma)";

    SlabNoEpiSigmaNga(Imt imt) {
      super(imt, NAME, 0.0);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      return calc(in, false, SigmaType.NGA);
    }
  }

  /* PRVI 2025 Intraslab - no epi, PRVI sigma */
  @Deprecated
  static final class SlabNoEpiSigmaPrvi extends Slab {

    static final String NAME = Slab.NAME + " (no epi, PRVI sigma)";

    SlabNoEpiSigmaPrvi(Imt imt) {
      super(imt, NAME, 0.0);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      return calc(in, false, SigmaType.PRVI);
    }
  }

  @Deprecated
  static final class SlabTotal implements GroundMotionModel {
    static final String NAME = Slab.NAME + " (total tree)";
    static final Constraints CONSTRAINTS = AbrahamsonGulerce_2020.CONSTRAINTS_SLAB;

    private final Imt imt;
    private final Slab gmm1;
    private final SlabAdjusted gmm2;

    SlabTotal(Imt imt) {
      this.imt = imt;
      gmm1 = new Slab(imt);
      gmm2 = new SlabAdjusted(imt);
    }

    @Override
    public Imt imt() {
      return imt;
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      LogicTree<GroundMotion> treeBase = gmm1.calc(in);
      LogicTree<GroundMotion> treeAdj = gmm2.calc(in);
      return combine(treeBase, treeAdj, "intraslab");
    }
  }

  /*
   * Hopefully temporary support for collapsed/combined GMM logic trees. For
   * PRVI, kmilner requested access to a single logic tree object for each
   * tectonic setting.
   */

  static final double WT_SCALE = 0.5;

  /* Merge base and adjusted model logic trees */
  static LogicTree<GroundMotion> combine(
      LogicTree<GroundMotion> base,
      LogicTree<GroundMotion> adj,
      String name) {

    LogicTree.Builder<GroundMotion> tree = LogicTree.builder("gmm-tree-prvi-" + name);
    for (Branch<GroundMotion> b : base) {
      tree.addBranch(MODEL_BASE_ID + "-" + b.id(), b.value(), b.weight() * 0.5);
    }
    for (Branch<GroundMotion> b : adj) {
      tree.addBranch(MODEL_ADJUST_ID + "-" + b.id(), b.value(), b.weight() * 0.5);
    }
    return tree.build();
  }

}
