package gov.usgs.earthquake.nshmp.internal;

import java.net.URL;

import com.google.common.base.Charsets;
import com.google.common.io.Resources;
import com.google.gson.Gson;

/**
 * Interface for reading in version info JSON files.
 *
 * @author U.S. Geological Survey
 */
public interface AppVersion {

  /**
   * Returns the version info for the repo.
   */
  public VersionInfo getVersionInfo();

  /**
   * Returns the version info for a repository from a JSON resource file.
   *
   * @param resource The JSON file with version info.
   */
  static VersionInfo versionInfo(URL resource) {
    try {
      return new Gson().fromJson(
          Resources.toString(resource, Charsets.UTF_8),
          VersionInfo.class);
    } catch (Exception e) {
      throw new RuntimeException(e.getMessage());
    }
  }

  /**
   * The version info of a repository.
   */
  public static class VersionInfo {
    public String projectName;
    public String url;
    public String version;
    public String gitHash;

    private VersionInfo(Builder builder) {
      projectName = builder.projectName;
      url = builder.url;
      version = builder.version;
      gitHash = builder.gitHash;
    }

    /**
     * Returns the version info builder.
     */
    public static Builder builder() {
      return new Builder();
    }

    public static class Builder {
      private String projectName;
      private String url;
      private String version;
      private String gitHash;

      private Builder() {}

      /**
       * Set the project name
       *
       * @param projectName The project name
       */
      public Builder projectName(String projectName) {
        this.projectName = projectName;
        return this;
      }

      /**
       * Set the url
       *
       * @param url The repo url
       */
      public Builder url(String url) {
        this.url = url;
        return this;
      }

      /**
       * Set the version.
       *
       * @param version The repo version
       */
      public Builder version(String version) {
        this.version = version;
        return this;
      }

      /**
       * Set appreviated git hash.
       *
       * @param gitHash Git hash
       */
      public Builder gitHash(String gitHash) {
        this.gitHash = gitHash;
        return this;
      }

      /**
       * Returns the version info.
       */
      public VersionInfo build() {
        return new VersionInfo(this);
      }
    }
  }
}
