package gov.usgs.earthquake.nshmp.model;

import static gov.usgs.earthquake.nshmp.geo.BorderType.MERCATOR_LINEAR;
import static gov.usgs.earthquake.nshmp.geo.GriddedRegion.ANCHOR_0_0;

import java.awt.geom.Area;
import java.awt.geom.PathIterator;
import java.net.URISyntaxException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;

import gov.usgs.earthquake.nshmp.geo.GriddedRegion;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.Region;
import gov.usgs.earthquake.nshmp.geo.Regions;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.FeatureCollection;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;

/*
 * Initial work to port Fortran codes amake, amakeshear, amakeshearMP for
 * on-the-fly generation of grids
 */

public class AreaSourceAmake {

  private static final boolean DEBUG_DUMP = true;

  public static void main(String[] args) throws URISyntaxException {

    /*
     * For now, read polygon from GeoJSON file filepath specified from root
     * directory of nshmp-lib project
     */
    String filepath = "../nshmp-haz-v2/etc/nshm/map-sf-bay.geojson";
    // filepath = "map-sf-bay.geojson";
    Path fJsonPath = Paths.get(filepath);

    // create Feature from json path
    // workaround for existing GeoJson file structured as FeatureCollection
    // instead of a single Feature
    FeatureCollection fc = GeoJson.from(fJsonPath).toFeatureCollection();
    List<Feature> fl = fc.features();
    Feature f = fl.get(0);

    // get spacing and title from GeoJSON file Feature
    // set defaults if not defined
    double spacing = f.properties().getDouble("spacing").orElse(0.1);
    String title = f.properties().getString("title").orElse("");

    // temp override spacing for testing
    spacing = 0.1;

    // get polygon border as LocationList
    LocationList border = f.asPolygonBorder();

    // get bounds of polygon to define (LL + UR corners) rectangular region
    // containing source
    double[] pbounds = border.bounds().toArray();

    // nudge corners to make sure all grid nodes covering source are returned
    // without this nudge, some node cells intersected by the area source
    // boundary
    // are omitted. For the SF-bay region, a nudge down to 0.3 still works,
    // other
    // sources tested required a nudge of 0.6
    double nudge = 0.6;
    pbounds[0] -= nudge * spacing;
    pbounds[1] -= nudge * spacing;
    pbounds[2] += nudge * spacing;
    pbounds[3] += nudge * spacing;

    // create Region for area source
    Region areaSourceRegion = Regions.create(title, border, MERCATOR_LINEAR);
    double areaSourceRegionExtent = areaSourceRegion.extent();

    // create rectangular GriddedRegion using LL + UR corners
    // it would be nice to have
    // Regions.createRectangularGridded(title,Region,dx,dy,anchor)
    // that will automatically expand LL+UR as needed to fully cover Region with
    // nodes
    GriddedRegion areaGrid = Regions.createRectangularGridded(
        String.format("%s-%s", title, "grid"),
        Location.create(pbounds[0], pbounds[1]),
        Location.create(pbounds[2], pbounds[3]),
        spacing, spacing, ANCHOR_0_0);

    // keep track of cells
    double nodesInAreaSourceExtent = 0.0;
    int cellsInside = 0;
    int cellsOutside = 0;
    int cellsIntersected = 0;

    // retain node weights (fraction of node cell inside Area source boundary)
    List<Double> nodeWts = new ArrayList<Double>();

    // iterate over grid nodes
    for (int i = 0; i < areaGrid.size(); i++) {

      // generate java.awt.geom.Area for grid node cell
      Area nodeArea = areaGrid.areaForIndex(i);

      // get LocationList for nodeArea (cell) boundary
      LocationList nodeBorder = geomAreaToLocationList(nodeArea);

      // generate Region for grid node cell and calculate its area (extent,
      // km^2)
      Region nodeRegion = Regions.create(null, nodeBorder, MERCATOR_LINEAR);
      double nodeExtent = nodeRegion.extent();

      // determine weight for grid node - fraction of the node cell inside the
      // Area source
      double wt = 0.0;
      double intersectExtent = 0.0;
      if (areaSourceRegion.contains(nodeRegion)) {
        // grid node cell entirely contained within Area source polygon
        wt = 1.0;
        intersectExtent = nodeExtent;
        cellsInside += 1;
      } else {
        // calculate intersection Region between node cell and Area source
        Region intersection = Regions.intersectionOf(null, areaSourceRegion, nodeRegion);
        if (intersection == null) {
          // intersection is empty - no overlap between node cell and Area
          // source
          wt = 0.0;
          cellsOutside += 1;
        } else {
          intersectExtent = intersection.extent();
          wt = intersectExtent / nodeExtent;
          cellsIntersected += 1;
        }
      }

      // save wt for node (fraction of node cell inside Area source boundary
      nodeWts.add(wt);

      // capture running total of node cell area to compare to total Area source
      // area
      nodesInAreaSourceExtent += intersectExtent;

    }

    // compare extent of Area source and total of node cell areas inside Area
    // source
    System.out.printf("%d Total cells, %d cells inside, %d cells outside, %d cells intersected\n",
        areaGrid.size(), cellsInside, cellsOutside, cellsIntersected);
    System.out.printf("    Area Source Extent: %.8f\n", areaSourceRegionExtent);
    System.out.printf("Total Node Cell Extent: %.8f\n", nodesInAreaSourceExtent);
    System.out.printf("                 Ratio: %.4f\n",
        nodesInAreaSourceExtent / areaSourceRegionExtent);

  }

  /*
   * Iterate over outline of java.awt.geom.Area object and create a LocationList
   * Make last point equal the first point if it doesn't already (easier
   * plotting in GMT)
   */
  public static LocationList geomAreaToLocationList(Area a) {
    PathIterator pI = a.getPathIterator(null);
    double[] coords = new double[6]; // 6-element array required for
                                     // pI.currentSegment()
    LocationList.Builder locations = LocationList.builder();

    // iterate over paths
    while (!pI.isDone()) {
      int segtype = pI.currentSegment(coords);
      if (segtype == PathIterator.SEG_LINETO || segtype == PathIterator.SEG_MOVETO) {
        locations.add(Location.create(coords[0], coords[1]));
      }
      pI.next();
    }
    return locations.build();
  }

  /***************************************************************************/
  /* Debug dumps code */

}
