package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.checkRupturesFile;
import static gov.usgs.earthquake.nshmp.model.ModelFiles.readSystemFeatures;
import static gov.usgs.earthquake.nshmp.model.SourceType.FAULT_SYSTEM;
import static gov.usgs.earthquake.nshmp.model.SourceType.INTERFACE_SYSTEM;
import static java.util.stream.Collectors.toUnmodifiableMap;

import java.io.BufferedReader;
import java.io.IOException;
import java.lang.reflect.Type;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.OptionalDouble;
import java.util.function.Function;
import java.util.stream.Stream;

import com.google.common.collect.Range;
import com.google.common.primitives.Ints;
import com.google.common.reflect.TypeToken;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonParser;

import gov.usgs.earthquake.nshmp.fault.FocalMech;
import gov.usgs.earthquake.nshmp.fault.surface.RuptureFloating;
import gov.usgs.earthquake.nshmp.fault.surface.RuptureScaling;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.FeatureCollection;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;
import gov.usgs.earthquake.nshmp.geo.json.Properties;
import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Properties.GutenbergRichter;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Properties.Single;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Properties.TaperedGr;
import gov.usgs.earthquake.nshmp.model.HazardModel.MapRegion;
import gov.usgs.earthquake.nshmp.model.MfdConfig.AleatoryProperties;
import gov.usgs.earthquake.nshmp.tree.LogicGroup;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Factory class for source model deserialization methods and helpers.
 *
 * Developer notes:
 *
 * Logic trees (*-tree) and other NSHM source model maps (*-map) are difficult
 * to deserialize using a Gson instance due to the use of generics and the
 * complexity of some nested objects. Some objects have conditional fields that
 * may or may not be present (e.g. MFDs), or that may be represented by either
 * an object or a string pointer to an object in a map. Gson.fromJson() methods
 * require the use of TypeTokens, which are difficult to create for generic
 * collections of objects that themselves include parameterized types. A
 * TypteToken based approach also requires either polluting the 'tree' package
 * with types outside the package, or exposing the various branch flavors, which
 * we'd like to avoid. We therefore opt to deserialize most *-tree and *-map
 * objects using a JsonParser.
 */
class Deserialize {

  private static final Gson GSON = new GsonBuilder().create();

  static final String ID = "id";
  static final String NAME = "name";
  static final String MODEL = "model";
  static final String SPACING = "spacing";
  static final String MFD_TREE = "mfd-tree";
  static final String RATE_TREE = "rate-tree";
  static final String SOURCE_TREE = "source-tree";
  static final String GMM_TREES = "gmm-trees";

  private static final String WEIGHT = "weight";
  private static final String SCALE = "scale";
  private static final String VALUE = "value";
  private static final String TYPE = "type";
  private static final String PROPERTIES = "properties";
  private static final String SECTIONS = "sections";
  private static final String FEATURE = "feature";
  private static final String GRID_MFDS = "grid-mfds";
  private static final String SPATIAL_PDF = "spatial-pdf";
  private static final String RUPTURE_SETS = "rupture-sets";
  private static final String GMM_TREE = "gmm-tree";
  private static final String SLIP_RATE_TREE = "slip-rate-tree";
  private static final String GRID_RATE_TREE = "grid-rate-tree";
  private static final String M_MIN = "mMin";
  private static final String M_MAX = "mMax";
  private static final String DEPTH_TREE = "depth-tree";
  private static final String SURFACE_SPACING = "surface-spacing";
  private static final String RUPTURE_FLOATING = "rupture-floating";
  private static final String RUPTURE_SCALING = "rupture-scaling";
  private static final String GRID_SPACING = "grid-spacing";
  private static final String SMOOTHING_DENSITY = "smoothing-density";
  private static final String SMOOTHING_LIMIT = "smoothing-limit";
  private static final String DISTANCE_BIN = "opt-distance-bin";
  private static final String POINT_SOURCE_TYPE = "point-source-type";
  private static final String MAX_DEPTH = "max-depth";
  private static final String MAX_WIDTH = "max-width";
  private static final String FOCAL_MECH_TREE = "focal-mech-tree";
  private static final String GRID_DEPTH_MAP = "grid-depth-map";
  private static final String DIP_SLIP_MODEL = "dip-slip-model";
  private static final String DIP_TREE = "dip-tree";

  /* Read Feature Collection */
  static FeatureCollection featureCollection(Path geojson) {
    return GeoJson.from(geojson).toFeatureCollection();
  }

  /* Read model info. */
  static HazardModel.Info modelInfo(Path json) {
    var info = fromJson(json, HazardModel.Info.class);
    checkNotNull(info.name, "Missing 'name' in %s", json.getFileName());
    checkNotNull(info.siteClassMap, "Missing 'site-class-vs30' in %s", json.getFileName());
    return info;
  }

  /* Read tectonic setting map-region. */
  static HazardModel.MapRegion mapRegion(Path geojson) {
    Feature feature = GeoJson.from(geojson).toFeature();
    return new MapRegion(feature);
  }

  /* Load the site-data directory. */
  static SiteData siteData(Path dir) {
    return SiteData.create(dir);
  }

  /* Read a MFD configuration. */
  static MfdConfig mfdConfig(Path json) {
    JsonObject config = jsonObject(json);
    MfdConfig.Builder builder = MfdConfig.builder();

    JsonElement mMin = config.get(MfdConfig.MIN_MAGNITUDE_KEY);
    if (!mMin.isJsonNull()) {
      builder.minimumMagnitude(mMin.getAsDouble());
    }

    JsonElement epiTree = config.get(MfdConfig.EPISTEMIC_TREE_KEY);
    if (!epiTree.isJsonNull()) {
      builder.epistemicTree(doubleTree(epiTree, MfdConfig.EPISTEMIC_TREE_KEY));
    }

    JsonElement aleaProps = config.get(MfdConfig.ALEATORY_PROPS_KEY);
    if (!aleaProps.isJsonNull()) {
      builder.aleatoryProperties(GSON.fromJson(aleaProps, AleatoryProperties.class));
    }

    JsonElement nshmBinModel = config.get(MfdConfig.NSHM_BIN_MODEL_KEY);
    if (!nshmBinModel.isJsonNull()) {
      builder.nshmBinModel(nshmBinModel.getAsBoolean());
    }

    return builder.build();
  }

  /* Read a GMM configuration. */
  static GmmConfig gmmConfig(Path json) {
    JsonObject config = jsonObject(json);
    GmmConfig.Builder builder = GmmConfig.builder();

    double rMax = config.get(GmmConfig.MAX_DISTANCE_KEY).getAsDouble();
    builder.maximumDistance(rMax);

    return builder.build();
  }

  /* Create a GMM logic tree. */
  static LogicTree<Gmm> gmmTree(Path json) {
    return enumValueTree(jsonArray(json), Gmm.class, GMM_TREE);
  }

  /* Create a GMM logic tree. */
  static LogicTree<Gmm> gmmTree(JsonElement json) {
    return enumValueTree(json.getAsJsonArray(), Gmm.class, GMM_TREE);
  }

  /* Create a source path logic tree. */
  static LogicTree<Path> sourceTree(Path json) {
    LogicTree.Builder<Path> builder = LogicTree.builder(SOURCE_TREE);
    for (JsonElement branch : jsonArray(json)) {
      JsonObject obj = branch.getAsJsonObject();
      /* null source logic tree branches are allowed. */
      JsonElement idElement = obj.get(ID);
      // Consider changing to "do-nothing" or "disabled"
      String id = idElement.isJsonNull() ? "null" : idElement.getAsString();
      builder.addBranch(
          id,
          json.resolveSibling(id),
          obj.get(WEIGHT).getAsDouble());
    }
    return builder.build();
  }

  /* Create a source tree ID. */
  static SourceTree.Info treeInfo(Path json) {
    JsonObject obj = jsonObject(json);
    int id = obj.get(ID).getAsInt();
    JsonElement nameElement = obj.get(NAME);
    Optional<String> name = nameElement != null
        ? Optional.of(nameElement.getAsString())
        : Optional.empty();
    return new SourceTree.Info(id, name);
  }

  /* Create a source path logic group. */
  static LogicGroup<Path> sourceGroup(Path json) {
    LogicGroup.Builder<Path> builder = LogicGroup.builder();
    for (JsonElement branch : jsonArray(json)) {
      JsonObject obj = branch.getAsJsonObject();
      String id = obj.get(ID).getAsString();
      double scale = obj.has(SCALE) ? obj.get(SCALE).getAsDouble() : 1.0;
      builder.addBranch(id, json.resolveSibling(id), scale);
    }
    return builder.build();
  }

  /* Create a numeric rate logic tree. */
  static LogicTree<Double> rateTree(Path json) {
    return doubleTreeWithNulls(jsonArray(json), RATE_TREE);
  }

  /* Create a grid source rate logic tree; paths to gridded rates. */
  static LogicTree<Path> gridRatesTree(Path json) {
    LogicTree.Builder<Path> builder = LogicTree.builder(GRID_RATE_TREE);
    for (JsonElement branch : jsonArray(json)) {
      JsonObject obj = branch.getAsJsonObject();
      builder.addBranch(
          obj.get(ID).getAsString(),
          Path.of(obj.get(VALUE).getAsString()),
          obj.get(WEIGHT).getAsDouble());
    }
    return builder.build();
  }

  /* Create a fault source rate logic tree; IDs of rate models in a rate-map. */
  static LogicTree<String> slipRateTree(Path json) {
    LogicTree.StringValueBuilder builder = LogicTree.stringValueBuilder(SLIP_RATE_TREE);
    for (JsonElement branch : jsonArray(json)) {
      JsonObject obj = branch.getAsJsonObject();
      builder.addBranch(
          obj.get(ID).getAsString(),
          obj.get(WEIGHT).getAsDouble());
    }
    return builder.build();
  }

  /* Create an id:fault-feature map from the files in a directory. */
  static Map<Integer, SourceFeature.NshmFault> faultFeatures(Path dir) {
    return features(dir, SourceFeature::newNshmFault);
  }

  /* Create an id:interface-feature map from the files in a directory. */
  static Map<Integer, SourceFeature.Interface> interfaceFeatures(Path dir) {
    return features(dir, SourceFeature::newInterfaceSection);
  }

  /* Create an id:grid-feature map from the files in a directory. */
  static Map<Integer, SourceFeature.Grid> gridFeatures(Path dir) {
    return features(dir, SourceFeature::newSourceGrid);
  }

  private static <T extends SourceFeature> Map<Integer, T> features(
      Path dir,
      Function<Path, T> mapper) {

    try (Stream<Path> paths = Files.list(dir)) {
      return paths.filter(ModelFiles::isGeoJson)
          .map(mapper::apply)
          .collect(toUnmodifiableMap(
              feature -> feature.id,
              Function.identity()));
    } catch (IOException ioe) {
      throw new RuntimeException(ioe);
    }
  }

  /* Create an id:system-feature map from a fault-sections.geojson. */
  static Map<Integer, SourceFeature.SystemSection> systemFeatures(Path json) {
    return GeoJson.from(json).toFeatureCollection().features().stream()
        .map(SourceFeature.SystemSection::new)
        .collect(toUnmodifiableMap(
            feature -> feature.index,
            Function.identity()));
  }

  /* Create a grid rupture set. */
  static List<GridRuptureSet> gridRuptureSets(Path json, ModelData data) {
    JsonArray ruptureSets = jsonArray(json);
    List<GridRuptureSet> rsList = new ArrayList<>();
    for (JsonElement ruptureSet : ruptureSets) {
      JsonObject obj = ruptureSet.getAsJsonObject();
      boolean spatialPdf = (obj.get(SPATIAL_PDF) != null);
      data.spatialPdf(spatialPdf);
      String gridFile = spatialPdf ? SPATIAL_PDF : GRID_MFDS;
      data.gridRateFile(Path.of(obj.get(gridFile).getAsString()));
      GridRuptureSet grs = GridLoader.createGridRuptureSet(
          obj.get(ID).getAsInt(),
          obj.get(NAME).getAsString(),
          obj.get(FEATURE).getAsInt(),
          obj.get(MFD_TREE).getAsString(),
          data);
      rsList.add(grs);
    }
    return rsList;
  }

  /* Create a slab rupture set. */
  static List<SlabRuptureSet> slabRuptureSets(Path json, ModelData data) {
    JsonArray ruptureSets = jsonArray(json);
    List<SlabRuptureSet> rsList = new ArrayList<>();
    for (JsonElement ruptureSet : ruptureSets) {
      JsonObject obj = ruptureSet.getAsJsonObject();
      data.gridRateFile(Path.of(obj.get(SPATIAL_PDF).getAsString()));
      SlabRuptureSet srs = GridLoader.createSlabRuptureSet(
          obj.get(ID).getAsInt(),
          obj.get(NAME).getAsString(),
          obj.get(FEATURE).getAsInt(),
          obj.get(MFD_TREE).getAsString(),
          data);
      rsList.add(srs);
    }
    return rsList;
  }

  /* Create an interface rupture set. */
  static InterfaceRuptureSet interfaceRuptureSet(
      Path json,
      ModelData data) {

    JsonObject obj = jsonObject(json);
    return interfaceRuptureSet(obj, data);
  }

  /* Create an interface rupture set. */
  private static InterfaceRuptureSet interfaceRuptureSet(
      JsonObject obj,
      ModelData data) {

    int id = obj.get(ID).getAsInt();

    InterfaceRuptureSet.Builder ruptureSet = InterfaceRuptureSet.builder()
        .modelData(data)
        .name(obj.get(NAME).getAsString())
        .id(id)
        .mfdTree(mfdTree(obj, data));

    /* Optional 'sections' array. */
    if (obj.has(SECTIONS)) {
      int[] sections = GSON.fromJson(obj.get(SECTIONS), int[].class);
      ruptureSet.sections(Ints.asList(sections));
    } else {
      ruptureSet.sections(List.of(id));
    }
    return ruptureSet.build();
  }

  /* Create a cluster rupture set, initialized with data from file. */
  static ClusterRuptureSet.Data interfaceClusterSet(
      Path json,
      ModelData data) {

    JsonObject obj = jsonObject(json);

    /* Set cluster flag so faultRuptureSet MFDs are built correctly. */
    data.clusterModel();

    List<InterfaceRuptureSet> interfaceRuptureSets = new ArrayList<>();
    JsonArray ruptures = obj.get(RUPTURE_SETS).getAsJsonArray();
    for (JsonElement rupture : ruptures) {
      InterfaceRuptureSet ruptureSet = interfaceRuptureSet(rupture.getAsJsonObject(), data);
      // set leaf weight set here for nested FaultRuptureSets as they
      // will not be set by source tree on build.
      ruptureSet.setLeafWeight(1.0);
      interfaceRuptureSets.add(ruptureSet);
    }

    return new ClusterRuptureSet.Data(
        obj.get(NAME).getAsString(),
        obj.get(ID).getAsInt(),
        data,
        interfaceRuptureSets);
  }

  /* Create a crustal fault system rupture set. */
  static SystemRuptureSet crustalSystemRuptureSet(Path json, ModelData data) {
    return systemRuptureSet(json, data, FAULT_SYSTEM);
  }

  /* Create an interface system rupture set. */
  static SystemRuptureSet interfaceSystemRuptureSet(Path json, ModelData data) {
    return systemRuptureSet(json, data, INTERFACE_SYSTEM);
  }

  private static SystemRuptureSet systemRuptureSet(
      Path json,
      ModelData data,
      SourceType type) {

    Path parent = checkNotNull(json.getParent());
    JsonObject obj = jsonObject(json);

    return SystemRuptureSet.builder(type)
        .name(obj.get(NAME).getAsString())
        .id(obj.get(ID).getAsInt())
        .modelData(data)
        .sections(readSystemFeatures(parent).orElseThrow())
        .ruptures(checkRupturesFile(parent).orElseThrow())
        .build();
  }

  /* Create a fault rupture set. */
  static FaultRuptureSet faultRuptureSet(
      Path json,
      ModelData data) {

    JsonObject obj = jsonObject(json);
    return faultRuptureSet(obj, data);
  }

  /* Create a fault rupture set. */
  private static FaultRuptureSet faultRuptureSet(
      JsonObject obj,
      ModelData data) {

    int id = obj.get(ID).getAsInt();

    FaultRuptureSet.Builder ruptureSet = FaultRuptureSet.builder()
        .name(obj.get(NAME).getAsString())
        .id(id)
        .modelData(data)
        .mfdTree(mfdTree(obj, data));

    /* Optional 'sections' array. */
    if (obj.has(SECTIONS)) {
      int[] sections = GSON.fromJson(obj.get(SECTIONS), int[].class);
      ruptureSet.sections(Ints.asList(sections));
    } else {
      ruptureSet.sections(List.of(id));
    }

    /* Optional additional or overriding 'properties'. */
    if (obj.has(PROPERTIES)) {
      Type propsType = new TypeToken<Map<String, ?>>() {}.getType();
      Map<String, ?> props = GSON.fromJson(obj.get(PROPERTIES), propsType);
      ruptureSet.properties(props);
    }

    return ruptureSet.build();
  }

  /* Create a cluster rupture set, initialized with data from file. */
  static ClusterRuptureSet.Data faultClusterSet(
      Path json,
      ModelData data) {

    JsonObject obj = jsonObject(json);

    /* Set cluster flag so faultRuptureSet MFDs are built correctly. */
    data.clusterModel();

    List<FaultRuptureSet> faultRuptureSets = new ArrayList<>();
    JsonArray ruptures = obj.get(RUPTURE_SETS).getAsJsonArray();
    for (JsonElement rupture : ruptures) {
      FaultRuptureSet ruptureSet = faultRuptureSet(rupture.getAsJsonObject(), data);
      // set leaf weight here for nested FaultRuptureSets as they
      // will not be set by source tree on build.
      ruptureSet.setLeafWeight(1.0);
      faultRuptureSets.add(ruptureSet);
    }

    return new ClusterRuptureSet.Data(
        obj.get(NAME).getAsString(),
        obj.get(ID).getAsInt(),
        data,
        faultRuptureSets);
  }

  /* Convenience method to use Deserialize.GSON instance external to class. */
  static <T> T fromJson(JsonElement json, Type typeOfT) {
    return GSON.fromJson(json, typeOfT);
  }

  /* Create an mfd-tree from a JSON element or via map lookup. */
  private static LogicTree<Mfd.Properties> mfdTree(
      JsonElement mfdElement,
      ModelData data) {

    if (mfdElement.isJsonPrimitive()) {
      var mfdMap = data.mfdMap().orElseThrow();
      var mfdId = mfdElement.getAsString();
      checkState(mfdMap.containsKey(mfdId), "MFD id '%s' missing in mfd-map", mfdId);
      return mfdMap.get(mfdId);
    }
    return mfdPropertiesTree(mfdElement);
  }

  /* Create an mfd-tree from a JSON object. */
  private static LogicTree<Mfd.Properties> mfdTree(JsonObject obj, ModelData data) {
    return mfdTree(obj.get(MFD_TREE), data);
  }

  /* Create an mfd-tree from Json.Properties. */
  static LogicTree<Mfd.Properties> mfdTree(Properties props, ModelData data) {
    return mfdTree(props.getJsonElement(MFD_TREE).orElseThrow(), data);
  }

  /* Create a string:mfd-tree map from a mfd-map. */
  static Map<String, LogicTree<Mfd.Properties>> mfdMap(Path json) {
    Map<String, LogicTree<Mfd.Properties>> mfdMap = new HashMap<>();
    for (Entry<String, JsonElement> entry : jsonObject(json).entrySet()) {
      mfdMap.put(entry.getKey(), mfdPropertiesTree(entry.getValue()));
    }
    return Map.copyOf(mfdMap);
  }

  /* Create an mfd-tree from in inline element. */
  static LogicTree<Mfd.Properties> mfdPropertiesTree(JsonElement json) {
    LogicTree.Builder<Mfd.Properties> builder = LogicTree.builder(MFD_TREE);
    for (JsonElement branch : json.getAsJsonArray()) {
      JsonObject obj = branch.getAsJsonObject();
      builder.addBranch(
          obj.get(ID).getAsString(),
          mfdProperties(obj.get(VALUE)),
          obj.get(WEIGHT).getAsDouble());
    }
    return builder.build();
  }

  private static Mfd.Properties mfdProperties(JsonElement json) {
    JsonObject obj = json.getAsJsonObject();
    Mfd.Type type = Mfd.Type.valueOf(obj.get(TYPE).getAsString());
    Mfd.Properties properties;
    switch (type) {
      case SINGLE:
        properties = GSON.fromJson(validateSingle(obj), Single.class);
        break;
      case GR:
      case GR_MMAX_GR:
      case GR_MMAX_SINGLE:
        properties = GSON.fromJson(validateGr(obj), GutenbergRichter.class);
        break;
      case GR_TAPER:
        properties = GSON.fromJson(validateGrTaper(obj), TaperedGr.class);
        break;
      case INCR:
        // Custom props object that we'll replace once MFDs built
        properties = GSON.fromJson(validateIncr(obj), Incremental.class);
        break;
      default:
        throw new IllegalArgumentException("Unknown MFD type: " + type);
    }
    return properties;
  }

  static class Incremental extends Mfd.Properties {

    double[] magnitudes;
    double[] rates;

    @Override
    public Mfd.Builder toBuilder() {
      return Mfd.Builder.from(
          magnitudes,
          rates == null ? new double[magnitudes.length] : rates);
    }
  }

  private static JsonElement validateSingle(JsonObject o) {
    checkValue(o, "type");
    return checkValue(o, "m");
  }

  private static JsonElement validateGr(JsonObject o) {
    checkValue(o, "type");
    checkValue(o, "b");
    checkValue(o, "Δm");
    checkValue(o, "mMin");
    return checkValue(o, "mMax");
  }

  private static JsonElement validateGrTaper(JsonObject o) {
    validateGr(o);
    return checkValue(o, "mc");
  }

  private static JsonElement validateIncr(JsonObject o) {
    checkValue(o, "type");
    return checkValue(o, "magnitudes");
  }

  private static JsonObject checkValue(JsonObject obj, String member) {
    checkState(obj.has(member), "'%s' not set", member);
    return obj;
  }

  /* Create a fault source configuration. */
  static SourceConfig.Fault faultConfig(Path json) {
    JsonObject obj = jsonObject(json);
    double spacing = obj.get(SURFACE_SPACING).getAsDouble();
    String floatingStr = obj.get(RUPTURE_FLOATING).getAsString();
    String scalingStr = obj.get(RUPTURE_SCALING).getAsString();

    SourceConfig.Fault.Builder config = SourceConfig.Fault.builder(json)
        .spacing(spacing)
        .floating(RuptureFloating.valueOf(floatingStr))
        .scaling(RuptureScaling.valueOf(scalingStr));

    JsonElement dipTree = obj.get(DIP_TREE);
    if (!dipTree.isJsonNull()) {
      config.dipTree(Deserialize.doubleTree(dipTree, DIP_TREE));
    }

    JsonElement dipSlipModel = obj.get(DIP_SLIP_MODEL);
    if (!dipSlipModel.isJsonNull()) {
      config.dipSlipModel(DipSlipModel.valueOf(dipSlipModel.getAsString()));
    }

    return config.build();
  }

  /* Create an interface source configuration. */
  static SourceConfig.Interface interfaceConfig(Path json) {
    JsonObject obj = jsonObject(json);
    double spacing = obj.get(SURFACE_SPACING).getAsDouble();
    String floatingStr = obj.get(RUPTURE_FLOATING).getAsString();
    String scalingStr = obj.get(RUPTURE_SCALING).getAsString();
    return SourceConfig.Interface.builder(json)
        .spacing(spacing)
        .floating(RuptureFloating.valueOf(floatingStr))
        .scaling(RuptureScaling.valueOf(scalingStr))
        .build();
  }

  /* Create a slab source configuration. */
  static SourceConfig.Slab slabConfig(Path json) {
    JsonObject obj = jsonObject(json);
    double spacing = obj.get(GRID_SPACING).getAsDouble();
    String typeStr = obj.get(POINT_SOURCE_TYPE).getAsString();
    String scalingStr = obj.get(RUPTURE_SCALING).getAsString();
    double maxWidth = obj.get(MAX_WIDTH).getAsDouble();
    JsonElement fmtElement = obj.get(FOCAL_MECH_TREE);
    return SourceConfig.Slab.builder(json)
        .spacing(spacing)
        .sourceType(PointSourceType.valueOf(typeStr))
        .scaling(RuptureScaling.valueOf(scalingStr))
        .maxWidth(maxWidth)
        .focalMechTree(enumValueTree(fmtElement, FocalMech.class, FOCAL_MECH_TREE))
        .build();
  }

  /* Create a grid source configuration. */
  static SourceConfig.Grid gridConfig(Path json) {
    JsonObject obj = jsonObject(json);
    double spacing = obj.get(GRID_SPACING).getAsDouble();
    int smoothingDensity = obj.get(SMOOTHING_DENSITY).getAsInt();
    double smoothingLimit = obj.get(SMOOTHING_LIMIT).getAsDouble();
    JsonElement distanceBinElement = obj.get(DISTANCE_BIN);
    OptionalDouble distanceBin = distanceBinElement.isJsonNull()
        ? OptionalDouble.empty()
        : OptionalDouble.of(distanceBinElement.getAsDouble());
    String typeStr = obj.get(POINT_SOURCE_TYPE).getAsString();
    String scalingStr = obj.get(RUPTURE_SCALING).getAsString();
    double maxDepth = obj.get(MAX_DEPTH).getAsDouble();
    JsonElement gdmElement = obj.get(GRID_DEPTH_MAP);
    JsonElement fmtElement = obj.get(FOCAL_MECH_TREE);
    return SourceConfig.Grid.builder(json)
        .spacing(spacing)
        .smoothingDensity(smoothingDensity)
        .smoothingLimit(smoothingLimit)
        .distanceBin(distanceBin)
        .sourceType(PointSourceType.valueOf(typeStr))
        .scaling(RuptureScaling.valueOf(scalingStr))
        .maxDepth(maxDepth)
        .focalMechTree(enumValueTree(fmtElement, FocalMech.class, FOCAL_MECH_TREE))
        .gridDepthMap(gridDepthMap(gdmElement))
        .build();
  }

  /* Create a gridded seismicity source depth model map. */
  private static Map<String, GridDepthModel> gridDepthMap(JsonElement json) {
    Map<String, GridDepthModel> depthMap = new HashMap<>();
    for (Entry<String, JsonElement> entry : json.getAsJsonObject().entrySet()) {
      depthMap.put(entry.getKey(), depthModel(entry.getValue()));
    }
    return Map.copyOf(depthMap);
  }

  /* Create a gridded seismicity source depth model. */
  private static GridDepthModel depthModel(JsonElement json) {
    JsonObject obj = json.getAsJsonObject();
    Range<Double> mRange = Range.closedOpen(
        obj.get(M_MIN).getAsDouble(),
        obj.get(M_MAX).getAsDouble());
    LogicTree<Double> depthTree = doubleTree(obj.get(DEPTH_TREE), DEPTH_TREE);
    return new GridDepthModel(mRange, depthTree);
  }

  /* Create an enum valued tree. */
  private static <E extends Enum<E>> LogicTree<E> enumValueTree(
      JsonElement json,
      Class<E> clazz,
      String name) {
    LogicTree.EnumValueBuilder<E> builder = LogicTree.enumValueBuilder(name);
    for (JsonElement branch : json.getAsJsonArray()) {
      JsonObject obj = branch.getAsJsonObject();
      builder.addBranch(
          Enum.valueOf(clazz, obj.get(ID).getAsString()),
          obj.get(WEIGHT).getAsDouble());
    }
    return builder.build();
  }

  /* Create double valued logic tree. */
  private static LogicTree<Double> doubleTree(JsonElement json, String name) {
    LogicTree.Builder<Double> builder = LogicTree.builder(name);
    for (JsonElement branch : json.getAsJsonArray()) {
      JsonObject obj = branch.getAsJsonObject();
      builder.addBranch(
          obj.get(ID).getAsString(),
          obj.get(VALUE).getAsDouble(),
          obj.get(WEIGHT).getAsDouble());
    }
    return builder.build();
  }

  /* Create double valued logic tree that permits null values. */
  private static LogicTree<Double> doubleTreeWithNulls(JsonElement json, String name) {
    LogicTree.Builder<Double> builder = LogicTree.builder(name);
    for (JsonElement branch : json.getAsJsonArray()) {
      JsonObject obj = branch.getAsJsonObject();
      JsonElement valueElement = obj.get(VALUE);
      double value = valueElement.isJsonNull() ? 0.0 : valueElement.getAsDouble();
      builder.addBranch(
          obj.get(ID).getAsString(),
          value,
          obj.get(WEIGHT).getAsDouble());
    }
    return builder.build();
  }

  /* Buffered reader wrapper for JSON array. */
  private static JsonArray jsonArray(Path json) {
    try (BufferedReader br = Files.newBufferedReader(json)) {
      return JsonParser.parseReader(br).getAsJsonArray();
    } catch (IOException ioe) {
      throw new RuntimeException(ioe);
    }
  }

  /* Buffered reader wrapper for JSON object. */
  private static JsonObject jsonObject(Path json) {
    try (BufferedReader br = Files.newBufferedReader(json)) {
      return JsonParser.parseReader(br).getAsJsonObject();
    } catch (IOException ioe) {
      throw new RuntimeException(ioe);
    }
  }

  /* Buffered reader wrapper for class. */
  private static <T> T fromJson(Path json, Class<T> clazz) {
    try (BufferedReader br = Files.newBufferedReader(json)) {
      return GSON.fromJson(br, clazz);
    } catch (IOException ioe) {
      throw new RuntimeException(ioe);
    }
  }

}
