package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.Earthquakes.MAG_RANGE;
import static gov.usgs.earthquake.nshmp.model.SourceType.FAULT;
import static java.util.stream.Collectors.toList;
import static java.util.stream.Collectors.toUnmodifiableList;

import java.util.AbstractList;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.data.XyPoint;
import gov.usgs.earthquake.nshmp.fault.surface.DefaultGriddedSurface;
import gov.usgs.earthquake.nshmp.fault.surface.GriddedSurface;
import gov.usgs.earthquake.nshmp.fault.surface.RuptureFloating;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Properties;
import gov.usgs.earthquake.nshmp.mfd.Mfds;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Fault source representation. This class wraps a model of a fault geometry and
 * a list of magnitude frequency distributions that characterize how the fault
 * might rupture (e.g. as one, single geometry-filling event, or as multiple
 * smaller events) during earthquakes. Smaller events are modeled as 'floating'
 * ruptures; they occur in multiple locations on the fault surface with
 * appropriately scaled rates.
 *
 * <p>A {@code FaultSource} cannot be created directly; it may only be created
 * by a private parser.
 *
 * @author U.S. Geological Survey
 */
public class FaultSource extends AbstractList<Rupture> implements Source {

  final SourceFeature.NshmFault feature;
  final SourceConfig.Fault config;
  final GriddedSurface surface;

  final LogicTree<Mfd> mfdTree;

  private final List<Mfd> scaledMfds; // includes ruptureSet weight
  private final List<List<Rupture>> ruptureLists; // 1:1 with Mfds
  private final List<Rupture> ruptureList; // concatenated rupture list
  final List<Boolean> floats;

  FaultSource(Builder builder) {
    this.feature = builder.feature;
    this.config = builder.config;
    this.mfdTree = builder.mfdTree;

    this.surface = DefaultGriddedSurface.builder()
        .trace(feature.trace)
        .depth(feature.upperDepth)
        .dip(feature.dip)
        .width(feature.width)
        .spacing(config.surfaceSpacing)
        .build();

    this.scaledMfds = ModelTrees.scaledMfdList(mfdTree);

    this.ruptureLists = new ArrayList<>(mfdTree.size());
    this.floats = new ArrayList<>(mfdTree.size());
    scaledMfds.forEach(mfd -> {
      boolean floats = floats(mfd);
      this.floats.add(floats);
      this.ruptureLists.add(createRuptureList(mfd, floats));
    });
    ruptureList = ruptureLists.stream()
        .flatMap(List::stream)
        .collect(toList());

    checkState(
        ruptureLists.stream().mapToInt(List::size).sum() > 0,
        "FaultSource has no ruptures");

    // long nRuptures = ruptureLists.stream().flatMap(List::stream).count();
    // System.out.println(nRuptures);

    // System.out.println(mfdTree);
    // System.out.println(Mfds.momentRate(scaledMfds.get(0)));
    // System.out.println(trace);
    // System.out.println(surface);

    // scaledMfds.stream().forEach(mfd -> System.out.println(mfd.properties()));
    // scaledMfds.stream().forEach(System.out::println);
    // System.out.println(mfdTree);
  }

  // further compact the scaledMfds list by combining all
  // SINGLE and GR MFDs for fault source trees

  static final Set<Mfd.Type> TYPES = EnumSet.of(Mfd.Type.GR, Mfd.Type.SINGLE);

  static void checkTypes(List<Mfd> mfds, String name) {
    // Set<Mfd.Type> types = mfds.stream()
    // .map(Mfd::properties)
    // .map(Properties::type)
    // .collect(Collectors.toSet());

    Map<Mfd.Type, Long> counted = mfds.stream()
        .map(Mfd::properties)
        .map(Properties::type)
        .collect(Collectors.groupingBy(Function.identity(), Collectors.counting()));

    // if (!counted.keySet().equals(TYPES)) {
    System.out.println(name + ": " + counted);
    // }
  }

  static List<Mfd> reduceMfdList(List<Mfd> mfds) {
    // mfds.stream().forEach(mfd ->
    // System.out.println(mfd.properties().type()));

    List<Mfd> grMfds = mfds.stream()
        .filter(mfd -> mfd.properties().type() == Mfd.Type.GR)
        .collect(Collectors.toList());
    Mfd grMfd = Mfds.combine(grMfds);

    double mMax = grMfds.stream()
        .mapToDouble(mfd -> mfd.properties().getAsGr().mMax())
        .average()
        .orElseThrow();

    Properties.GutenbergRichter grProps = grMfds.get(0).properties().getAsGr();
    grProps = new Properties.GutenbergRichter(
        grProps.a(), grProps.b(),
        grProps.Δm(), grProps.mMin(), Maths.round(mMax, 2));
    grMfd = Mfd.create(grMfd.data(), grProps);

    List<Mfd> singleMfds = mfds.stream()
        .filter(mfd -> mfd.properties().type() == Mfd.Type.SINGLE)
        .collect(Collectors.toList());
    Mfd singleMfd = Mfds.combine(singleMfds);

    double m = singleMfds.stream()
        .mapToDouble(mfd -> mfd.properties().getAsSingle().magnitude())
        .average()
        .orElseThrow();

    Properties.Single singleProps = singleMfds.get(0).properties().getAsSingle();
    singleProps = new Properties.Single(Maths.round(m, 2));
    singleMfd = Mfd.create(singleMfd.data(), singleProps);

    return List.of(grMfd, singleMfd);
  }

  @Override
  public String name() {
    return feature.name;
  }

  @Override
  public int id() {
    return feature.id;
  }

  @Override
  public SourceType type() {
    return FAULT;
  }

  @Override
  public int size() {
    return ruptureList.size();
  }

  @Override
  public Rupture get(int index) {
    return ruptureList.get(index);
  }

  /**
   * The closest point on the fault trace, relative to the supplied site
   * {@code Location}.
   */
  @Override
  public Location location(Location site) {
    return Locations.closestPoint(site, feature.trace);
  }

  @Override
  public List<Mfd> mfds() {
    return mfdTree.stream()
        .map(Branch::value)
        .collect(toUnmodifiableList());
  }

  @Override
  public String toString() {
    Map<String, ?> data = Map.of(
        "name", feature.name,
        "dip", feature.dip,
        "width", feature.width,
        "rake", feature.rake,
        "mfds", mfdTree.size(),
        "top", feature.trace.first().depth);
    return getClass().getSimpleName() + " " + data;
  }

  /*
   * Notes on floating ruptures:
   *
   * By default, GR always floats and SINGLE never floats, but there are
   * exceptions (e.g. regional or historic mMax < characteristic (WC-Length) M.
   *
   * If 'floats' is false, rupture list builder will create a single rupture for
   * each magnitude in the supplied MFD.
   *
   * If 'floats' is true, rupture list builder will make an additional check
   * that rupture length is shorter than the trace and only float those that are
   * shorter.
   *
   * Width is irrelevant as we know current WUS models only use WC-Length and
   * NSHM down dip floating model where rupture base is constant regardless of
   * magnitude.
   *
   * Computed lengths of faults in the CONUS 2014/2018 NSHMs are inconsistent
   * between this and legacy codes/inputs. Most WUS fault source features
   * therefore include a length property. The supplied length is then used to
   * compute 'characteristic' magnitudes that agree with magnitudes in the
   * prescriptive MFDs of the XML CONUS model.
   */

  private boolean floats(Mfd mfd) {

    double magRef = feature.magnitude.orElse(MAG_RANGE.upperEndpoint());
    double mLength = config.ruptureScaling.dimensions(magRef, feature.width).length;
    double traceLength = feature.trace.length();
    Mfd.Type mfdType = mfd.properties().type();

    boolean floats = (mfdType == Mfd.Type.GR);

    /*
     * NSHM floating rules for SINGLE MFDs: Ruptures for SINGLE MFDs do not
     * float by default. First float check; use reference length from source to
     * determine if true for SINGLE MFDs, otherwise epistemic branches end up
     * floating.
     */
    if (mfdType == Mfd.Type.SINGLE && config.ruptureFloating == RuptureFloating.NSHM) {
      if (feature.magnitudeNote.isEmpty()) {
        floats = false;
      } else {
        MagnitudeNote mNote = feature.magnitudeNote.orElseThrow();
        /*
         * MagnitudeNotes provide hints for alternate behavior.
         *
         * REGIONAL_M_MAX is a cap and always reduces M below char.
         * HISTORIC_M_MAX may be higher or lower than char so check lengths.
         */
        floats = (mNote == MagnitudeNote.REGIONAL_M_MAX) ||
            ((mNote == MagnitudeNote.HISTORIC_M_MAX) && mLength < traceLength);
      }
    } else {
      /* Unless NSHM, use floating model from config. */
      floats = true;
    }
    return floats;
  }

  private List<Rupture> createRuptureList(Mfd mfd, boolean floats) {

    List<Rupture> ruptureList = new ArrayList<>();
    for (XyPoint xy : mfd.data()) {
      double mag = xy.x();
      double rate = xy.y();

      // Check if low rate cutoff is used; get value from config
      if (rate < 1e-14) {
        continue; // shortcut low rates
      }

      RuptureFloating floating = (floats)
          ? config.ruptureFloating
          : RuptureFloating.OFF;

      List<Rupture> floaters = floating.createFloatingRuptures(
          surface,
          config.ruptureScaling,
          mag,
          rate,
          feature.rake);

      ruptureList.addAll(floaters);
    }

    // We now have rates of zero for null branches that preserve
    // logic tree structure but that result in empty rupture lists
    // checkState(ruptureList.size() > 0, "Rupture list for %s is empty", name);

    return ruptureList;
  }

  /* Single use builder */
  static class Builder {

    private boolean built = false;

    SourceFeature.NshmFault feature;
    SourceConfig.Fault config;
    LogicTree<Mfd> mfdTree;
    Mfd mfdTotal;

    Builder feature(SourceFeature.NshmFault feature) {
      this.feature = feature;
      return this;
    }

    Builder config(SourceConfig.Fault config) {
      this.config = config;
      return this;
    }

    Builder mfdTree(LogicTree<Mfd> mfdTree) {
      checkArgument(checkNotNull(mfdTree).size() > 0, "MFD list is empty");
      this.mfdTree = mfdTree;
      return this;
    }

    Builder mfdTotal(Mfd mfdTotal) {
      this.mfdTotal = checkNotNull(mfdTotal);
      return this;
    }

    void validateState(String label) {
      checkState(!built, "Single use builder");
      checkNotNull(feature, "%s feature", label);
      checkNotNull(config, "%s config", label);
      checkNotNull(mfdTree, "%s MFD tree", label);
      checkNotNull(mfdTotal, "%s total MFD", label);
      built = true;
    }

    FaultSource build() {
      validateState("FaultSource.Builder");
      return new FaultSource(this);
    }
  }
}
