package gov.usgs.earthquake.nshmp.model;

import static java.util.stream.Collectors.toUnmodifiableMap;

import java.util.Map;
import java.util.Optional;
import java.util.OptionalDouble;
import java.util.Set;

import com.google.gson.JsonElement;

import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * GMM logic tree convenience container class that holds a
 * {@code LogicTree<Gmm>} and properties derived from a {@code GmmConfig}.
 *
 * @author U.S. Geological Survey
 *
 */
public class GmmTree {

  /*
   * GmmTree is a replactement for GmmSet. GmmSet supported multiple GMM logic
   * trees for different distances. This behavior is not needed for current
   * models but will need to be resurrected in some format when we get around to
   * converting CONUS 2014 and earlier. Currently when a model is parsed,
   * GmmSets are initialized for each tectonic setting and source type by
   * reading a gmm-tree and gmm-config. However, we now also support region
   * specific GMMs, but the gmm-config data (max distance) isn't stored in with
   * the site data polygons and we'd rather not repeat it. We therefore store
   * maxDistance as an optional; it is only used for initializing grid source
   * tables and iterableForLocation. TODO Can we remove it to be a property of a
   * source tree or rupture set.
   *
   * EnumSet or ImmutableEnumSet?
   */

  private final LogicTree<Gmm> tree;
  private final Map<Gmm, Double> gmms;

  /* GmmConfig fields. */
  private final OptionalDouble maxDistance;

  private GmmTree(LogicTree<Gmm> tree, Optional<GmmConfig> config) {
    this.tree = tree;
    this.gmms = tree.stream()
        .collect(toUnmodifiableMap(
            Branch::value,
            Branch::weight));
    this.maxDistance = config.isPresent()
        ? OptionalDouble.of(config.orElseThrow().maxDistance)
        : OptionalDouble.empty();
  }

  /** Create a new GMM logic tree container. */
  public static GmmTree create(LogicTree<Gmm> tree, Optional<GmmConfig> config) {
    return new GmmTree(tree, config);
  }

  /** The GMM logic tree. */
  public LogicTree<Gmm> tree() {
    return tree;
  }

  /** The set of GMMs in the logic tree. */
  public Set<Gmm> gmms() {
    return gmms.keySet();
  }

  /**
   * The logic tree weight of the requested GMM.
   * @throws NullPointerException if the supplied GMM is not in this tree
   */
  public double gmmWeight(Gmm gmm) {
    return gmms.get(gmm);
  }

  /** The maximum applicable distance for the GMMs in the logic tree. */
  public OptionalDouble maxDistance() {
    return maxDistance;
  }

  /**
   * Convenience method to parse a JSON logic tree of GMMs and associated
   * weights.
   *
   * @param e JsonElement to parse
   * @return a logic tree of GMMs
   */
  public static LogicTree<Gmm> parseTree(JsonElement e) {
    return Deserialize.gmmTree(e);
  }

}
