package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.model.Deserialize.MFD_TREE;
import static java.lang.Math.log10;
import static java.lang.Math.pow;
import static java.util.stream.Collectors.toUnmodifiableList;

import java.nio.file.Path;
import java.util.ArrayList;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.OptionalDouble;
import java.util.stream.Stream;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.data.DelimitedData;
import gov.usgs.earthquake.nshmp.data.DelimitedData.Record;
import gov.usgs.earthquake.nshmp.data.XyPoint;
import gov.usgs.earthquake.nshmp.fault.FocalMech;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Properties.GutenbergRichter;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Type;
import gov.usgs.earthquake.nshmp.mfd.Mfds;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;
import gov.usgs.earthquake.nshmp.tree.Trees;

class GridLoader {

  private GridLoader() {}

  /*
   * Currently specialized flavors of GR are used for double-counting processing
   * but it would be preferable to do it when the model loads. This would permit
   * structural changes to a fault model without having to regnerate a gridded
   * source model. Note in documentation.
   */

  /*
   * Reduce location mfd-tree to total MFD. All MFDs should be GR* with
   * variations in mMax or GR-type and backed by common magnitude arrays that
   * are not copied when MFDs are combined.
   */

  /*
   * 9/2022 - given other separations of slab and grid handling (e.g.
   * SlabRuptureSet no longer extends GridRuptureSet and slab config no longer
   * uses SourceConfig.Grid) we've added a custom GR loader for slab sources
   * that allows customization without the complexity of handling the growing
   * number of differences between gridded sources in different tectonix
   * settings. For example, SlabLoader processes depths from spatial PDFs and
   * removes handling for mMax truncation. This allows for greater flexibility
   * and reduces risk of inadvertant side effects to other gridded sources when
   * new behavior is required.
   */

  // TODO deal with GR_TAPER mix in in WUS

  /* Grid csv file column header and property keys. */
  static class Key {
    static final String LON = "lon";
    static final String LAT = "lat";
    static final String DEPTH = "depth";
    static final String STRIKE = "strike";
    static final String RATE = "rate";
    static final String NAME = "name";
    static final String PDF = "pdf";
    static final String MMAX = "mMax";
    static final String STRIKE_SLIP_WEIGHT = "ss_wt";
    static final String REVERSE_WEIGHT = "r_wt";
    static final String NORMAL_WEIGHT = "n_wt";
  }

  static SlabRuptureSet createSlabRuptureSet(
      int id,
      String name,
      int featureId,
      String mfdTreeKey,
      ModelData data) {

    SlabLoader loader = new SlabLoader(id, name, featureId, mfdTreeKey, data);
    return loader.createSlabRuptureSet();
  }

  static GridRuptureSet createGridRuptureSet(
      int id,
      String name,
      int featureId,
      String mfdTreeKey,
      ModelData data) {

    /*
     * If rate-tree is absent at a grid logic tree branch tip,
     * data.gridRateFile() is a grid-mfds file that defines explicit INCR MFDs
     * (e.g. UCERF3, CA 2008 NSHM). Otherwise data.gridRateFile() is a
     * spatial-pdf that will be scaled by values in the rate-tree.
     */
    if (data.spatialPdf()) {
      GrLoader loader = new GrLoader(id, name, featureId, mfdTreeKey, data);
      return loader.createGridRuptureSet();
    }
    IncrLoader loader = new IncrLoader(id, name, featureId, mfdTreeKey, data);
    return loader.createRuptureSet();
  }

  /* Explicit INCR MFD (fault systems, e.g. UCERF3). */
  private static class IncrLoader {

    private final int id;
    private final String name;
    private final ModelData data;
    private final SourceFeature.Grid feature;
    private final Path mfdsPath;
    private final String mfdTreeKey;
    private final FeatureData featureData;

    private IncrLoader(
        int id,
        String name,
        int featureId,
        String mfdTreeKey,
        ModelData data) {

      this.id = id;
      this.name = name;
      this.data = data;
      this.feature = data.gridFeatureMap().orElseThrow().get(featureId);
      this.mfdsPath = data.gridRateFile().orElseThrow();
      this.mfdTreeKey = mfdTreeKey;
      this.featureData = new FeatureData();
      processNodes();
      checkState(!featureData.focalMechMaps.isEmpty());
    }

    private GridRuptureSet createRuptureSet() {
      return GridRuptureSet.builder()
          .name(name)
          .id(id)
          .modelData(data)
          .feature(feature)
          .nodeData(
              featureData.locations,
              featureData.mfds,
              featureData.mfdsTree,
              Optional.of(featureData.focalMechMaps))
          .build();
    }

    private void processNodes() {

      DelimitedData ratefile = DelimitedData.comma(mfdsPath);

      /* Because we have explicit MFDs, we expect a single branch mfd-tree. */
      LogicTree<Mfd.Properties> mfdTree = data.mfdMap().orElseThrow().get(mfdTreeKey);
      checkState(mfdTree.size() == 1);
      Branch<Mfd.Properties> propsBranch = mfdTree.get(0);
      Mfd.Builder mfdBuilder = propsBranch.value().toBuilder();

      try (Stream<Record> s = ratefile.records()) {
        s.forEach(record -> processRecord(record, mfdBuilder));
      }

      featureData.mfdsTree = LogicTree.singleton(
          feature.name,
          propsBranch.id(),
          featureData.mfds);
    }

    /* Create node MFD trees skipping nodes outside feature. */
    private void processRecord(Record record, Mfd.Builder builder) {

      Location loc = Location.create(
          record.getDouble(Key.LON),
          record.getDouble(Key.LAT));
      if (!feature.contains(loc)) {
        return;
      }

      Map<FocalMech, Double> focalMechMap = createFocalMechMap(
          record.getDouble(Key.STRIKE_SLIP_WEIGHT),
          record.getDouble(Key.REVERSE_WEIGHT),
          record.getDouble(Key.NORMAL_WEIGHT));

      IncrRates incrRates = new IncrRates(record);
      builder.transform(incrRates::set);

      featureData.locations.add(loc);
      featureData.focalMechMaps.add(focalMechMap);
      featureData.mfds.add(builder.build());
    }

    private static class IncrRates {

      final Record record;

      IncrRates(Record record) {
        this.record = record;
      }

      void set(XyPoint mBin) {
        mBin.set(record.getDouble(Double.toString(mBin.x())));
      }
    }
  }

  private static Map<FocalMech, Double> createFocalMechMap(
      double ssWeight,
      double rWeight,
      double nWeight) {

    // This should have validation of weight values and totals
    EnumMap<FocalMech, Double> fmMap = new EnumMap<>(FocalMech.class);
    fmMap.put(FocalMech.STRIKE_SLIP, ssWeight);
    fmMap.put(FocalMech.REVERSE, rWeight);
    fmMap.put(FocalMech.NORMAL, nWeight);
    return fmMap;
  }

  /* Handles logic trees of GR MFDs and multiple features. */
  private static class GrLoader {

    private final int id;
    private final String name;
    private final ModelData data;
    private final SourceFeature.Grid feature;
    private final Path pdfPath;
    private final Optional<LogicTree<Double>> rateTree;
    private final LogicTree<Mfd> modelMfdTree;
    private final FeatureData featureData;

    private GrLoader(
        int id,
        String name,
        int featureId,
        String mfdTreeKey,
        ModelData data) {

      this.id = id;
      this.name = name;
      this.data = data;
      this.feature = data.gridFeatureMap().orElseThrow().get(featureId);
      this.pdfPath = data.gridRateFile().orElseThrow();
      this.rateTree = data.rateTree();
      // System.out.println(mfdTreeKey);
      LogicTree<Mfd.Properties> mfdPropsTree = data.mfdMap().orElseThrow().get(mfdTreeKey);
      // System.out.println(mfdPropsTree);
      this.modelMfdTree = ModelTrees.grPropsTreeToMmaxMfdTree(mfdPropsTree);
      this.featureData = new FeatureData();
      processNodes();
    }

    private GridRuptureSet createGridRuptureSet() {
      return GridRuptureSet.builder()
          .name(name)
          .id(id)
          .modelData(data)
          .feature(feature)
          .nodeData(
              featureData.locations,
              featureData.mfds,
              featureData.mfdsTree,
              Optional.empty())
          .build();
    }

    private void processNodes() {
      DelimitedData data = DelimitedData.comma(pdfPath);
      /* May or may not have mMax column in rate file. */
      boolean mMaxPresent = data.columnKeys().contains(Key.MMAX);
      try (Stream<Record> s = data.records()) {
        s.forEach(record -> processRecord(record, mMaxPresent));
      }

      /* Create total MFDs and transpose mfd tree lists. */
      featureData.mfds = reduceMfds(featureData.mfdTrees);
      featureData.mfdsTree = Trees.transpose(featureData.mfdTrees);
    }

    private List<Mfd> reduceMfds(List<LogicTree<Mfd>> mfdTrees) {
      return mfdTrees.stream()
          .map(ModelTrees::reduceMfdTree)
          .collect(toUnmodifiableList());
    }

    /* Create node MFD trees skipping nodes outside feature. */
    private void processRecord(Record record, boolean mMaxPresent) {

      Location loc = Location.create(
          record.getDouble(Key.LON),
          record.getDouble(Key.LAT));
      if (!feature.contains(loc)) {
        return;
      }

      featureData.locations.add(loc);
      double pdfValue = record.getDouble(Key.PDF);
      OptionalDouble mMax = mMaxPresent
          ? record.getOptionalDouble(Key.MMAX)
          : OptionalDouble.empty();
      LogicTree<Mfd> nodeMfds = createNodeMfdTree(
          modelMfdTree, rateTree,
          pdfValue, mMax);
      featureData.mfdTrees.add(nodeMfds);
    }
  }

  /* Handles logic trees of GR MFDs for slab sources; single feature. */
  private static class SlabLoader {

    private final int id;
    private final String name;
    private final ModelData data;
    private final SourceFeature.Grid feature;
    private final Path pdfPath;
    private final Optional<LogicTree<Double>> rateTree;
    private final LogicTree<Mfd> modelMfdTree;
    private final FeatureData featureData;

    private SlabLoader(
        int id,
        String name,
        int featureId,
        String mfdTreeKey,
        ModelData data) {

      this.id = id;
      this.name = name;
      this.data = data;
      this.feature = data.gridFeatureMap().orElseThrow().get(featureId);
      this.pdfPath = data.gridRateFile().orElseThrow();
      this.rateTree = data.rateTree();
      // System.out.println(mfdTreeKey);
      LogicTree<Mfd.Properties> mfdPropsTree = data.mfdMap().orElseThrow().get(mfdTreeKey);
      // System.out.println(mfdPropsTree);
      this.modelMfdTree = ModelTrees.grPropsTreeToMmaxMfdTree(mfdPropsTree);
      this.featureData = new FeatureData();
      processNodes();
    }

    private SlabRuptureSet createSlabRuptureSet() {
      SlabRuptureSet.Builder b = SlabRuptureSet.builder();
      b.name(name)
          .id(id)
          .modelData(data)
          .feature(feature)
          .nodeData(
              featureData.locations,
              featureData.mfds,
              featureData.mfdsTree);
      return b.build();
    }

    private void processNodes() {
      DelimitedData data = DelimitedData.comma(pdfPath);
      try (Stream<Record> s = data.records()) {
        s.forEach(record -> processRecord(record));
      }

      /* Create total MFDs and transpose mfd tree lists. */
      featureData.mfds = reduceMfds(featureData.mfdTrees);
      featureData.mfdsTree = Trees.transpose(featureData.mfdTrees);
    }

    /* Create node MFD trees from each record in a spatial PDF. */
    private void processRecord(Record record) {

      Location loc = Location.create(
          record.getDouble(Key.LON),
          record.getDouble(Key.LAT),
          record.getDouble(Key.DEPTH));

      featureData.locations.add(loc);
      double pdfValue = record.getDouble(Key.PDF);
      LogicTree<Mfd> nodeMfds = createNodeMfdTree(
          modelMfdTree, rateTree,
          pdfValue, OptionalDouble.empty());
      featureData.mfdTrees.add(nodeMfds);
    }
  }

  /* Collapse the node MFDs logic trees. */
  private static List<Mfd> reduceMfds(List<LogicTree<Mfd>> mfdTrees) {
    return mfdTrees.stream()
        .map(ModelTrees::reduceMfdTree)
        .collect(toUnmodifiableList());
  }

  /* Shared by Grid GrLoader and SlabLoader. */
  private static LogicTree<Mfd> createNodeMfdTree(
      LogicTree<Mfd> modelTree,
      Optional<LogicTree<Double>> rateTree,
      double pdfValue,
      OptionalDouble mMax) {

    LogicTree.Builder<Mfd> nodeMfdTree = LogicTree.builder(MFD_TREE);
    for (Branch<Mfd> mBranch : modelTree) {
      Mfd.Properties.GutenbergRichter grProps = mBranch.value().properties().getAsGr();
      if (rateTree.isPresent()) {
        for (Branch<Double> rBranch : rateTree.orElseThrow()) {
          double mMin = grProps.mMin() + grProps.Δm() * 0.5;
          double a = log10(rBranch.value() * pdfValue);
          double rate = Mfds.gutenbergRichterRate(a, grProps.b(), mMin);
          Mfd.Builder nodeMfd = Mfd.Builder.from(mBranch.value())
              .scaleToIncrementalRate(rate);
          mMax.ifPresent(m -> truncate(grProps, nodeMfd, m));
          String id = String.join(":", Mfd.Type.GR.name(), mBranch.id(), rBranch.id());
          double weight = mBranch.weight() * rBranch.weight();
          nodeMfdTree.addBranch(id, nodeMfd.build(), weight);
        }
      } else {
        // all branches are fully specified in the MFD
        // (i.e. the MFD contains a-values)
        double mMin = grProps.mMin() + grProps.Δm() * 0.5;
        // note conversion from a --> A for PDF scaling and then back to log
        double a = log10(pow(10, grProps.a()) * pdfValue);
        double rate = Mfds.gutenbergRichterRate(a, grProps.b(), mMin);
        Mfd.Builder nodeMfd = Mfd.Builder.from(mBranch.value())
            .scaleToIncrementalRate(rate);
        mMax.ifPresent(m -> truncate(grProps, nodeMfd, m));
        String id = String.join(":", Mfd.Type.GR.name(), mBranch.id());
        nodeMfdTree.addBranch(id, nodeMfd.build(), mBranch.weight());
      }
    }
    return nodeMfdTree.build();
  }

  /*
   * Data container for feature polygon.
   *
   * Different fields are used by GrLoader and IncrLoader.
   */
  private static class FeatureData {

    /* node specific data */
    List<Location> locations = new ArrayList<>();
    List<LogicTree<Mfd>> mfdTrees = new ArrayList<>();
    List<Map<FocalMech, Double>> focalMechMaps = new ArrayList<>();

    /* transposed and combined data */
    LogicTree<List<Mfd>> mfdsTree;
    List<Mfd> mfds = new ArrayList<>();
  }

  /* Truncate special GR cases; e.g. WUS double counting. */
  private static void truncate(
      GutenbergRichter grProps,
      Mfd.Builder builder,
      double mMax) {
    Mfd.Type type = grProps.type();
    if (type == Type.GR_MMAX_GR) {
      builder.transform(xy -> zeroAboveM(xy, Math.min(mMax, 6.5)));
    } else if (type == Type.GR_MMAX_SINGLE) {
      /*
       * Keeps the magnitude bin just above mMax if mMax > m_gr + Δm/2. Rounding
       * to 5 decimal places ensures that mMax/Δm doesn't equal *.00...001,
       * which ceil() would take up to *.1
       *
       * We really should just be cutting out the uppermost bin with overlap but
       * this is consistent with what was being done in legacy fortran codes.
       */
      double Δm = grProps.Δm();
      double m = Math.ceil(Maths.round(mMax / Δm, 5)) * Δm;
      builder.transform(xy -> zeroAboveM(xy, m));
    }
  }

  static GridRuptureSet createGrid(
      String name,
      int id,
      ModelData data,
      SourceType type, // remove for zone specific method instead??
      SourceFeature feature,
      List<Location> locations,
      List<Mfd> mfds,
      LogicTree<List<Mfd>> mfdsTree,
      Optional<List<Map<FocalMech, Double>>> focalMechMaps) {

    GridRuptureSet.Builder builder = new GridRuptureSet.Builder(type)
        .name(name)
        .id(id)
        .feature(feature)
        .modelData(data);

    // System.out.println(mfds.get(0).properties().type());
    builder.nodeData(locations, mfds, mfdsTree, focalMechMaps);

    return builder.build();
  }

  private static void zeroAboveM(XyPoint xy, double m) {
    if (xy.x() > m) {
      xy.set(0.0);
    }
  }

  /* Read a lon,lat,rate file into the supplied arrays */
  static void readZoneRateFile(
      Path csv,
      List<Location> locations,
      List<Double> rates) {

    try (Stream<Record> s = DelimitedData.comma(csv).records()) {
      s.forEach(record -> {
        locations.add(Location.create(
            record.getDouble(Key.LON),
            record.getDouble(Key.LAT)));
        rates.add(record.getDouble(Key.RATE));
      });
    }
  }

}
