package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.Earthquakes.MU;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.stream.Collectors;

import gov.usgs.earthquake.nshmp.Earthquakes;
import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.data.Sequences;
import gov.usgs.earthquake.nshmp.fault.surface.GriddedSurface;
import gov.usgs.earthquake.nshmp.fault.surface.RuptureSurface;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.tree.Branch;

class InterfaceSlipCalc {

  public static void main(String[] args) throws IOException {
    calcSlip();
  }

  private static void calcSlip() throws IOException {
    // Path testModel = Paths.get("../nshm-conus-2018-5.x-maint");
    Path testModel = Paths.get("../nshm-conus");
    HazardModel model = ModelLoader.load(testModel);

    double[] lats = Sequences.arrayBuilder(40.0, 49.5, 0.02)
        .centered()
        .scale(2)
        .build();

    // map of slip by latitude
    Map<Double, Double> latSlipMap = new TreeMap<>();
    for (int i = 0; i < lats.length; i++) {
      latSlipMap.put(lats[i], 0.0);
    }

    // get Cascadia tree
    SourceTree tree = model.tree(3199).orElseThrow();

    for (Branch<RuptureSet<? extends Source>> branch : tree) {
      double weight = branch.weight();

      if (branch.value() instanceof ClusterRuptureSet) {
        System.out.println("Cluster BranchID: " + branch.id() + " " + branch.value().name());
        ClusterRuptureSet crs = (ClusterRuptureSet) branch.value();
        ClusterSource csrc = crs.get(0);
        for (RuptureSet<? extends Source> rs : csrc.ruptureSets()) {
          InterfaceRuptureSet irs = (InterfaceRuptureSet) rs;
          InterfaceSource isrc = irs.get(0);
          for (Rupture rup : isrc) {
            computeSlip(rup, weight * csrc.rate(), latSlipMap);
          }
        }
        continue;
      }
      InterfaceRuptureSet irs = (InterfaceRuptureSet) branch.value();

      // System.out.println("Wt: " + weight);
      String id = branch.id();
      System.out.println("BranchID: " + id + " " + irs.name());
      // System.out.println(irs);
      // System.out.println(irs.mfdTree());
      // System.out.println(irs.features());
      checkState(irs.size() == 1);

      InterfaceSource src = irs.get(0);

      // System.out.println(src.size());
      // System.out.println(src.mfds());

      for (Rupture rup : src) {
        // double m = rup.mag();
        // double rate = rup.rate();
        // pretty sure rate has been scaled for rupture set
        // but still need to include source tre weights
        // RuptureSurface surf = rup.surface();
        // System.out.println(surf.getClass());
        computeSlip(rup, weight, latSlipMap);
        // break;
      }

      // break;
    }
    System.out.println(latSlipMap);

    // Path out = Path.of("tmp/cascadia-slip/cascadia-lat-slip-2018.csv");
    // Path out =
    // Path.of("tmp/cascadia-slip/cascadia-lat-slip-2023-middle-unseg-cluster-out.csv");
    Path out = Path.of("tmp/cascadia-slip/cascadia-lat-slip-2023-top.csv");

    List<String> lines = latSlipMap.entrySet().stream()
        .map(e -> e.getKey() + "," + Maths.round(e.getValue(), 2))
        .collect(Collectors.toList());

    Files.write(out, lines);

    // System.out.println(tree.weights().values().stream().mapToDouble(Double::doubleValue).sum());

    // for (Branch<RuptureSet<? extends Source>> branch : tree) {
    // System.out.println(branch);
    // }
  }

  // Populates <latitude,slip> map with values from rupture
  private static void computeSlip(Rupture rup, double weight, Map<Double, Double> slipMap) {
    double m = rup.mag();
    double rate = rup.rate();
    double mo = Earthquakes.magToMoment(m);
    // double moRate = mo * rate;

    RuptureSurface surface = rup.surface();
    List<RuptureDimensions> dimensions = computeColumnAreas((GriddedSurface) surface);

    /* Summation of Wi^2 * Li for 'c' constant from Frankel cszslipm.f */
    double afSumWidth = dimensions.stream()
        .mapToDouble(rd -> rd.length * rd.width * rd.width)
        .sum();
    double c = mo / (MU * afSumWidth);

    // double totalArea = areas.stream().mapToDouble(ra -> ra.area).sum();

    for (RuptureDimensions rd : dimensions) {
      // double areaFraction = ra.area / totalArea;
      // double moRateFraction = moRate * areaFraction;
      // double weightedSlipRate = weight * Earthquakes.slip(ra.area,
      // moRateFraction) / 1000.0;

      double slip = c * rd.width;
      double weightedSlipRate = (slip * weight * rate) / 1000.0;

      for (Double lat : slipMap.keySet()) {
        if (lat >= rd.minLat && lat < rd.maxLat) {
          slipMap.merge(lat, weightedSlipRate, Double::sum);
        }
      }
    }
  }

  private static List<RuptureDimensions> computeColumnAreas(GriddedSurface surf) {
    LocationList upperTrace = surf.getEvenlyDiscritizedUpperEdge();
    // System.out.println(upperTrace);
    LocationList lowerTrace = surf.getEvenlyDiscritizedLowerEdge();
    checkState(upperTrace.size() == lowerTrace.size());
    List<RuptureDimensions> dimensionsList = new ArrayList<>();

    for (int i = 0; i < upperTrace.size() - 1; i++) {
      Location ut1 = upperTrace.get(i);
      Location ut2 = upperTrace.get(i + 1);
      Location lt1 = lowerTrace.get(i);
      Location lt2 = lowerTrace.get(i + 1);

      // lengths
      double ul = Locations.linearDistanceFast(ut1, ut2);
      double ll = Locations.linearDistanceFast(lt1, lt2);
      double length = (ul + ll) / 2.0;

      // widths
      double w1 = Locations.linearDistanceFast(ut1, lt1);
      double w2 = Locations.linearDistanceFast(ut2, lt2);
      double width = (w1 + w2) / 2.0;

      // double area = ((ul + ll) / 2.0) * ((w1 + w2) / 2.0);
      RuptureDimensions dimensions = new RuptureDimensions(
          ut1.latitude, ut2.latitude,
          length, width);
      dimensionsList.add(dimensions);
    }
    return dimensionsList;
  }

  /* Holds the average length and width of a rupture strip */
  private static class RuptureDimensions {

    final double minLat;
    final double maxLat;
    final double length;
    final double width;

    RuptureDimensions(double minLat, double maxLat, double length, double width) {
      this.minLat = minLat;
      this.maxLat = maxLat;
      this.length = length;
      this.width = width;
    }
  }

}
