package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static com.google.common.base.StandardSystemProperty.LINE_SEPARATOR;
import static gov.usgs.earthquake.nshmp.model.SourceType.INTERFACE;
import static java.util.stream.Collectors.toList;
import static java.util.stream.Collectors.toUnmodifiableList;

import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.google.common.collect.Iterables;

import gov.usgs.earthquake.nshmp.data.XyPoint;
import gov.usgs.earthquake.nshmp.fault.surface.ApproxGriddedSurface;
import gov.usgs.earthquake.nshmp.fault.surface.GriddedSurface;
import gov.usgs.earthquake.nshmp.fault.surface.RuptureFloating;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Subduction source representation. This class wraps a model of a subduction
 * interface geometry and a list of magnitude frequency distributions that
 * characterize how the fault might rupture (e.g. as one, single
 * geometry-filling event, or as multiple smaller events) during earthquakes.
 * Smaller events are modeled as 'floating' ruptures; they occur in multiple
 * locations on the fault surface with appropriately scaled rates.
 *
 * <p>An {@code InterfaceSource} cannot be created directly; it may only be
 * created by a private parser.
 *
 * @author U.S. Geological Survey
 */
public class InterfaceSource extends AbstractList<Rupture> implements Source {

  final SourceFeature.Interface feature;
  final SourceConfig.Interface config;
  final GriddedSurface surface;

  // temp remove
  final LocationList upperTrace;
  final double upperTraceLength;
  final LocationList lowerTrace;
  final double lowerTraceLength;

  final LogicTree<Mfd> mfdTree;

  private final List<Mfd> scaledMfds; // includes ruptures weight
  private final List<List<Rupture>> ruptureLists; // 1:1 with Mfds
  private final List<Rupture> ruptureList; // concatenated rupture list

  private InterfaceSource(Builder builder) {
    this.feature = builder.feature;
    this.config = builder.config;

    this.mfdTree = builder.mfdTree;

    upperTrace = feature.traces.get(0);
    upperTraceLength = upperTrace.length();
    lowerTrace = feature.traces.get(feature.traces.size() - 1);
    lowerTraceLength = lowerTrace.length();

    surface = new ApproxGriddedSurface(upperTrace, lowerTrace, config.surfaceSpacing);

    this.scaledMfds = ModelTrees.scaledMfdList(mfdTree);
    this.ruptureLists = scaledMfds.stream()
        .map(this::createRuptureList)
        .collect(toUnmodifiableList());
    ruptureList = ruptureLists.stream()
        .flatMap(List::stream)
        .collect(toList());

    // refactor to use size()
    checkState(Iterables.size(Iterables.concat(ruptureLists)) > 0,
        "FaultSource has no ruptures");

  }

  @Override
  public String name() {
    return feature.name;
  }

  @Override
  public int id() {
    return feature.id;
  }

  @Override
  public SourceType type() {
    return INTERFACE;
  }

  @Override
  public int size() {
    return ruptureList.size();
  }

  @Override
  public Rupture get(int index) {
    return ruptureList.get(index);
  }

  /**
   * The closest point on the upper or lower fault trace, relative to the
   * supplied site {@code Location}.
   */
  @Override
  public Location location(Location site) {
    return Locations.closestPoint(site,
        LocationList.builder()
            .addAll(upperTrace)
            .addAll(lowerTrace)
            .build());
  }

  @Override
  public List<Mfd> mfds() {
    return mfdTree.stream()
        .map(Branch::value)
        .collect(toUnmodifiableList());
  }

  @Override
  public Iterator<Rupture> iterator() {
    return Iterables.concat(ruptureLists).iterator();
  }

  // We do want to be able to return width and everage dip properties of an
  // interface source; checkInterfaceDepth checkInterfaceWidth should
  // be used on deserialize, perhaps to SourceFeature.NshmInterface with
  // additional derived fields not in the interface-fault-section

  @Override
  public String toString() {
    return new StringBuilder()
        .append("==========  Interface Source  ==========")
        .append(LINE_SEPARATOR.value())
        .append("  Source name: ").append(name())
        // .append(LINE_SEPARATOR.value())
        // .append(" dip: ").append(dip)
        // .append(LINE_SEPARATOR.value())
        // .append(" width: ").append(width)
        // .append(LINE_SEPARATOR.value())
        // .append(" rake: ").append(rake)
        .append(LINE_SEPARATOR.value())
        .append("         mfds: ").append(mfdTree.size())
        .append(LINE_SEPARATOR.value())
        .append("          top: ").append(feature.traces.get(0).first().depth)
        .append(LINE_SEPARATOR.value()).toString();
  }

  /* Single use builder. */
  static class Builder {

    private static final String ID = "InterfaceSource.Builder";
    private boolean built = false;

    SourceFeature.Interface feature;
    SourceConfig.Interface config;

    LogicTree<Mfd> mfdTree;
    Mfd mfdTotal;

    Builder feature(SourceFeature.Interface feature) {
      this.feature = feature;
      return this;
    }

    Builder config(SourceConfig.Interface config) {
      this.config = config;
      return this;
    }

    Builder mfdTree(LogicTree<Mfd> mfdTree) {
      checkArgument(checkNotNull(mfdTree).size() > 0, "MFD list is empty");
      this.mfdTree = mfdTree;
      return this;
    }

    Builder mfdTotal(Mfd mfdTotal) {
      checkNotNull(mfdTotal, "MFD total is null");
      this.mfdTotal = mfdTotal;
      return this;
    }

    InterfaceSource build() {

      /*
       * Either upper and lower trace must be set, or upper trace depth, dip,
       * and width. If upper and lower are set, depth, dip and width (if also
       * set) will ultimately be ignored and computed lazily by the subduction
       * surface implementation; otherwise the three fields are set to NaN to
       * satisfy parent builder.
       */
      checkState(!built, "This %s instance has already been used", ID);
      checkState(feature != null, "%s feature not set", ID);
      checkState(config != null, "%s config not set", ID);
      checkState(mfdTree != null, "%s MFD tree not set", ID);
      checkState(mfdTotal != null, "%s total MFD not set", ID);
      built = true;

      return new InterfaceSource(this);
    }

  }

  private List<Rupture> createRuptureList(Mfd mfd) {
    List<Rupture> ruptureList = new ArrayList<>();
    // Be careful when working with combined MFDs
    RuptureFloating floatModel = (mfd.properties().type() == Mfd.Type.SINGLE)
        ? RuptureFloating.OFF
        : config.ruptureFloating;
    for (XyPoint xy : mfd.data()) {
      double mag = xy.x();
      double rate = xy.y();

      // Is this necessary? derive from config and apply elsewhere
      if (rate < 1e-14) {
        continue; // shortcut low rates
      }

      // Revisit, floating rupture behavior controlled by
      // magnitude and RuptureFloating; OFF forces full rupture
      // at whatever magnitude

      // Rake should be explicitely picked up from a feature
      // this default isn't good (although correct, it isn't necessary
      // for interface, and it works for Hawaii decllements, but
      // wouldn't for low angle listric normal faults)
      List<Rupture> floaters = floatModel.createFloatingRuptures(
          surface, config.ruptureScaling, mag, rate, 90.0);
      ruptureList.addAll(floaters);
    }
    return List.copyOf(ruptureList);
  }

}
