package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;

import java.util.Optional;
import java.util.OptionalDouble;

import gov.usgs.earthquake.nshmp.Earthquakes;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/*
 * Magnitude frequency distribution (MFD) configuration. Epistemic uncertainty
 * and aleatory variability on maximum magnitude. If any epistemic {@code mMax <
 * minMagnitude} uncertainty is not applied.
 *
 * @author U.S. Geological Survey
 */
class MfdConfig {

  static final String EPISTEMIC_TREE_KEY = "epistemic-tree";
  static final String ALEATORY_PROPS_KEY = "aleatory-properties";
  static final String MIN_MAGNITUDE_KEY = "minimum-magnitude";
  static final String NSHM_BIN_MODEL_KEY = "nshm-bin-model";

  final Optional<LogicTree<Double>> epistemicTree;
  final OptionalDouble minEpiOffset;
  final Optional<AleatoryProperties> aleatoryProperties;
  final OptionalDouble minMagnitude;
  final boolean nshmBinModel;

  private MfdConfig(Builder builder) {
    this.epistemicTree = builder.epistemicTree;
    this.minEpiOffset = builder.minEpiOffset;
    this.aleatoryProperties = builder.aleatoryProperties;
    this.minMagnitude = builder.minMagnitude;
    this.nshmBinModel = builder.nshmBinModel;
  }

  static Builder builder() {
    return new Builder();
  }

  /* Single use builder. */
  static class Builder {

    private Optional<LogicTree<Double>> epistemicTree = Optional.empty();
    private OptionalDouble minEpiOffset = OptionalDouble.empty();
    private Optional<AleatoryProperties> aleatoryProperties = Optional.empty();
    private OptionalDouble minMagnitude = OptionalDouble.empty();
    private boolean nshmBinModel = true;
    private boolean built = false;

    Builder epistemicTree(LogicTree<Double> tree) {
      this.epistemicTree = Optional.of(checkNotNull(tree));
      if (this.epistemicTree.isPresent()) {
        this.minEpiOffset = OptionalDouble.of(
            this.epistemicTree.orElseThrow().stream()
                .mapToDouble(b -> b.value())
                .min()
                .orElseThrow());
      }
      return this;
    }

    Builder aleatoryProperties(AleatoryProperties props) {
      this.aleatoryProperties = Optional.of(checkNotNull(props));
      return this;
    }

    Builder minimumMagnitude(double mMin) {
      Earthquakes.checkMagnitude(mMin);
      this.minMagnitude = OptionalDouble.of(mMin);
      return this;
    }

    Builder nshmBinModel(boolean nshmBinModel) {
      this.nshmBinModel = nshmBinModel;
      return this;
    }

    MfdConfig build() {
      checkState(!built, "This MfdConfig.Builder has already been used");
      if (epistemicTree.isPresent() || aleatoryProperties.isPresent()) {
        checkState(minMagnitude.isPresent(), "%s required", MIN_MAGNITUDE_KEY);
      }
      built = true;
      return new MfdConfig(this);
    }
  }

  static class AleatoryProperties {

    final int size;
    final int nσ;
    final double σ;

    AleatoryProperties(
        int size,
        int nσ,
        double σ) {

      this.size = size;
      this.nσ = nσ;
      this.σ = σ;
    }
  }
}
