package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static java.util.stream.Collectors.toMap;

import java.nio.file.Path;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;

import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Stateful container of model data.
 *
 * This class is NOT thread safe. Use of copyOf() when parsing a model permits
 * overriding settings in nested model directories.
 *
 * @author U.S. Geological Survey
 */
class ModelData {

  private TectonicSetting setting;
  private GmmTree gmms;
  private SourceConfig sourceConfig;
  private boolean clusterModel = false;
  private boolean faultSourceTree = false;

  private Optional<MfdConfig> mfdConfig = Optional.empty();
  private Optional<Map<String, LogicTree<Mfd.Properties>>> mfdMap = Optional.empty();

  private Optional<LogicTree<Double>> rateTree = Optional.empty();
  private boolean spatialPdf = false;

  private Optional<LogicTree<String>> faultRateTree = Optional.empty();

  private Optional<Map<Integer, SourceFeature.NshmFault>> faultFeatureMap = Optional.empty();
  private Optional<Map<String, SourceFeature.SystemSection>> systemFeatureMap = Optional.empty();
  private Optional<Map<Integer, SourceFeature.Interface>> interfaceFeatureMap = Optional.empty();
  private Optional<Map<Integer, SourceFeature.Grid>> gridFeatureMap = Optional.empty();

  private Optional<Path> gridDataDir = Optional.empty();
  private Optional<Path> gridRateFile = Optional.empty();

  private boolean distanceFilterUpdate = false;

  private ModelData() {}

  /* Initial instances require GMMs and source config. */
  static ModelData create(
      TectonicSetting setting,
      GmmTree gmms,
      SourceConfig sourceConfig) {

    ModelData data = new ModelData();
    data.tectonicSetting(setting);
    data.gmms(gmms);
    data.sourceConfig(sourceConfig);
    return data;
  }

  static ModelData copyOf(ModelData original) {
    ModelData copy = new ModelData();
    copy.setting = original.setting;
    copy.gmms = original.gmms;
    copy.sourceConfig = original.sourceConfig;
    copy.clusterModel = original.clusterModel;
    copy.faultSourceTree = original.faultSourceTree;
    copy.mfdConfig = original.mfdConfig;
    copy.mfdMap = original.mfdMap;
    copy.rateTree = original.rateTree;
    copy.spatialPdf = original.spatialPdf;
    copy.faultRateTree = original.faultRateTree;
    copy.gridDataDir = original.gridDataDir;
    copy.gridRateFile = original.gridRateFile;
    copy.faultFeatureMap = original.faultFeatureMap;
    copy.interfaceFeatureMap = original.interfaceFeatureMap;
    copy.gridFeatureMap = original.gridFeatureMap;
    copy.distanceFilterUpdate = original.distanceFilterUpdate;
    return copy;
  }

  void tectonicSetting(TectonicSetting setting) {
    this.setting = checkNotNull(setting);
  }

  TectonicSetting tectonicSetting() {
    return setting;
  }

  void gmms(GmmTree gmms) {
    this.gmms = checkNotNull(gmms);
  }

  GmmTree gmms() {
    return gmms;
  }

  void sourceConfig(SourceConfig sourceConfig) {
    this.sourceConfig = checkNotNull(sourceConfig);
  }

  SourceConfig sourceConfig() {
    return sourceConfig;
  }

  void clusterModel() {
    this.clusterModel = true;
  }

  boolean isClusterModel() {
    return clusterModel;
  }

  void faultSourceTree() {
    this.faultSourceTree = true;
  }

  boolean isFaultSourceTree() {
    return faultSourceTree;
  }

  void mfdConfig(MfdConfig mfdConfig) {
    this.mfdConfig = Optional.of(checkNotNull(mfdConfig));
  }

  Optional<MfdConfig> mfdConfig() {
    return mfdConfig;
  }

  void mfdMap(Map<String, LogicTree<Mfd.Properties>> mfdMap) {
    this.mfdMap = Optional.of(checkNotNull(mfdMap));
  }

  Optional<Map<String, LogicTree<Mfd.Properties>>> mfdMap() {
    return mfdMap;
  }

  void gridDataDirectory(Path gridDataDir) {
    checkState(this.gridDataDir.isEmpty());
    this.gridDataDir = Optional.of(checkNotNull(gridDataDir));
  }

  Optional<Path> gridDataDirectory() {
    return gridDataDir;
  }

  /* Resolves against grid-data dir. */
  void gridRateFile(Path gridRateFile) {
    checkNotNull(gridRateFile);
    this.gridRateFile = Optional.of(gridDataDir.orElseThrow().resolve(gridRateFile));
  }

  Optional<Path> gridRateFile() {
    return gridRateFile;
  }

  void rateTree(LogicTree<Double> rateTree) {
    this.rateTree = Optional.of(checkNotNull(rateTree));
  }

  Optional<LogicTree<Double>> rateTree() {
    return rateTree;
  }

  void spatialPdf(boolean spatialPdf) {
    this.spatialPdf = spatialPdf;
  }

  boolean spatialPdf() {
    return spatialPdf;
  }

  void faultRateTree(LogicTree<String> faultRateTree) {
    this.faultRateTree = Optional.of(checkNotNull(faultRateTree));
  }

  Optional<LogicTree<String>> faultRateTree() {
    return faultRateTree;
  }

  void faultFeatureMap(Map<Integer, SourceFeature.NshmFault> faultFeatureMap) {
    checkNotNull(faultFeatureMap);
    checkArgument(faultFeatureMap.size() > 0);
    this.faultFeatureMap = Optional.of(faultFeatureMap);
  }

  Optional<Map<Integer, SourceFeature.NshmFault>> faultFeatureMap() {
    return faultFeatureMap;
  }

  // used for faults; see issue #210
  // if support for the old/new model is required longer term this flag
  // should be moved to SourceConfig.Fault
  void useDistanceFilterUpdate(boolean use) {
    this.distanceFilterUpdate = use;
  }

  boolean useDistanceFilterUpdate() {
    return distanceFilterUpdate;
  }

  void systemFeatureMap(Map<Integer, SourceFeature.SystemSection> systemFeatureMap) {
    /*
     * TODO: revisit, this isn't great
     *
     * Multiple system section feature maps may be added and combined. Currently
     * we want the consolidated list of unique features across fault models.
     * Because the secion IDs may not be common to different branches, we
     * compare names when combining.
     *
     * We would really like to consolidate UCERF3 section features but this
     * requires remapping all the rupture section lists, which is difficult.
     * Alternatively we could store the parent fault feaures with the model and
     * then have a 'features' directory like other source elements.
     */
    checkArgument(systemFeatureMap.size() > 0);
    if (this.systemFeatureMap.isEmpty()) {
      Map<String, SourceFeature.SystemSection> sectionMap = systemFeatureMap.values().stream()
          .collect(toMap(
              section -> section.name,
              section -> section));
      this.systemFeatureMap = Optional.of(sectionMap);
    } else {
      Map<String, SourceFeature.SystemSection> sectionMap = this.systemFeatureMap.orElseThrow();
      systemFeatureMap.values().stream()
          .forEach(section -> sectionMap.putIfAbsent(
              section.name,
              section));
    }
  }

  Optional<Map<String, SourceFeature.SystemSection>> systemFeatureMap() {
    return systemFeatureMap;
  }

  void interfaceFeatureMap(Map<Integer, SourceFeature.Interface> interfaceFeatureMap) {
    checkNotNull(interfaceFeatureMap);
    checkArgument(interfaceFeatureMap.size() > 0);
    this.interfaceFeatureMap = Optional.of(interfaceFeatureMap);
  }

  Optional<Map<Integer, SourceFeature.Interface>> interfaceFeatureMap() {
    return interfaceFeatureMap;
  }

  void gridFeatureMap(Map<Integer, SourceFeature.Grid> gridFeatureMap) {
    checkNotNull(gridFeatureMap);
    checkArgument(gridFeatureMap.size() > 0);
    this.gridFeatureMap = Optional.of(gridFeatureMap);
  }

  Optional<Map<Integer, SourceFeature.Grid>> gridFeatureMap() {
    return gridFeatureMap;
  }

  /* Convert feature map to List of original features. */
  static List<Feature> featureMapToList(Collection<? extends SourceFeature> features) {
    return features.stream()
        .map(feature -> feature.source)
        .collect(Collectors.toUnmodifiableList());
  }

}
