package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static gov.usgs.earthquake.nshmp.mfd.Mfd.Type.GR;
import static gov.usgs.earthquake.nshmp.mfd.Mfd.Type.GR_MMAX_GR;
import static gov.usgs.earthquake.nshmp.mfd.Mfd.Type.GR_MMAX_SINGLE;
import static gov.usgs.earthquake.nshmp.mfd.Mfd.Type.GR_TAPER;
import static java.util.stream.Collectors.toMap;
import static java.util.stream.Collectors.toSet;
import static java.util.stream.Collectors.toUnmodifiableList;

import java.nio.file.Path;
import java.util.EnumSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import gov.usgs.earthquake.nshmp.data.MutableXySequence;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Properties;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Properties.GutenbergRichter;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Properties.TaperedGr;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Type;
import gov.usgs.earthquake.nshmp.mfd.Mfds;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicGroup;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Factory class for working with model logic trees. Primarily logic trees of
 * MFDs.
 *
 * @author U.S. Geological Survey
 */
class ModelTrees {

  /* All Gutenberg-Richter MFD types. */
  private static final Set<Type> GR_TYPES = EnumSet.of(
      GR,
      GR_MMAX_GR,
      GR_MMAX_SINGLE,
      GR_TAPER);

  /*
   * Convert tree of Gutenberg-Richter MFD properties (with different mMax and
   * b-values possible) to a tree of model MFDs with identical x-value arrays.
   * This supports optimizations in XySequence.combine().
   */
  static LogicTree<Mfd> grPropsTreeToMmaxMfdTree(LogicTree<Mfd.Properties> tree) {
    checkGrMinAndDelta(tree);
    double mMax = mfdTreeMaxMagnitude(tree);

    /*
     * First create logic tree of mfds with xValue arrays that are the same size
     * but unique objects, and for which yValues are set based on the GR mfd
     * parameters or set to zero at the appropriate mMax.
     */
    LogicTree.Builder<Mfd> mfds1 = LogicTree.builder(tree.name());
    tree.forEach(branch -> mfds1.addBranch(
        branch.id(),
        mMaxGrMfd(branch.value(), mMax),
        branch.weight()));
    LogicTree<Mfd> tree1 = mfds1.build();

    /*
     * Next, create a reusable XySequence and build a new logic tree of MFDs,
     * copying over the y-values from the first MFD logic tree.
     */
    double[] xValues = tree1.get(0).value().data().xValues().toArray();
    MutableXySequence xy = MutableXySequence.create(xValues, Optional.empty());
    LogicTree.Builder<Mfd> mfds2 = LogicTree.builder(tree1.name());
    for (Branch<Mfd> branch : tree1) {
      Mfd branchMfd = branch.value();
      XySequence mfdXy = setRates(xy, branchMfd.data().yValues().toArray());
      Mfd newMfd = Mfd.create(mfdXy, branchMfd.properties());
      mfds2.addBranch(branch.id(), newMfd, branch.weight());
    }
    return mfds2.build();
  }

  /* Convenience method to set the y-values in a xy-sequence. */
  private static XySequence setRates(MutableXySequence xy, double[] ys) {
    checkArgument(xy.size() == ys.length);
    for (int i = 0; i < xy.size(); i++) {
      xy.set(i, ys[i]);
    }
    return xy;
  }

  /* Find the maximum mMax, checking that tree is all GR MFDs. */
  private static double mfdTreeMaxMagnitude(LogicTree<Mfd.Properties> tree) {
    return tree.stream()
        .mapToDouble(b -> b.value().getAsGr().mMax())
        .max()
        .orElseThrow();
  }

  /*
   * Create a GR mfd with values extending up to mMax but with zeroed rates
   * above gr.mMax. Zero rate bins will not scale in later modifications to MFD.
   * The rate zeroing done below permits combining of GR_TAPER and GR branches,
   * where GR_TAPER might go as high as M8. It should not be confused with the
   * rate zeroing that occurs later on as individual node MFDs are being created
   * for GR_MMAX_GR and GR_MMAX_SINGLE to avoid double counting.
   */
  private static Mfd mMaxGrMfd(Mfd.Properties props, double mMax) {
    Mfd.Properties newProps = (props.type() == Type.GR_TAPER)
        ? new TaperedGr(props.getAsGrTaper(), mMax)
        : GutenbergRichter.copyOfWithMax(props.getAsGr(), mMax);
    return newProps.toBuilder()
        .transform(p -> p.set((p.x() > props.getAsGr().mMax()) ? 0.0 : p.y()))
        .build();
  }

  /* Ensure all MFDs in a tree are the same as specified type. */
  static void checkType(LogicTree<Mfd.Properties> tree, Mfd.Type type) {
    Mfd.Type refType = tree.get(0).value().type();
    checkArgument(
        checkTreeTypes(tree, refType) && refType == type,
        "mfd-tree types are not type: %s", type);
  }

  /* Check if all MFDs in a tree are the specified type. */
  static boolean checkTreeTypes(LogicTree<Mfd.Properties> tree, Mfd.Type type) {
    return tree.stream()
        .map(Branch::value)
        .map(Mfd.Properties::type)
        .allMatch(type::equals);
  }

  /* Ensure all MFDs in a tree are some Gutenberg-Richter type. */
  private static void checkGrTypes(LogicTree<Mfd.Properties> tree) {
    Set<Mfd.Type> types = tree.stream()
        .map(Branch::value)
        .map(Mfd.Properties::type)
        .collect(toSet());
    checkArgument(
        GR_TYPES.containsAll(types),
        "mfd-tree types are not all Gutenberg-Richter: %s", types);
  }

  /* Ensure all MFDs are Gutenberg-Richter and that mMin and Δm match. */
  private static void checkGrMinAndDelta(LogicTree<Mfd.Properties> tree) {
    checkGrTypes(tree);
    GutenbergRichter refProps = tree.get(0).value().getAsGr();
    boolean minAndDeltaEqual = tree.stream()
        .skip(1)
        .map(Branch::value)
        .map(Properties::getAsGr)
        .allMatch(props -> checkMinAndDelta(refProps, props));
    checkArgument(
        minAndDeltaEqual,
        "Gutenberg-Richter mfd-tree has different mMin and Δm");
  }

  private static boolean checkMinAndDelta(
      GutenbergRichter props1,
      GutenbergRichter props2) {
    return (props1.mMin() == props2.mMin() && props1.Δm() == props2.Δm());
  }

  /* Check if the IDs and weights of two mfd-trees are the same. */
  static void checkTreeIdsAndWeights(LogicTree<Mfd> tree1, LogicTree<Mfd> tree2) {
    checkArgument(
        mfdTreeWeightMap(tree1).equals(mfdTreeWeightMap(tree2)),
        "mfd-trees do not have identical IDs and wieghts\n Tree1: %s \nTree2: %s",
        tree1, tree2);
  }

  /* Create map of mfd-tree ID:weight. */
  private static Map<String, Double> mfdTreeWeightMap(LogicTree<Mfd> tree) {
    return tree.stream().collect(toMap(Branch::id, Branch::weight));
  }

  /* Get a list of branch IDs from a tree. */
  static List<String> treeIds(LogicTree<?> tree) {
    return tree.stream()
        .map(Branch::id)
        .collect(toUnmodifiableList());
  }

  /*
   * Reduce a logic tree of Mfd.Properties to a MFD with rate of each magnitude
   * set to it's weight. This MFD can then be used as a model for a builder that
   * can be scaled by rate.
   */
  static Mfd reduceSingleMfdTree(LogicTree<Mfd.Properties> tree) {
    double[] magnitudes = new double[tree.size()];
    double[] weights = new double[tree.size()];
    for (int i = 0; i < tree.size(); i++) {
      Branch<Mfd.Properties> mfd = tree.get(i);
      magnitudes[i] = mfd.value().getAsSingle().magnitude();
      weights[i] = mfd.weight();
    }
    return Mfd.create(magnitudes, weights);
  }

  /* Reduce a MFD logic tree; weighted branch combiner. */
  static Mfd reduceMfdTree(LogicTree<Mfd> tree) {
    return Mfds.combine(scaledMfdList(tree));
  }

  /* LogicTree<MFD> --> List<MFD * branchWeight> */
  static List<Mfd> scaledMfdList(LogicTree<Mfd> tree) {
    return tree.stream()
        .map(ModelTrees::reduceMfdBranch)
        .collect(toUnmodifiableList());
  }

  /* LogicTree<MFD> --> List<MFD * branchWeight * weight> */
  @Deprecated // this supports weight collapsing which we're trying to avoid
  static List<Mfd> scaledMfdList(LogicTree<Mfd> tree, double weight) {
    return tree.stream()
        .map(branch -> reduceMfdBranch(branch, weight))
        .collect(toUnmodifiableList());
  }

  /* Branch<MFD> --> MFD * branchWeight */
  private static Mfd reduceMfdBranch(Branch<Mfd> branch) {
    return Mfd.Builder.from(branch.value()).scale(branch.weight()).build();
  }

  /* Branch<MFD> --> MFD * branchWeight * weight */
  private static Mfd reduceMfdBranch(Branch<Mfd> branch, double weight) {
    return Mfd.Builder.from(branch.value()).scale(branch.weight() * weight).build();
  }

  static XySequence scaleMfd(XySequence mfd, double scale) {
    return MutableXySequence.copyOf(mfd)
        .multiply(scale)
        .toImmutable();
  }

  /*
   * Convert any LogicTree to a tree of pseudo-paths based on the id() field of
   * the supplied tree. This is used to accomodate rate-tree and dip-tree
   * branches in a source-tree.
   */
  static <T> LogicTree<Path> toPathTree(LogicTree<T> tree) {
    LogicTree.Builder<Path> pathTree = LogicTree.builder(tree.name());
    tree.forEach(branch -> pathTree.addBranch(
        branch.id(),
        Path.of(branch.id()),
        branch.weight()));
    return pathTree.build();
  }

  /*
   * Convert a SourceTree to a logic tree of the total MFDs for each branch
   * (RuptureSet). A LogicGroup accomodates null (do-nothing) branches in the
   * source tree.
   */
  static LogicGroup<Mfd> toMfdTree(SourceTree sourceTree) {
    LogicGroup.Builder<Mfd> mfdTree = LogicGroup.builder();
    sourceTree.forEach(branch -> mfdTree.addBranch(
        branch.id(),
        branch.value().totalMfd(),
        branch.weight()));
    return mfdTree.build();
  }

  /* Join the supplied IDs on ' : ' */
  static String combineBranchIds(String... ids) {
    return String.join(" : ", ids);
  }

  /*
   * Combine IDs and weights of a tree and branch, preserving tree values.
   * Branches in resultant list do not likely sum to one.
   */
  static <T, R> List<Branch<T>> toBranchList(
      LogicTree<T> tree,
      Branch<R> branch) {

    return tree.stream()
        .map(treeBranch -> new IdWeightBranch<>(
            combineBranchIds(branch.id(), treeBranch.id()),
            treeBranch.value(),
            branch.weight() * treeBranch.weight()))
        .collect(toUnmodifiableList());
  }

  private static class IdWeightBranch<T> implements Branch<T> {

    final String id;
    final double weight;
    final T value;

    @Override
    public String id() {
      return id;
    }

    @Override
    public double weight() {
      return weight;
    }

    @Override
    public T value() {
      return value;
    }

    IdWeightBranch(String id, T value, double weight) {
      this.id = id;
      this.weight = weight;
      this.value = value;
    }
  }

}
