package gov.usgs.earthquake.nshmp.model;

import static java.util.stream.Collectors.groupingBy;
import static java.util.stream.Collectors.mapping;
import static java.util.stream.Collectors.toList;
import static java.util.stream.Collectors.toMap;

import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NoSuchElementException;
import java.util.Optional;
import java.util.function.Predicate;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.FeatureCollection;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.mfd.Mfds;
import gov.usgs.earthquake.nshmp.model.SourceTree.Leaf;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Factory class for querying source models.
 *
 * @author U.S. Geological Survey
 */
public class Models {

  /**
   * Returns an object for JSON serialization with the name and ID of all source
   * logic tree groups in the supplied model organized by tectonic setting and
   * source type.
   *
   * @param model to extract logic tree data from
   */
  public static Object trees(HazardModel model) {
    return model.trees().asMap().entrySet().stream()
        .map(Models::toSettingGroup)
        .collect(toList());
  }

  /**
   * Returns an object containing the branches of a source logic tree for JSON
   * serialization.
   *
   * @param model to query
   * @param id of desired source tree
   * @throws NoSuchElementException if the id is not present in the model
   */
  public static Object tree(HazardModel model, int id) {
    return model.tree(id).map(t -> new Tree(t, model.root())).orElseThrow();
  }

  /**
   * Returns the list of Feature(s) associated with the branches of a source
   * logic tree. This will include all stitched and recombined features as well
   * as dip variants for normal faults that have been created from the raw
   * GeoJSON features that would be returned by
   * {@link #features(HazardModel, Optional)}.
   *
   * @param model to query
   * @param id of desired source tree feature(s)
   * @throws NoSuchElementException if the id is not present in the model
   */
  public static Map<String, List<Feature>> features(HazardModel model, int id) {
    return model.tree(id).orElseThrow().branches().entrySet().stream()
        .collect(toMap(
            e -> weightsString(e.getValue()),
            e -> e.getKey().ruptures.features()));
  }

  /**
   * Returns an {@link Optional} describing a FeatureCollection of the GeoJSON
   * features used to represent sources in the supplied model, or an empty
   * {@code Optional} if {@code type} is not present in supplied model. Method
   * only returns features used for finite fault representation and source zones
   * (i.e. it does not return features for grid or instraslab sources).
   *
   * @param model to query
   * @param type optional source type of the features; if empty all features are
   *        returned
   */
  public static Optional<FeatureCollection> features(
      HazardModel model,
      Optional<SourceType> type) {
    try {
      GeoJson.Builder fc = GeoJson.builder();
      if (type.isPresent()) {
        SourceType t = type.orElseThrow();
        if (t == SourceType.FAULT) {
          model.features().get(SourceType.FAULT_SYSTEM).forEach(fc::add);
        }
        model.features().get(t).forEach(fc::add);
      } else {
        model.features().get(SourceType.DECOLLEMENT).forEach(fc::add);
        model.features().get(SourceType.INTERFACE).forEach(fc::add);
        model.features().get(SourceType.FAULT).forEach(fc::add);
        model.features().get(SourceType.FAULT_SYSTEM).forEach(fc::add);
        model.features().get(SourceType.ZONE).forEach(fc::add);
      }
      return Optional.of(fc.build());
    } catch (IllegalStateException ise) {
      // build() throws ISE for no features added
      return Optional.empty();
    }
  }

  /**
   * Returns an {@link Optional} describing a total magnitude frequency
   * distribution (MFD) for a tectonic setting and optionally for a specific
   * source type in the supplied model, or an empty {@code Optional} if
   * {@code setting} or {@code type} is not represented in supplied model.
   *
   * @param model to query
   * @param setting tectonic setting of MFD
   * @param type optional source type of MFD
   */
  public static Optional<Mfd> mfd(
      HazardModel model,
      TectonicSetting setting,
      Optional<SourceType> type) {

    if (!model.settings().contains(setting)) {
      return Optional.empty();
    }

    if (type.isPresent() && !model.types().contains(type.orElseThrow())) {
      return Optional.empty();
    }

    Predicate<SourceTree> filter = type.isPresent()
        ? ((SourceTree st) -> st.type() == type.orElseThrow())
        : ((SourceTree st) -> true);

    List<Mfd> mfds = model.trees().get(setting).stream()
        .filter(filter)
        .map(ModelTrees::toMfdTree)
        .map(ModelTrees::reduceMfdTree)
        .collect(toList());

    return Optional.of(Mfds.combine(mfds));
  }

  private static SettingGroup toSettingGroup(Entry<TectonicSetting, Collection<SourceTree>> entry) {
    return new SettingGroup(
        entry.getKey(),
        toSourceGroup(entry.getValue()));
  }

  private static List<SourceGroup> toSourceGroup(Collection<SourceTree> trees) {
    return trees.stream()
        .collect(groupingBy(
            SourceTree::type,
            mapping(
                TreeInfo::new,
                Collectors.toList())))
        .entrySet().stream()
        .map(e -> new SourceGroup(e.getKey(), e.getValue()))
        .collect(toList());
  }

  static final class SettingGroup {

    final TectonicSetting setting;
    final List<SourceGroup> data;

    SettingGroup(TectonicSetting setting, List<SourceGroup> data) {
      this.setting = setting;
      this.data = data;
    }
  }

  static final class SourceGroup {

    final SourceType type;
    final List<TreeInfo> data;

    SourceGroup(SourceType type, List<TreeInfo> data) {
      this.type = type;
      this.data = data;
    }
  }

  static final class TreeInfo {

    final int id;
    final String name;

    TreeInfo(SourceTree tree) {
      this.id = tree.id();
      this.name = tree.name();
    }
  }

  static final class Tree {

    final int id;
    final String name;
    final String path;
    final TectonicSetting setting;
    final SourceType type;
    final XySequence totalMfd;
    final List<SourceBranch> branches;

    Tree(SourceTree tree, Path modelRoot) {
      this.id = tree.id();
      this.name = tree.name();
      this.path = modelRoot.relativize(tree.path()).toString();
      this.setting = tree.setting();
      this.type = tree.type();

      /*
       * Map source tree into a logic tree of MFDs. Note that the mfdTree
       * created below is actually a LogicGroup because a SourceTree may contain
       * null (do-nothing) branches such that weights do not sum to one.
       */
      LogicTree<Mfd> mfdTree = ModelTrees.toMfdTree(tree);
      this.totalMfd = ModelTrees.reduceMfdTree(mfdTree).data();

      /* Build the list of branches. */
      List<SourceBranch> branches = new ArrayList<>();
      for (Entry<Leaf, List<Branch<Path>>> e : tree.branches().entrySet()) {
        List<Branch<Path>> pathList = e.getValue();
        String branchWts = weightsString(pathList);
        Branch<Path> leafBranch = pathList.get(pathList.size() - 1);
        // System.out.println(pathList.size());
        // System.out.println(leafBranch.value() + " " +
        // leafBranch.value().getNameCount());

        Path branchPath = leafBranch.value().getNameCount() == 1
            ? leafBranch.value()
            : tree.path().relativize(leafBranch.value());

        SourceBranch branch = new SourceBranch(
            branchWts,
            branchPath.toString(),
            leafBranch.id(),
            tree.weights().get(e.getKey()),
            e.getKey().ruptures);
        branches.add(branch);
      }
      this.branches = branches;
    }
  }

  private static String weightsString(List<Branch<Path>> branches) {
    return branches.stream()
        .map(b -> b.id() + "[" + b.weight() + "]")
        .collect(Collectors.joining(":"));
  }

  static final class SourceBranch {

    final String branch;
    final String path;
    final String name;
    final double weight;
    final List<MfdBranch> mfds;

    SourceBranch(
        String branch,
        String path,
        String name,
        double weight,
        RuptureSet<? extends Source> ruptureSet) {

      this.branch = branch;
      this.path = path;
      this.name = name;
      this.weight = weight;
      this.mfds = createMfdBranches(ruptureSet);
    }
  }

  /* Source-type specific rules for rupture floating. */
  static List<MfdBranch> createMfdBranches(RuptureSet<? extends Source> ruptureSet) {
    if (ruptureSet.type() == SourceType.FAULT) {
      FaultRuptureSet frs = (FaultRuptureSet) ruptureSet;
      return IntStream.range(0, frs.mfdTree().size())
          .mapToObj(i -> new MfdBranch(
              frs.mfdTree().get(i),
              frs.mfdFloats.get(i)))
          .collect(toList());
    } else {
      return ruptureSet.mfdTree().stream()
          .map(mfdBranch -> new MfdBranch(mfdBranch, false))
          .collect(toList());
    }
  }

  static final class MfdBranch {
    final String id; // TODO change to name
    final double weight;
    final Mfd mfd;
    final boolean floats;

    MfdBranch(Branch<Mfd> branch, boolean floats) {
      this.id = branch.id();
      this.weight = branch.weight();
      this.mfd = branch.value();
      this.floats = floats;
    }
  }
}
