package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.CaseFormat.LOWER_HYPHEN;
import static com.google.common.base.CaseFormat.UPPER_UNDERSCORE;
import static gov.usgs.earthquake.nshmp.Text.NEWLINE;
import static gov.usgs.earthquake.nshmp.model.Deserialize.GMM_TREES;
import static gov.usgs.earthquake.nshmp.model.Deserialize.ID;
import static gov.usgs.earthquake.nshmp.model.Deserialize.MODEL;
import static gov.usgs.earthquake.nshmp.model.Deserialize.NAME;
import static gov.usgs.earthquake.nshmp.model.Deserialize.SPACING;
import static java.util.stream.Collectors.toCollection;
import static java.util.stream.Collectors.toUnmodifiableMap;
import static java.util.stream.Collectors.toUnmodifiableSet;

import java.awt.geom.Area;
import java.io.IOException;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.EnumMap;
import java.util.EnumSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.OptionalDouble;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

import com.google.gson.JsonElement;
import com.google.gson.JsonObject;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.data.DelimitedData;
import gov.usgs.earthquake.nshmp.data.DelimitedData.Record;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;
import gov.usgs.earthquake.nshmp.geo.json.Properties;
import gov.usgs.earthquake.nshmp.gmm.Gmm;

/**
 * Site data wrapper class. This class provides one method {@code get(Location)}
 * that returns the superset of all site data parameters required by the current
 * suite of supported NSHMs.
 *
 * This class assumes that if a location 'hit test' succeeds on an area for
 * which there are site data, then non null values will exist for the location.
 *
 * @author U.S. Geological Survey
 */
public class SiteData {

  /** Empty site data instance. */
  public static final SiteData EMPTY = new SiteData(null, null, null) {
    @Override
    public Values get(Location loc) {
      return Values.EMPTY;
    }
  };

  private static Path BASIN_DIR = Path.of("basin");
  private static Path MARGIN_DIR = Path.of("margin");
  private static Path GMM_DIR = Path.of("gmm-region");

  private Set<Basin> basins;
  private Set<Margin> margins;
  private Set<GmmRegion> gmmRegions;

  private SiteData(
      Set<Basin> basins,
      Set<Margin> margins,
      Set<GmmRegion> gmmRegions) {

    this.basins = basins;
    this.margins = margins;
    this.gmmRegions = gmmRegions;
  }

  static SiteData create(Path dir) {
    Set<Basin> basins = init(dir.resolve(BASIN_DIR), Basin::new);
    Set<Margin> margins = init(dir.resolve(MARGIN_DIR), Margin::new);
    Set<GmmRegion> gmmRegions = init(dir.resolve(GMM_DIR), GmmRegion::new);
    return (basins.isEmpty() && margins.isEmpty() && gmmRegions.isEmpty())
        ? EMPTY
        : new SiteData(basins, margins, gmmRegions);
  }

  /**
   * Return the site data values for the specified location.
   *
   * @param location of interest
   */
  public Values get(Location location) {

    /*
     * For all site data once a polygon 'hit' occurs, method breaks out of
     * search loops.
     *
     * TODO Current CONUS model we know we have non-overlapping regions, but
     * there could conceivably be situations where different overlapping
     * polygons for gmm-regions are used or required. At present, all regions
     * after the first will be ignored. Multiple regions should be supported,
     * but this requires that tectonic settings and source types are unique to
     * each region.
     */

    Values.Builder builder = new Values.Builder();
    for (Basin basin : basins) {
      Location snapped = snapToGrid(location, basin.spacing, basin.scale);
      // handle sparse data within polygon
      // TODO looking up in the hashmap might be just as fast as contains
      if (basin.contains(snapped) && basin.data.containsKey(snapped)) {
        builder.basinValues(basin.data.get(snapped));
        break;
      }
    }
    for (Margin margin : margins) {
      Location snapped = snapToGrid(location, margin.spacing, margin.scale);
      // check map because margin data is sparse within polygon
      // TODO looking up in the hashmap might be just as fast as contains
      if (margin.contains(snapped) && margin.data.containsKey(snapped)) {
        builder.marginValue(margin.data.get(snapped));
        break;
      }
    }
    for (GmmRegion gmmRegion : gmmRegions) {
      if (gmmRegion.contains(location)) {
        builder.gmmTrees(gmmRegion.trees);
        break;
      }
    }

    return builder.build();
  }

  /** The (lazily created) set of all GMMs defined in GMM regions. */
  public Set<Gmm> gmms() {
    if (this == EMPTY) {
      return Set.of();
    }
    return gmmRegions.stream()
        .flatMap(region -> region.trees.values().stream()) // tectonic seting
        .flatMap(typeMap -> typeMap.values().stream()) // source type
        .flatMap(gmmTree -> gmmTree.gmms().stream()) // gmm tree
        .collect(toCollection(() -> EnumSet.noneOf(Gmm.class)));
  }

  /* Snap location to closest multiple of spacing. */
  private static Location snapToGrid(Location loc, double spacing, int scale) {
    return Location.create(
        snapToGrid(loc.longitude, spacing, scale),
        snapToGrid(loc.latitude, spacing, scale));
  }

  /*
   * Snap value to closest multiple of spacing.
   *
   * Double precision rounding errors results in grid midpoints rounding both up
   * and down in the initial Math.round() call.
   */
  private static double snapToGrid(double value, double spacing, int scale) {
    return Maths.round(
        Math.round(value / spacing) * spacing,
        scale, RoundingMode.HALF_UP);
  }

  @Override
  public String toString() {
    StringBuilder sb = new StringBuilder(ModelFiles.SITE_DATA_DIR).append("/");
    if (this == EMPTY) {
      return sb.append(" (empty)").toString();
    }
    sb.append(NEWLINE);
    basins.forEach(basin -> sb.append("  basin-data-region: ")
        .append(basin.name)
        .append(NEWLINE));
    margins.forEach(margin -> sb.append("  margin-data-region: ")
        .append(margin.name)
        .append(NEWLINE));
    gmmRegions.forEach(region -> sb.append("  gmm-region: ")
        .append(region.name)
        // uncomment to log gmm-region logic trees
        .append(NEWLINE)
        .append(treesToString(region.trees))
        .append(NEWLINE));
    return sb.toString();
  }

  private static String treesToString(Map<TectonicSetting, Map<SourceType, GmmTree>> trees) {
    StringBuilder sb = new StringBuilder();
    for (TectonicSetting setting : trees.keySet()) {
      String settingStr = UPPER_UNDERSCORE.to(LOWER_HYPHEN, setting.name());
      for (Entry<SourceType, GmmTree> entry : trees.get(setting).entrySet()) {
        String sourceTypeStr = UPPER_UNDERSCORE.to(LOWER_HYPHEN, entry.getKey().name());
        sb.append(NEWLINE).append(settingStr).append("/").append(sourceTypeStr)
            .append(NEWLINE).append(NEWLINE)
            .append(entry.getValue().tree())
            .append(NEWLINE).append(NEWLINE);
      }
    }
    return sb.toString();
  }

  /** Optional site data values associated with a location. */
  public static class Values {

    private static final Values EMPTY = new Builder().build();

    /** Depth to the shear-wave velocity horizon of 1.0 km/s, in km. */
    public final OptionalDouble z1p0;

    /** Depth to the shear-wave velocity horizon of 2.5 km/s, in km. */
    public final OptionalDouble z2p5;

    /** Continental margin sediment thickness, in km. */
    public final OptionalDouble zSed;

    /** Region-specific GMM logic trees. */
    public final Map<TectonicSetting, Map<SourceType, GmmTree>> gmmTrees;

    private Values(Builder builder) {
      this.z1p0 = builder.z1p0;
      this.z2p5 = builder.z2p5;
      this.zSed = builder.zSed;
      this.gmmTrees = builder.gmmTrees;
    }

    private static class Builder {

      OptionalDouble z1p0 = OptionalDouble.empty();
      OptionalDouble z2p5 = OptionalDouble.empty();
      OptionalDouble zSed = OptionalDouble.empty();
      Map<TectonicSetting, Map<SourceType, GmmTree>> gmmTrees = Map.of();

      Builder basinValues(Basin.Values values) {
        this.z1p0 = OptionalDouble.of(values.z1p0);
        this.z2p5 = OptionalDouble.of(values.z2p5);
        return this;
      }

      Builder marginValue(Margin.Value value) {
        this.zSed = OptionalDouble.of(value.zSed);
        return this;
      }

      Builder gmmTrees(Map<TectonicSetting, Map<SourceType, GmmTree>> gmmTrees) {
        this.gmmTrees = gmmTrees;
        return this;
      }

      Values build() {
        return new Values(this);
      }
    }
  }

  private static Location toLocation(Record record) {
    return Location.create(
        record.getDouble("lon"),
        record.getDouble("lat"));
  }

  private static <T> Set<T> init(Path dir, Function<Path, T> mapper) {
    if (Files.exists(dir)) {
      try (DirectoryStream<Path> ds = Files.newDirectoryStream(dir, "*.geojson")) {
        return StreamSupport.stream(ds.spliterator(), false)
            .map(mapper)
            .collect(toUnmodifiableSet());
      } catch (IOException ioe) {
        throw new RuntimeException(ioe);
      }
    }
    return Set.of();
  }

  /* Container for a basin region and depth data. */
  static final class Basin {

    final String name;
    final String id;
    final String model;
    final double spacing;
    final int scale;
    final Feature feature;
    final Area area;
    final Map<Location, Values> data;

    private Basin(Path geojson) {
      Feature feature = GeoJson.from(geojson).toFeature();
      Properties properties = feature.properties();
      this.name = properties.getString(NAME).orElseThrow();
      this.id = properties.getString(ID).orElseThrow();
      this.model = properties.getString(MODEL).orElseThrow();
      this.spacing = properties.getDouble(SPACING).orElseThrow();
      this.scale = BigDecimal.valueOf(spacing).scale();
      this.feature = feature;
      this.area = new Area(Locations.toPath(feature.asPolygonBorder()));
      Path csvpath = geojson.resolveSibling(id + ".csv");
      this.data = loadData(csvpath);
    }

    /* Return whether the basin contains the supplied location. */
    boolean contains(Location loc) {
      return area.contains(loc.longitude, loc.latitude);
    }

    private Map<Location, Values> loadData(Path file) {
      DelimitedData csv = DelimitedData.comma(file);
      try (Stream<Record> stream = csv.records()) {
        return stream.collect(toUnmodifiableMap(
            SiteData::toLocation,
            Basin::toValues));
      }
    }

    private static Values toValues(Record record) {
      return new Values(
          record.getDouble("z1p0"),
          record.getDouble("z2p5"));
    }

    static final class Values {

      final double z1p0;
      final double z2p5;

      Values(double z1p0, double z2p5) {
        this.z1p0 = z1p0;
        this.z2p5 = z2p5;
      }
    }
  }

  /* Container for continental margin sediment thickness data. */
  static final class Margin {

    final String name;
    final String id;
    final String model;
    final double spacing;
    final int scale;
    final Feature feature;
    final Area area;
    final Map<Location, Value> data;

    private Margin(Path geojson) {
      Feature feature = GeoJson.from(geojson).toFeature();
      Properties properties = feature.properties();
      this.name = properties.getString(NAME).orElseThrow();
      this.id = properties.getString(ID).orElseThrow();
      this.model = properties.getString(MODEL).orElseThrow();
      this.spacing = properties.getDouble(SPACING).orElseThrow();
      this.scale = BigDecimal.valueOf(spacing).scale();
      this.feature = feature;
      this.area = new Area(Locations.toPath(feature.asPolygonBorder()));
      Path csvpath = geojson.resolveSibling(id + ".csv");
      this.data = loadData(csvpath);
    }

    /* Return whether the basin contains the supplied location. */
    boolean contains(Location loc) {
      return area.contains(loc.longitude, loc.latitude);
    }

    private Map<Location, Value> loadData(Path file) {
      DelimitedData csv = DelimitedData.comma(file);
      try (Stream<Record> stream = csv.records()) {
        return stream.collect(toUnmodifiableMap(
            SiteData::toLocation,
            Margin::toValue));
      }
    }

    private static Value toValue(Record record) {
      double value = Maths.round(record.getDouble("zsed") / 1000.0, 3);
      return new Value(value);
    }

    static final class Value {

      final double zSed;

      Value(double zSed) {
        this.zSed = zSed;
      }
    }
  }

  /* Container for alternative GMM logic trees (applicability region). */
  public static final class GmmRegion {

    final String name;
    final Feature feature;
    final Area area;
    final Map<TectonicSetting, Map<SourceType, GmmTree>> trees;

    private GmmRegion(Path geojson) {
      Feature feature = GeoJson.from(geojson).toFeature();
      Properties properties = feature.properties();
      this.name = properties.getString(NAME).orElseThrow();
      this.feature = feature;
      this.area = new Area(Locations.toPath(feature.asPolygonBorder()));
      this.trees = readGmmTrees(properties.getJsonElement(GMM_TREES).orElseThrow());
    }

    /* Return whether the region contains the supplied location. */
    boolean contains(Location loc) {
      return area.contains(loc.longitude, loc.latitude);
    }
  }

  private static final String ALL = "all";

  private static Map<TectonicSetting, Map<SourceType, GmmTree>> readGmmTrees(JsonElement e) {
    JsonObject trees = e.getAsJsonObject();
    Map<TectonicSetting, Map<SourceType, GmmTree>> gmmTrees =
        new EnumMap<>(TectonicSetting.class);
    for (String key : trees.keySet()) {
      TectonicSetting setting = parseSetting(key);
      gmmTrees.put(setting, parseGmmTrees(trees.get(key), setting));
    }
    return gmmTrees;
  }

  private static Map<SourceType, GmmTree> parseGmmTrees(
      JsonElement e,
      TectonicSetting setting) {

    JsonObject trees = e.getAsJsonObject();
    Map<SourceType, GmmTree> gmmTrees = new EnumMap<>(SourceType.class);

    // If 'all' is present, other source types are ignored.
    if (trees.keySet().contains(ALL)) {
      GmmTree gmmTree = GmmTree.create(
          Deserialize.gmmTree(trees.get(ALL)),
          Optional.empty());
      setting.sourceTypes().forEach(type -> gmmTrees.put(type, gmmTree));
    } else {
      trees.keySet().forEach(key -> gmmTrees.put(
          parseType(key),
          GmmTree.create(
              Deserialize.gmmTree(trees.get(key)),
              Optional.empty())));
    }
    return gmmTrees;
  }

  private static TectonicSetting parseSetting(String setting) {
    return TectonicSetting.valueOf(LOWER_HYPHEN.to(UPPER_UNDERSCORE, setting));
  }

  private static SourceType parseType(String type) {
    return SourceType.valueOf(LOWER_HYPHEN.to(UPPER_UNDERSCORE, type));
  }

}
