package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.Earthquakes.checkCrustalDepth;
import static gov.usgs.earthquake.nshmp.Earthquakes.checkSlabDepth;
import static gov.usgs.earthquake.nshmp.Faults.checkStrike;
import static gov.usgs.earthquake.nshmp.model.PointSourceType.FIXED_STRIKE;
import static gov.usgs.earthquake.nshmp.model.SourceType.SLAB;

import java.util.Arrays;
import java.util.Collections;
import java.util.EnumMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.NavigableMap;
import java.util.Optional;
import java.util.OptionalDouble;
import java.util.TreeMap;
import java.util.function.Predicate;

import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.primitives.Doubles;

import gov.usgs.earthquake.nshmp.Earthquakes;
import gov.usgs.earthquake.nshmp.fault.FocalMech;
import gov.usgs.earthquake.nshmp.fault.surface.RuptureScaling;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.Properties;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.mfd.Mfds;
import gov.usgs.earthquake.nshmp.model.GridLoader.Key;
import gov.usgs.earthquake.nshmp.model.PointSource.DepthModel;
import gov.usgs.earthquake.nshmp.tree.LogicTree;
import gov.usgs.earthquake.nshmp.tree.Trees;

/*
 * Subduction intraslab rupture set implemention.
 *
 * @author U.S. Geological Survey
 */
class SlabRuptureSet extends AbstractRuptureSet<PointSource> {

  private final SourceFeature feature;

  private final List<Location> locations;
  private final List<Mfd> mfds;
  private final LogicTree<List<Mfd>> mfdsTree;
  private final LogicTree<Mfd> mfdTree;

  final List<Map<FocalMech, Double>> mechMaps; // may be nCopies
  private final OptionalDouble strike;

  final RuptureScaling rupScaling;
  private final double maxWidth;
  private final PointSourceType pointSourceType;

  final DepthModel depthModel;
  // final boolean optimizable;

  final double gridSpacing;

  private static final NavigableMap<Double, Map<Double, Double>> MAG_DEPTH_MAP;

  static {
    MAG_DEPTH_MAP = new TreeMap<>();
    MAG_DEPTH_MAP.put(10.0, Map.of(1.0, 1.0));
  }

  /*
   * Enforce expectation that all MFDs are GR*
   */

  SlabRuptureSet(Builder builder) {
    super(builder);
    this.feature = builder.feature;

    this.locations = builder.locations;
    this.mfds = builder.mfds;
    this.mfdsTree = builder.mfdsTree;
    this.mfdTree = builder.mfdTree;

    this.rupScaling = builder.rupScaling;
    this.mechMaps = builder.mechMaps;
    this.pointSourceType = builder.pointSourceType;
    this.strike = builder.strike;
    this.maxWidth = builder.maxWidth;

    double[] depthMags = this.mfds.get(0).data().xValues().toArray();

    this.depthModel = DepthModel.create(
        MAG_DEPTH_MAP,
        Doubles.asList(depthMags),
        OptionalDouble.empty(),
        OptionalDouble.of(maxWidth));

    this.gridSpacing = builder.gridSpacing;
  }

  @Override
  public int size() {
    return locations.size();
  }

  @Override
  public PointSource get(int index) {
    return getSource(index);
  }

  @Override
  public List<Feature> features() {
    return List.of(feature.source);
  }

  @Override
  public LogicTree<Mfd> mfdTree() {
    return mfdTree;
  }

  @Override
  public Location location(Location site) {
    throw new UnsupportedOperationException();
  }

  /**
   * The {@link PointSource} representation used by this source set.
   */
  public PointSourceType pointSourceType() {
    return pointSourceType;
  }

  @Override
  public Predicate<PointSource> distanceFilter(Location loc, double distance) {
    return new DistanceFilter(loc, distance);
  }

  /* Not inlined for use by area sources */
  static final class DistanceFilter implements Predicate<PointSource> {
    private final Predicate<Location> filter;

    DistanceFilter(Location loc, double distance) {
      filter = Locations.distanceAndRectangleFilter(loc, distance);
    }

    @Override
    public boolean test(PointSource source) {
      return filter.test(source.loc);
    }
  }

  @Override
  public Iterator<PointSource> iterator() {
    return new Iterator<PointSource>() {
      int caret = 0;

      @Override
      public boolean hasNext() {
        return caret < locations.size();
      }

      @Override
      public PointSource next() {
        return getSource(caret++);
      }

      @Override
      public void remove() {
        throw new UnsupportedOperationException();
      }
    };
  }

  private PointSource getSource(int index) {

    /*
     * Stricter rules regarding what sorts of default mfds can be used with grid
     * sources (in an individual grid source set) will allow parsers to create
     * XySequence based MFDs directly using copyOf so as to not create zillions
     * of mag arrays.
     */
    Location loc = locations.get(index);
    Mfd mfd = mfds.get(index);
    Map<FocalMech, Double> mechMap = mechMaps.get(index);

    switch (pointSourceType) {
      case POINT:
        return new PointSource(this,
            SLAB, loc, mfd, mechMap, rupScaling, depthModel, Optional.empty());

      case FINITE:
        return new PointSourceFinite(this,
            SLAB, loc, mfd, mechMap, rupScaling, depthModel, Optional.empty());

      case FIXED_STRIKE:
        return new PointSourceFixedStrike(this,
            SLAB, loc, mfd, mechMap, rupScaling, depthModel, strike.orElseThrow());

      default:
        throw new IllegalStateException("Unhandled point source type");
    }
  }

  static Builder builder() {
    return new Builder();
  }

  /* Single use builder. */
  static class Builder extends AbstractRuptureSet.Builder {

    private SourceFeature feature; // Grid or Zone
    private ModelData data;

    private List<Location> locations;
    private List<Mfd> mfds;
    private LogicTree<List<Mfd>> mfdsTree;
    private LogicTree<Mfd> mfdTree;

    // from GSSb
    private RuptureScaling rupScaling;
    private List<Map<FocalMech, Double>> mechMaps = Lists.newArrayList();
    // private NavigableMap<Double, Map<Double, Double>> magDepthMap;
    private PointSourceType pointSourceType;
    private OptionalDouble strike = OptionalDouble.empty();
    private Double maxWidth;
    // private Double mMin;
    // private Double mMax;
    // private Double Δm;

    private Map<FocalMech, Double> mechMap;

    private Double gridSpacing;

    private Builder() {
      super(SourceType.SLAB);
    }

    Builder(SourceType type) {
      super(type);
    }

    @Override
    Builder name(String name) {
      super.name(name);
      return this;
    }

    @Override
    Builder id(int id) {
      super.id(id);
      return this;
    }

    private Builder slabConfig(SourceConfig.Slab slabConfig) {

      this.maxWidth = slabConfig.maxWidth;

      // makes bad assumption of GRID
      // reenable but need slab handling
      // validateDepth(this.maxDepth, SourceType.GRID);

      this.rupScaling = slabConfig.ruptureScaling;
      if (strike.isEmpty()) { // already set to FIXED_STRIKE
        this.pointSourceType = slabConfig.pointSourceType;
      }

      this.mechMap = convertFocalMechTree(slabConfig.focalMechTree);
      checkArgument(this.mechMap.size() == 3);

      // validateMagCutoffs(this.magDepthMap);

      // makes bad assumption of GRID
      // reenable but need slab handling
      // validateDepthMap(this.magDepthMap, SourceType.GRID);

      this.gridSpacing = slabConfig.gridSpacing;
      // this.smoothingDensity = slabConfig.smoothingDensity;
      // this.smoothingLimit = slabConfig.smoothingLimit;
      // this.distanceBin = slabConfig.distanceBin;

      return this;
    }

    private Map<FocalMech, Double> convertFocalMechTree(LogicTree<FocalMech> tree) {
      Map<FocalMech, Double> fmMmap = new EnumMap<FocalMech, Double>(FocalMech.class);
      Arrays.stream(FocalMech.values()).forEach(v -> fmMmap.put(v, 0.0));
      tree.forEach(branch -> fmMmap.put(
          FocalMech.valueOf(branch.id()),
          branch.weight()));
      return Maps.immutableEnumMap(fmMmap);
    }

    Builder feature(SourceFeature feature) {
      this.feature = feature;
      Properties props = feature.source.properties();
      OptionalDouble strike = props.getDouble(Key.STRIKE);
      if (strike.isPresent()) {
        checkStrike(strike.orElseThrow());
        this.strike = strike;
        this.pointSourceType = FIXED_STRIKE;
      }
      return this;
    }

    Builder modelData(ModelData data) {
      this.data = data;
      super.setting(data.tectonicSetting());
      super.gmms(data.gmms());
      slabConfig(data.sourceConfig().asSlab());
      return this;
    }

    Builder nodeData(
        List<Location> locations,
        List<Mfd> mfds,
        LogicTree<List<Mfd>> mfdsTree) {

      checkNotNull(locations);
      checkNotNull(mfds);
      checkNotNull(mfdsTree);
      checkArgument(locations.size() > 0);
      checkArgument(mfds.size() == locations.size());
      mfdsTree.forEach(branch -> checkArgument(branch.value().size() == locations.size()));

      this.locations = locations;
      this.mfds = mfds;
      this.mfdsTree = mfdsTree;
      this.mfdTree = Trees.transform(mfdsTree, Mfds::combine, Optional.empty());

      return this;
    }

    static void validateDepthMap(
        Map<Double, Map<Double, Double>> magDepthMap,
        SourceType type) {
      for (Map<Double, Double> magMap : magDepthMap.values()) {
        for (double depth : magMap.keySet()) {
          validateDepth(depth, type);
        }
      }
    }

    static void validateDepth(double depth, SourceType type) {
      switch (type) {
        case GRID:
          checkCrustalDepth(depth);
          break;
        case SLAB:
          checkSlabDepth(depth);
          break;
        case ZONE:
          checkCrustalDepth(depth);
          break;
        default:
          throw new IllegalStateException(type + " not a grid or related source type");
      }
    }

    static void validateMaxAndMapDepths(Map<Double, Map<Double, Double>> magDepthMap,
        double maxDepth, String id) {
      for (Map<Double, Double> magMap : magDepthMap.values()) {
        for (double depth : magMap.keySet()) {
          checkState(depth <= maxDepth, "%s mag-depth-weight map depth %s > %s", id,
              depth, maxDepth);
        }
      }
    }

    static void validateMagCutoffs(Map<Double, Map<Double, Double>> magDepthMap) {
      double mMax = Earthquakes.MAG_RANGE.upperEndpoint();
      for (double mag : magDepthMap.keySet()) {
        if (mag > mMax) {
          return;
        }
      }
      throw new IllegalStateException("MagDepthMap must contain at least one M > " + mMax);
    }

    void validateAndInit(String label) {

      validateState(label);

      checkNotNull(feature, "%s feature", label);
      checkNotNull(data, "%s MFD reference data", label);

      checkNotNull(locations, "%s locations", label);
      checkNotNull(mfds, "%s total MFDs", label);
      checkState(!mfds.isEmpty(), "%s has no MFDs", label);
      checkNotNull(mfdsTree, "%s MFDs logic tree", label);
      checkNotNull(mfdTree, "%s collapsed MFD logic tree", label);

      checkNotNull(pointSourceType != null, "%s point source type", label);
      checkNotNull(rupScaling, "%s rupture-scaling relation", label);
      checkNotNull(maxWidth, "%s max depth", label);
      checkNotNull(mechMap, "%s focal mech map", label);

      checkNotNull(gridSpacing, "%s grid spacing", label);

      /*
       * There are too many assumptions built into this; whose to say ones bin
       * spacing should be only be in the hundredths?
       *
       * Where did this come from anyway? Are mag deltas really all that strange
       *
       * We should read precision of supplied mMin and mMax and delta and use
       * largest for formatting
       *
       * In the case of single combined/flattened MFDs, mags may not be
       * uniformly spaced. Can this be refactored
       */

      /*
       * Validate size of mechMaps; size could get out of sync if mixed calls to
       * location(...) were made; one can imagine a future use case where a
       * default is required with an override in a few locations; for now, if
       * custom mechMaps are required, there must be one for each node. If no
       * custom maps supplied populate mechMaps with nCopies (singleton list
       * with multiple elements)
       */

      mechMaps = Collections.nCopies(locations.size(), mechMap);

      /* Validate width. */
      checkArgument(maxWidth >= 0.0 && maxWidth < 20.0);
      // validateMaxAndMapDepths(magDepthMap, maxDepth, label);

      /* Validate strike value and type agreement. */
      if (strike.isPresent()) {
        checkState(pointSourceType == FIXED_STRIKE,
            "Source type must be FIXED_STRIKE for strike [%s]", strike.orElseThrow());
      } else {
        checkState(pointSourceType != FIXED_STRIKE,
            "Source type FIXED_STRIKE invalid for strike [%s]", strike);
      }

    }

    /* Create a new rupture set. */
    SlabRuptureSet build() {
      validateAndInit("SlabRuptureSet.Builder");
      return new SlabRuptureSet(this);
    }
  }
}
