package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.data.DoubleData.checkInRange;

import java.nio.file.Path;
import java.util.Map;
import java.util.Optional;
import java.util.OptionalDouble;

import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.fault.FocalMech;
import gov.usgs.earthquake.nshmp.fault.surface.RuptureFloating;
import gov.usgs.earthquake.nshmp.fault.surface.RuptureScaling;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/*
 * Source model configuration wrapper class.
 *
 * @author U.S. Geological Survey
 */
abstract class SourceConfig {

  final Path resource;

  private SourceConfig(Path resource) {
    this.resource = resource;
  }

  Fault asFault() {
    return (Fault) this;
  }

  Grid asGrid() {
    return (Grid) this;
  }

  Interface asInterface() {
    return (Interface) this;
  }

  Slab asSlab() {
    return (Slab) this;
  }

  static class Fault extends SourceConfig {

    final double surfaceSpacing;
    final RuptureFloating ruptureFloating;
    final RuptureScaling ruptureScaling;
    final Optional<LogicTree<Double>> dipTree;
    final Optional<DipSlipModel> dipSlipModel;

    private Fault(Builder builder) {
      super(builder.resource);
      this.surfaceSpacing = builder.surfaceSpacing;
      this.ruptureFloating = builder.ruptureFloating;
      this.ruptureScaling = builder.ruptureScaling;
      this.dipTree = builder.dipTree;
      this.dipSlipModel = builder.dipSlipModel;
    }

    static Builder builder(Path resource) {
      return new Builder(resource);
    }

    static class Builder {

      private boolean built = false;

      private Path resource;
      private Double surfaceSpacing;
      private RuptureFloating ruptureFloating;
      private RuptureScaling ruptureScaling;
      private Optional<LogicTree<Double>> dipTree = Optional.empty();
      private Optional<DipSlipModel> dipSlipModel = Optional.empty();

      private Builder(Path resource) {
        this.resource = resource;
      }

      Builder spacing(double spacing) {
        checkArgument(spacing > 0.0);
        this.surfaceSpacing = spacing;
        return this;
      }

      Builder floating(RuptureFloating floating) {
        this.ruptureFloating = floating;
        return this;
      }

      Builder scaling(RuptureScaling scaling) {
        this.ruptureScaling = scaling;
        return this;
      }

      Builder dipTree(LogicTree<Double> dipTree) {
        this.dipTree = Optional.of(dipTree);
        return this;
      }

      Builder dipSlipModel(DipSlipModel dipSlipModel) {
        this.dipSlipModel = Optional.of(dipSlipModel);
        return this;
      }

      Fault build() {
        checkState(!built, "This builder has already been used");
        checkNotNull(resource);
        checkNotNull(surfaceSpacing);
        checkNotNull(ruptureFloating);
        checkNotNull(ruptureScaling);
        this.built = true;
        return new Fault(this);
      }
    }
  }

  static class Interface extends SourceConfig {

    final double surfaceSpacing;
    final RuptureFloating ruptureFloating;
    final RuptureScaling ruptureScaling;

    private Interface(Builder builder) {
      super(builder.resource);
      this.surfaceSpacing = builder.surfaceSpacing;
      this.ruptureFloating = builder.ruptureFloating;
      this.ruptureScaling = builder.ruptureScaling;
    }

    static Builder builder(Path resource) {
      return new Builder(resource);
    }

    static class Builder {

      private boolean built = false;

      private Path resource;
      private Double surfaceSpacing;
      private RuptureFloating ruptureFloating;
      private RuptureScaling ruptureScaling;

      private Builder(Path resource) {
        this.resource = resource;
      }

      Builder spacing(double spacing) {
        checkArgument(spacing > 0.0);
        this.surfaceSpacing = spacing;
        return this;
      }

      Builder floating(RuptureFloating floating) {
        this.ruptureFloating = floating;
        return this;
      }

      Builder scaling(RuptureScaling scaling) {
        this.ruptureScaling = scaling;
        return this;
      }

      Interface build() {
        checkState(!built, "This builder has already been used");
        checkNotNull(resource);
        checkNotNull(surfaceSpacing);
        checkNotNull(ruptureFloating);
        checkNotNull(ruptureScaling);
        this.built = true;
        return new Interface(this);
      }
    }
  }

  static class Slab extends SourceConfig {

    final double gridSpacing;
    final PointSourceType pointSourceType;
    final RuptureScaling ruptureScaling;
    final double maxWidth;
    final LogicTree<FocalMech> focalMechTree;

    private Slab(Builder builder) {
      super(builder.resource);
      this.gridSpacing = builder.gridSpacing;
      this.pointSourceType = builder.pointSourceType;
      this.ruptureScaling = builder.ruptureScaling;
      this.maxWidth = builder.maxWidth;
      this.focalMechTree = builder.focalMechTree;
    }

    static Builder builder(Path resource) {
      return new Builder(resource);
    }

    static class Builder {

      private boolean built = false;

      private Path resource;
      private Double gridSpacing;
      private PointSourceType pointSourceType;
      private RuptureScaling ruptureScaling;
      private Double maxWidth;
      private LogicTree<FocalMech> focalMechTree;

      private Builder(Path resource) {
        this.resource = resource;
      }

      Builder spacing(double spacing) {
        this.gridSpacing = spacing;
        return this;
      }

      Builder sourceType(PointSourceType type) {
        this.pointSourceType = type;
        return this;
      }

      Builder scaling(RuptureScaling scaling) {
        this.ruptureScaling = scaling;
        return this;
      }

      Builder maxWidth(double width) {
        this.maxWidth = width;
        return this;
      }

      Builder focalMechTree(LogicTree<FocalMech> tree) {
        this.focalMechTree = tree;
        return this;
      }

      Slab build() {
        checkState(!built, "This builder has already been used");
        checkNotNull(resource);
        checkNotNull(gridSpacing);
        checkNotNull(pointSourceType);
        checkNotNull(ruptureScaling);
        checkNotNull(maxWidth);
        checkNotNull(focalMechTree);
        this.built = true;
        return new Slab(this);
      }
    }
  }

  static class Grid extends SourceConfig {

    final double gridSpacing;
    final int smoothingDensity;
    final double smoothingLimit;
    final OptionalDouble distanceBin;
    final PointSourceType pointSourceType;
    final RuptureScaling ruptureScaling;
    final double maxDepth;
    final LogicTree<FocalMech> focalMechTree;
    final Map<String, GridDepthModel> gridDepthMap;

    private Grid(Builder builder) {
      super(builder.resource);
      this.gridSpacing = builder.gridSpacing;
      this.smoothingDensity = builder.smoothingDensity;
      this.smoothingLimit = builder.smoothingLimit;
      this.distanceBin = builder.distanceBin;
      this.pointSourceType = builder.pointSourceType;
      this.ruptureScaling = builder.ruptureScaling;
      this.maxDepth = builder.maxDepth;
      this.focalMechTree = builder.focalMechTree;
      this.gridDepthMap = builder.gridDepthMap;
    }

    static Builder builder(Path resource) {
      return new Builder(resource);
    }

    static class Builder {

      private boolean built = false;

      private Path resource;
      private Double gridSpacing;
      private Integer smoothingDensity;
      private Double smoothingLimit;
      private OptionalDouble distanceBin;
      private PointSourceType pointSourceType;
      private RuptureScaling ruptureScaling;
      private Double maxDepth;
      private LogicTree<FocalMech> focalMechTree;
      private Map<String, GridDepthModel> gridDepthMap;

      private Builder(Path resource) {
        this.resource = resource;
      }

      Builder spacing(double spacing) {
        this.gridSpacing = spacing;
        return this;
      }

      Builder smoothingDensity(int density) {
        checkArgument(density == 10 || density == 4, "Only 4 and 10 supported");
        this.smoothingDensity = density;
        return this;
      }

      private static final Range<Double> SMOOTHING_LIMIT_RANGE = Range.closed(10.0, 100.0);

      Builder smoothingLimit(double limit) {
        checkInRange(SMOOTHING_LIMIT_RANGE, "Smoothing limit", limit);
        this.smoothingLimit = limit;
        return this;
      }

      private static final Range<Double> DISTANCE_RANGE = Range.closed(0.01, 5.0);

      Builder distanceBin(OptionalDouble distance) {
        distance.ifPresent(d -> checkInRange(DISTANCE_RANGE, "Distance bin", d));
        this.distanceBin = distance;
        return this;
      }

      Builder sourceType(PointSourceType type) {
        this.pointSourceType = type;
        return this;
      }

      Builder scaling(RuptureScaling scaling) {
        this.ruptureScaling = scaling;
        return this;
      }

      Builder maxDepth(double depth) {
        this.maxDepth = depth;
        return this;
      }

      Builder focalMechTree(LogicTree<FocalMech> tree) {
        this.focalMechTree = tree;
        return this;
      }

      Builder gridDepthMap(Map<String, GridDepthModel> depthMap) {
        this.gridDepthMap = depthMap;
        return this;
      }

      Grid build() {
        checkState(!built, "This builder has already been used");
        checkNotNull(resource);
        checkNotNull(gridSpacing);
        checkNotNull(smoothingDensity);
        checkNotNull(smoothingLimit);
        checkNotNull(distanceBin);
        checkNotNull(pointSourceType);
        checkNotNull(ruptureScaling);
        checkNotNull(maxDepth);
        checkNotNull(focalMechTree);
        checkNotNull(gridDepthMap);
        this.built = true;
        return new Grid(this);
      }
    }
  }
}
