package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.CaseFormat.UPPER_CAMEL;
import static com.google.common.base.CaseFormat.UPPER_UNDERSCORE;

import javax.xml.transform.Source;

import com.google.common.base.CaseFormat;

/**
 * Identifier for different earthquake {@link Source} types. These are declared
 * in a preferred order of iteration, with potentially longer running source
 * types coming first. {@code *_CLUSTER} and {@code *_SYSTEM} source types are
 * not top level types in a model directory and file tree. They are specialized
 * versions of {@code FAULT} and {@code INTERFACE} source types and are used
 * internally to route the sources down alternate computational pathways.
 *
 * @author U.S. Geological Survey
 */
public enum SourceType {

  // TODO check if 'preferred iteration order' is actually doing anything.

  /** Décollement (low angle detachment) source. */
  DECOLLEMENT,

  /** Finite fault. */
  FAULT,

  /** Fault cluster. */
  FAULT_CLUSTER,

  /** Fault system. */
  FAULT_SYSTEM,

  /** Gridded (background) seismicity. */
  GRID,

  /** Subduction interface. */
  INTERFACE,

  /** Subduction interface cluster. */
  INTERFACE_CLUSTER,

  /** Subduction interface system. */
  INTERFACE_SYSTEM,

  /** Subduction interface gridded seismicity. */
  INTERFACE_GRID,

  /** Subduction intraslab. */
  @Deprecated
  SLAB,

  /** Subduction intraslab. */
  SLAB_GRID,

  /** Zone (a.k.a. Area). */
  ZONE;

  /**
   * Returns the supertype of this source type. Source types such as
   * {@code FAULT_CLUSTER} require special handling but are <i>always</i>
   * grouped with {@code FAULT} (i.e. the supertype) sources in a model logic
   * tree. Supertypes are their own supertype (e.g. the supertype of
   * {@code FAULT} is {@code FAULT}. Rupture sets, which are the tips of source
   * tree branches, in a hazard model will always return a sub-type if
   * appropriate. However, ground motion model logic trees are mapped to
   * supertypes exclusively and this method supports that mapping.
   */
  public SourceType supertype() {
    switch (this) {
      case DECOLLEMENT:
        return DECOLLEMENT;
      case FAULT:
        return FAULT;
      case FAULT_CLUSTER:
        return FAULT;
      case FAULT_SYSTEM:
        return FAULT;
      case GRID:
        return GRID;
      case INTERFACE:
        return INTERFACE;
      case INTERFACE_CLUSTER:
        return INTERFACE;
      case INTERFACE_GRID:
        return INTERFACE_GRID;
      case INTERFACE_SYSTEM:
        return INTERFACE;
      case SLAB:
        return SLAB;
      case SLAB_GRID:
        return SLAB_GRID;
      case ZONE:
        return ZONE;
      default:
        throw new IllegalArgumentException("Unknown source type: " + this);
    }
  }

  /**
   * Returns a {@code CaseFormat#UPPER_CAMEL} {@code String} representation of
   * this {@code SourceType}.
   */
  @Override
  public String toString() {
    return UPPER_UNDERSCORE.to(UPPER_CAMEL, name());
  }

  /**
   * Converts supplied {@code String} to equivalent {@code SourceType}. Method
   * expects a {@code String} with {@code CaseFormat#UPPER_CAMEL}
   * @param s {@code String} to convert
   * @return the corresponding {@code SourceType}
   * @see CaseFormat
   * @throws IllegalArgumentException if supplied {@code String} is incorrectly
   *         formatted or no matching {@code SourceType} exists
   */
  public static SourceType fromString(String s) {
    return valueOf(UPPER_CAMEL.to(UPPER_UNDERSCORE, s));
  }

}
