package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static gov.usgs.earthquake.nshmp.Earthquakes.checkCrustalDepth;
import static gov.usgs.earthquake.nshmp.Earthquakes.checkCrustalWidth;
import static gov.usgs.earthquake.nshmp.Earthquakes.checkInterfaceDepth;
import static gov.usgs.earthquake.nshmp.Earthquakes.checkInterfaceWidth;
import static gov.usgs.earthquake.nshmp.Earthquakes.checkMagnitude;
import static gov.usgs.earthquake.nshmp.Faults.checkDip;
import static gov.usgs.earthquake.nshmp.Faults.checkRake;
import static gov.usgs.earthquake.nshmp.geo.Locations.horzDistanceFast;
import static gov.usgs.earthquake.nshmp.model.Deserialize.MFD_TREE;
import static gov.usgs.earthquake.nshmp.model.SourceFeature.Key.DEPTH;
import static gov.usgs.earthquake.nshmp.model.SourceFeature.Key.DIP;
import static gov.usgs.earthquake.nshmp.model.SourceFeature.Key.INDICES;
import static gov.usgs.earthquake.nshmp.model.SourceFeature.Key.M;
import static gov.usgs.earthquake.nshmp.model.SourceFeature.Key.RAKE;
import static gov.usgs.earthquake.nshmp.model.SourceFeature.Key.RATE;
import static gov.usgs.earthquake.nshmp.model.SourceFeature.Key.WIDTH;
import static java.lang.Math.min;
import static java.util.stream.Collectors.toList;

import java.math.RoundingMode;
import java.nio.file.Path;
import java.util.AbstractList;
import java.util.BitSet;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Iterables;
import com.google.common.primitives.Doubles;

import gov.usgs.earthquake.nshmp.Faults;
import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.calc.HazardInput;
import gov.usgs.earthquake.nshmp.calc.InputList;
import gov.usgs.earthquake.nshmp.calc.Site;
import gov.usgs.earthquake.nshmp.calc.SystemInputList;
import gov.usgs.earthquake.nshmp.data.DelimitedData;
import gov.usgs.earthquake.nshmp.data.DelimitedData.Record;
import gov.usgs.earthquake.nshmp.data.Indexing;
import gov.usgs.earthquake.nshmp.data.IntervalArray;
import gov.usgs.earthquake.nshmp.fault.surface.ApproxGriddedSurface;
import gov.usgs.earthquake.nshmp.fault.surface.DefaultGriddedSurface;
import gov.usgs.earthquake.nshmp.fault.surface.GriddedSurface;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.model.SourceFeature.SystemSection;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Crustal fault-system rupture set.
 *
 * @author U.S. Geological Survey
 */
public class SystemRuptureSet extends AbstractRuptureSet<SystemRuptureSet.SystemSource> {

  /*
   * Developer notes:
   *
   * Work is required to improve indexing of fault systems. Currently The two
   * UCERF3 fault model subsections consist of overlapping sets of IDs most of
   * which are common to both models. There are also many shared ruptures across
   * branches of the UCERF3 logic tree, but it is not clear if averaged
   * properties are consistent.
   *
   * For now, the fault model subsection IDs have been bumped up to 31**** and
   * 32**** until we can work on reducing redundancy. There should then be a
   * mapping in rupture-set between index and relevant fault section. This will
   * better support more logic tree branches beyond the two fault models.
   *
   * Unfiltered UCERF3 rupture counts: FM31 = 253,706 FM32 = 305,709
   */

  final Map<Integer, SystemSection> sectionMap;
  private final Path ruptures;

  final GriddedSurface[] sections;
  final String[] sectionNames;

  private final BitSet[] bitsets;
  private final double[] mags;
  private final double[] rates;
  private final double[] depths;
  private final double[] dips;
  private final double[] widths;
  private final double[] rakes;

  private final LogicTree<Mfd> mfdTree;

  public final Statistics stats;

  /*
   * Developer notes:
   *
   * Revisit the fact that BitSets are mutable and could potentially be altered
   * via a SystemSource.
   *
   * Don't like the fact that original trace data for sections is lost; same for
   * other attributes
   */

  SystemRuptureSet(Builder builder) {
    super(builder);
    this.sectionMap = builder.sectionMap;
    this.ruptures = builder.ruptures;

    SourceConfig config = builder.data.sourceConfig();

    this.sections = sectionMap.keySet().stream()
        .sorted()
        .map(sectionMap::get)
        .map(f -> {
          return (type() == SourceType.FAULT_SYSTEM)
              ? featureToCrustalSurface(f, config.asFault().surfaceSpacing)
              : featureToInterfaceSurface(f, config.asInterface().surfaceSpacing);
        })
        .toArray(GriddedSurface[]::new);

    this.sectionNames = sectionMap.keySet().stream()
        .sorted()
        .map(sectionMap::get)
        .map(feature -> feature.name)
        .toArray(String[]::new);

    /* Initialize rupture attribute arrays. */
    List<Record> ruptureRecords;
    try (Stream<Record> s = DelimitedData.comma(ruptures).records()) {
      ruptureRecords = s.collect(toList());
    }
    int nRups = ruptureRecords.size();
    this.bitsets = new BitSet[nRups];
    this.mags = new double[nRups];
    this.rates = new double[nRups];
    this.depths = new double[nRups];
    this.dips = new double[nRups];
    this.widths = new double[nRups];
    this.rakes = new double[nRups];

    /* Populate rupture attribute arrays. */
    double mMin = Double.POSITIVE_INFINITY;
    double mMax = Double.NEGATIVE_INFINITY;;
    for (int i = 0; i < ruptureRecords.size(); i++) {
      Record r = ruptureRecords.get(i);

      int[] indices = indexStringToIndices(r.get(INDICES));
      checkArgument(indices.length > 1, "Ruptures must consist of 2 or more sections");
      this.bitsets[i] = Indexing.indicesToBits(sections.length, indices);

      double m = checkMagnitude(r.getDouble(M));
      mMin = (m < mMin) ? m : mMin;
      mMax = (m > mMax) ? m : mMax;
      this.mags[i] = m;

      double rate = r.getDouble(RATE);
      checkArgument(Doubles.isFinite(rate), "Non-finite rate value");
      this.rates[i] = rate;

      this.dips[i] = checkDip(r.getDouble(DIP));
      this.rakes[i] = checkRake(r.getDouble(RAKE));
      if (type() == SourceType.FAULT_SYSTEM) {
        this.depths[i] = checkCrustalDepth(r.getDouble(DEPTH));
        this.widths[i] = checkCrustalWidth(r.getDouble(WIDTH));
      } else { // interface
        this.depths[i] = checkInterfaceDepth(r.getDouble(DEPTH));
        this.widths[i] = checkInterfaceWidth(r.getDouble(WIDTH));

      }
    }
    this.stats = new Statistics(mMin, mMax);

    // build total MFD for fault system
    IntervalArray.Builder mfdBuilder = IntervalArray.Builder.withRows(
        Maths.round(stats.mMin, 1, RoundingMode.FLOOR),
        Maths.round(stats.mMax, 1, RoundingMode.CEILING),
        0.1);
    for (int i = 0; i < mags.length; i++) {
      mfdBuilder.add(mags[i], rates[i]);
    }
    Mfd mfd = Mfd.create(mfdBuilder.build().values());
    this.mfdTree = LogicTree.singleton(MFD_TREE, name(), mfd);
  }

  private static int[] indexStringToIndices(String s) {
    return Stream.of(s.split("-"))
        .flatMapToInt(SystemRuptureSet::toIndices)
        .toArray();
  }

  private static IntStream toIndices(String s) {
    return s.contains(":") ? rangeToIndices(s) : IntStream.of(Integer.valueOf(s));
  }

  /*
   * Must be string of the form #:#
   *
   * NOTE: the string provided to ruptureIndexStringToIndices replicates the
   * order of the declared subsections in the original UCERF3 solutions. This
   * stitching order is ont relevant for how nshmp-lib processes hazard and in
   * the method below, the order of indices is reversed to always be ascending.
   */
  private static IntStream rangeToIndices(String s) {
    String[] minmax = s.split(":");
    int min = Integer.valueOf(minmax[0]);
    int max = Integer.valueOf(minmax[1]);
    if (min > max) {
      int newMax = min;
      min = max;
      max = newMax;
    }
    return IntStream.rangeClosed(min, max);
  }

  private static GriddedSurface featureToCrustalSurface(
      SourceFeature.SystemSection section,
      double spacing) {

    return DefaultGriddedSurface.builder()
        .depth(section.upperDepth)
        .lowerDepth(section.lowerDepth)
        .aseis(section.aseismicity)
        .dip(section.dip)
        .dipDir(section.dipDirection)
        .trace(section.traces.get(0))
        .spacing(spacing)
        .build();
  }

  private static GriddedSurface featureToInterfaceSurface(
      SourceFeature.SystemSection section,
      double spacing) {

    return new ApproxGriddedSurface(
        section.traces.get(0),
        section.traces.get(1),
        spacing);
  }

  @Override
  public int size() {
    return bitsets.length;
  }

  @Override
  public SystemSource get(int index) {
    return new SystemSource(index);
  }

  @Override
  public List<Feature> features() {
    return sectionMap.keySet().stream()
        .sorted()
        .map(sectionMap::get)
        .map(section -> section.source)
        .collect(toList());
  }

  @Override
  public LogicTree<Mfd> mfdTree() {
    return mfdTree;
  }

  /*
   * With the switch to RuptureSet implementing List, when we add a rupture set
   * to a Map, hashCode() is called on the contents of the rupture set. System
   * sources have special handling and so we revert to the system hash code and
   * equals. See also ClusterRuptureSet.
   */

  @Override
  public int hashCode() {
    return System.identityHashCode(this);
  }

  @Override
  public boolean equals(Object obj) {
    return (this == obj);
  }

  @Override
  public Iterator<SystemSource> iterator() {
    return new Iterator<SystemSource>() {
      int caret = 0;

      @Override
      public boolean hasNext() {
        return caret < size();
      }

      @Override
      public SystemSource next() {
        return new SystemSource(caret++);
      }

      @Override
      public void remove() {
        throw new UnsupportedOperationException();
      }
    };
  }

  @Override
  public Predicate<SystemSource> distanceFilter(Location loc, double distance) {
    BitSet siteBitset = bitsetForLocation(loc, distance);
    return new BitsetFilter(siteBitset);
  }

  /**
   * The fault section surface corresponding to the supplied {@code index}.
   *
   * <p>This method exists because system source sets are complex and commonly
   * encapsulate 100K+ sources. The results of a hazard calculation and
   * disaggregation are therefore better represented in the context of
   * individual fault sections, rather than on a per-source basis.
   *
   * @param index of fault section surface to retrieve
   */
  public GriddedSurface section(int index) {
    return sections[index];
  }

  /**
   * The name of the fault section corresponding to the supplied {@code index}.
   *
   * @param index of fault section name to retrieve
   */
  public String sectionName(int index) {
    return sectionNames[index];
  }

  /**
   * A single source in a fault system. These sources do not currently support
   * rupture iteration.
   *
   * <p>We skip the notion of a {@code Rupture} for now. Aleatory uncertainty on
   * magnitude isn't required, but if it is, we'll alter this implementation to
   * return List<GmmInput> per source rather than one GmmInput.
   */
  public final class SystemSource extends AbstractList<Rupture> implements Source {

    private final int index;

    private SystemSource(int index) {
      this.index = index;
    }

    @Override
    public String name() {
      // How to create name? RuptureSet will need parent section names
      return "Unnamed fault system source";
    }

    @Override
    public int size() {
      return 1;
    }

    @Override
    public Rupture get(int index) {
      throw new UnsupportedOperationException();
    }

    @Override
    public int id() {
      return SystemRuptureSet.this.id();
    }

    @Override
    public SourceType type() {
      return SystemRuptureSet.this.type();
    }

    /**
     * This method is not required for disaggregation and currently throws an
     * {@code UnsupportedOperationException}.
     */
    @Override
    public Location location(Location location) {
      throw new UnsupportedOperationException();
    }

    @Override
    public List<Mfd> mfds() {
      throw new UnsupportedOperationException();
    }

    @Override
    public Iterator<Rupture> iterator() {
      throw new UnsupportedOperationException();
    }

    BitSet bitset() {
      return bitsets[index];
    }

    double magnitude() {
      return mags[index];
    }

    double rate() {
      return rates[index];
    }

    double depth() {
      return depths[index];
    }

    double dip() {
      return dips[index];
    }

    double width() {
      return widths[index];
    }

    double rake() {
      return rakes[index];
    }
  }

  /**
   * Container of summary data for this sytem source set.
   */
  public static final class Statistics {

    /* Currently used to build section participation MFDs for disagg. */

    /** Minimum magnitude over all ruptures. */
    public final double mMin;

    /** Maximum magnitude over all ruptures. */
    public final double mMax;

    Statistics(
        double mMin,
        double mMax) {

      this.mMin = mMin;
      this.mMax = mMax;
    }
  }

  /**
   * The closest point on the fault trace, relative to the supplied site
   * {@code Location}.
   */
  @Override
  public Location location(Location site) {
    throw new UnsupportedOperationException();
    // return null; // Locations.closestPoint(site, feature.trace);
  }

  static Builder builder(SourceType type) {
    return new Builder(type);
  }

  // Temporary name generating method, called externally
  // makes assumptions about ucerf3 and two fault models
  // Spotless path null checks are really annoying
  static String ruptureSetName(Path branchPath) {
    String parent = checkNotNull(checkNotNull(branchPath.getParent()).getFileName()).toString();
    String branch = checkNotNull(branchPath.getFileName()).toString();
    return parent + "-" + branch;
  }

  /* Single use builder */
  static class Builder extends AbstractRuptureSet.Builder {

    private ModelData data;

    private Map<Integer, SystemSection> sectionMap;
    private Path ruptures;

    static final String ID = "SystemRuptureSet.Builder";

    private Builder(SourceType type) {
      super(type);
    }

    @Override
    Builder name(String name) {
      super.name(name);
      return this;
    }

    @Override
    Builder id(int id) {
      super.id(id);
      return this;
    }

    Builder sections(Map<Integer, SystemSection> sectionMap) {
      checkArgument(checkNotNull(sectionMap).size() >= 1);
      this.sectionMap = sectionMap;

      return this;
    }

    Builder ruptures(Path ruptures) {
      this.ruptures = checkNotNull(ruptures);
      return this;
    }

    /* Set MFD helper data. */
    Builder modelData(ModelData data) {
      this.data = checkNotNull(data);
      super.setting(data.tectonicSetting());
      super.gmms(data.gmms());
      return this;
    }

    private void validateAndInit(String label) {
      validateState(label);
      checkNotNull(data, "%s model data", label);
      checkNotNull(sectionMap, "%s subsection map", label);
      checkNotNull(ruptures, "%s rupture file", label);
    }

    SystemRuptureSet build() {
      validateAndInit("SystemRuptureSet.Builder");
      return new SystemRuptureSet(this);
    }
  }

  /*
   * Handle rate calculations internally as SystemSource is not fully
   * implemented. If/when it is, this should be removed in favor using
   * iterableForLocation and getRupture(0).
   */

  /**
   * Return an instance of a {@code Function} that converts a
   * {@code SystemRuptureSet} to a ground motion model {@code InputList}.
   *
   * @param location with which to initialize instance.
   * @param distance if interest (relevant source radius)
   * @param modelMfd MFD to populate
   */
  public static Function<SystemRuptureSet, IntervalArray> toRatesFunction(
      Location location,
      double distance,
      IntervalArray modelMfd) {
    return new ToRates(location, distance, modelMfd);
  }

  private static final class ToRates implements Function<SystemRuptureSet, IntervalArray> {

    private final Location location;
    private final double distance;
    private final IntervalArray modelMfd;

    ToRates(
        Location location,
        double distance,
        IntervalArray modelMfd) {

      this.location = location;
      this.distance = distance;
      this.modelMfd = modelMfd;
    }

    @Override
    public IntervalArray apply(SystemRuptureSet ruptures) {
      IntervalArray.Builder mfdForLocation = IntervalArray.Builder.fromModel(modelMfd);
      BitSet siteBitset = ruptures.bitsetForLocation(location, distance);
      if (siteBitset.isEmpty()) {
        return modelMfd;
      }
      Predicate<SystemSource> distanceFilter = new BitsetFilter(siteBitset);
      for (SystemSource source : Iterables.filter(ruptures, distanceFilter::test)) {
        mfdForLocation.add(source.magnitude(), source.rate());
      }
      return mfdForLocation.build();
    }
  }

  /*
   * System source calculation pipeline.
   *
   * Rather than expose highly mutable bitsets and attendant logic that is used
   * to generate HazardInputs from SystemRuptureSets, we opt to locate transform
   * Functions and related classes here.
   *
   * System sources (e.g. UCERF3 ruptures) are composed of multiple small (~7km
   * long x ~15km wide) adjacent fault sections in a large and dense fault
   * network. Each source is defined by the indices of the participating fault
   * sections, magnitude, average width, average rake, and average dip, among
   * other parameters.
   *
   * In order to filter ruptures and calculate distance parameters when doing a
   * hazard calculation, one can treat each source as a finite fault. This
   * ultimately requires careful handling of site-to-source calculations when
   * trying to reduce redundant calculations because many sources share the same
   * fault sections and sources will be handled one-at-a-time, in sequence.
   *
   * Alternatively, one can approach the problem from an indexing standpoint,
   * precomuting that data which will be required, and then mining it on a
   * per-source basis, as follows:
   *
   * 1) For each source, create a BitSet with size = nSections. Set the bits for
   * each section that the source uses. [sourceBitSet]
   *
   * 2) Create another BitSet with size = nSections. Set the bits for each
   * section within the distance cutoff for a Site. Do this quickly using only
   * the centroid of each fault section. [siteBitSet]
   *
   * 3) Create and populate a Map<SectionIndex, double[rJB, rRup, rX]> of
   * distance metrics for each section in the siteBitSet. This is created
   * pre-sorted ascending on rRup (the closest sections to a site come first).
   *
   * 4) For each sourceBitSet, 'siteBitSet.intersects(sourceBitSet)' returns
   * whether a source is close enough to the site to be considered.
   *
   * 5) For each considered source, 'sourceBitSet AND siteBitSet' yields a
   * BitSet that only includes set bits with indices in the distance metric
   * table.
   *
   * 6) For each source, loop the ascending indices, checking whether the bit at
   * 'index' is set in the sources bitset. The first hit will be the closest
   * section in a source, relative to a site. (the rX value used is keyed to the
   * minimum rRup).
   *
   * 7) Build GmmInputs and proceed with hazard calculation.
   *
   * Note on the above. Although one could argue that only rRup or rJb be
   * calculated first, there are geometries for which min(rRup) != min(rJB);
   * e.g. location on hanging wall of dipping fault that abuts a vertical
   * fault... vertical might yield min(rRup) but min(rJB) would be 0 (over
   * dipping fault). While checking the bits in a source, we therefore look at
   * the three closest sections.
   */

  /*
   * Concurrency
   *
   * The use case assumed here is that 1 or 2 fault systems (e.g. UCERF3
   * branch-averaged solutions) will most commonly be run when supporting
   * web-services. We therefore compute hazard by first creating a (large) input
   * list and then distribute the more time consuming curve calculation.
   * However, if many fault systems were to be run, it probably makes sense to
   * farm each onto an independent thread.
   */

  /**
   * Return an instance of a {@code Function} that converts a
   * {@code SystemRuptureSet} to a ground motion model {@code InputList}.
   *
   * @param site with which to initialize instance.
   */
  public static Function<SystemRuptureSet, InputList> toInputsFunction(Site site) {
    return new ToInputs(site);
  }

  private static final class ToInputs implements Function<SystemRuptureSet, InputList> {

    private final Site site;

    ToInputs(Site site) {
      this.site = site;
    }

    @Override
    public InputList apply(SystemRuptureSet ruptures) {

      /* Create Site BitSet. */
      double maxDistance = ruptures.gmmTree().maxDistance().orElseThrow();
      BitSet siteBitset = ruptures.bitsetForLocation(site.location(), maxDistance);
      if (siteBitset.isEmpty()) {
        return SystemInputList.empty(ruptures);
      }

      /* Create and fill distance map. */
      int[] siteIndices = Indexing.bitsToIndices(siteBitset);
      ImmutableMap.Builder<Integer, double[]> rMapBuilder =
          ImmutableMap.<Integer, double[]> builder()
              .orderEntriesByValue(new DistanceTypeSorter(R_RUP_INDEX));
      for (int i : siteIndices) {
        Distance r = ruptures.sections[i].distanceTo(site.location());
        rMapBuilder.put(i, new double[] { r.rJB, r.rRup, r.rX });
      }

      /* Create inputs. */
      Map<Integer, double[]> rMap = rMapBuilder.build();
      Function<SystemSource, HazardInput> inputGenerator = new InputGenerator(rMap, site);
      Predicate<SystemSource> rFilter = new BitsetFilter(siteBitset);
      Iterable<SystemSource> sources = Iterables.filter(ruptures, rFilter::test);

      /* Fill input list. */
      SystemInputList inputs = new SystemInputList(ruptures, rMap.keySet());
      for (SystemSource source : sources) {
        inputs.add(inputGenerator.apply(source));
        // for disagg
        inputs.addBitset(source.bitset());
      }

      return inputs;
    }
  }

  private static final class DistanceTypeSorter implements Comparator<double[]> {

    final int rTypeIndex;

    DistanceTypeSorter(int rTypeIndex) {
      this.rTypeIndex = rTypeIndex;
    }

    @Override
    public int compare(double[] left, double[] right) {
      if (left != null && right != null) {
        return Double.compare(left[rTypeIndex], right[rTypeIndex]);
      } else {
        throw new IllegalStateException();
      }
    }
  }

  /*
   * Predicate that tests the intersection of a site bitset (fault sections
   * within a specified distance of a site) with source bitsets (fault sections
   * that participate in a SystemSource/Rupture).
   */
  private static class BitsetFilter implements Predicate<SystemSource> {

    private final BitSet siteBitset;

    BitsetFilter(BitSet siteBitset) {
      this.siteBitset = siteBitset;
    }

    @Override
    public boolean test(SystemSource source) {
      return siteBitset.intersects(source.bitset());
    }
  }

  private static final int R_JB_INDEX = 0;
  private static final int R_RUP_INDEX = 1;
  private static final int R_X_INDEX = 2;

  private static final int R_HIT_LIMIT = 3;

  private static final class InputGenerator implements Function<SystemSource, HazardInput> {

    private final Map<Integer, double[]> rMap;
    private final Site site;

    InputGenerator(
        Map<Integer, double[]> rMap,
        Site site) {

      this.rMap = rMap;
      this.site = site;
    }

    @Override
    public HazardInput apply(SystemSource source) {

      /* Find r minima. */
      BitSet sections = source.bitset();
      double rJB = Double.MAX_VALUE;
      double rRup = Double.MAX_VALUE;
      double rX = Double.MAX_VALUE;
      int hitCount = 0;
      for (int sectionIndex : rMap.keySet()) {
        if (sections.get(sectionIndex)) {
          double[] distances = rMap.get(sectionIndex);
          rJB = min(rJB, distances[R_JB_INDEX]);
          double rRupNew = distances[R_RUP_INDEX];
          if (rRupNew < rRup) {
            rRup = rRupNew;
            rX = distances[R_X_INDEX];
          }
          if (++hitCount > R_HIT_LIMIT) {
            break;
          }
        }
      }

      double dip = source.dip();
      double width = source.width();
      double zTor = source.depth();
      double zHyp = Faults.hypocentralDepth(dip, width, zTor);

      return new HazardInput(
          source.rate(),
          source.magnitude(),
          rJB,
          rRup,
          rX,
          dip,
          width,
          zTor,
          zHyp,
          source.rake(),
          site.vs30(),
          site.z1p0(),
          site.z2p5(),
          site.zSed());
    }
  }

  private final BitSet bitsetForLocation(Location loc, double r) {
    BitSet bits = new BitSet(sections.length);
    int count = 0;
    for (GriddedSurface surface : sections) {
      bits.set(count++, horzDistanceFast(loc, surface.centroid()) <= r);
    }
    return bits;
  }
}
