package gov.usgs.earthquake.nshmp.tree;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.function.Function;

/**
 * Factory class for manipulating and querying logic trees.
 *
 * @author U.S. Geological Survey
 */
public class Trees {

  private Trees() {}

  /**
   * Transpose a list of logic trees to a new logic tree of immutable lists.
   * Supplied logic trees are assumed to all have same branch names, IDs and
   * weights.
   *
   * @param <T> the value type of the supplied logic tree
   * @param treeList the list of logic trees to transpose
   */
  public static <T> LogicTree<List<T>> transpose(List<LogicTree<T>> treeList) {
    LogicTree<T> model = treeList.get(0);

    /* Init branch value lists. */
    List<List<T>> valueLists = new ArrayList<>(model.size());
    model.forEach(b -> valueLists.add(new ArrayList<T>()));

    /* Populate value lists. */
    for (LogicTree<T> tree : treeList) {
      for (int i = 0; i < tree.size(); i++) {
        valueLists.get(i).add(tree.get(i).value());
      }
    }

    /* Create tree of lists. */
    LogicTree.Builder<List<T>> treeOfLists = LogicTree.builder(model.name());
    for (int i = 0; i < model.size(); i++) {
      Branch<T> branch = model.get(i);
      treeOfLists.addBranch(
          branch.id(),
          List.copyOf(valueLists.get(i)),
          branch.weight());
    }
    return treeOfLists.build();
  }

  /**
   * Transform a logic tree with one value type into a new logic tree with a
   * different value type, preserving branch IDs and weights.
   *
   * @param <T> the value type of the supplied logic tree
   * @param <R> the value type of the transformed logic tree
   * @param tree to transform
   * @param function to apply
   * @param name optional new name for the transformed tree; if empty the name
   *        of the supplied tree is used
   */
  public static <T, R> LogicTree<R> transform(
      LogicTree<T> tree,
      Function<T, R> function,
      Optional<String> name) {

    LogicTree.Builder<R> builder = LogicTree.builder(name.orElse(tree.name()));
    tree.forEach(branch -> builder.addBranch(
        branch.id(),
        function.apply(branch.value()),
        branch.weight()));
    return builder.build();
  }

}
