package gov.usgs.earthquake.nshmp;

import static gov.usgs.earthquake.nshmp.Text.Delimiter.COLON;
import static gov.usgs.earthquake.nshmp.Text.Delimiter.COMMA;
import static gov.usgs.earthquake.nshmp.Text.Delimiter.DASH;
import static gov.usgs.earthquake.nshmp.Text.Delimiter.PERIOD;
import static gov.usgs.earthquake.nshmp.Text.Delimiter.SLASH;
import static gov.usgs.earthquake.nshmp.Text.Delimiter.SPACE;
import static gov.usgs.earthquake.nshmp.Text.Delimiter.UNDERSCORE;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;

import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.function.Function;

import org.junit.jupiter.api.Test;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;

import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;

class TextTests {

  private enum TestEnum {
    ABC,
    DEF,
    G_HI_J;

    @Override
    public String toString() {
      return this.name().toLowerCase();
    }
  }

  static final Gson GSON = new GsonBuilder()
      .setPrettyPrinting()
      .disableHtmlEscaping()
      .serializeNulls()
      .create();

  public static void main(String[] args) {
    Path geojson = Path.of("../nshm-fault-sections/fault-sections/AZ/Algodones.geojson");
    System.out.println(Files.exists(geojson));
    Feature f = GeoJson.from(geojson).toFeature();
    String s = GSON.toJson(f);
    System.out.println(s);
    String s2 = Text.cleanCoordinates(s);
    System.out.println(s2);
    System.out.println(dirtyCoords);
    String cleanCoords = Text.cleanCoordinates(dirtyCoords);
    System.out.println(cleanCoords);
  }

  @Test
  final void testCheckName() {
    assertThrows(IllegalArgumentException.class, () -> {
      Text.checkName("   ", "empty name");
    });

    assertEquals("test name", Text.checkName("test name", "test name"));
  }

  @Test
  final void testSplitsAndJoin() {

    String str = "Test string";
    Iterable<String> split = SPACE.splitter().split(str);
    assertEquals(str, SPACE.joiner().join(split));

    str = "test";
    split = SPACE.splitter().split(str);
    assertEquals(str, SPACE.joiner().join(split));

    // duplicate delimiters do not produce an empty field
    str = "test_double__underscore_delimited";
    split = UNDERSCORE.splitter().split(str);
    assertNotEquals(str, UNDERSCORE.joiner().join(split));

    str = "Test string 1 with   extra spaces";
    split = SPACE.splitter().split(str);
    assertNotEquals(str, SPACE.joiner().join(split));

    str = "test string with μ β θν° unicode";
    split = SPACE.splitter().split(str);
    assertEquals(str, SPACE.joiner().join(split));

    str = "test:colon:delimited";
    split = COLON.splitter().split(str);
    assertEquals(str, COLON.joiner().join(split));

    str = "test,comma,delimited";
    split = COMMA.splitter().split(str);
    assertEquals(str, COMMA.joiner().join(split));

    str = "test-dash-delimited";
    split = DASH.splitter().split(str);
    assertEquals(str, DASH.joiner().join(split));

    str = "test.period.delimited";
    split = PERIOD.splitter().split(str);
    assertEquals(str, PERIOD.joiner().join(split));

    str = "test/slash/delimited";
    split = SLASH.splitter().split(str);
    assertEquals(str, SLASH.joiner().join(split));

    str = "test_underscore_delimited";
    split = UNDERSCORE.splitter().split(str);
    assertEquals(str, UNDERSCORE.joiner().join(split));

    str = "1 2 3 4";
    split = SPACE.splitter().split(str);
    assertEquals(str, SPACE.joiner().join(split));
  }

  @Test
  final void testEnumLabels() {
    assertEquals("abc", Text.enumLabelWithSpaces(TestEnum.ABC, false));
    assertEquals("Abc", Text.enumLabelWithSpaces(TestEnum.ABC, true));
    assertEquals("g hi j", Text.enumLabelWithSpaces(TestEnum.G_HI_J, false));
    assertEquals("G Hi J", Text.enumLabelWithSpaces(TestEnum.G_HI_J, true));

    assertEquals("abc", Text.enumLabelWithDashes(TestEnum.ABC, false));
    assertEquals("Abc", Text.enumLabelWithDashes(TestEnum.ABC, true));
    assertEquals("g-hi-j", Text.enumLabelWithDashes(TestEnum.G_HI_J, false));
    assertEquals("G-Hi-J", Text.enumLabelWithDashes(TestEnum.G_HI_J, true));
  }

  @Test
  final void testToStringMethods() {
    List<Double> doubles = List.of(1e-5, 2e-4, 3e-3, 0.0, 1.0001, 10.0);
    String fmt_f = "%f";
    String expected_f0 = "[0.000010,0.000200,0.003000,0.0,1.000100,10.000000]";

    assertEquals(expected_f0, Text.toString(doubles, fmt_f, ",", true));

    assertEquals(
        expected_f0.substring(1, expected_f0.length() - 1),
        Text.toString(doubles, fmt_f, ",", false));
    assertEquals(
        expected_f0.replaceAll(",", "|"),
        Text.toString(doubles, fmt_f, "|", true));
    assertEquals(
        expected_f0,
        Text.toString(doubles.stream().mapToDouble(Double::doubleValue), fmt_f, ",", true));
  }

  @Test
  final void testFormatDoubleFunction() {
    Function<Double, String> fmtFunc = Text.formatDoubleFunction("%.3f");
    assertEquals("1.001", fmtFunc.apply(1.0011));
    assertEquals("0.0", fmtFunc.apply(0.00000000));

    // compact (clean) zero-values are only written if the original value is
    // zero, not if it is zero after formatting
    assertEquals("0.000", fmtFunc.apply(0.000000001));
  }

  @Test
  final void testCleanCoordinates() {
    assertEquals(cleanCoords, Text.cleanCoordinates(dirtyCoords));
  }

  static final String dirtyCoords =
      "    \"coordinates\": [\n" +
          "      [\n" +
          "        -114.40759,\n" +
          "        32.36229\n" +
          "      ],\n" +
          "      [\n" +
          "        -114.42494,\n" +
          "        32.39011\n" +
          "      ],\n" +
          "      [\n" +
          "        -114.44586,\n" +
          "        32.42941\n" +
          "      ],\n" +
          "      [\n" +
          "        -114.47051,\n" +
          "        32.45656\n" +
          "      ],\n" +
          "      [\n" +
          "        -114.5132,\n" +
          "        32.48507\n" +
          "      ],\n" +
          "      [\n" +
          "        -114.55289,\n" +
          "        32.5366\n" +
          "      ],\n" +
          "      [\n" +
          "        -114.57051,\n" +
          "        32.55287\n" +
          "      ],\n" +
          "      [\n" +
          "        -114.60994,\n" +
          "        32.58949\n" +
          "      ]\n" +
          "    ]\n" +
          "  },\n" +
          "";

  static final String cleanCoords =
      "    \"coordinates\": [\n" +
          "      [-114.40759, 32.36229],\n" +
          "      [-114.42494, 32.39011],\n" +
          "      [-114.44586, 32.42941],\n" +
          "      [-114.47051, 32.45656],\n" +
          "      [-114.5132, 32.48507],\n" +
          "      [-114.55289, 32.5366],\n" +
          "      [-114.57051, 32.55287],\n" +
          "      [-114.60994, 32.58949]\n" +
          "    ]\n" +
          "  },\n" +
          "";

}
