package gov.usgs.earthquake.nshmp.calc;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;

import org.junit.jupiter.api.Test;

import gov.usgs.earthquake.nshmp.NamedLocation;
import gov.usgs.earthquake.nshmp.geo.Location;

class SiteTests {

  static final String NAME = "TestSite";
  static final Location LOC = Location.create(-118, 34);
  static final double VS_30 = 530.0;
  static final double Z1P0 = 0.5;
  static final double Z2P5 = 3.0;
  static final double ZSED = 7.5;

  static final String TO_STRING_1 =
      "Unnamed                     -118.000 34.000 vs30=760.0 " +
          "z1.0=NaN z2.5=NaN zSed=NaN";
  static final String TO_STRING_2 =
      "TestSite                    -118.000 34.000 vs30=530.0 " +
          "z1.0=0.5 z2.5=3.0 zSed=7.5";

  static Site.Builder siteBuilder() {
    return Site.builder()
        .name(NAME)
        .location(LOC)
        .vs30(VS_30)
        .z1p0(Z1P0)
        .z2p5(Z2P5)
        .zSed(ZSED);
  }

  @Test
  final void testBuilder() {

    Site site1 = siteBuilder().build();

    assertEquals(NAME, site1.name());
    assertEquals(LOC, site1.location());
    assertEquals(VS_30, site1.vs30());
    assertEquals(Z1P0, site1.z1p0());
    assertEquals(Z2P5, site1.z2p5());
    assertEquals(ZSED, site1.zSed());

    // check single use
    assertThrows(IllegalStateException.class, () -> {
      Site.Builder site2 = siteBuilder();
      site2.build();
      site2.build();
    });

    // z-value NaNs
    Site site3 = siteBuilder()
        .z1p0(Double.NaN)
        .z2p5(Double.NaN)
        .zSed(Double.NaN)
        .build();

    assertEquals(Double.NaN, site3.z1p0());
    assertEquals(Double.NaN, site3.z2p5());
    assertEquals(Double.NaN, site3.zSed());

    // named location
    NamedLocation loc = new NamedLocation() {
      @Override
      public Location location() {
        return LOC;
      }

      @Override
      public String name() {
        return "test named";
      }
    };

    Site site4 = siteBuilder()
        .location(loc)
        .build();

    assertEquals("test named", site4.name());
    assertEquals(LOC, site4.location());
  }

  @Test
  final void testMethods() {

    assertEquals(TO_STRING_2, siteBuilder().build().toString());
    Site site1 = Site.builder().location(LOC).build();
    assertEquals(TO_STRING_1, site1.toString());

    // trimming long site name
    String longname = "ReallyLoooooooooooooooooooooooooooooooooooooooooooooooooooooooooongSiteName";
    String truncated = "ReallyLoooooooooooooooooooooooooooooooooooooooooooooooooooooooooongSiteN";

    Site site2 = siteBuilder().name(longname).build();
    assertEquals(truncated, site2.name());
  }

}
