package gov.usgs.earthquake.nshmp.data;

import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertDoesNotThrow;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotSame;
import static org.junit.jupiter.api.Assertions.assertSame;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

import org.junit.jupiter.api.Test;

import com.google.common.collect.Range;
import com.google.common.primitives.Doubles;

final class DoubleDataTests {

  private static final double[] VALUES = { 1.0, 10.0, 100.0 };

  private static double[] valueArray() {
    return Arrays.copyOf(VALUES, VALUES.length);
  }

  private static List<Double> valueList() {
    return Doubles.asList(valueArray());
  }

  static final void testArrayAndList(
      double[] expectArray,
      double[] actualArray,
      List<Double> actualList) {

    assertArrayEquals(expectArray, actualArray, 0.0);
    List<Double> expectList = Doubles.asList(expectArray);
    assertEquals(expectList, actualList);
  }

  /* * * * * * * * * * * * * OPERATORS * * * * * * * * * * * * */

  @Test
  final void testAddTerm() {
    // 1D
    double[] expectArray = { 2.0, 11.0, 101.0 };
    double[] actualArray = DoubleData.add(1.0, valueArray());
    List<Double> actualList = DoubleData.add(1.0, valueList());
    testArrayAndList(expectArray, actualArray, actualList);
    // 2D
    double[][] d2_expect = { expectArray, expectArray };
    double[][] d2_input = { valueArray(), valueArray() };
    double[][] d2_actual = DoubleData.add(1.0, d2_input);
    for (int i = 0; i < d2_expect.length; i++) {
      assertArrayEquals(d2_expect[i], d2_actual[i], 0.0);
    }
    // 3D
    double[][][] d3_expect = { { expectArray, expectArray }, { expectArray, expectArray } };
    double[][][] d3_input = { { valueArray(), valueArray() }, { valueArray(), valueArray() } };
    double[][][] d3_actual = DoubleData.add(1.0, d3_input);
    for (int i = 0; i < d3_expect.length; i++) {
      for (int j = 0; j < d3_expect[1].length; j++) {
        assertArrayEquals(d3_expect[i][j], d3_actual[i][j], 0.0);
      }
    }
  }

  @Test
  final void testAddArrays() {
    // 1D array and list
    double[] expectArray = { 2.0, 20.0, 200.0 };
    double[] actualArray = DoubleData.add(valueArray(), valueArray());
    List<Double> actualList = DoubleData.add(valueList(), valueList());
    testArrayAndList(expectArray, actualArray, actualList);
    // 2D primitive arrays
    double[][] d2_expect = { { 2.0, 20.0, 200.0 }, { 2.0, 20.0, 200.0 } };
    double[][] d2_1 = { valueArray(), valueArray() };
    double[][] d2_2 = { valueArray(), valueArray() };
    double[][] d2_actual = DoubleData.add(d2_1, d2_2);
    for (int i = 0; i < d2_expect.length; i++) {
      assertArrayEquals(d2_expect[i], d2_actual[i], 0.0);
    }
    // 3D primitive arrays
    double[][][] d3_expect = {
        { { 2.0, 20.0, 200.0 }, { 2.0, 20.0, 200.0 } },
        { { 2.0, 20.0, 200.0 }, { 2.0, 20.0, 200.0 } } };
    double[][][] d3_1 = { { valueArray(), valueArray() }, { valueArray(), valueArray() } };
    double[][][] d3_2 = { { valueArray(), valueArray() }, { valueArray(), valueArray() } };
    double[][][] d3_actual = DoubleData.add(d3_1, d3_2);
    for (int i = 0; i < d3_expect.length; i++) {
      for (int j = 0; j < d3_expect[1].length; j++) {
        assertArrayEquals(d3_expect[i][j], d3_actual[i][j], 0.0);
      }
    }
  }

  @Test
  final void testAddArraysUnchecked() {
    // 1D checked covariant passes through to unchecked
    // 2D primitive arrays
    double[][] d2_expect = { { 2.0, 20.0, 200.0 }, { 2.0, 20.0, 200.0 } };
    double[][] d2_1 = { valueArray(), valueArray() };
    double[][] d2_2 = { valueArray(), valueArray() };
    double[][] d2_actual = DoubleData.uncheckedAdd(d2_1, d2_2);
    for (int i = 0; i < d2_expect.length; i++) {
      assertArrayEquals(d2_expect[i], d2_actual[i], 0.0);
    }
    // 3D primitive arrays
    double[][][] d3_expect = {
        { { 2.0, 20.0, 200.0 }, { 2.0, 20.0, 200.0 } },
        { { 2.0, 20.0, 200.0 }, { 2.0, 20.0, 200.0 } } };
    double[][][] d3_1 = { { valueArray(), valueArray() }, { valueArray(), valueArray() } };
    double[][][] d3_2 = { { valueArray(), valueArray() }, { valueArray(), valueArray() } };
    double[][][] d3_actual = DoubleData.uncheckedAdd(d3_1, d3_2);
    for (int i = 0; i < d3_expect.length; i++) {
      for (int j = 0; j < d3_expect[1].length; j++) {
        assertArrayEquals(d3_expect[i][j], d3_actual[i][j], 0.0);
      }
    }
  }

  @Test
  final void testAddIAE() {
    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.add(valueList(), new ArrayList<Double>());
    });

    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.add(valueArray(), new double[0]);
    });

    assertThrows(IllegalArgumentException.class, () -> {
      double[][] d2_1 = { valueArray(), valueArray() };
      // 1st level lengths different
      double[][] d2_2 = { valueArray() };
      DoubleData.add(d2_1, d2_2);
    });

    assertThrows(IllegalArgumentException.class, () -> {
      double[][] d2_1 = { valueArray(), valueArray() };
      // 2nd level lengths different
      double[][] d2_2 = { valueArray(), {} };
      DoubleData.add(d2_1, d2_2);
    });

    assertThrows(IllegalArgumentException.class, () -> {
      double[][][] d3_1 = { { valueArray(), valueArray() }, { valueArray(), valueArray() } };
      // 1st level lengths different
      double[][][] d3_2 = { { valueArray(), valueArray() } };
      DoubleData.add(d3_1, d3_2);
    });

    assertThrows(IllegalArgumentException.class, () -> {
      double[][][] d3_1 = { { valueArray(), valueArray() }, { valueArray(), valueArray() } };
      // 2nd level lengths different
      double[][][] d3_2 = { { valueArray(), valueArray() }, { valueArray() } };
      DoubleData.add(d3_1, d3_2);
    });

    assertThrows(IllegalArgumentException.class, () -> {
      double[][][] d3_1 = { { valueArray(), valueArray() }, { valueArray(), valueArray() } };
      // 3rd level lengths different
      double[][][] d3_2 = { { valueArray(), valueArray() }, { valueArray(), {} } };
      DoubleData.add(d3_1, d3_2);
    });
  }

  @Test
  final void testSubtract() {
    // 1D array and list
    double[] expectArray = { 0.0, 0.0, 0.0 };
    double[] actualArray = DoubleData.subtract(valueArray(), valueArray());
    List<Double> actualList = DoubleData.subtract(valueList(), valueList());
    testArrayAndList(expectArray, actualArray, actualList);
  }

  @Test
  final void testSubtractIAE() {
    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.subtract(valueList(), new ArrayList<Double>());
    });

    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.subtract(valueArray(), new double[0]);
    });
  }

  @Test
  final void testMultiplyTerm() {
    // 1D
    double[] expectArray = { 5.0, 50.0, 500.0 };
    double[] actualArray = DoubleData.multiply(5.0, valueArray());
    List<Double> actualList = DoubleData.multiply(5.0, valueList());
    testArrayAndList(expectArray, actualArray, actualList);
    // 2D
    double[][] d2_expect = { expectArray, expectArray };
    double[][] d2_input = { valueArray(), valueArray() };
    double[][] d2_actual = DoubleData.multiply(5.0, d2_input);
    for (int i = 0; i < d2_expect.length; i++) {
      assertArrayEquals(d2_expect[i], d2_actual[i], 0.0);
    }
    // 3D
    double[][][] d3_expect = { { expectArray, expectArray }, { expectArray, expectArray } };
    double[][][] d3_input = { { valueArray(), valueArray() }, { valueArray(), valueArray() } };
    double[][][] d3_actual = DoubleData.multiply(5.0, d3_input);
    for (int i = 0; i < d3_expect.length; i++) {
      for (int j = 0; j < d3_expect[1].length; j++) {
        assertArrayEquals(d3_expect[i][j], d3_actual[i][j], 0.0);
      }
    }
  }

  @Test
  final void testMultiplyArrays() {
    // 1D array and list
    double[] expectArray = { 1.0, 100.0, 10000.0 };
    double[] actualArray = DoubleData.multiply(valueArray(), valueArray());
    List<Double> actualList = DoubleData.multiply(valueList(), valueList());
    testArrayAndList(expectArray, actualArray, actualList);
  }

  @Test
  final void testMultiplyIAE() {
    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.multiply(valueList(), new ArrayList<Double>());
    });

    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.multiply(valueArray(), new double[0]);
    });
  }

  @Test
  final void testAbs() {
    double[] expectArray = valueArray();
    double[] absArray = DoubleData.multiply(-1, valueArray());
    double[] actualArray = DoubleData.abs(absArray);
    assertArrayEquals(expectArray, actualArray, 0.0);
  }

  @Test
  final void testExp() {
    double[] expectArray = new double[3];
    for (int i = 0; i < 3; i++) {
      expectArray[i] = Math.exp(VALUES[i]);
    }
    double[] actualArray = DoubleData.exp(valueArray());
    List<Double> actualList = DoubleData.exp(valueList());
    testArrayAndList(expectArray, actualArray, actualList);
  }

  @Test
  final void testLn() {
    double[] expectArray = new double[3];
    for (int i = 0; i < 3; i++) {
      expectArray[i] = Math.log(VALUES[i]);
    }
    double[] actualArray = DoubleData.ln(valueArray());
    List<Double> actualList = DoubleData.ln(valueList());
    testArrayAndList(expectArray, actualArray, actualList);
  }

  @Test
  final void testPow10() {
    double[] expectArray = new double[3];
    for (int i = 0; i < 3; i++) {
      expectArray[i] = Math.pow(10, VALUES[i]);
    }
    double[] actualArray = DoubleData.pow10(valueArray());
    assertArrayEquals(expectArray, actualArray, 0.0);
  }

  @Test
  final void testLog() {
    double[] expectArray = new double[3];
    for (int i = 0; i < 3; i++) {
      expectArray[i] = Math.log10(VALUES[i]);
    }
    double[] actualArray = DoubleData.log(valueArray());
    assertArrayEquals(expectArray, actualArray, 0.0);
  }

  @Test
  final void testComplement() {
    double[] expectArray = new double[] { 0.2, 0.5, 0.8 };
    double[] actualArray = new double[] { 0.8, 0.5, 0.2 };
    DoubleData.complement(actualArray);
    assertArrayEquals(expectArray, actualArray, 0.01);
  }

  @Test
  final void testFlip() {
    double[] expectArray = { -1.0, -10.0, -100.0 };
    double[] actualArray = DoubleData.flip(valueArray());
    assertArrayEquals(expectArray, actualArray, 0.0);
  }

  @Test
  final void testSum() {
    double expect = 111.0;
    assertEquals(expect, DoubleData.sum(valueArray()), 0.0);
    assertEquals(expect, DoubleData.sum(valueList()), 0.0);
  }

  @Test
  final void testCollapse() {
    // 2D
    double[] d2_expect = { 111.0, 111.0 };
    double[][] d2_input = { valueArray(), valueArray() };
    double[] d2_actual = DoubleData.collapse(d2_input);
    assertArrayEquals(d2_expect, d2_actual, 0.0);
    // 3D
    double[][] d3_expect = { { 111.0, 111.0 }, { 111.0, 111.0 } };
    double[][][] d3_input = { { valueArray(), valueArray() }, { valueArray(), valueArray() } };
    double[][] d3_actual = DoubleData.collapse(d3_input);
    for (int i = 0; i < d3_expect.length; i++) {
      assertArrayEquals(d3_expect[i], d3_actual[i], 0.0);
    }
  }

  @Test
  final void testCumulate() {
    double[] data = { 0.0, 2.0, 4.0, 6.0, 8.0 };
    double[] expected = { 0.0, 2.0, 6.0, 12.0, 20.0 };
    assertArrayEquals(expected, DoubleData.cumulate(data), 0.0);
  }

  @Test
  final void testNormalize() {
    double[] expectArray = { 0.2, 0.3, 0.5 };
    double[] inputArray = { 20, 30, 50 };
    double[] actualArray = DoubleData.normalize(inputArray);
    assertArrayEquals(expectArray, actualArray, 0.0);
  }

  @Test
  final void testRound() {
    double[] expectArray = { 0.23, 1.32 };
    double[] inputArray = { 0.23449999, 1.3150001 };
    List<Double> inputList = Doubles.asList(Arrays.copyOf(inputArray, inputArray.length));
    double[] actualArray = DoubleData.round(2, inputArray);
    List<Double> actualList = DoubleData.round(2, inputList);
    testArrayAndList(expectArray, actualArray, actualList);
  }

  @Test
  final void testPositivize() {
    double[] empty = {};
    assertArrayEquals(empty, DoubleData.positivize(empty), 0.0);
    double[] values = valueArray();
    assertArrayEquals(values, DoubleData.positivize(values), 0.0);
    double[] expect = { 99.0, 90.0, 0.0 };
    double[] actual = DoubleData.positivize(DoubleData.flip(values));
    assertArrayEquals(expect, actual, 0.0);
  }

  @Test
  final void testDiff() {
    double[] increasing_dupes = { -10, -1, 0, 0, 1, 10 };
    double[] increasing_nodupes = { -10, -1, 0, 1, 10 };
    double[] decreasing_dupes = { 10, 1, 0, 0, -1, -10 };
    double[] decreasing_nodupes = { 10, 1, 0, -1, -10 };
    double[] expect = new double[] { 9, 1, 0, 1, 9 };
    assertArrayEquals(expect, DoubleData.diff(increasing_dupes), 0.0);
    assertArrayEquals(DoubleData.flip(expect), DoubleData.diff(decreasing_dupes), 0.0);
    expect = new double[] { 9, 1, 1, 9 };
    assertArrayEquals(expect, DoubleData.diff(increasing_nodupes), 0.0);
    assertArrayEquals(DoubleData.flip(expect), DoubleData.diff(decreasing_nodupes), 0.0);

    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.diff(new double[1]);
    });
  }

  @Test
  final void testPercentDiff() {
    assertEquals(5.0, DoubleData.percentDiff(95.0, 100.0), 0.0);
    assertEquals(5.0, DoubleData.percentDiff(105.0, 100.0), 0.0);
    assertEquals(0.0, DoubleData.percentDiff(0.0, 0.0), 0.0);
    assertEquals(Double.POSITIVE_INFINITY, DoubleData.percentDiff(1.0, 0.0), 0.0);

    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.percentDiff(Double.NaN, 1.0);
    });

    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.percentDiff(1.0, Double.NaN);
    });
  }

  @Test
  final void testWeightedSum() {
    double expected = valueList().stream()
        .map(x -> x * x)
        .collect(Collectors.summingDouble(x -> x));

    double actual = DoubleData.weightedSum(valueArray(), valueArray());
    assertEquals(expected, actual, 0.0);
  }

  @Test
  final void testWeightedSumLn() {
    double expected = valueList().stream()
        .map(x -> Math.exp(x) * x)
        .collect(Collectors.summingDouble(x -> x));
    double actual = DoubleData.weightedSumLn(valueArray(), valueArray());
    assertEquals(Math.log(expected), actual, 0.0);

    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.weightedSum(new double[] { 1.0 }, new double[] { 1.0, 2.0 });
    });

    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.weightedSumLn(new double[] { 1.0 }, new double[] { 1.0, 2.0 });
    });
  }

  /* * * * * * * * * * * * * * STATE * * * * * * * * * * * * * */

  @Test
  final void testState() {
    // isPositiveAndReal
    assertTrue(DoubleData.isPositiveAndReal(10.0));
    assertFalse(DoubleData.isPositiveAndReal(0.0));
    assertFalse(DoubleData.isPositiveAndReal(Double.POSITIVE_INFINITY));
    assertFalse(DoubleData.isPositiveAndReal(Double.NaN));
    // isPositiveAndRealOrZero
    assertTrue(DoubleData.isPositiveAndRealOrZero(10.0));
    assertTrue(DoubleData.isPositiveAndRealOrZero(0.0));
    assertFalse(DoubleData.isPositiveAndRealOrZero(Double.POSITIVE_INFINITY));
    assertFalse(DoubleData.isPositiveAndRealOrZero(Double.NaN));
    // arePositiveAndRealOrZero
    assertTrue(DoubleData.arePositiveAndRealOrZero(valueArray()));
    assertFalse(DoubleData.arePositiveAndRealOrZero(DoubleData.flip(valueArray())));
    // areZeroValued
    assertTrue(DoubleData.areZeroValued(new double[] { 0, 0 }));
    assertFalse(DoubleData.areZeroValued(new double[] { 0, 1 }));
    // areMonotonic
    double[] increasing_dupes = { -10, -1, 0, 0, 1, 10 };
    double[] increasing_nodupes = { -10, -1, 0, 1, 10 };
    double[] increasing_bad = { -10, -1, 0, -1, -10 };
    double[] decreasing_dupes = { 10, 1, 0, 0, -1, -10 };
    assertTrue(DoubleData.areMonotonic(true, false, increasing_dupes));
    assertFalse(DoubleData.areMonotonic(true, true, increasing_dupes));
    assertTrue(DoubleData.areMonotonic(true, true, increasing_nodupes));
    assertFalse(DoubleData.areMonotonic(true, false, increasing_bad));
    assertTrue(DoubleData.areMonotonic(false, false, decreasing_dupes));
    assertFalse(DoubleData.areMonotonic(false, true, decreasing_dupes));
  }

  @Test
  final void testStateIAE() {
    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.areZeroValued();
    });
  }

  /* * * * * * * * * * * * PRECONDITIONS * * * * * * * * * * * */

  @Test
  final void testCheckDelta() {
    assertEquals(2.0, DoubleData.checkDelta(0.0, 10.0, 2.0), 0.0);

    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.checkDelta(10.0, 10.0, 0.0);
    });

    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.checkDelta(10.0, 0.0, 2.0);
    });

    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.checkDelta(0.0, 10.0, -2.0);
    });

    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.checkDelta(0.0, 10.0, 0.0);
    });

    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.checkDelta(0.0, 10.0, 11.0);
    });
  }

  @Test
  final void testCheckInRange() {
    // also tests checkInRange(double)
    double[] expectArray = { 5.0, 2.0 };
    Range<Double> r = Range.open(0.0, 10.0);
    assertArrayEquals(expectArray, DoubleData.checkInRange(r, expectArray), 0.0);
    assertSame(expectArray, DoubleData.checkInRange(r, expectArray));
  }

  @Test
  final void testCheckInRangeIAE() {
    assertThrows(IllegalArgumentException.class, () -> {
      Range<Double> r = Range.open(0.0, 10.0);
      DoubleData.checkInRange(r, new double[] { -1.0 });
    });
  }

  @Test
  final void testCheckIsPositiveAndReal() {
    assertEquals(DoubleData.checkIsPositiveAndReal(1.0), 1.0, 0.0);
    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.checkIsPositiveAndReal(0.0);
    });
    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.checkIsPositiveAndReal(-1.0);
    });
  }

  @Test
  final void testCheckIsPositiveAndRealOrZero() {
    assertEquals(DoubleData.checkIsPositiveAndRealOrZero(1.0), 1.0, 0.0);
    assertEquals(DoubleData.checkIsPositiveAndRealOrZero(0.0), 0.0, 0.0);
    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.checkIsPositiveAndReal(-1.0);
    });
  }

  /*
   * checkSize overloads are checked via operator tests
   */

  @Test
  final void testCheckWeight() {
    assertEquals(0.5, DoubleData.checkWeight(0.5), 0.0);

    assertThrows(IllegalArgumentException.class, () -> {
      DoubleData.checkWeight(0.0);
    });
  }

  @Test
  final void testCheckWeights() {
    double[] wtArray = { 0.4, 0.600001 };
    List<Double> wtList = Doubles.asList(wtArray);

    assertEquals(wtList, DoubleData.checkWeights(wtList));
    assertSame(wtList, DoubleData.checkWeights(wtList));

    assertArrayEquals(wtArray, DoubleData.checkWeights(wtArray));
    assertSame(wtArray, DoubleData.checkWeights(wtArray));

    assertArrayEquals(wtArray, DoubleData.checkWeights(wtArray, true));
    assertSame(wtArray, DoubleData.checkWeights(wtArray, true));

    assertThrows(IllegalArgumentException.class, () -> {
      double[] wtArrayBadValueHi = { 1.0001 };
      DoubleData.checkWeights(Doubles.asList(wtArrayBadValueHi));
    });

    assertThrows(IllegalArgumentException.class, () -> {
      double[] wtArrayBadValueLo = { 0.0 };
      DoubleData.checkWeights(Doubles.asList(wtArrayBadValueLo));
    });

    assertThrows(IllegalArgumentException.class, () -> {
      double[] wtArrayBadSum = { 0.4, 0.6002 };
      DoubleData.checkWeights(Doubles.asList(wtArrayBadSum));
    });

    assertDoesNotThrow(() -> {
      double[] wtArrayGoodSum = { 0.4, 0.6 };
      DoubleData.checkWeights(Doubles.asList(wtArrayGoodSum));
    });

  }

  /* * * * * * * * 2D & 3D ARRAYS EXTENSIONS * * * * * * * * */

  @Test
  final void testCopyOf() {
    // 2D
    double[][] expect2D = { valueArray(), valueArray(), {}, { 1.0, Double.NaN } };
    double[][] actual2D = DoubleData.copyOf(expect2D);
    assertNotSame(expect2D, actual2D);
    for (int i = 0; i < expect2D.length; i++) {
      assertArrayEquals(expect2D[i], actual2D[i], 0.0);
    }
    // 3D
    double[][][] expect3D = {
        { valueArray(), valueArray() },
        { valueArray() },
        { {}, { 1.0, Double.NaN } } };
    double[][][] actual3D = DoubleData.copyOf(expect3D);
    for (int i = 0; i < expect3D.length; i++) {
      for (int j = 0; j < expect3D[i].length; j++) {
        assertArrayEquals(expect3D[i][j], actual3D[i][j], 0.0);
      }
    }
  }

}
