package gov.usgs.earthquake.nshmp.data;

import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.util.Arrays;
import java.util.stream.IntStream;

import org.junit.jupiter.api.Test;

import com.google.common.primitives.Doubles;

import gov.usgs.earthquake.nshmp.data.IntervalArray.Builder;

class IntervalArrayTests {

  static final double MIN = 2.0;
  static final double MAX = 6.0;
  static final double Δ = 1.0;
  static final double[] ROWS = IntervalData.keys(MIN, MAX, Δ);
  static final XySequence XY = XySequence.create(ROWS, ROWS);

  static final IntervalArray ARRAY = Builder
      .withRows(MIN, MAX, Δ)
      .addEach(XY)
      .build();

  @Test
  final void getTests() {
    IntStream.range(0, ROWS.length).forEach(i -> {
      double row = ROWS[i];
      assertEquals(row, ARRAY.get(row), 0);
      assertEquals(row, ARRAY.get(i), 0);
    });

    assertThrows(IndexOutOfBoundsException.class, () -> {
      ARRAY.get(1.0);
    });
  }

  @Test
  final void rowTests() {
    assertEquals(MIN, ARRAY.rowMin(), 0);
    assertEquals(MAX, ARRAY.rowMax(), 0);
    assertEquals(Δ, ARRAY.rowΔ(), 0);

    assertArrayEquals(ROWS, Doubles.toArray(ARRAY.rows()), 0);
  }

  @Test
  final void valuesTests() {
    assertEquals(XY, ARRAY.values());
    assertArrayEquals(XY.xValues().toArray(), ARRAY.values().xValues().toArray(), 0);
    assertArrayEquals(XY.yValues().toArray(), ARRAY.values().yValues().toArray(), 0);
  }

  @Test
  final void sumTests() {
    double expected = Arrays.stream(ROWS).sum();
    assertEquals(expected, ARRAY.sum(), 0);
  }

  @Test
  final void indexTests() {
    double[] valuesA = { 0, 1.0, 2.0, 3.0 };
    double[] valuesB = { 3.0, 2.0, 1.0, 0.0 };
    double[] valuesC = { 1.0, 3.0, 0.0, 2.0 };

    int expectedMinA = 0;
    int expectedMaxA = 3;
    int expectedMinB = 3;
    int expectedMaxB = 0;
    int expectedMinC = 2;
    int expectedMaxC = 1;

    IntervalArray actualA = Builder.fromModel(ARRAY).add(valuesA).build();
    IntervalArray actualB = Builder.fromModel(ARRAY).add(valuesB).build();
    IntervalArray actualC = Builder.fromModel(ARRAY).add(valuesC).build();

    assertEquals(expectedMinA, actualA.minIndex(), 0);
    assertEquals(expectedMaxA, actualA.maxIndex(), 0);

    assertEquals(expectedMinB, actualB.minIndex(), 0);
    assertEquals(expectedMaxB, actualB.maxIndex(), 0);

    assertEquals(expectedMinC, actualC.minIndex(), 0);
    assertEquals(expectedMaxC, actualC.maxIndex(), 0);
  }

  @Test
  final void toStringTests() {
    String expected = ARRAY.toString();
    String actual = Builder.copyOf(ARRAY).build().toString();

    assertEquals(expected, actual);
    assertTrue(expected.equals(actual));
  }

  @Test
  final void builderCopyOfTests() {
    IntervalArray actual = Builder.copyOf(ARRAY).build();

    assertEquals(ARRAY.minIndex(), actual.minIndex(), 0);
    assertEquals(ARRAY.maxIndex(), actual.maxIndex(), 0);

    assertEquals(ARRAY.rowMin(), actual.rowMin(), 0);
    assertEquals(ARRAY.rowMax(), actual.rowMax(), 0);
    assertEquals(ARRAY.rows(), actual.rows());
    assertEquals(ARRAY.rowΔ(), actual.rowΔ(), 0);

    assertEquals(ARRAY.sum(), actual.sum(), 0);
    assertEquals(ARRAY.values(), actual.values());
    assertTrue(ARRAY.values().equals(actual.values()));
  }

  @Test
  final void builderFromModelTests() {
    IntervalArray actual = Builder.fromModel(ARRAY).build();

    assertEquals(ARRAY.rowMin(), actual.rowMin(), 0);
    assertEquals(ARRAY.rowMax(), actual.rowMax(), 0);
    assertEquals(ARRAY.rows(), actual.rows());
    assertEquals(ARRAY.rowΔ(), actual.rowΔ(), 0);
  }

  @Test
  final void builderWithRowsTests() {
    IntervalArray actual = Builder.withRows(MIN, MAX, Δ).build();

    assertEquals(ARRAY.rowMin(), actual.rowMin(), 0);
    assertEquals(ARRAY.rowMax(), actual.rowMax(), 0);
    assertEquals(ARRAY.rows(), actual.rows());
    assertEquals(ARRAY.rowΔ(), actual.rowΔ(), 0);
  }

  @Test
  final void builderRowIndexTests() {
    int expected = 3;
    int actual = Builder.fromModel(ARRAY).rowIndex(ROWS[expected]);

    assertEquals(expected, actual);
  }

  @Test
  final void builderSetTests() {
    int index = 0;
    double expected = 1.25;
    IntervalArray actual = Builder.copyOf(ARRAY)
        .set(ROWS[index], expected)
        .build();

    assertEquals(expected, actual.get(ROWS[index]), 0);
    assertEquals(expected, actual.get(index), 0);
  }

  @Test
  final void builderAddTests() {
    int index = 0;
    double[] expectedArray = Arrays.stream(ROWS).map(x -> x + x).toArray();
    double expected = expectedArray[index];

    // add(double row, double value)
    IntervalArray actualA = Builder.copyOf(ARRAY)
        .add(ROWS[index], ROWS[index])
        .build();
    assertEquals(expected, actualA.get(index), 0);

    // add(int row, double value)
    IntervalArray actualB = Builder.copyOf(ARRAY)
        .add(index, ROWS[index])
        .build();
    assertEquals(expected, actualB.get(index), 0);

    // add(double[] values)
    IntervalArray actualC = Builder.copyOf(ARRAY)
        .add(ROWS)
        .build();
    assertArrayEquals(expectedArray, actualC.values().yValues().toArray(), 0);

    // add(XySequence values)
    XySequence xy = ARRAY.values();
    IntervalArray actualD = Builder.copyOf(ARRAY)
        .add(xy)
        .build();
    assertArrayEquals(expectedArray, actualD.values().yValues().toArray(), 0);

    // add(IntervalArray array)
    IntervalArray actualE = Builder.copyOf(ARRAY)
        .add(ARRAY)
        .build();
    assertArrayEquals(expectedArray, actualE.values().yValues().toArray(), 0);
  }

  @Test
  final void builderSetData() {
    double[] data = { 5, 4, 3, 2 };
    IntervalArray ia = Builder.copyOf(ARRAY)
        .data(data)
        .build();
    // really should be assertSame, but data is not visible
    assertArrayEquals(data, ia.values().yValues().toArray());
  }

  @Test
  final void builderMultiplyTests() {
    double scale = 2.25;
    double[] expected = Arrays.stream(ROWS).map(x -> x * scale).toArray();
    double[] actual = Builder.copyOf(ARRAY)
        .multiply(scale)
        .build()
        .values()
        .yValues()
        .toArray();

    assertArrayEquals(expected, actual, 0);
  }

  @Test
  final void builderBuiltTests() {
    assertThrows(IllegalStateException.class, () -> {
      Builder b = Builder.copyOf(ARRAY);
      b.build();
      b.build();
    });
  }

}
