package gov.usgs.earthquake.nshmp.data;

import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.util.Arrays;
import java.util.stream.IntStream;

import org.junit.jupiter.api.Test;

import com.google.common.primitives.Doubles;

import gov.usgs.earthquake.nshmp.data.IntervalVolume.Builder;

class IntervalVolumeTests {

  static final double SCALE = 2.0;
  static final double Δ = 1.0;
  static final double ROW_MIN = 2.0;
  static final double ROW_MAX = 6.0;
  static final double COL_MIN = ROW_MIN + SCALE;
  static final double COL_MAX = ROW_MAX + SCALE;
  static final double LVL_MIN = COL_MIN + SCALE;
  static final double LVL_MAX = COL_MAX + SCALE;

  static final double[] ROWS = IntervalData.keys(ROW_MIN, ROW_MAX, Δ);
  static final double[] COLS = IntervalData.keys(COL_MIN, COL_MAX, Δ);
  static final double[] LVLS = IntervalData.keys(LVL_MIN, LVL_MAX, Δ);

  static final IntervalVolume VOLUME;
  static final XySequence XY = XySequence.create(COLS, LVLS);

  static {
    Builder builder = new Builder()
        .rows(ROW_MIN, ROW_MAX, Δ)
        .columns(COL_MIN, COL_MAX, Δ)
        .levels(LVL_MIN, LVL_MAX, Δ);

    IntStream.range(0, ROWS.length).forEach(i -> {
      IntStream.range(0, COLS.length).forEach(j -> {
        IntStream.range(0, LVLS.length).forEach(k -> {
          builder.add(i, j, k, LVLS[k]);
        });
      });
    });

    VOLUME = builder.build();
  }

  @Test
  final void getTests() {
    IntStream.range(0, ROWS.length).forEach(i -> {
      double row = ROWS[i];
      IntStream.range(0, COLS.length).forEach(j -> {
        double col = COLS[j];
        IntStream.range(0, LVLS.length).forEach(k -> {
          double lvl = LVLS[k];
          assertEquals(lvl, VOLUME.get(row, col, lvl), 0);
          assertEquals(lvl, VOLUME.get(i, j, k), 0);
        });
      });
    });

    assertThrows(IndexOutOfBoundsException.class, () -> {
      VOLUME.get(1.0, 1.0, 1.0);
    });
  }

  @Test
  final void rowTests() {
    assertEquals(ROW_MIN, VOLUME.rowMin(), 0);
    assertEquals(ROW_MAX, VOLUME.rowMax(), 0);
    assertEquals(Δ, VOLUME.rowΔ(), 0);
    assertArrayEquals(ROWS, Doubles.toArray(VOLUME.rows()), 0);
  }

  @Test
  final void columnTests() {
    assertEquals(COL_MIN, VOLUME.columnMin(), 0);
    assertEquals(COL_MAX, VOLUME.columnMax(), 0);
    assertEquals(Δ, VOLUME.columnΔ(), 0);
    assertArrayEquals(COLS, Doubles.toArray(VOLUME.columns()), 0);

    IntStream.range(0, ROWS.length).forEach(i -> {
      double row = ROWS[i];
      IntStream.range(0, COLS.length).forEach(j -> {
        double col = COLS[j];

        assertArrayEquals(LVLS, VOLUME.column(row, col).xValues().toArray(), 0);
        assertArrayEquals(LVLS, VOLUME.column(i, j).xValues().toArray(), 0);

        assertArrayEquals(LVLS, VOLUME.column(row, col).yValues().toArray(), 0);
        assertArrayEquals(LVLS, VOLUME.column(i, j).yValues().toArray(), 0);
      });
    });
  }

  @Test
  final void levelTests() {
    assertEquals(LVL_MIN, VOLUME.levelMin(), 0);
    assertEquals(LVL_MAX, VOLUME.levelMax(), 0);
    assertEquals(Δ, VOLUME.levelΔ(), 0);
    assertArrayEquals(LVLS, Doubles.toArray(VOLUME.levels()), 0);
  }

  @Test
  final void collapseTests() {
    double[] expected = Arrays.stream(LVLS).map(col -> Arrays.stream(LVLS).sum()).toArray();
    IntervalTable table = VOLUME.collapse();

    for (double row : ROWS) {
      XySequence values = table.row(row);
      assertArrayEquals(COLS, values.xValues().toArray(), 0);
      assertArrayEquals(expected, values.yValues().toArray(), 0);
    }
  }

  @Test
  final void indexTests() {
    int index = 0;

    double[] valuesA = { 0, 1.0, 2.0, 3.0 };
    double[] valuesB = { 3.0, 2.0, 1.0, 0.0 };
    double[] valuesC = { 1.0, 3.0, 0.0, 2.0 };

    int[] expectedMinA = { 0, 0, 0 };
    int[] expectedMaxA = { 0, 0, 3 };;
    int[] expectedMinB = { 0, 0, 3 };
    int[] expectedMaxB = { 0, 0, 0 };
    int[] expectedMinC = { 0, 0, 2 };
    int[] expectedMaxC = { 0, 0, 1 };

    Builder builderA = Builder.fromModel(VOLUME);
    Builder builderB = Builder.fromModel(VOLUME);
    Builder builderC = Builder.fromModel(VOLUME);

    IntStream.range(0, valuesA.length).forEach(lvlIndex -> {
      builderA.add(index, index, lvlIndex, valuesA[lvlIndex]);
      builderB.add(index, index, lvlIndex, valuesB[lvlIndex]);
      builderC.add(index, index, lvlIndex, valuesC[lvlIndex]);
    });

    IntervalVolume actualA = builderA.build();
    IntervalVolume actualB = builderB.build();
    IntervalVolume actualC = builderC.build();

    assertArrayEquals(expectedMinA, actualA.minIndex());
    assertArrayEquals(expectedMaxA, actualA.maxIndex());

    assertArrayEquals(expectedMinB, actualB.minIndex());
    assertArrayEquals(expectedMaxB, actualB.maxIndex());

    assertArrayEquals(expectedMinC, actualC.minIndex());
    assertArrayEquals(expectedMaxC, actualC.maxIndex());
  }

  @Test
  final void toStringTests() {
    String expected = VOLUME.toString();
    String actual = Builder.copyOf(VOLUME).build().toString();

    assertEquals(expected, actual);
    assertTrue(expected.equals(actual));
  }

  @Test
  final void builderInitTests() {
    // test branch where columns have been initialized but not rows.
    IntervalVolume iv = new Builder()
        .columns(COL_MIN, COL_MAX, Δ)
        .rows(ROW_MIN, ROW_MAX, Δ)
        .levels(LVL_MIN, LVL_MAX, Δ)
        .build();
    assertEquals(iv.rows(), VOLUME.rows());
    assertEquals(iv.columns(), VOLUME.columns());
    assertEquals(iv.levels(), VOLUME.levels());
  }

  @Test
  final void builderCopyOfTests() {
    IntervalVolume actual = Builder.copyOf(VOLUME).build();

    assertArrayEquals(VOLUME.minIndex(), actual.minIndex());
    assertArrayEquals(VOLUME.maxIndex(), actual.maxIndex());

    assertEquals(VOLUME.columnMin(), actual.columnMin(), 0);
    assertEquals(VOLUME.columnMax(), actual.columnMax(), 0);
    assertEquals(VOLUME.columns(), actual.columns());
    assertTrue(VOLUME.columns().equals(actual.columns()));
    assertEquals(VOLUME.columnΔ(), actual.columnΔ(), 0);

    assertEquals(VOLUME.rowMin(), actual.rowMin(), 0);
    assertEquals(VOLUME.rowMax(), actual.rowMax(), 0);
    assertEquals(VOLUME.rows(), actual.rows());
    assertTrue(VOLUME.rows().equals(actual.rows()));
    assertEquals(VOLUME.rowΔ(), actual.rowΔ(), 0);

    assertEquals(VOLUME.levelMin(), actual.levelMin(), 0);
    assertEquals(VOLUME.levelMax(), actual.levelMax(), 0);
    assertEquals(VOLUME.levels(), actual.levels());
    assertTrue(VOLUME.levels().equals(actual.levels()));
    assertEquals(VOLUME.levelΔ(), actual.levelΔ(), 0);

    assertEquals(VOLUME.collapse().columns(), actual.collapse().columns());
    assertEquals(VOLUME.collapse().rows(), actual.collapse().rows());
  }

  @Test
  final void builderFromModelTests() {
    IntervalVolume actual = Builder.fromModel(VOLUME).build();

    assertEquals(VOLUME.columnMax(), actual.columnMax(), 0);
    assertEquals(VOLUME.columnMin(), actual.columnMin(), 0);
    assertEquals(VOLUME.columns(), actual.columns());
    assertEquals(VOLUME.columnΔ(), actual.columnΔ(), 0);

    assertEquals(VOLUME.rowMax(), actual.rowMax(), 0);
    assertEquals(VOLUME.rowMin(), actual.rowMin(), 0);
    assertEquals(VOLUME.rows(), actual.rows());
    assertEquals(VOLUME.rowΔ(), actual.rowΔ(), 0);

    assertEquals(VOLUME.levelMin(), actual.levelMin(), 0);
    assertEquals(VOLUME.levelMax(), actual.levelMax(), 0);
    assertEquals(VOLUME.levels(), actual.levels());
    assertEquals(VOLUME.levelΔ(), actual.levelΔ(), 0);
  }

  @Test
  final void builderIndexTests() {
    int expected = 3;
    int actualRow = Builder.fromModel(VOLUME).rowIndex(ROWS[expected]);
    int actualCol = Builder.fromModel(VOLUME).columnIndex(COLS[expected]);
    int actualLvl = Builder.fromModel(VOLUME).levelIndex(LVLS[expected]);

    assertEquals(expected, actualRow);
    assertEquals(expected, actualCol);
    assertEquals(expected, actualLvl);
  }

  @Test
  final void builderSetTests() {
    int i = 0;
    double expected = 1.25;
    IntervalVolume actual = Builder.copyOf(VOLUME)
        .set(ROWS[i], COLS[i], LVLS[i], expected)
        .build();

    assertEquals(expected, actual.get(ROWS[i], COLS[i], LVLS[i]), 0);
    assertEquals(expected, actual.get(i, i, i), 0);
  }

  @Test
  final void builderAddTests() {
    int index = 0;

    // add by index
    Builder builder = Builder.copyOf(VOLUME);

    IntStream.range(0, LVLS.length).forEach(lvlIndex -> {
      builder.add(index, index, lvlIndex, LVLS[lvlIndex]);
    });

    IntervalVolume actual = builder.build();

    IntStream.range(0, LVLS.length).forEach(lvlIndex -> {
      double expected = LVLS[lvlIndex] * 2;
      assertEquals(expected, actual.get(index, index, lvlIndex), 0);
    });

    // add(double row, double column, double[] values)
    double[] expectedA = DoubleData.multiply(2, Arrays.copyOf(LVLS, LVLS.length));
    IntervalVolume actualA = Builder.copyOf(VOLUME)
        .add(ROWS[index], COLS[index], LVLS)
        .build();
    assertArrayEquals(expectedA, actualA.column(index, index).yValues().toArray(), 0);

    // add(double row, double column, XySequence values)
    double[] expectedB = DoubleData.multiply(2, Arrays.copyOf(LVLS, LVLS.length));
    IntervalVolume actualB = Builder.copyOf(VOLUME)
        .add(ROWS[index], COLS[index], XY)
        .build();
    assertArrayEquals(expectedB, actualB.column(index, index).yValues().toArray(), 0);

    // add by value
    IntervalVolume iv = Builder.copyOf(VOLUME)
        .add(4.7, 5.3, 8.7, 42.0)
        .build();
    assertEquals(50.5, iv.get(2, 1, 2), 0.0);

    // add(IntervalVolume)
    IntervalVolume volumeToAdd = Builder.copyOf(VOLUME).build();
    IntervalVolume actual2 = Builder.copyOf(VOLUME)
        .add(volumeToAdd)
        .build();
    assertEquals(13.0, actual2.get(0, 0, 0), 0.0);
    assertEquals(19.0, actual2.get(3, 3, 3), 0.0);
  }

  @Test
  final void builderSetData() {
    double[][][] data = {
        { { 5, 4, 3 }, { 6, 7, 8 }, { 1, 0, -1 } },
        { { 5, 4, 3 }, { 6, 7, 8 }, { 1, 0, -1 } },
        { { 5, 4, 3 }, { 6, 7, 8 }, { 1, 0, -1 } }
    };

    IntervalVolume iv = Builder.copyOf(VOLUME)
        .data(data)
        .build();

    // really should be assertSame, but data is not visible
    assertArrayEquals(data[0][0], iv.column(0, 0).yValues().toArray());
    assertArrayEquals(data[1][1], iv.column(1, 1).yValues().toArray());
    assertArrayEquals(data[2][2], iv.column(2, 2).yValues().toArray());
  }

  @Test
  final void builderMultiplyTests() {
    IntervalVolume actual = Builder.copyOf(VOLUME)
        .multiply(3.0)
        .build();
    assertEquals(19.5, actual.get(0, 0, 0));
    assertEquals(28.5, actual.get(3, 3, 3));
  }

  @Test
  final void builderBuiltTests() {
    assertThrows(IllegalStateException.class, () -> {
      Builder b = Builder.copyOf(VOLUME);
      b.build();
      b.build();
    });
  }

}
