package gov.usgs.earthquake.nshmp.data;

import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotSame;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.function.Consumer;

import org.junit.jupiter.api.Test;

import com.google.common.primitives.Doubles;

class MutableXySequenceTests {

  /* Tests MutableXySequence and array backed implementation. */

  private static double[] xs = new double[] { 0, 1, 2, 3 };
  private static double[] ys = new double[] { -1, 10.5, 5.25, 2.5 };
  private static MutableXySequence xy = MutableXySequence.create(xs, Optional.of(ys));

  @Test
  final void factoryConstructorTests() {
    // arrays
    assertArrayEquals(xs, ((MutableArrayXySequence) xy).xs, 0);
    assertArrayEquals(ys, ((MutableArrayXySequence) xy).ys, 0);
    // collections; guava Doubles copies array on toArray()
    List<Double> xList = Doubles.asList(xs);
    List<Double> yList = Doubles.asList(ys);
    MutableXySequence xy2 = MutableXySequence.create(xList, Optional.of(yList));
    assertArrayEquals(xs, ((MutableArrayXySequence) xy2).xs, 0);
    assertArrayEquals(ys, ((MutableArrayXySequence) xy2).ys, 0);
    // test new sequence is created on copy for mutable
    MutableXySequence xy3 = MutableXySequence.copyOf(xy);
    assertNotSame(xy, xy3);
    // test empty ys
    MutableArrayXySequence xy4 =
        (MutableArrayXySequence) MutableXySequence.create(xs, Optional.empty());
    assertArrayEquals(new double[4], xy4.ys, 0);
    MutableArrayXySequence xy5 =
        (MutableArrayXySequence) MutableXySequence.create(xList, Optional.empty());
    assertArrayEquals(new double[4], xy5.ys, 0);
    // test emptyCopyOf
    MutableArrayXySequence xy6 = (MutableArrayXySequence) MutableXySequence.emptyCopyOf(xy);
    assertArrayEquals(new double[4], xy6.ys, 0);

    assertThrows(IllegalArgumentException.class, () -> {
      MutableXySequence.create(new double[0], Optional.of(new double[0]));
    });

    assertThrows(IllegalArgumentException.class, () -> {
      MutableXySequence.create(xs, Optional.of(new double[0]));
    });

    assertThrows(IllegalArgumentException.class, () -> {
      double[] reversed = Arrays.copyOf(xs, xs.length);
      Doubles.reverse(reversed);
      MutableXySequence.create(reversed, Optional.of(ys));
    });
  }

  @Test
  final void setTests() {
    /* set */
    MutableXySequence xySet = MutableXySequence.copyOf(xy).set(0, 5.5);
    assertEquals(xySet.y(0), 5.5, 0);
    /* point.set */
    MutableXySequence xyCopy = MutableXySequence.copyOf(xy);
    double setTerm = 5.5;
    xyCopy.forEach(p -> {
      p.set(setTerm);
      assertEquals(p.y(), setTerm, 0);
    });
  }

  @Test
  final void minMaxTest() {
    MutableXySequence xy1 = MutableXySequence.copyOf(xy);
    XyPoint first = xy1.min();
    assertEquals(xs[0], first.x(), 0);
    assertEquals(ys[0], first.y(), 0);
    XyPoint last = xy1.max();
    assertEquals(xs[xs.length - 1], last.x(), 0);
    assertEquals(ys[ys.length - 1], last.y(), 0);
    // test correct class (mutable)
    assertEquals(first.getClass(), MutableArrayXySequence.MutablePoint.class);
    assertEquals(last.getClass(), MutableArrayXySequence.MutablePoint.class);
    // and that it's actually mutable
    first.set(4.0);
    assertEquals(4.0, xy1.y(0), 0);
    last.set(4.0);
    assertEquals(4.0, xy1.y(3), 0);
  }

  @Test
  final void streamTest() {
    // check that stream consists of mutable points
    assertEquals(
        xy.stream().findFirst().orElseThrow().getClass(),
        MutableArrayXySequence.MutablePoint.class);
  }

  @Test
  final void addTests() {

    MutableArrayXySequence xyAddTerm =
        (MutableArrayXySequence) MutableXySequence.copyOf(xy).add(5.5);
    double[] xyAddTermExpected = new double[] { 4.5, 16.0, 10.75, 8.0 };
    assertArrayEquals(xyAddTermExpected, xyAddTerm.ys, 0.0);

    MutableArrayXySequence xyAddArray =
        (MutableArrayXySequence) MutableXySequence.copyOf(xy).add(ys);
    double[] xyAddArrayExpected = new double[] { -2, 21.0, 10.5, 5.0 };
    assertArrayEquals(xyAddArrayExpected, xyAddArray.ys, 0.0);

    MutableArrayXySequence xyAddSequence =
        (MutableArrayXySequence) MutableXySequence.copyOf(xy).add(xy);
    assertArrayEquals(xyAddArrayExpected, xyAddSequence.ys, 0.0);
  }

  @Test
  final void multiplyTests() {

    MutableArrayXySequence xyMultiplyTerm =
        (MutableArrayXySequence) MutableXySequence.copyOf(xy).multiply(2);
    double[] xyMultiplyExpected = new double[] { -2, 21.0, 10.5, 5.0 };
    assertArrayEquals(xyMultiplyExpected, xyMultiplyTerm.ys, 0.0);

    double[] scaleArray = new double[] { 2, 2, 2, 2 };
    MutableArrayXySequence xyMultiplyArray =
        (MutableArrayXySequence) MutableXySequence.copyOf(xy).multiply(scaleArray);
    assertArrayEquals(xyMultiplyExpected, xyMultiplyArray.ys, 0.0);

    XySequence scaleSequence = XySequence.create(xs, scaleArray);
    MutableArrayXySequence xyMultiplySequence =
        (MutableArrayXySequence) MutableXySequence.copyOf(xy).multiply(scaleSequence);
    assertArrayEquals(xyMultiplyExpected, xyMultiplySequence.ys, 0.0);
  }

  @Test
  final void complementTest() {
    double[] xs = new double[] { 1, 2, 3, 4, 5 };
    double[] ys = new double[] { 0.0, 0.25, 0.5, 0.75, 1.0 };
    double[] ysExpected = new double[] { 1.0, 0.75, 0.5, 0.25, 0.0 };
    MutableArrayXySequence xy =
        (MutableArrayXySequence) MutableXySequence.create(xs, Optional.of(ys));
    xy.complement();
    assertArrayEquals(ysExpected, xy.ys, 0);
  }

  @Test
  final void clearTest() {
    assertFalse(xy.isClear());
    assertTrue(MutableXySequence.copyOf(xy).clear().isClear());
  }

  @Test
  final void trimTest() {
    double[] xs = { 0, 1, 2, 3, 4, 5 };
    double[] ys = { 0, 0, 1, 2, 0, 0 };
    XySequence xyTrimmed = MutableXySequence.create(xs, Optional.of(ys)).trim();
    assertEquals(xyTrimmed.getClass(), MutableArrayXySequence.class);
  }

  @Test
  final void toImmutableTest() {
    XySequence immutable = xy.toImmutable();
    assertEquals(immutable.getClass(), ArrayXySequence.class);
  }

  @Test
  final void testIteratorRemoveUOE() {
    assertThrows(UnsupportedOperationException.class, () -> {
      xy.iterator().remove();
    });
  }

  @Test
  final void transformConsumerTest() {
    double[] yExpected = new double[] { -1, 11.5, 7.25, 5.5 };
    Consumer<XyPoint> consumer = xy -> xy.set(xy.x() + xy.y());
    MutableXySequence xyActual = MutableXySequence.copyOf(xy).transform(consumer);
    assertArrayEquals(yExpected, xyActual.yValues().toArray(), 0);
  }

}
