package gov.usgs.earthquake.nshmp.geo.json;

import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotSame;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.OptionalInt;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.io.TempDir;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Lists;
import com.google.common.io.Resources;
import com.google.gson.JsonElement;
import com.google.gson.JsonIOException;
import com.google.gson.JsonParser;

import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson.Type;

class GeoJsonTests {

  static final Path RESOURCES = Path.of("src/test/resources/geo/json/");

  @TempDir
  static Path testFolder;

  private static final String FEATURE_COLLECTION_FILENAME = "feature-collection.geojson";
  private static final String FEATURE_FILENAME = "feature.geojson";
  private static final String TEST_FILENAME = "test.geojson";
  private static final String NOFILE_FILENAME = "nofile.geojson";

  private static final Path FEATURE_PROPERTIES_PATH = getResourcePathForFile("properties.geojson");
  private static final Path FEATURE_PROPERTIES_ABSENT_PATH =
      getResourcePathForFile("properties-prop-absent.geojson");
  private static final Path FEATURE_PROPERTIES_NULL_PATH =
      getResourcePathForFile("properties-prop-null.geojson");
  private static final Path FEATURE_PROPERTIES_OBJECT_PATH =
      getResourcePathForFile("properties-prop-object.geojson");
  private static final Path FEATURE_PROPERTIES_STRING_PATH =
      getResourcePathForFile("properties-prop-string.geojson");

  static final Feature TEST_FEATURE_PROPERTIES =
      GeoJson.from(FEATURE_PROPERTIES_PATH).toFeature();
  static final Feature TEST_FEATURE_PROPERTIES_ABSENT =
      GeoJson.from(FEATURE_PROPERTIES_ABSENT_PATH).toFeature();
  static final Feature TEST_FEATURE_PROPERTIES_NULL =
      GeoJson.from(FEATURE_PROPERTIES_NULL_PATH).toFeature();
  static final Feature TEST_FEATURE_PROPERTIES_OBJECT =
      GeoJson.from(FEATURE_PROPERTIES_OBJECT_PATH).toFeature();
  static final Feature TEST_FEATURE_PROPERTIES_STRING =
      GeoJson.from(FEATURE_PROPERTIES_STRING_PATH).toFeature();

  static final Location TEST_POINT = Location.create(-117, 34);

  static final LocationList TEST_LINE = LocationList.of(
      Location.create(-117, 34),
      Location.create(-118, 35),
      Location.create(-116, 37),
      Location.create(-117, 38));

  static final List<LocationList> TEST_MULTI_LINE = Lists.newArrayList(
      TEST_LINE,
      LocationList.of(
          Location.create(-117, 44),
          Location.create(-118, 45),
          Location.create(-116, 47),
          Location.create(-117, 48)));

  private static final double[] BBOX = new double[] { -118, 34, -116, 38 };

  private static final Feature TEST_LINE_STRING_FEATURE = Feature.polygon(TEST_LINE)
      .id(3)
      .bbox(BBOX)
      .build();

  private static final FeatureCollection TEST_FEATURE_COLLECTION = GeoJson.builder()
      .bbox(BBOX)
      .add(Feature.point(TEST_POINT)
          .id("featureId")
          .properties(ImmutableMap.of(
              "id", 1,
              "title", "Feature Title",
              "color", "#ff0080"))
          .build())
      .add(Feature.lineString(TEST_LINE)
          .build())
      .add(TEST_LINE_STRING_FEATURE)
      .add(Feature.multiLineString(TEST_MULTI_LINE)
          .build())
      .build();

  private static final Feature.Builder F_BUILDER_1 = Feature.point(TEST_POINT)
      .id("featureId")
      .properties(ImmutableMap.of(
          "id", 1,
          "title", "Feature Title",
          "color", "#ff0080"));

  private static final Feature.Builder F_BUILDER_2 = Feature.point(TEST_POINT)
      .id("featureId")
      .properties(ImmutableMap.of(
          "id", 1,
          "title", "Feature Title",
          "color", "#ff0080"));

  private static final Map<String, ?> PROPS_FOR_COPY = ImmutableMap.of(
      "id", 1,
      "title", "Feature Title",
      "color", "#ff0080");

  private static final Feature FEATURE_FOR_COPY = Feature.point(TEST_POINT)
      .id("featureId")
      .properties(PROPS_FOR_COPY)
      .build();

  private static Path getResourcePathForFile(String file) {
    return RESOURCES.resolve(file);
  }

  private static URL getResourceUrlForFile(String file) throws MalformedURLException {
    return getResourcePathForFile(file).toUri().toURL();
  }

  /* This hits most read methods in classes of the json package. */
  private void checkFeatureCollection(FeatureCollection fc) {

    List<Feature> features = fc.features();
    assertEquals(3, features.size());
    checkBboxIsEqualCopy(BBOX, fc.bbox().orElseThrow());

    Feature f1 = features.get(0);
    assertEquals("featureId", f1.idAsString().orElseThrow());
    assertEquals(Type.POINT, f1.type());
    checkBboxIsOptionalEmpty(f1.bbox());
    Properties f1Props = f1.properties();
    assertEquals("#ff0080", f1Props.getString("color").orElseThrow());
    assertEquals(1.0, f1Props.getDouble("id").orElseThrow(), 0.0);
    assertEquals(1, f1Props.getInt("id").orElseThrow());

    Feature f3 = features.get(2);
    assertEquals(3, f3.idAsInt().orElseThrow());
    checkBboxIsEqualCopy(BBOX, f3.bbox().orElseThrow());
  }

  private void checkBboxIsOptionalEmpty(Optional<double[]> actualBbox) {
    assertEquals(Optional.empty(), actualBbox);
  }

  private void checkBboxIsEqualCopy(double[] expectedBbox, double[] actualBbox) {
    assertNotSame(expectedBbox, actualBbox);
    assertArrayEquals(expectedBbox, actualBbox, 0.0);
  }

  private GeoJson.Builder dummyFeatureCollBuilder() {
    return GeoJson.builder()
        .bbox(BBOX)
        .add(dummyFeaturePoint())
        .add(dummyFeatureLineString());
  }

  private Feature dummyFeaturePoint() {
    return Feature.point(TEST_POINT)
        .id("featureId")
        .properties(PROPS_FOR_COPY)
        .build();
  }

  private Feature dummyFeatureLineString() {
    return Feature.lineString(TEST_LINE).id(3).build();
  }

  private Feature dummyFeaturePolygon() {
    return Feature.polygon(TEST_LINE).id(5).bbox(BBOX).build();
  }

  private Feature dummyFeatureMultiLineString() {
    return Feature.multiLineString(TEST_MULTI_LINE).id("id").bbox(BBOX).build();
  }

  private Feature dummyGetFullCopyOf(Feature f) {
    return Feature.copyOf(f)
        .properties(Properties.fromFeature(f).build())
        .build();
  }

  @Nested
  @DisplayName("Feature Tests")
  class FeatureTests {
    @Test
    void testIsFeatureType() {
      assertEquals(GeoJson.Type.FEATURE, dummyFeaturePoint().topLevelType());
    }

    @Test
    void testBboxIsInvalid() {
      // Feature bbox validation
      assertThrows(IllegalArgumentException.class, () -> {
        Feature.point(TEST_POINT).bbox(new double[0]);
      });
    }

    @Test
    void testWrite() throws IOException {
      /*
       * This also tests the removal of null 'id' and 'bbox' members on
       * serialization.
       */

      /* Feature JSON direct from builder. */
      JsonElement jsonExpected = JsonParser.parseString(F_BUILDER_1.build().toJson());

      /* Write to file and read back in. */
      Path out = testFolder.resolve(TEST_FILENAME);
      F_BUILDER_2.build().write(out);
      JsonElement jsonActual = JsonParser.parseString(new String(Files.readAllBytes(out)));
      assertEquals(jsonExpected, jsonActual);
    }

    @Test
    void testAsMethods_FeatureConversion() {
      assertEquals(TEST_POINT, dummyFeaturePoint().asPoint());
      assertEquals(TEST_LINE, dummyFeatureLineString().asLineString());
      assertEquals(TEST_LINE, dummyFeaturePolygon().asPolygonBorder());
      assertEquals(TEST_MULTI_LINE, dummyFeatureMultiLineString().asMultiLineString());
    }

    @Test
    void testAsMethods_Exceptions() {
      /*
       * The exceptions, i.e. requesting a Feature of Type A as Type B, are
       * handled by the Geometry class, so do not need to be tested exhaustively
       * here.
       */
      assertThrows(UnsupportedOperationException.class, () -> {
        dummyFeaturePoint().asLineString();
      });
      assertThrows(UnsupportedOperationException.class, () -> {
        dummyFeatureLineString().asPoint();
      });
      assertThrows(UnsupportedOperationException.class, () -> {
        dummyFeaturePolygon().asMultiLineString();
      });
      assertThrows(UnsupportedOperationException.class, () -> {
        dummyFeatureMultiLineString().asLineString();
      });
    }

    @Test
    void testToMethod_ThrowsUOE() {
      assertThrows(UnsupportedOperationException.class, () -> {
        GeoJson.from(getResourcePathForFile(FEATURE_COLLECTION_FILENAME)).toFeature();
      });
    }

    @Test
    void testFromMethodExceptions() {
      assertThrows(JsonIOException.class, () -> {
        URL badUrl = getResourceUrlForFile(NOFILE_FILENAME);
        GeoJson.from(badUrl).toFeatureCollection();
      });
    }

    @Test
    void testFromMethodsF() throws IOException {
      Feature f;

      /* URL */
      URL fJsonUrl = getResourceUrlForFile(FEATURE_FILENAME);
      f = GeoJson.from(fJsonUrl).toFeature();
      assertEquals(TEST_POINT, f.asPoint());

      /* Path */
      Path fJsonPath = getResourcePathForFile(FEATURE_FILENAME);
      f = GeoJson.from(fJsonPath).toFeature();
      assertEquals(TEST_POINT, f.asPoint());

      /* String */
      String fJsonString = Resources.toString(fJsonUrl, StandardCharsets.UTF_8);
      f = GeoJson.from(fJsonString).toFeature();
      assertEquals(TEST_POINT, f.asPoint());
    }

    @Test
    void testFeatureToJsonTree() {
      JsonElement expected = JsonParser.parseString(dummyFeaturePoint().toJson());
      JsonElement actual = dummyFeaturePoint().toJsonTree();
      assertEquals(expected, actual);
    }

    @Test
    void testFeatureId() {
      // toFeature() is implicitly tested here, pull that out to a testToMethods
      // test...
      Feature f = GeoJson.from(Feature.point(TEST_POINT).build().toJson()).toFeature();
      assertEquals(OptionalInt.empty(), f.idAsInt());
      assertEquals(Optional.empty(), f.idAsString());
      f = GeoJson.from(Feature.point(TEST_POINT).id("3").build().toJson()).toFeature();
      assertEquals(3, f.idAsInt().getAsInt());
      assertEquals("3", f.idAsString().orElseThrow());
    }

    @Test
    void testBbox_shouldReturnEqualCopyOrOptionalEmpty() {
      // Feature bbox
      checkBboxIsEqualCopy(BBOX, TEST_LINE_STRING_FEATURE.bbox().orElseThrow());
      // Missing bbox is Optional.empty
      checkBboxIsOptionalEmpty(dummyFeatureLineString().bbox());
    }

    @Test
    void testFeatureUsedBuilder() {
      assertThrows(IllegalStateException.class, () -> {
        Feature.Builder f = Feature.point(TEST_POINT);
        f.build();
        f.build();
      });
    }

    @Test
    void testFeatureCopyOf() {

      Feature copy = Feature.copyOf(FEATURE_FOR_COPY).build();

      assertEquals("featureId", copy.idAsString().orElseThrow());
      assertEquals(TEST_POINT, copy.asPoint());
      assertTrue(copy.bbox().isEmpty());
      assertTrue(copy.properties().map().isEmpty());
    }

    @Test
    void testPropertiesFromFeature() {

      Map<String, ?> props = Properties.fromFeature(FEATURE_FOR_COPY).build();

      Feature copy = Feature.copyOf(FEATURE_FOR_COPY)
          .properties(props)
          .build();

      assertEquals(props, copy.properties().map());
    }

    @Test
    void testFeatureHashCode() {
      // Point Feature
      assertEquals(
          FEATURE_FOR_COPY.hashCode(),
          dummyGetFullCopyOf(FEATURE_FOR_COPY).hashCode());
      // Line Feature
      assertEquals(
          dummyFeatureLineString().hashCode(),
          dummyFeatureLineString().hashCode());
      // Polygon Feature
      assertEquals(
          dummyFeaturePolygon().hashCode(),
          dummyFeaturePolygon().hashCode());
      // MultiLine Feature
      assertEquals(
          dummyFeatureMultiLineString().hashCode(),
          dummyFeatureMultiLineString().hashCode());
    }

    @SuppressWarnings("unlikely-arg-type")
    @Test
    void testFeatureEquals() {
      Feature targetSame = FEATURE_FOR_COPY;

      Feature targetCopy = dummyGetFullCopyOf(FEATURE_FOR_COPY);
      // make a modified copy
      Feature targetCopyMod = Feature.copyOf(FEATURE_FOR_COPY)
          .properties(Properties.fromFeature(FEATURE_FOR_COPY).build())
          .id(FEATURE_FOR_COPY.idAsString().orElseThrow() + "2")
          .build();

      // test logic branches
      assertFalse(FEATURE_FOR_COPY.equals(null));
      assertTrue(FEATURE_FOR_COPY.equals(targetSame));
      assertFalse(FEATURE_FOR_COPY.equals(TEST_FEATURE_COLLECTION));

      // copy is equal to original
      assertEquals(FEATURE_FOR_COPY, targetCopy);
      assertTrue(FEATURE_FOR_COPY.equals(targetCopy));
      // modified copy is unequal to original
      assertFalse(FEATURE_FOR_COPY.equals(targetCopyMod));

      // LineString Feature
      assertTrue(
          dummyFeatureLineString().equals(
              dummyFeatureLineString()));
      // Polygon Feature
      assertTrue(
          dummyFeaturePolygon().equals(
              dummyFeaturePolygon()));
      // MultiLineString Feature
      assertTrue(
          dummyFeatureMultiLineString().equals(
              dummyFeatureMultiLineString()));
    }

    /* Test geometry array parsing exception */
    @Test
    void testGeometryReadArrayFeature() {
      assertThrows(IllegalArgumentException.class, () -> {
        String coords4d = "{\"type\": \"Feature\",\"geometry\": {\"type\": \"Point\"," +
            "\"coordinates\": [[[[-117,34]]]]},\"properties\": {}}";
        GeoJson.from(coords4d).toFeature();
      });

    }
  }

  @Nested
  @DisplayName("FeatureCollection Tests")
  class FeatureCollectionTests {

    @Test
    void testBuilder_WhenEmptyThrowsISE() {
      assertThrows(IllegalStateException.class, () -> {
        GeoJson.builder().build().toJson();
      });
    }

    @Test
    void testIsFeatureCollectionType() {
      assertEquals(GeoJson.Type.FEATURE_COLLECTION, TEST_FEATURE_COLLECTION.topLevelType());
    }

    @Test
    void testBboxIsInvalid() {
      // Feature collection bbox validation
      assertThrows(IllegalArgumentException.class, () -> {
        GeoJson.builder().bbox(new double[] {});
      });
    }

    @Test
    void testFeatureCollectionWrite() throws IOException {
      /*
       * This also tests the removal of null 'id' and 'bbox' members on
       * serialization.
       */

      /* Feature collection JSON direct from builder. */
      JsonElement jsonExpected = JsonParser.parseString(TEST_FEATURE_COLLECTION.toJson());

      /* Write to file and read back in. */
      Path out = testFolder.resolve(TEST_FILENAME);
      TEST_FEATURE_COLLECTION.write(out);
      JsonElement jsonActual = JsonParser.parseString(new String(Files.readAllBytes(out)));

      /*
       * Note that the JsonObject.equals() is tolerant of object members being
       * reordered as a Map comparison is taking place under the hood. However,
       * if JsonArray elements are reordered, equals tests fail because a List
       * is being compared.
       */
      assertEquals(jsonExpected, jsonActual);
    }

    /* Test the toFeatureCollection() and Feature.as*() methods. */
    // where are the testToMethod tests? scattered within other tests, pull
    // these
    // out to make tests simpler with a single purpose
    @Test
    void testToMethod_ThrowsUOE() {
      assertThrows(UnsupportedOperationException.class, () -> {
        GeoJson.from(getResourcePathForFile(FEATURE_FILENAME)).toFeatureCollection();
      });
    }

    @Test
    void testFromMethodExceptions() {
      assertThrows(JsonIOException.class, () -> {
        Path badPath = getResourcePathForFile(NOFILE_FILENAME).toAbsolutePath();
        GeoJson.from(badPath).toFeatureCollection();
      });
    }

    @Test
    void testFromMethods() throws URISyntaxException, IOException {

      FeatureCollection fc;

      /* URL */
      URL fcJsonUrl = getResourceUrlForFile(FEATURE_COLLECTION_FILENAME);
      fc = GeoJson.from(fcJsonUrl).toFeatureCollection();
      checkFeatureCollection(fc);

      /* Path */
      Path fcJsonPath = getResourcePathForFile(FEATURE_COLLECTION_FILENAME);
      fc = GeoJson.from(fcJsonPath).toFeatureCollection();
      checkFeatureCollection(fc);

      /* String */
      String fcJsonString = Resources.toString(fcJsonUrl, StandardCharsets.UTF_8);
      fc = GeoJson.from(fcJsonString).toFeatureCollection();
      checkFeatureCollection(fc);
    }

    @Test
    void testFeatureCollectionBboxRemove() {
      String json = GeoJson.builder()
          .add(Feature.point(TEST_POINT).build())
          .build()
          .toJson();
      FeatureCollection fc = GeoJson.from(json).toFeatureCollection();
      checkBboxIsOptionalEmpty(fc.bbox());
    }

    @Test
    void testFeatureCollectionToJsonTree() {

      GeoJson.Builder b1 = dummyFeatureCollBuilder();
      String fcString = b1.build().toJson();
      JsonElement expected = JsonParser.parseString(fcString);

      GeoJson.Builder b2 = dummyFeatureCollBuilder();
      JsonElement actual = b2.build().toJsonTree();

      assertEquals(expected, actual);
    }

    @Test
    void testBbox_shouldReturnEqualCopyOrOptionalEmpty() {
      // Feature Collection bbox
      checkBboxIsEqualCopy(BBOX, TEST_FEATURE_COLLECTION.bbox().orElseThrow());
      // Missing bbox is Optional.empty
      checkBboxIsOptionalEmpty(TEST_FEATURE_COLLECTION.features().get(0).bbox());
    }

    @Test
    void testFeatureCollectionUsedBuilder() {
      assertThrows(IllegalStateException.class, () -> {
        GeoJson.Builder fc = GeoJson.builder().add(TEST_LINE_STRING_FEATURE);
        fc.build();
        fc.build();
      });
    }

    /* Test geometry array parsing exception */
    @Test
    void testGeometryReadArray() {
      assertThrows(IllegalArgumentException.class, () -> {
        String coords4d = "{\"type\": \"FeatureCollection\"," +
            "\"features\":[{\"type\": \"Feature\",\"geometry\": {\"type\": \"Point\"," +
            "\"coordinates\": [[[[-117,34]]]]},\"properties\": {}}]}";
        GeoJson.from(coords4d).toFeatureCollection();
      });

    }

  }

}
