package gov.usgs.earthquake.nshmp.geo.json;

import static gov.usgs.earthquake.nshmp.geo.json.GeoJsonTests.TEST_FEATURE_PROPERTIES;
import static gov.usgs.earthquake.nshmp.geo.json.GeoJsonTests.TEST_FEATURE_PROPERTIES_ABSENT;
import static gov.usgs.earthquake.nshmp.geo.json.GeoJsonTests.TEST_FEATURE_PROPERTIES_NULL;
import static gov.usgs.earthquake.nshmp.geo.json.GeoJsonTests.TEST_FEATURE_PROPERTIES_OBJECT;
import static gov.usgs.earthquake.nshmp.geo.json.GeoJsonTests.TEST_FEATURE_PROPERTIES_STRING;
import static gov.usgs.earthquake.nshmp.geo.json.Properties.Style.DESCRIPTION;
import static gov.usgs.earthquake.nshmp.geo.json.Properties.Style.FILL;
import static gov.usgs.earthquake.nshmp.geo.json.Properties.Style.FILL_OPACITY;
import static gov.usgs.earthquake.nshmp.geo.json.Properties.Style.MARKER_COLOR;
import static gov.usgs.earthquake.nshmp.geo.json.Properties.Style.MARKER_SIZE;
import static gov.usgs.earthquake.nshmp.geo.json.Properties.Style.MARKER_SYMBOL;
import static gov.usgs.earthquake.nshmp.geo.json.Properties.Style.STROKE;
import static gov.usgs.earthquake.nshmp.geo.json.Properties.Style.STROKE_OPACITY;
import static gov.usgs.earthquake.nshmp.geo.json.Properties.Style.STROKE_WIDTH;
import static gov.usgs.earthquake.nshmp.geo.json.Properties.Style.TITLE;
import static gov.usgs.earthquake.nshmp.geo.json.Properties.checkStyle;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.Test;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;

import gov.usgs.earthquake.nshmp.geo.json.Properties.Style;

class PropertiesTests {

  /*
   * Expected exceptions for get methods.
   *
   * When reading a property that is expected to be of a certain type (e.g.
   * double), GSON may throw a variety of exceptions: member missing
   * (NoSuchElementException), member present but null
   * (UnsupportedOperationException), member present but not a JsonPrimitive
   * (ClassCastException), member present and represented by an array of size>1
   * (IllegalStateException), or member present but not parseable into the
   * desired number (NumberFOrmatException).
   *
   * Collectively these are caught and an Optional.empty() is returned; we test
   * a few such cases below.
   */

  private static class TestObject {
    final String str;
    final List<Double> list;

    TestObject(String str, List<Double> list) {
      this.str = str;
      this.list = list;
    }

    @Override
    public boolean equals(Object obj) {
      if (obj == null) {
        return false;
      }

      TestObject other = (TestObject) obj;
      return this.list.equals(other.list) && this.str.equals(other.str);
    }
  }

  @Test
  void testGetMethods() {
    Properties propAbsent = TEST_FEATURE_PROPERTIES_ABSENT.properties();
    Properties propNull = TEST_FEATURE_PROPERTIES_NULL.properties();
    Properties propObject = TEST_FEATURE_PROPERTIES_OBJECT.properties();
    Properties propString = TEST_FEATURE_PROPERTIES_STRING.properties();

    String key = "test-property";

    assertTrue(propAbsent.getDouble(key).isEmpty());
    assertTrue(propNull.getDouble(key).isEmpty());
    assertTrue(propObject.getDouble(key).isEmpty());
    assertTrue(propString.getDouble(key).isEmpty());

    assertTrue(propAbsent.getInt(key).isEmpty());
    assertTrue(propNull.getInt(key).isEmpty());
    assertTrue(propObject.getInt(key).isEmpty());
    assertTrue(propString.getInt(key).isEmpty());

    assertTrue(propAbsent.getString(key).isEmpty());
    assertTrue(propNull.getString(key).isEmpty());
    assertTrue(propObject.getString(key).isEmpty());
    assertEquals("test1234", propString.getString(key).orElseThrow());
  }

  @Test
  void testProperties() { // throws MalformedURLException {

    Properties props = TEST_FEATURE_PROPERTIES.properties();

    assertTrue(props.getString(DESCRIPTION.toString()).isPresent());
    assertTrue(props.getBoolean("boolean").orElseThrow());
    assertEquals(42, props.getInt("id").getAsInt());
    assertEquals(0.5, props.getDouble(FILL_OPACITY.toString()).orElseThrow(), 0.0);
    assertEquals("description", props.getString(DESCRIPTION.toString()).orElseThrow());
    TestObject actual = props.get("object", TestObject.class).orElseThrow();
    TestObject expected = new TestObject("test", ImmutableList.of(1.0, 2.0));
    assertEquals(expected.getClass(), actual.getClass());
    assertEquals(expected, actual);

  }

  @Test
  void testBuilder() {
    /*
     * test {@code Properties.put()} for {@code String} and {@code Style} keys,
     * and String, number, boolean, and object values.
     */

    Map<String, ?> bMap = Properties.builder()
        .put(DESCRIPTION, "description")
        .put(FILL, "#112233")
        .put(FILL_OPACITY, 0.5)
        .put(MARKER_COLOR, "#332211")
        .put(MARKER_SIZE, "small")
        .put(MARKER_SYMBOL, "a")
        .put(STROKE, "#555555")
        .put(STROKE_OPACITY, 0.7)
        .put(STROKE_WIDTH, 1.0)
        .put(TITLE, "title")
        .put("id", (double) 42) // mimic GSON deserialization
        .put("boolean", true)
        .put("object", new TestObject("test", ImmutableList.of(1.0, 2.0)))
        .build();

    // verify that all keys were added to map
    assertTrue(bMap.containsKey(DESCRIPTION.toString()));
    assertTrue(bMap.containsKey(FILL.toString()));
    assertTrue(bMap.containsKey(FILL_OPACITY.toString()));
    assertTrue(bMap.containsKey(MARKER_COLOR.toString()));
    assertTrue(bMap.containsKey(MARKER_SIZE.toString()));
    assertTrue(bMap.containsKey(MARKER_SYMBOL.toString()));
    assertTrue(bMap.containsKey(STROKE.toString()));
    assertTrue(bMap.containsKey(STROKE_OPACITY.toString()));
    assertTrue(bMap.containsKey(STROKE_WIDTH.toString()));
    assertTrue(bMap.containsKey(TITLE.toString()));
    assertTrue(bMap.containsKey("id"));
    assertTrue(bMap.containsKey("boolean"));
    assertTrue(bMap.containsKey("object"));
  }

  @Test
  void testBuilderExceptions() {
    // put wrong type of value or invalid value into builder
    assertThrows(
        IllegalArgumentException.class,
        () -> Properties.builder().put(Style.DESCRIPTION, 0.0));
    assertThrows(
        IllegalArgumentException.class,
        () -> Properties.builder().put(Style.FILL_OPACITY, "string"));
  }

  @Test
  void testBuilderRemove() {
    Map<String, ?> expected = Properties.builder()
        .put("id", 1)
        .put("title", "Feature Title")
        .build();

    Map<String, ?> actual = Properties.builder()
        .put("id", 1)
        .put("title", "Feature Title")
        .put("removeme", "please")
        .remove("removeme")
        .build();

    assertEquals(expected, actual);
  }

  @Test
  void testCheckStyleColor() {
    var colors = Lists.newArrayList(
        "f00",
        "ff0000",
        "#f00",
        "#ff0000");

    colors.forEach(color -> {
      assertEquals(color, checkStyle(Style.FILL, color));
      assertEquals(color, checkStyle(Style.MARKER_COLOR, color));
      assertEquals(color, checkStyle(Style.STROKE, color));
    });

    assertThrows(IllegalArgumentException.class, () -> checkStyle(Style.FILL, "red"));
    assertThrows(IllegalArgumentException.class, () -> checkStyle(Style.MARKER_COLOR, "red"));
    assertThrows(IllegalArgumentException.class, () -> checkStyle(Style.STROKE, "red"));
  }

  @Test
  void testCheckStyleOpacity() {
    var opacity = 0.5;
    assertEquals(opacity, (double) checkStyle(Style.FILL_OPACITY, opacity), 0);
    assertEquals(opacity, (double) checkStyle(Style.STROKE_OPACITY, opacity), 0);

    assertThrows(IllegalArgumentException.class, () -> checkStyle(Style.FILL_OPACITY, -0.1));
    assertThrows(IllegalArgumentException.class, () -> checkStyle(Style.FILL_OPACITY, 1.1));
    assertThrows(IllegalArgumentException.class, () -> checkStyle(Style.STROKE_OPACITY, -0.1));
    assertThrows(IllegalArgumentException.class, () -> checkStyle(Style.STROKE_OPACITY, 1.1));
  }

  @Test
  void testCheckStyleMarkerSize() {
    assertThrows(IllegalArgumentException.class, () -> checkStyle(Style.MARKER_SIZE, "bad_size"));
    assertEquals("small", checkStyle(Style.MARKER_SIZE, "small"));
  }

  @Test
  void testCheckStyleMarkerSymbol() {
    assertEquals("a", checkStyle(Style.MARKER_SYMBOL, "a"));
    assertEquals("car-15", checkStyle(Style.MARKER_SYMBOL, "car-15"));
    assertEquals(5, checkStyle(Style.MARKER_SYMBOL, 5));

    assertThrows(IllegalArgumentException.class, () -> checkStyle(Style.MARKER_SYMBOL, 10));
    assertThrows(IllegalArgumentException.class, () -> checkStyle(Style.MARKER_SYMBOL, -1));
    assertThrows(IllegalArgumentException.class, () -> checkStyle(Style.MARKER_SYMBOL, "1"));
    assertThrows(IllegalArgumentException.class, () -> checkStyle(Style.MARKER_SYMBOL, 1.0));
  }

  @Test
  void testCheckStyleStrokeWidth() {
    assertEquals(0.0, (double) checkStyle(Style.STROKE_WIDTH, 0.0), 0);

    assertThrows(IllegalArgumentException.class, () -> checkStyle(Style.STROKE_WIDTH, "0.0"));
    assertThrows(IllegalArgumentException.class, () -> checkStyle(Style.STROKE_WIDTH, -1.0));
  }

  @Test
  void testCheckStyle() {
    assertEquals("text", checkStyle(Style.DESCRIPTION, "text"));
    assertEquals("text", checkStyle(Style.TITLE, "text"));

    assertThrows(IllegalArgumentException.class, () -> checkStyle(Style.DESCRIPTION, 0.0));
    assertThrows(IllegalArgumentException.class, () -> checkStyle(Style.TITLE, 0.0));
  }

  @Test
  void testCheckStyleNull() {
    assertThrows(NullPointerException.class, () -> checkStyle(null, ""));
    assertThrows(NullPointerException.class, () -> checkStyle(Style.DESCRIPTION, null));
  }

}
