package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.DIP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RAKE;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RJB;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RRUP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RX;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.WIDTH;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.Z1P0;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.Z2P5;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.ZHYP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.ZSED;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.ZTOR;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import org.junit.jupiter.api.Test;

import gov.usgs.earthquake.nshmp.gmm.GmmInput.Builder;
import gov.usgs.earthquake.nshmp.gmm.GmmInput.Field;

class GmmInputTest {

  private static final double dip = DIP.defaultValue;
  private static final double Mw = MW.defaultValue;
  private static final double rake = RAKE.defaultValue;
  private static final double rJB = RJB.defaultValue;
  private static final double rRup = RRUP.defaultValue;
  private static final double rX = RX.defaultValue;
  private static final double vs30 = VS30.defaultValue;
  private static final double width = WIDTH.defaultValue;
  private static final double z1p0 = Z1P0.defaultValue;
  private static final double z2p5 = Z2P5.defaultValue;
  private static final double zSed = ZSED.defaultValue;
  private static final double zHyp = ZHYP.defaultValue;
  private static final double zTor = ZTOR.defaultValue;

  @Test
  final void builderFromCopy() {
    GmmInput model = GmmInput.builder().withDefaults().build();
    GmmInput gmmCopy = GmmInput.builder().fromCopy(model).build();

    assertEquals(gmmCopy.dip, model.dip, 0);
    assertEquals(gmmCopy.Mw, model.Mw, 0);
    assertEquals(gmmCopy.rake, model.rake, 0);
    assertEquals(gmmCopy.rJB, model.rJB, 0);
    assertEquals(gmmCopy.rRup, model.rRup, 0);
    assertEquals(gmmCopy.rX, model.rX, 0);
    assertEquals(gmmCopy.vs30, model.vs30, 0);
    assertEquals(gmmCopy.width, model.width, 0);
    assertEquals(gmmCopy.z1p0, model.z1p0, 0);
    assertEquals(gmmCopy.z2p5, model.z2p5, 0);
    assertEquals(gmmCopy.zSed, model.zSed, 0);
    assertEquals(gmmCopy.zHyp, model.zHyp, 0);
    assertEquals(gmmCopy.zTor, model.zTor, 0);
  }

  @Test
  final void builderWithDefaults() {
    GmmInput gmm = GmmInput.builder().withDefaults().build();

    assertEquals(gmm.dip, dip, 0);
    assertEquals(gmm.Mw, Mw, 0);
    assertEquals(gmm.rake, rake, 0);
    assertEquals(gmm.rJB, rJB, 0);
    assertEquals(gmm.rRup, rRup, 0);
    assertEquals(gmm.rX, rX, 0);
    assertEquals(gmm.vs30, vs30, 0);
    assertEquals(gmm.width, width, 0);
    assertEquals(gmm.z1p0, z1p0, 0);
    assertEquals(gmm.z2p5, z2p5, 0);
    assertEquals(gmm.zSed, zSed, 0);
    assertEquals(gmm.zHyp, zHyp, 0);
    assertEquals(gmm.zTor, zTor, 0);
  }

  @Test
  final void fieldFromString() {
    Field mwField = Field.fromString("mw");

    assertEquals(MW, mwField);
    assertEquals(MW.defaultValue, mwField.defaultValue, 0);
    assertEquals(MW.id, mwField.id);
    assertEquals(MW.info, mwField.info);
    assertEquals(MW.label, mwField.label);
    assertEquals(MW.units, mwField.units);

    assertThrows(IllegalArgumentException.class, () -> {
      Field.fromString("test");
    });
  }

  @Test
  final void gmmInput() {
    GmmInput gmm = GmmInput.builder()
        .mag(Mw)
        .rJB(rJB)
        .rRup(rRup)
        .rX(rX)
        .dip(dip)
        .width(width)
        .zTor(zTor)
        .zHyp(zHyp)
        .rake(rake)
        .vs30(vs30)
        .z1p0(z1p0)
        .z2p5(z2p5)
        .zSed(zSed)
        .build();

    assertEquals(gmm.dip, dip, 0);
    assertEquals(gmm.Mw, Mw, 0);
    assertEquals(gmm.rake, rake, 0);
    assertEquals(gmm.rJB, rJB, 0);
    assertEquals(gmm.rRup, rRup, 0);
    assertEquals(gmm.rX, rX, 0);
    assertEquals(gmm.vs30, vs30, 0);
    assertEquals(gmm.width, width, 0);
    assertEquals(gmm.z1p0, z1p0, 0);
    assertEquals(gmm.z2p5, z2p5, 0);
    assertEquals(gmm.zSed, zSed, 0);
    assertEquals(gmm.zHyp, zHyp, 0);
    assertEquals(gmm.zTor, zTor, 0);
  }

  @Test
  final void gmmInputEquals() {
    double v = 25.0;
    Builder gmmBuilder = GmmInput.builder();
    GmmInput gmm = gmmBuilder.withDefaults().build();
    GmmInput gmmCopy = GmmInput.builder().fromCopy(gmm).build();

    assertTrue(gmm.equals(gmmCopy));
    assertTrue(gmm.equals(gmm));
    assertFalse(gmm.equals(null));
    assertFalse(gmm.equals(new Object()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().dip(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().mag(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().rake(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().rJB(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().rRup(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().rX(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().vs30(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().width(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().z1p0(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().z2p5(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().zSed(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().zHyp(v).build()));
    assertFalse(gmm.equals(gmmBuilder.withDefaults().zTor(v).build()));
  }

  @Test
  final void gmmInputHashCode() {
    GmmInput gmm = GmmInput.builder().withDefaults().build();
    GmmInput gmmCopy = GmmInput.builder().fromCopy(gmm).build();
    GmmInput gmmDiff = GmmInput.builder().withDefaults()
        .mag(10.0).z1p0(10.0).z2p5(10.0).build();

    assertEquals(gmm.hashCode(), gmmCopy.hashCode());
    assertNotEquals(gmm.hashCode(), gmmDiff.hashCode());
  }

  @Test
  final void gmmInputToString() {
    GmmInput gmm = GmmInput.builder().withDefaults().build();
    GmmInput gmmCopy = GmmInput.builder().fromCopy(gmm).build();
    String toString = gmm.toString();
    String toStringCopy = gmmCopy.toString();

    assertTrue(toString.contains("dip="));
    assertTrue(toString.contains("Mw="));
    assertTrue(toString.contains("rake="));
    assertTrue(toString.contains("rJB="));
    assertTrue(toString.contains("rRup="));
    assertTrue(toString.contains("rX="));
    assertTrue(toString.contains("vs30="));
    assertTrue(toString.contains("width="));
    assertTrue(toString.contains("z1p0="));
    assertTrue(toString.contains("z2p5="));
    assertTrue(toString.contains("zSed="));
    assertTrue(toString.contains("zHyp="));
    assertTrue(toString.contains("zTor="));
    assertEquals(toString, toStringCopy);
  }

}
