package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.Imt.AI;
import static gov.usgs.earthquake.nshmp.gmm.Imt.ASI;
import static gov.usgs.earthquake.nshmp.gmm.Imt.CAV;
import static gov.usgs.earthquake.nshmp.gmm.Imt.DS575;
import static gov.usgs.earthquake.nshmp.gmm.Imt.DS595;
import static gov.usgs.earthquake.nshmp.gmm.Imt.DSI;
import static gov.usgs.earthquake.nshmp.gmm.Imt.PGA;
import static gov.usgs.earthquake.nshmp.gmm.Imt.PGD;
import static gov.usgs.earthquake.nshmp.gmm.Imt.PGV;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P01;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P02;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P03;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P05;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P075;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P1;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P15;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P2;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P25;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P3;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P4;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P5;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P75;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA10P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA1P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA1P5;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA2P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA3P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA4P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA5P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA7P5;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SI;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertThrows;

import java.util.EnumSet;
import java.util.Set;

import org.junit.jupiter.api.Test;

class ImtTest {

  /*
   * IMT fields, particularly PGA and spectral accelerations are hit by a
   * variety of tests. The tests here cover edge case exceptions, less
   * frequently used IMTs, and other helper methods.
   */

  @Test
  void testToString() {
    assertEquals(PGA.toString(), "Peak Ground Acceleration");
    assertEquals(PGV.toString(), "Peak Ground Velocity");
    assertEquals(PGD.toString(), "Peak Ground Displacement");
    assertEquals(AI.toString(), "Arias Intensity");
    assertEquals(ASI.toString(), "Acceleration Spectrum Intensity");
    assertEquals(DSI.toString(), "Displacement Spectrum Intensity");
    assertEquals(SI.toString(), "Spectrum intensity");
    assertEquals(CAV.toString(), "Cumulative Absolute Velocity");
    assertEquals(DS575.toString(), "Significant Duration 5-75%");
    assertEquals(DS595.toString(), "Significant Duration 5-95%");
    assertEquals(SA1P0.toString(), "1.0 s Spectral Acceleration");
  }

  @Test
  void testUnits() {
    assertEquals(PGA.units(), "g");
    assertEquals(PGV.units(), "cm/s");
    assertEquals(PGD.units(), "cm");
    assertEquals(AI.units(), "m/s");
    assertEquals(ASI.units(), "g⋅s");
    assertEquals(DSI.units(), "cm⋅s");
    assertEquals(SI.units(), "cm⋅s/s");
    assertEquals(CAV.units(), "g⋅s");
    assertEquals(DS575.units(), "s");
    assertEquals(DS595.units(), "s");
    assertEquals(SA1P0.units(), "g");
  }

  @Test
  void testPeriod() {
    assertEquals(0.2, SA0P2.period());
    assertThrows(UnsupportedOperationException.class, () -> {
      PGV.period();
    });
  }

  @Test
  void testFromPeriod() {
    assertEquals(SA0P2, Imt.fromPeriod(0.2));
    assertThrows(IllegalArgumentException.class, () -> {
      Imt.fromPeriod(0.1234);
    });
  }

  @Test
  void testFrequency() {
    assertEquals(100, PGA.frequency());
    assertEquals(5.0, SA0P2.frequency());
    assertEquals(3.3333, SA0P3.frequency(), 0.0001);
    assertEquals(0.1, SA10P0.frequency());
    assertThrows(UnsupportedOperationException.class, () -> {
      PGV.frequency();
    });
  }

  @Test
  void testSaImts() {
    assertEquals(Imt.saImts(), EnumSet.range(SA0P01, SA10P0));
    assertFalse(Imt.saImts().contains(PGV));
  }

  @Test
  void testMprs() {
    Set<Imt> mprsImts = EnumSet.of(
        PGA,
        SA0P01, SA0P02, SA0P03, SA0P05, SA0P075,
        SA0P1, SA0P15, SA0P2, SA0P25, SA0P3, SA0P4, SA0P5, SA0P75,
        SA1P0, SA1P5, SA2P0, SA3P0, SA4P0, SA5P0, SA7P5,
        SA10P0);
    assertEquals(mprsImts, Imt.mprsImts());
  }

}
