package gov.usgs.earthquake.nshmp.mfd;

import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.io.BufferedReader;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;

import org.junit.jupiter.api.Test;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonObject;
import com.google.gson.JsonParser;

import gov.usgs.earthquake.nshmp.Text;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Builder;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Properties;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Properties.GutenbergRichter;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Properties.Single;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Properties.TaperedGr;
import gov.usgs.earthquake.nshmp.mfd.Mfd.Type;

class MfdTests {

  // Add test to see if we can slip in negative a value
  // via GR properties

  static final double[] M = { 5.05, 5.15, 5.25, 5.35, 5.45 };
  static final double[] R = { 0.1, 0.08, 0.06, 0.04, 0.02 };

  @Test
  void testCreate() {
    /* Covers array and xy constructors. */
    Mfd mfd = Mfd.create(M, R);
    assertArrayEquals(M, mfd.data().xValues().toArray());
    assertArrayEquals(R, mfd.data().yValues().toArray());
    assertEquals(Type.INCR, mfd.properties().type());
  }

  @Test
  void testBuilderFrom() {
    Mfd mfd = Builder.from(M, R).build();
    assertEquals(Type.INCR, mfd.properties().type());
    assertArrayEquals(M, mfd.data().xValues().toArray());
    assertArrayEquals(R, mfd.data().yValues().toArray());

    mfd = Builder.from(XySequence.create(M, R)).build();
    assertEquals(Type.INCR, mfd.properties().type());
    assertArrayEquals(M, mfd.data().xValues().toArray());
    assertArrayEquals(R, mfd.data().yValues().toArray());

    Mfd gr = Mfd.newGutenbergRichterBuilder(1.0, 0.1, 5.0, 5.5).build();
    mfd = Builder.from(gr).build();
    assertEquals(Type.GR, mfd.properties().type());
    assertArrayEquals(GR_M, mfd.data().xValues().toArray());
    assertArrayEquals(GR_R, mfd.data().yValues().toArray());
  }

  @Test
  void mfdToString() {
    Mfd mfd = Builder.from(M, R).build();
    assertEquals("MFD:" + mfd.properties() + Text.NEWLINE + mfd.data(), mfd.toString());
  }

  @Test
  void testPropsToBuilderException() {
    /* Base INCR builder may not support toBuilder */
    Properties props = new Properties();
    assertThrows(UnsupportedOperationException.class, () -> {
      props.toBuilder();
    });
  }

  @Test
  void testPropsToString() {

  }

  static final double[] GAUSS_M = {
      6.76,
      6.808,
      6.856,
      6.904,
      6.952,
      7.0,
      7.048,
      7.096,
      7.144,
      7.192,
      7.24 };

  static final double[] GAUSS_R = {
      0.022190548492442928,
      0.04558899978527847,
      0.07981140824009195,
      0.11906462996609925,
      0.1513608096777362,
      0.16396720767670236,
      0.1513608096777362,
      0.11906462996609925,
      0.07981140824009195,
      0.04558899978527847,
      0.022190548492442928 };

  @Test
  void testSingle() {

    /* Factory builder; covers props.toBuilder() */
    Mfd mfd = Mfd.newSingleBuilder(6.0).build();
    XySequence xy = mfd.data();
    assertTrue(xy.size() == 1);
    assertEquals(6.0, xy.min().x());
    assertEquals(1.0, xy.min().y());

    /* Factory builder; covers props.toGaussianBuilder() */
    mfd = Mfd.newGaussianBuilder(7.0, 11, 0.12, 2).build();
    xy = mfd.data();
    assertTrue(xy.size() == 11);
    assertArrayEquals(GAUSS_M, xy.xValues().toArray());
    assertArrayEquals(GAUSS_R, xy.yValues().toArray());
    assertEquals(1.0, xy.yValues().sum());

    /* Properties */
    Properties props = new Single(5.0, 2.0);
    Single singleProps = props.getAsSingle(); // cover getAs
    assertEquals(5.0, singleProps.magnitude());
    assertEquals(2.0, singleProps.rate());
    singleProps = new Single(7.0);
    assertEquals(7.0, singleProps.magnitude());
    assertEquals(1.0, singleProps.rate());

    /* props.toString() */
    assertEquals(
        singleProps.type() +
            " {magnitude=" + singleProps.magnitude() +
            ", rate=" + singleProps.rate() + "}",
        singleProps.toString());

    // Changing rate=0 to 1 is probably not correct
    // and this check should/will go away
    singleProps = new Single(7.0, 0.0);
    mfd = singleProps.toBuilder().build();
    xy = mfd.data();
    assertEquals(7.0, xy.min().x());
    assertEquals(1.0, xy.min().y());
  }

  static final double[] GR_M = {
      5.05,
      5.15,
      5.25,
      5.35,
      5.45 };

  static final double[] GR_R = {
      8.912509381337459E-5,
      7.079457843841373E-5,
      5.623413251903491E-5,
      4.466835921509635E-5,
      3.5481338923357534E-5 };

  @Test
  void testGutenbergRichter() {

    /* Factory builder; covers props.toBuilder() */
    Mfd mfd = Mfd.newGutenbergRichterBuilder(1.0, 0.1, 5.0, 5.5).build();
    XySequence xy = mfd.data();
    assertTrue(xy.size() == 5);
    assertArrayEquals(GR_M, xy.xValues().toArray());
    assertArrayEquals(GR_R, xy.yValues().toArray());

    /* Properties */
    Properties props = new GutenbergRichter(1.0, 1.0, 0.1, 5.0, 5.5);
    GutenbergRichter grProps = props.getAsGr(); // cover getAs
    assertEquals(1.0, grProps.a());
    assertEquals(1.0, grProps.b());
    assertEquals(0.1, grProps.Δm());
    assertEquals(5.0, grProps.mMin());
    assertEquals(5.5, grProps.mMax());

    grProps = GutenbergRichter.copyOfWithMax(grProps, 7.5);
    assertEquals(1.0, grProps.a());
    assertEquals(1.0, grProps.b());
    assertEquals(0.1, grProps.Δm());
    assertEquals(5.0, grProps.mMin());
    assertEquals(7.5, grProps.mMax());

    grProps = GutenbergRichter.copyOfWithRate(grProps, 1.5);
    assertEquals(1.5, grProps.a());
    assertEquals(1.0, grProps.b());
    assertEquals(0.1, grProps.Δm());
    assertEquals(5.0, grProps.mMin());
    assertEquals(7.5, grProps.mMax());

    /* props.toString() */
    assertEquals(
        grProps.type() + " {a=" + grProps.a() +
            ", b=" + grProps.b() +
            ", Δm=" + grProps.Δm() +
            ", mMin=" + grProps.mMin() +
            ", mMax=" + grProps.mMax() + "}",
        grProps.toString());
  }

  /*
   * Something in the TaperedGR moment based scaling leads to differences in the
   * 4th to 5th significant figure that should be better understood, and then
   * this main method cleaned out. The results are still pretty close overall.
   */
  public static void main(String[] args) {
    double bVal = 0.8;
    double aVal = Math.log10(4.0);
    double Δm = 0.1;
    double mMin = 5.0;
    double mMax = 8.0;
    double mc = 7.5;

    // /* GR direct vs scaled builder comparison */
    // System.out.println("Gutenberg Richter");
    // Mfd mfd = new GutenbergRichter(aVal, bVal, Δm, mMin, mMax)
    // .toBuilder()
    // .build();
    //
    // System.out.println(Arrays.toString(mfd.data().xValues().toArray()));
    // System.out.println(Arrays.toString(mfd.data().yValues().toArray()));
    //
    // double incrRate = Mfds.gutenbergRichterRate(aVal, bVal, 5.05);
    // mfd = Mfd.newGutenbergRichterBuilder(bVal, Δm, mMin, mMax)
    // .scaleToIncrementalRate(incrRate)
    // .build();
    //
    // System.out.println(Arrays.toString(mfd.data().yValues().toArray()));

    /* GR direct vs scaled builder comparison */
    System.out.println("Tapered Gutenberg Richter");
    Mfd mfd = new TaperedGr(aVal, bVal, Δm, mMin, mMax, mc)
        .toBuilder()
        .build();

    double[] directRates = mfd.data().yValues().toArray();
    System.out.println(Arrays.toString(mfd.data().xValues().toArray()));
    System.out.println(Arrays.toString(directRates));

    double incrRate = Mfds.gutenbergRichterRate(aVal, bVal, 5.05);
    mfd = Mfd.newTaperedGutenbergRichterBuilder(bVal, Δm, mMin, mMax, mc)
        .scaleToIncrementalRate(incrRate)
        .build();

    double[] scaledRates = mfd.data().yValues().toArray();
    System.out.println(Arrays.toString(scaledRates));

    int size = directRates.length;
    double[] ratios = new double[size];
    for (int i = 0; i < size; i++) {
      double ratio = 100.0 * (directRates[i] / scaledRates[i]);
      ratios[i] = ratio;
    }
    System.out.println(Arrays.toString(ratios));

  }

  private static final double[] TAPERED_GR_M = {
      5.05, 5.15, 5.25, 5.35, 5.45,
      5.55, 5.65, 5.75, 5.85, 5.95,
      6.05, 6.15, 6.25, 6.35, 6.45,
      6.55, 6.65, 6.75, 6.85, 6.95,
      7.05, 7.15, 7.25, 7.35, 7.45,
      7.55, 7.65, 7.75, 7.85, 7.95 };

  /* Generated using props.toBuilder().build() */
  private static final double[] TAPERED_GR_R = {
      3.6480433574236396E-4,
      3.0345403881799134E-4,
      2.5242909958337484E-4,
      2.099931150361607E-4,
      1.7470192588897718E-4,
      1.4535446859752395E-4,
      1.2095189730212227E-4,
      1.0066358352203966E-4,
      8.37988347826301E-5,
      6.978336683158799E-5,
      5.813972403773392E-5,
      4.8470974011098916E-5,
      4.04471093306309E-5,
      3.37936743657631E-5,
      2.8282200517277988E-5,
      2.3722080373146463E-5,
      1.9953542732997602E-5,
      1.6841412644916558E-5,
      1.4269370950942086E-5,
      1.213450863835374E-5,
      1.0342192289241769E-5,
      8.80177763108668E-6,
      7.4247305061395814E-6,
      6.1282556531428345E-6,
      4.848713278025381E-6,
      3.56685822196953E-6,
      2.3358516405761174E-6,
      1.2823409762916184E-6,
      5.437829885686743E-7,
      1.5949675112240326E-7 };

  /* These match nshmp-haz and should be removed */
  private static final double[] TAPERED_GR_R_TEMP = {
      3.648734771264746E-4,
      3.0351155247723867E-4,
      2.5247694248332143E-4,
      2.1003291504182058E-4,
      1.7473503715378214E-4,
      1.4538201763727166E-4,
      1.2097482132130438E-4,
      1.0068266229609009E-4,
      8.381471718000333E-5,
      6.979659287661517E-5,
      5.815074326255645E-5,
      4.848016071724236E-5,
      4.0454775273297684E-5,
      3.3800079282566184E-5,
      2.8287560844165223E-5,
      2.3726576420233305E-5,
      1.9957324528955872E-5,
      1.6844604598702632E-5,
      1.4272075425536471E-5,
      1.213680849238659E-5,
      1.0344152445466782E-5,
      8.803445832444013E-6,
      7.426137715686031E-6,
      6.1294171417451564E-6,
      4.849632254896957E-6,
      3.5675342487878286E-6,
      2.3362943546550987E-6,
      1.2825840184413834E-6,
      5.438860517858617E-7,
      1.5952698054966957E-7 };

  @Test
  void testTaperedGr() {

    double aVal = Math.log10(4.0);
    double bVal = 0.8;
    double Δm = 0.1;
    double mMin = 5.0;
    double mMax = 8.0;
    double mc = 7.5;

    /* Factory builder; covers props.toBuilder() */
    double incrRate = Mfds.gutenbergRichterRate(aVal, bVal, 5.05);
    Mfd mfd = Mfd.newTaperedGutenbergRichterBuilder(bVal, Δm, mMin, mMax, mc)
        .scaleToIncrementalRate(incrRate)
        .build();
    XySequence xy = mfd.data();
    assertTrue(xy.size() == 30);
    assertArrayEquals(TAPERED_GR_M, xy.xValues().toArray());
    /* These end up being very slightly different due to scaling */
    assertArrayEquals(TAPERED_GR_R_TEMP, xy.yValues().toArray(), 1e-17);

    /* Properties */
    Properties props = new TaperedGr(aVal, bVal, Δm, mMin, mMax, mc);
    TaperedGr grProps = props.getAsGrTaper(); // cover getAs
    assertEquals(aVal, grProps.a());
    assertEquals(bVal, grProps.b());
    assertEquals(Δm, grProps.Δm());
    assertEquals(mMin, grProps.mMin());
    assertEquals(mMax, grProps.mMax());
    assertEquals(mc, grProps.mc());
    assertTrue(xy.size() == 30);
    mfd = props.toBuilder().build();
    xy = mfd.data();
    assertArrayEquals(TAPERED_GR_M, xy.xValues().toArray());
    assertArrayEquals(TAPERED_GR_R, xy.yValues().toArray(), 1e-19);

    grProps = new TaperedGr(grProps, 8.5);
    assertEquals(aVal, grProps.a());
    assertEquals(bVal, grProps.b());
    assertEquals(Δm, grProps.Δm());
    assertEquals(mMin, grProps.mMin());
    assertEquals(8.5, grProps.mMax());
    assertEquals(mc, grProps.mc());

    /* props.toString() */
    assertEquals(
        grProps.type() + " {a=" + grProps.a() +
            ", b=" + grProps.b() +
            ", Δm=" + grProps.Δm() +
            ", mMin=" + grProps.mMin() +
            ", mMax=" + grProps.mMax() +
            ", mc=" + grProps.mc() + "}",
        grProps.toString());
  }

  @Test
  void testBuilderTransforms() {

    /* tolerances set based on experimentaion with each transform */

    double bVal = 1.0;
    double aVal = Math.log10(150.0);
    double Δm = 0.1;
    double mMin = 5.0;
    double mMax = 8.0;

    Mfd mfd = new GutenbergRichter(aVal, bVal, Δm, mMin, mMax)
        .toBuilder()
        .build();

    /* Scale to incremental */
    double incrRate = Mfds.gutenbergRichterRate(aVal, bVal, 5.05);
    Mfd scaleIncrMfd = Mfd.newGutenbergRichterBuilder(bVal, Δm, mMin, mMax)
        .scaleToIncrementalRate(incrRate)
        .build();

    assertArrayEquals(
        mfd.data().yValues().toArray(),
        scaleIncrMfd.data().yValues().toArray(),
        1e-18);

    /* Scale to cumulative */
    double cumulativeRateExpected = mfd.data().yValues().sum() * 2.0;
    Mfd scaleCumulativeMfd = Mfd.newGutenbergRichterBuilder(bVal, Δm, mMin, mMax)
        .scaleToCumulativeRate(cumulativeRateExpected)
        .build();
    double cumulativeRateActual = scaleCumulativeMfd.data().yValues().sum();
    assertEquals(cumulativeRateExpected, cumulativeRateActual, 1e-17);

    /* Scale to moment */
    double momentRateExpected = Mfds.momentRate(mfd.data()) * 2.0;
    Mfd scaleMomentMfd = Mfd.newGutenbergRichterBuilder(bVal, Δm, mMin, mMax)
        .scaleToMomentRate(momentRateExpected)
        .build();
    double momentRateActual = Mfds.momentRate(scaleMomentMfd.data());
    /* very small delta relative to 1e16 moment rate values */
    assertEquals(momentRateExpected, momentRateActual, 10);

    /* Transform */
    double transformRateExpected = cumulativeRateExpected;
    Mfd transformMfd = new GutenbergRichter(aVal, bVal, Δm, mMin, mMax)
        .toBuilder()
        .transform(xy -> xy.set(xy.y() * 2.0))
        .build();
    double transformRateActual = transformMfd.data().yValues().sum();
    /* these values happen to match exactly */
    assertEquals(transformRateExpected, transformRateActual);
  }

  /* No-arg constructor tests */
  static final Path MFDS_JSON = Paths.get("src/test/resources/mfd/mfd-map.json");
  static final Gson GSON = new GsonBuilder().create();

  @Test
  void testNoArgConstructors() {
    JsonObject obj = readJson(MFDS_JSON);
    Single singleProps = GSON.fromJson(obj.get("mfd-single"), Single.class);
    assertEquals(Type.SINGLE, singleProps.type());
    assertEquals(6.75, singleProps.magnitude());
    assertEquals(1.0, singleProps.rate());
    GutenbergRichter grProps = GSON.fromJson(obj.get("mfd-gr"), GutenbergRichter.class);
    assertEquals(Type.GR, grProps.type());
    assertEquals(1.0, grProps.a());
    assertEquals(1.0, grProps.b());
    assertEquals(0.1, grProps.Δm());
    assertEquals(4.7, grProps.mMin());
    assertEquals(6.5, grProps.mMax());
    TaperedGr grTaperProps = GSON.fromJson(obj.get("mfd-gr-taper"), TaperedGr.class);
    assertEquals(Type.GR_TAPER, grTaperProps.type());
    assertEquals(1.0, grTaperProps.a());
    assertEquals(0.8, grTaperProps.b());
    assertEquals(0.1, grTaperProps.Δm());
    assertEquals(5.0, grTaperProps.mMin());
    assertEquals(8.0, grTaperProps.mMax());
    assertEquals(7.5, grTaperProps.mc());
    /* Check default behavior */
    Properties incrProps = GSON.fromJson(obj.get("mfd-empty"), Properties.class);
    assertEquals(Type.INCR, incrProps.type());
  }

  private static JsonObject readJson(Path path) {
    try (BufferedReader br = Files.newBufferedReader(path)) {
      return JsonParser.parseReader(br).getAsJsonObject();
    } catch (IOException ioe) {
      throw new RuntimeException(ioe);
    }
  }
}
