package gov.usgs.earthquake.nshmp.model;

import static gov.usgs.earthquake.nshmp.calc.HazardExport.CURVE_FILE;
import static gov.usgs.earthquake.nshmp.calc.HazardExport.GMM_DIR;
import static gov.usgs.earthquake.nshmp.calc.HazardExport.MAG_DIR;
import static gov.usgs.earthquake.nshmp.calc.HazardExport.TYPE_DIR;
import static java.lang.Math.abs;
import static java.util.stream.Collectors.toMap;
import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.Collection;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.OptionalDouble;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.stream.Stream;

import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.MethodSource;

import gov.usgs.earthquake.nshmp.calc.CalcConfig;
import gov.usgs.earthquake.nshmp.calc.Hazard;
import gov.usgs.earthquake.nshmp.calc.HazardCalcs;
import gov.usgs.earthquake.nshmp.calc.HazardExport;
import gov.usgs.earthquake.nshmp.calc.Site;
import gov.usgs.earthquake.nshmp.calc.Sites;
import gov.usgs.earthquake.nshmp.data.MutableXySequence;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.gmm.Imt;

class LoaderTests {

  /*
   * These tests ensure that the basic model structure is in place. Model
   * components (e.g. model-config.json, gmm-tree.json) are tested by their
   * respective deserializing classes.
   */

  private static final double TOLERANCE = 1e-8;

  static final Path MODEL_PATH = Paths.get("src/test/resources/model/test-model");
  static final Path RESULTS_PATH = MODEL_PATH.resolve("results");
  static final Path SITES_PATH = MODEL_PATH.resolve("test-sites.csv");

  static HazardModel model;
  static List<Site> sites;
  static Expecteds expecteds;
  static ExecutorService exec;

  @BeforeAll
  static void setUpBeforeClass() throws IOException {
    model = ModelLoader.load(MODEL_PATH);
    sites = Sites.fromCsv(SITES_PATH, model.siteData(), OptionalDouble.empty());
    expecteds = new Expecteds();
    int cores = Runtime.getRuntime().availableProcessors();
    exec = Executors.newFixedThreadPool(cores);
  }

  @AfterAll
  static void tearDownAfterClass() {
    exec.shutdown();
  }

  @ParameterizedTest
  @MethodSource("siteStream")
  final void testLocation(Site site) {
    Location loc = site.location();
    // Map<Imt, XySequence> totalExpected = expecteds.totalCurves.get(loc);
    Hazard actual = HazardCalcs.hazard(model, model.config(), site, exec);
    assertCurvesEqual(loc, expecteds, actual, TOLERANCE);
  }

  private static Stream<Site> siteStream() {
    return sites.stream();
  }

  private static void assertCurvesEqual(
      Location loc,
      Expecteds expecteds,
      Hazard actuals,
      double tol) {

    var gmmActualsImtMap = HazardExport.curvesByGmm(actuals);
    var typeActualsImtMap = HazardExport.curvesBySource(actuals);

    for (Imt imt : expecteds.totalCurves.keySet()) {

      // total curve
      XySequence totalExpected = expecteds.totalCurves.get(imt).get(loc);
      XySequence totalActual = actuals.curves().get(imt);
      assertCurveEquals(totalExpected, totalActual, tol);

      /*
       * For decomposed curves we need to loop actuals because expecteds will
       * have additional zero-values curves for missing GMMs, source types and
       * magnitudes.
       */

      // GMM curves
      var gmmExpecteds = expecteds.gmmCurves.get(imt);
      var gmmActuals = gmmActualsImtMap.get(imt);
      for (Gmm gmm : gmmActuals.keySet()) {
        XySequence gmmExpected = gmmExpecteds.get(gmm).get(loc);
        XySequence gmmActual = gmmActuals.get(gmm);
        assertCurveEquals(gmmExpected, gmmActual, tol);
      }

      // Source type curves
      var typeExpecteds = expecteds.typeCurves.get(imt);
      var typeActuals = typeActualsImtMap.get(imt);
      for (SourceType type : typeActuals.keySet()) {
        XySequence typeExpected = typeExpecteds.get(type).get(loc);
        XySequence typeActual = typeActuals.get(type);
        assertCurveEquals(typeExpected, typeActual, tol);
      }

      // Magnitude curves
      var magExpecteds = expecteds.magCurves.get(imt).get(loc);
      var magActuals = actuals.magCurves().get(imt);
      for (Double m : magActuals.keySet()) {
        // if mag bin not used actuals will be null
        // and expecteds will be array of zeros
        if (magActuals.get(m) == null) {
          continue;
        }
        assertCurveEquals(magExpecteds.get(m), magActuals.get(m), tol);
      }

      // Check that actual magnitude curves sum to expected total curve for IMT
      XySequence magCombinedActual = combineMagCurves(
          actuals.curves().get(imt),
          actuals.magCurves().get(imt).values());
      assertCurveEquals(totalExpected, magCombinedActual, tol);
    }
  }

  private static void assertCurveEquals(XySequence expected, XySequence actual, double tol) {

    // IMLs close but not exact due to exp() transform
    assertArrayEquals(
        expected.xValues().toArray(),
        actual.xValues().map(Math::exp).toArray(),
        tol);

    double[] expectedYs = expected.yValues().toArray();
    double[] actualYs = actual.yValues().toArray();

    // y-value difference relative to tolerance
    assertArrayEquals(expectedYs, actualYs, tol);

    // y-value ratio relative to tolerance
    for (int i = 0; i < expectedYs.length; i++) {
      String message = String.format(
          "arrays differ at [%s] expected:<[%s]> but was:<[%s]>",
          i, expectedYs[i], actualYs[i]);
      assertTrue(compare(expectedYs[i], actualYs[i], tol), message);
    }
  }

  private static boolean compare(double expected, double actual, double tolerance) {
    return abs(actual - expected) / expected < tolerance ||
        Double.valueOf(expected).equals(Double.valueOf(actual));
  }

  private static class Expecteds {

    Map<Imt, Map<Location, XySequence>> totalCurves = new EnumMap<>(Imt.class);
    Map<Imt, Map<Gmm, Map<Location, XySequence>>> gmmCurves = new EnumMap<>(Imt.class);
    Map<Imt, Map<SourceType, Map<Location, XySequence>>> typeCurves = new EnumMap<>(Imt.class);
    Map<Imt, Map<Location, Map<Double, XySequence>>> magCurves = new EnumMap<>(Imt.class);

    Expecteds() {
      try {
        Map<Imt, Path> imtDirs = Files.list(RESULTS_PATH)
            .filter(Files::isDirectory)
            .collect(toMap(
                p -> Imt.valueOf(p.getFileName().toString()),
                p -> p));
        imtDirs.forEach(this::loadImtDir);
      } catch (IOException ioe) {
        throw new RuntimeException(ioe);
      }
    }

    void loadImtDir(Imt imt, Path path) {
      try {

        totalCurves.put(imt, readLocationCurves(path.resolve(CURVE_FILE)));

        var imtGmmCurves = Files.list(path.resolve(GMM_DIR))
            .filter(Files::isDirectory)
            .collect(toMap(
                p -> Gmm.valueOf(p.getFileName().toString()),
                p -> readLocationCurves(p.resolve(CURVE_FILE))));
        gmmCurves.put(imt, imtGmmCurves);

        var imtTypeCurves = Files.list(path.resolve(TYPE_DIR))
            .filter(Files::isDirectory)
            .collect(toMap(
                p -> SourceType.valueOf(p.getFileName().toString()),
                p -> readLocationCurves(p.resolve(CURVE_FILE))));
        typeCurves.put(imt, imtTypeCurves);

        var imtMagCurves = Files.list(path.resolve(MAG_DIR))
            .filter(p -> p.getFileName().toString().endsWith(".csv"))
            .collect(toMap(
                p -> readLocation(p),
                p -> readMagnitudeCurves(p)));

        magCurves.put(imt, imtMagCurves);

      } catch (IOException ioe) {
        throw new RuntimeException(ioe);
      }
    }
  }

  private static Map<Location, XySequence> readLocationCurves(Path path) {
    int offset = 3;
    try {
      List<String> lines = Files.readAllLines(path);
      double[] imls = readValues(lines.get(0), offset);
      return lines.stream()
          .skip(1)
          .collect(toMap(
              line -> readLocation(line),
              line -> XySequence.create(imls, readValues(line, 3))));
    } catch (IOException ioe) {
      throw new RuntimeException(ioe);
    }
  }

  /* Read a location from a line string. */
  private static Location readLocation(String line) {
    String[] s = line.split(",");
    return Location.create(
        Double.parseDouble(s[1]),
        Double.valueOf(s[2]));
  }

  /* Read a location from a CSV filename. */
  private static Location readLocation(Path path) {
    String f = path.getFileName().toString();
    String[] s = f.substring(0, f.length() - 4).split(",");
    return Location.create(
        Double.parseDouble(s[1]),
        Double.parseDouble(s[2]));
  }

  private static Map<Double, XySequence> readMagnitudeCurves(Path path) {
    int offset = 1;
    try {
      List<String> lines = Files.readAllLines(path);
      double[] imls = readValues(lines.get(0), offset);
      Map<Double, XySequence> pp = lines.stream()
          .skip(1)
          .collect(toMap(
              line -> Double.valueOf(line.substring(0, line.indexOf(","))),
              line -> readMagnitudeCurve(line, imls, offset)));
      return pp;
    } catch (IOException ioe) {
      throw new RuntimeException(ioe);
    }
  }

  private static XySequence readMagnitudeCurve(
      String line,
      double[] xs,
      int offset) {

    double[] ys = readValues(line, offset);
    return XySequence.create(xs, ys);
  }

  /* Read values from a comma-delimited string. */
  private static double[] readValues(String line, int offset) {
    return Arrays.stream(line.split(","))
        .map(String::trim)
        .skip(offset)
        .mapToDouble(Double::parseDouble)
        .toArray();
  }

  /* --- mag hazard testing --- */
  private static XySequence combineMagCurves(
      XySequence model,
      Collection<MutableXySequence> curves) {

    MutableXySequence xyOut = MutableXySequence.emptyCopyOf(model);
    for (XySequence curve : curves) {
      if (curve != null) {
        xyOut.add(curve);
      }
    }
    return xyOut;
  }

  public static void main(String[] args) throws IOException {
    model = ModelLoader.load(MODEL_PATH);
    List<Site> sites = Sites.fromCsv(SITES_PATH, model.siteData(), OptionalDouble.empty());
    calcWrite(model, model.config(), sites);
  }

  /* For creating/updating results */
  private static Path calcWrite(
      HazardModel model,
      CalcConfig config,
      List<Site> sites) throws IOException {

    int cores = Runtime.getRuntime().availableProcessors();
    exec = Executors.newFixedThreadPool(cores);
    ExecutorService exec = Executors.newFixedThreadPool(cores);
    boolean namedSites = sites.get(0).name() != Site.NO_NAME;
    Path out = HazardExport.createDirectory(config.output.directory);
    HazardExport handler = HazardExport.create(model, config, namedSites, out);
    for (Site site : sites) {
      Hazard hazard = HazardCalcs.hazard(model, config, site, exec);
      handler.write(hazard);
    }
    exec.shutdown();
    return config.output.directory;
  }
}
