package gov.usgs.earthquake.nshmp.model;

import static gov.usgs.earthquake.nshmp.model.SourceFeature.newFaultSection;
import static gov.usgs.earthquake.nshmp.model.SourceFeature.newInterfaceSection;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;

import java.nio.file.Path;
import java.util.List;
import java.util.NoSuchElementException;

import org.junit.jupiter.api.Test;

import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;

class SourceFeatureTests {

  static final Path RESOURCES = Path.of("src/test/resources/model/section/");

  @Test
  final void testCommmonMembers() {

    /* Properties common to fault and interface sections */

    SourceFeature.Fault section = SourceFeature.newFaultSection(
        RESOURCES.resolve("fault-section.geojson"));

    assertEquals(100000, section.id);
    assertEquals("Test Fault Section", section.name);
    assertEquals("NV", section.state.orElseThrow());
    assertEquals(List.of("NV", "CA"), section.states);

    /* Test no states array */

    SourceFeature.Fault noStatesSection = SourceFeature.newFaultSection(
        RESOURCES.resolve("fault-section-no-states.geojson"));

    assertEquals(List.of("NV"), noStatesSection.states);
  }

  @Test
  final void testFaultSection() {

    SourceFeature.Fault section = SourceFeature.newFaultSection(
        RESOURCES.resolve("fault-section.geojson"));

    assertEquals(100000, section.id);
    assertEquals("Test Fault Section", section.name);
    assertEquals("NV", section.state.orElseThrow());
    assertEquals(List.of("NV", "CA"), section.states);
    assertEquals(0.0, section.upperDepth);
    assertEquals(15.0, section.lowerDepth);
    assertEquals(-90.0, section.rake);
    assertEquals(2.0, section.rate);
    assertEquals(RateType.VERTICAL_SLIP, section.rateType);

    LocationList trace = LocationList.of(
        Location.create(-119.8, 38.6),
        Location.create(-119.8, 39.1));
    assertEquals(trace, section.trace);

    /* Fautl section with zone */

    SourceFeature.Fault zoneSection = SourceFeature.newFaultSection(
        RESOURCES.resolve("fault-section-with-zone.geojson"));

    LocationList zoneBorder = LocationList.of(
        Location.create(-119.9, 38.6),
        Location.create(-119.7, 38.6),
        Location.create(-119.7, 39.1),
        Location.create(-119.9, 39.1),
        Location.create(-119.9, 38.6));
    assertEquals(zoneBorder, zoneSection.zone.orElseThrow());

  }

  @Test
  final void testInterfaceSection() {

    SourceFeature.Interface section =
        SourceFeature.newInterfaceSection(
            RESOURCES.resolve("interface-section.geojson"));

    assertEquals(100001, section.id);
    assertEquals("Test Interface Section",
        section.name);
    assertEquals("WA", section.state.orElseThrow());
    assertEquals(List.of("WA", "OR"), section.states);

    LocationList upperTrace = LocationList.of(
        Location.create(-125.53, 45.00, 5.0),
        Location.create(-125.57, 45.45, 5.0),
        Location.create(-125.73, 46.30, 5.0));
    assertEquals(upperTrace, section.traces.get(0));

    LocationList lowerTrace = LocationList.of(
        Location.create(-123.70, 45.00, 31.0),
        Location.create(-123.60, 45.50, 33.1),
        Location.create(-123.41, 46.30, 34.8));
    assertEquals(lowerTrace, section.traces.get(1));
  }

  @Test
  final void testSectionExceptions() {

    /* Missing members */

    assertThrows(NoSuchElementException.class, () -> {
      newFaultSection(RESOURCES.resolve("fault-section-no-id.geojson"));
    });

    assertThrows(NoSuchElementException.class, () -> {
      newFaultSection(RESOURCES.resolve("fault-section-no-name.geojson"));
    });

    assertThrows(NoSuchElementException.class, () -> {
      newFaultSection(RESOURCES.resolve("fault-section-no-state.geojson"));
    });

    assertThrows(NoSuchElementException.class, () -> {
      newFaultSection(RESOURCES.resolve("fault-section-no-dip.geojson"));
    });

    assertThrows(NoSuchElementException.class, () -> {
      newFaultSection(RESOURCES.resolve("fault-section-no-upper-depth.geojson"));
    });

    assertThrows(NoSuchElementException.class, () -> {
      newFaultSection(RESOURCES.resolve("fault-section-no-lower-depth.geojson"));
    });

    assertThrows(NoSuchElementException.class, () -> {
      newFaultSection(RESOURCES.resolve("fault-section-no-rake.geojson"));
    });

    assertThrows(NoSuchElementException.class, () -> {
      newFaultSection(RESOURCES.resolve("fault-section-no-rate.geojson"));
    });

    assertThrows(NoSuchElementException.class, () -> {
      newFaultSection(RESOURCES.resolve("fault-section-no-rate-type.geojson"));
    });

    /* Bad member values */

    assertThrows(IllegalArgumentException.class, () -> {
      newFaultSection(RESOURCES.resolve("fault-section-bad-id.geojson"));
    });

    assertThrows(IllegalArgumentException.class, () -> {
      newFaultSection(RESOURCES.resolve("fault-section-bad-name.geojson"));
    });

    assertThrows(IllegalArgumentException.class, () -> {
      newFaultSection(RESOURCES.resolve("fault-section-bad-states.geojson"));
    });

    assertThrows(IllegalArgumentException.class, () -> {
      newFaultSection(RESOURCES.resolve("fault-section-bad-dip.geojson"));
    });

    assertThrows(IllegalArgumentException.class, () -> {
      newFaultSection(RESOURCES.resolve("fault-section-bad-upper-depth.geojson"));
    });

    assertThrows(IllegalArgumentException.class, () -> {
      newFaultSection(RESOURCES.resolve("fault-section-bad-lower-depth.geojson"));
    });

    assertThrows(IllegalArgumentException.class, () -> {
      newFaultSection(RESOURCES.resolve("fault-section-bad-rake.geojson"));
    });

    assertThrows(IllegalArgumentException.class, () -> {
      newFaultSection(RESOURCES.resolve("fault-section-bad-rate.geojson"));
    });

    assertThrows(IllegalArgumentException.class, () -> {
      newFaultSection(RESOURCES.resolve("fault-section-bad-rate-type.geojson"));
    });

    /* Wrong or disallowed geometry */

    assertThrows(UnsupportedOperationException.class, () -> {
      newFaultSection(RESOURCES.resolve("fault-section-wrong-geometry.geojson"));
    });

    assertThrows(UnsupportedOperationException.class, () -> {
      newInterfaceSection(RESOURCES.resolve("interface-section-wrong-geometry.geojson"));
    });

    // assertThrows(IllegalArgumentException.class, () -> {
    // newInterfaceSection(RESOURCES.resolve("interface-section-bad-traces.geojson"));
    // });
  }
}
